#
# Copyright 2013 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

class ActivitySourceController < ProjectScopedController
  include RabbitServices
  before_filter :setup_config
  before_filter :canonicalize_server_params, only: [:create, :update]
  before_filter :find_source, only: [:edit, :update, :active_status]
  before_filter :establish_rabbitmq
  before_filter :validate_queue_auth, only: [:create]

  before_action :authorize_for_sources_edit!, only: [:edit, :update, :active_status, :index]
  before_action :authorize_for_sources_create!, only: [:new, :create]

  # Constructs the source for arbitrary types and stores the queue username/password in it.
  # This does not save the source!
  #
  # @param [Hash] options Information on what and how to build the source.
  # REQUIRED:
  #   :source_type (constant) - The type of the object to construct.
  #   :source_attributes (Hash) - The parameters to the type's constructor
  #
  # @return [<ActivitySource] An unsaved ActivitySource, constructed with the parameters given, and with the queue
  #   username and password optionally set up.
  def construct_source(options)
    options[:source_type].new(options[:source_attributes]).tap do |source|
      if @queue_username && @queue_password
        source.queue_username = @queue_username
        source.queue_password = @queue_password
      end
    end
  end

  # Constructs a new source for a given project, including SAK + Rabbit Creds
  #
  # @param [Hash] options the options to supply.
  # REQUIRED: :source_type (constant), :source_attributes (hash), :saved_message (string).
  # OPTIONAL: :redirect_path useful when creating a source and the redirect needs to be something other than the
  # current_project's project path
  def add_source_to_project(options)
    source = construct_source(options)
    source.project = current_project
    saved = source.save
    created_everything = saved && create_queue_account(@queue_username, @queue_password)

    if @queue_username && @queue_password
      @rabbitmq_username = @queue_username
      @rabbitmq_password = @queue_password
    end
    @source_server = source

    available_tags if source.is_a? CustomActivitySource
    redirect_path = options[:redirect_path] ||= polymorphic_path(@source_server.class, :project => current_project)

    if created_everything
      flash[:success] = options[:saved_message]
      redirect_to redirect_path
    elsif saved
      flash[:alert] = options[:not_created_message]
      source.delete
      redirect_to redirect_path
    else
      flash[:alert] = t('project.source.not_saved')
      render_crud_form(source)
    end
  end

  # Re-renders the crud form and sets the flash message if need be
  def render_crud_form(source, inhibit_messaging = false)
    set_warning_messages(source) unless inhibit_messaging
    render template: 'shared/crud_form', locals: {source_server: source}
  end

  # Sets the flash messages, separated by line breaks as needed.
  #
  # All text is escaped unless it has already been flagged as html_safe (watch out for variable substitutions!)
  # @param source the source to check for messages
  def set_warning_messages(source)
    messages = warning_messages(source).compact
    messages << flash[:alert] if flash[:alert]

    if messages.present?
      flash.now[:alert] = messages.map { |line| ERB::Util.html_escape(line) }.join('<br>').html_safe
    end
  end

  protected
  # returns hash of breadcrumbs navigation for activity sources
  # @return [Hash] of display text/href for the breadcrumbs
  def breadcrumbs
    Hash.new.tap do |breadcrumbs_hash|
      breadcrumbs_hash[current_project.name] = project_path(current_project)
      breadcrumbs_hash[t("menubar.sources_overview")] = sources_overview_path(current_project)
      if params.present?
        breadcrumbs_hash[params[:controller].singularize.titleize] = nil unless params[:action] == 'index'
      end
    end
  end

  def canonicalize_server_params
    # Override this.
  end

  def find_source
    # Override this
  end

  private
  def validate_queue_auth
    if params[:queue_username] && params[:revealed_password] && params[:obscured_password]
      display_name = @server_params[:display_name].to_s
      short_name = best_key
      @queue_username = "#{short_name}_#{canonicalize(display_name)}"
      @queue_password = params[:obscured_password]

      # Work around an IE8-ism, where it won't copy the obscured password over from the hidden form. [artf153923]
      bad_queue_info = params[:obscured_password].present? && params[:revealed_password] != params[:obscured_password]
      bad_queue_info ||= (params[:queue_username] != @queue_username && params[:queue_username] != @rabbitmq_username)

      if bad_queue_info
        # Password and usernames didn't match; not sure how to convey this better.  TODO(mschweers) - Figure out wording.
        flash[:alert] = t('project.source.not_saved')
                              logger.warn "User #{current_user.uid} edited the source passwords poorly."
        redirect_to :back
      else
        if ActivitySource.find_by_queue_username(@queue_username) != nil
          flash[:alert] = t('project.source.mq_user_already_exists')
          redirect_to :back
        end
      end
    end

    # We do not delete the user on deactivate, or re-create the user on activate because some adapters use one user to
    # converse with multiple sources.
  end

  def canonicalize(display_name)
    display_name.gsub(/^\s+|\s+$/, '').gsub(/\s+/, '_').gsub(/[^a-zA-Z0-9_]/, '').downcase
  end

  # Returns a flash message if the item is not active
  # @param item the item to check activation status
  def deactivated_msg(item)
    I18n.t('form.messages.not_active') if !item.active? && item.persisted?
  end

  # @return [Array][String] messages to be displayed
  def warning_messages(source)
    [deactivated_msg(source)]
  end

  # Updates this object's active flag with the given value.
  # @param item the item being updated
  # @param [String] new_status either "active" to activate the object or "inactive" to deactivate it.
  def update_active_from_param(item, new_status)
    if new_status == "active" && item.valid?
      item.update_attribute(:active, true)
    elsif new_status == "inactive"
      item.update_attribute(:active, false)
    end
  end

  # Set up the view variables for rabbitmq settings.
  def establish_rabbitmq
    uri = URI.parse(@rabbit_config['uri'])

    @rabbitmq_username_enabled = false

    if @source_server
      @rabbitmq_username = @source_server.queue_username || uri.user || 'guest'
      @rabbitmq_password = @source_server.queue_password || uri.password || 'guest'
    else
      config_user = uri.user
      @rabbitmq_username = ''
      @rabbitmq_password = SecureRandom.hex(32).to_s
      @rabbitmq_username_enabled = true
    end
    uri.password = uri.user = nil
    @rabbitmq_uri = uri.to_s
  end

  def available_tags
    @available_tags = ActivitySourceTag.all.sort_by { |tag| tag.name.downcase }
  end

  def sources_index(params)
    @source_types = ActivitySource.source_types
    @selected_source = @source_types.key(params["controller"].classify)
    @project_selected_sources = params["controller"].classify.constantize.all(:project_id => current_project.id)

    render 'project/sources_overview'
  end
end
