#
# Permissions for Reports Controller:
#   successfully authenticated through TeamForge &
#   :reporting_api || :site_admin permissions
#
#
# Copyright 2015 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.

class Api::V1::ReportsController < Api::BaseController

  #Authenticates API calls with TeamForge
  include CtfApiAuthentication

  before_action :authenticate_with_teamforge_and_get_permissions
  before_action :set_reporting_access


  def index
    query_string = request.raw_post
    query_string.insert(0, "in_project #{privileged_projects}\n") unless current_user.super_user
    query_results = QueryLanguage.run(query_string)
    render json: processed_results(query_results, params[:version])
  rescue ArgumentError, SyntaxError, JSON::ParserError, NoMethodError, TypeError => ex
    logger.error(ex)
    render json: { :error => ex.class.to_s+":"+ex.message }, status: 500
  end

  private

  # Preprocess the query results.
  #
  # @param [DecoratedPluckyQuery] query_results Query results
  # @param [String] version API version string.
  # @return [Hash] Query results suitable for rendering as json.
  def processed_results(query_results, version)
    query_results
  end

  # Aggregate source display data from all the ActivitySummaries found in the
  # query results.
  #
  # @param [DecoratedPluckyQuery, Array, Hash] results Data structure to search
  # @return [Hash] Hash map of display data, keyed by source id.
  def add_metadata_to_results(results)
    sources = {}
    work_item_urls = {}
    source_ids = []

    iterator = results.respond_to?(:all) ? results.all : results
    deep_find_sources_and_work_item_urls(iterator, source_ids, work_item_urls)

    ActivitySource.where({id: {:$in => source_ids}}).map { |source|
      # Build basic source display information:
      sources[source.id] = source_display_info(source)
    }

    { source_display_data: sources, work_item_urls: work_item_urls }
  end

  # Build the source display data for a given source
  def source_display_info(source)
    source_data = {
        _type: source._type,
        _links: {
            type_icon: {
                href: icon_url(source),
                type: icon_type(source)
            },
        },
        name: source.display_name,
        background_color: icon_color(source)
    }

    if source.is_a? WorkItemSource
      # Inject information on the source icon, if present (i.e., Tracker sources):
      source_icon = source_icon_url(source)
      if source_icon
        source_data[:_links][:source_icon] = {
            href: source_icon,
            type: source_icon_type(source)
        }
      end
    end
    source_data
  end

  # Recurse over the results pulling out activity sources and work item urls for each item
  def deep_find_sources_and_work_item_urls(results, source_ids, work_item_urls)
    results.each do |item|
      if item.is_a? ActivitySummary
        # Construct and return URLs for work item summaries
        work_item_urls["#{item.id}"] = item.url if item.is_a? WorkItemSummary
        source_ids << item.source_id
      else
        deep_find_sources_and_work_item_urls(item, source_ids, work_item_urls) if [Array, Hash].include? item.class
      end
    end
  end

  # Join two parts of a URL.  This is to work around some odd behaviors in URI.join affected by
  # leading and trailing slashes.
  # @param [String] host Host portion of the URL (possibly including path data)
  # @param [String] path Path portion of the URL
  # @return [String] Cleanly joined URL
  def self.join_url_parts(host, path)
    host.sub(/\/+$/, '') + '/' + path.sub(/^\/+/, '')
  end

  # Same as above.
  def join_url_parts(host, path)
    Api::V1::ReportsController.join_url_parts(host, path)
  end

  # TODO: Tragically duplicated from includes.scss
  ICON_COLOR_MAP = {
      BuildServer => '#1E0F72',
      ReviewServer => '#093420',
      WorkItemSource => '#f2511b',
      CtfRepository => '#275BA6',
      ExternalRepository => '#2c2c2c',
      CustomActivitySource => '#61c0e5'
  }

  BASE_URL = begin
    Rails.configuration.teamforge[:iaf][:orc_url]
  rescue
    'http://invalid.example.com/'
  end

  # "Left-hand-side" icons; the orc icons associated with a general type.  Only customizable
  # from within orchestrate.
  ICON_URL_MAP = {
      BuildServer => join_url_parts(BASE_URL, '/assets/icons/build.png'),
      ReviewServer => join_url_parts(BASE_URL, '/assets/icons/review-request.png'),
      WorkItemSource => join_url_parts(BASE_URL, '/assets/icons/work-item.png'),
      CtfRepository => join_url_parts(BASE_URL, '/assets/icons/commit.png'),
      ExternalRepository => join_url_parts(BASE_URL, '/assets/icons/external.png'),
      CustomActivitySource => join_url_parts(BASE_URL, '/assets/icons/custom.png')
      # In an alternate reality, we will be serving up these miniature assets our outside
      # design firm, but these *cannot* replace the equivalent elements that orchestrate
      # uses:
      #
      # BuildServer          => join_url_parts(BASE_URL, '/assets/icons/source_icons/build.png'),
      # ReviewServer         => join_url_parts(BASE_URL, '/assets/icons/source_icons/review.png'),
      # WorkItemSource       => join_url_parts(BASE_URL, '/assets/icons/source_icons/workitem.png'),
      # CtfRepository        => join_url_parts(BASE_URL, '/assets/icons/source_icons/commit.png'),
      # ExternalRepository   => join_url_parts(BASE_URL, '/assets/icons/source_icons/commit_external.png'),
      # CustomActivitySource => join_url_parts(BASE_URL, '/assets/icons/source_icons/custom.png')
  }

  # @param [ActivitySource] source ActivitySource to examine.
  # @return [String] Color code to use for background (including leading '#').
  def icon_color(source)
    source.respond_to?(:background_color_value) ? source.background_color_value : ICON_COLOR_MAP[source.class]
  end

  # @param [ActivitySource] source ActivitySource to examine.
  # @return [String] URL for icon.
  def icon_url(source)
    if (source.respond_to?(:icon_uri) && source.icon_uri)
      source.icon_uri.start_with?('http') ? source.icon_uri : join_url_parts(BASE_URL, source.icon_uri.sub(/^\/+/, ''))
    else
      ICON_URL_MAP[source.class]
    end
  end

  # @param [ActivitySource] source ActivitySource to examine.
  # @return [String] Content-type for icon.
  def icon_type(source)
    get_type_from_url(icon_url(source))
  end

  # @param [ActivitySource] source ActivitySource to examine.
  # @return [String, Nil] URL for icon (or nil if no icon defined).
  def source_icon_url(source)
    (source.respond_to?(:tracker) && source.tracker.respond_to?(:icon) && source.tracker.icon) ? source.tracker.icon : nil
  end

  # @param [ActivitySource] source ActivitySource to examine.
  # @return [String] Content-type for icon.
  def source_icon_type(source)
    get_type_from_url(source_icon_url(source))
  end

  # @param [String] url URL to examine.
  # @return [String] Content-type for icon.
  def get_type_from_url(url)
    return 'application/unknown' unless url.is_a?(String)
    extension = url.split('.').last
    case extension
      when /(png|gif|bmp|svg)/i
        "image/#{$1}"
      when /jpe?g|jif|jpe|jfif|jfi/i
        'image/jpeg'
      when /tif/i
        'image/tiff'
      else
        'application/octet-stream'
    end
  end
end
