#
# Copyright 2015 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#


# Controller for viewing and creation of Schemas via HTTP API
class Api::V1::SourcesController < Api::BaseController

  class NoPermissionError < Exception
  end


  include RabbitServices
  before_action :authenticate_with_teamforge_and_get_permissions
  before_action :get_project
  before_action :require_crud_permissions

  def index
    sources = ActivitySource.by_project(current_project.id)
    sources = sources.by_tool(params[:tool_id]) if params[:tool_id]
    render json: { sources: sources.all }, status: :ok
  end

  def new
    render json: { source: ActivitySource.generated_columns }, status: :ok
  end

  def show
    source = ActivitySource.where(id: params[:id]).by_project(current_project.id).first
    if source
      render json: { source: source }, status: :ok
    else
      head :not_found
    end
  end

  def create
    source = ApiSourceAdapter.new(params[:source], current_project)

    verify_trusted_sources(nil, params[:source][:trusted_sources]) if params[:source].include?(:trusted_sources)

    if source.save
      create_queue_account(source.native_source.queue_username, source.native_source.queue_password)
      render json: { source: source }, status: :ok
    else
      render json: { errors: source }, status: :unprocessable_entity
    end
  rescue NoPermissionError => npe
    render json: { errors: npe.message }, status: :forbidden
  rescue ArgumentError => arg
    render json: { errors: arg.message }, status: :not_found
  end

  def update

    if params[:source].include?(:trusted_sources)
      verify_trusted_sources(params['id'], params[:source][:trusted_sources])
    end

    source_updater(params[:source])
  rescue NoPermissionError => npe
    render json: { errors: npe.message }, status: :forbidden
  rescue ArgumentError => arg
    render json: { errors: arg.message }, status: :not_found
  end

  def activate
    source_updater({active: true})
  end

  def deactivate
    source_updater({active: false})
  end

  def source_updater(values={})
    source = ActivitySource.where(id: params[:id]).by_project(current_project.id).first
    if source 
      if source.update_attributes(ApiSourceAdapter::ParamsCurator.new(source._type.constantize, values).curated_params)
        head :ok
      else
        render json: { errors: source }, status: :unprocessable_entity
      end
    else
      head :not_found
    end
  end
  private :source_updater

  # base class implements this method to always return false
  # This prevents any previous versions to v6 from being able to set trusted_sources
  def verify_trusted_sources(target_source = nil, trusted_sources = nil)
    raise response_for_lack_of_permissions
  end

  def response_for_lack_of_permissions
    # FIXME: refactor/restructure so method name is either less generic, or less
    # tightly coupled with the specific permission.
    required_perm = :pipeline_trusted_association.to_s
    NoPermissionError.new("Missing required permission (#{required_perm}) on one or more projects")
  end

  def response_for_lack_of_sources
    ArgumentError.new('One or more trusted_sources did not exist')
  end
end
