#
# Copyright 2015 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#
require 'time'

# Controller for
class Api::V3::ActivitiesController < Api::BaseController

  include CtfApiAuthentication

  before_action :authenticate_with_teamforge_and_get_permissions
  before_action :get_project
  before_filter :authorize_for_project!

  TIMESTAMP_FORMAT_REGEX = /\A\d{12}\z/

  # Return all the activities as per requested criteria
  def index

    query_condition = generate_query(params)

    # Get requested response limits or set a limit
    limit = params.has_key?(:limit) ? params[:limit] : 50

    activities = Activity.where(query_condition).limit(limit)

    render json: { activities: activities.all }, status: :ok
  rescue ArgumentError, SyntaxError, JSON::ParserError, NoMethodError, TypeError => ex
    logger.error(ex)
    render json: { :error => ex.class.to_s+': '+ex.message }, status: 400
  end



  #
  # generate the query based on the paramaters passed in to index
  def generate_query(params)
    query_condition = {}

    # Check for requested sources in query params
    query_condition[:source_id] = params.has_key?(:sources) ? add_sources_constraint(params[:sources]) : add_sources_constraint

    # Check for requested types in query params
    query_condition[:_type] = add_type_constraint(params[:types]) if params[:types]

    # Check for requested users in query params. Restrict to activities from a particular project's sources
    query_condition[:created_by] = add_users_constraint(params[:created_by]) if params[:created_by] && query_condition.has_key?(:source_id)

    # Check for requested time range.
    if params[:start_date] && params[:end_date]

      from_time = format_and_validate_timestamp(params[:start_date])
      to_time = format_and_validate_timestamp(params[:end_date])
      gte_time, lt_time = add_time_constraint(from_time, to_time)
      query_condition[:time_created.gte] = gte_time
      query_condition[:time_created.lt] = lt_time

    elsif (params[:start_date] && !params[:end_date]) || (!params[:start_date] && params[:end_date])
      raise ArgumentError, 'Both start and end timestamps are required, if either is provided.'
    end

    # List the requested activities
    query_condition[:_id] = add_activities_constraint(params[:activities]) if params[:activities]
    query_condition
  end

  private

  # Identify and verify parameter sources or get all sources under the project
  def add_sources_constraint(requested_sources = '')

    project_sources = current_project.sources.all.map(&:id)
    project_sources.map! {|source_id| source_id.to_s}

    if requested_sources.empty?
      sources = project_sources
    else
      sources = (requested_sources).split(',').to_a
      # Intersection of sources requested in parameters and sources belonging to that project
      sources = sources & project_sources
    end
    sources
  end


  # Identify types of activities requested
  def add_type_constraint(requested_types)
    requested_types.split(',').to_a
  end


  # Identify users whose activities are requested
  def add_users_constraint(requested_users)
    requested_users.split(',').to_a
  end


  # Identify specific activities requested
  def add_activities_constraint(requested_activities)
    requested_activities.split(',').to_a
  end


  # Validate the time string and convert it to the user's time zone format.
  # UTC is the fallback timezone in case user's timezone is not available.
  # The time string should be in the following format: yyyymmddhhmm
  #   yyyy: Year
  #   mm: Month
  #   dd: Day
  #   hh: Hour
  #   mm: Minutes
  #   Seconds are set to 00 by default
  # Returns Time object in local time zone format. Eg: Sun, 01 Nov 2015 10:15:00 PST -08:00
  def format_and_validate_timestamp(time_string)

    unless time_string.=~(TIMESTAMP_FORMAT_REGEX)
      raise ArgumentError, 'Invalid timestamp format. Timestamp should be a 12 digit string in the format yyyymmddhhmm.'
    end

    year = time_string[0..3].to_i
    month = time_string[4..5].to_i
    day = time_string[6..7].to_i
    hour = time_string[8..9].to_i
    minutes = time_string[10..11].to_i
    seconds = 0

    if (hour > 23 || minutes > 59)
      raise ArgumentError, 'Invalid timestamp format. Hours should be less than 24 and minutes less than 60.'
    end

    Time.zone = (current_user.time_zone if current_user) || 'UTC'
    Time.zone.local(year, month, day, hour, minutes, seconds)
  end


  # Convert the time objects to UTC, ensure start and end times form a valid time range.
  def add_time_constraint(from_time = nil, to_time = nil)

    from_time_utc = from_time.utc unless from_time.nil?
    to_time_utc = to_time.utc unless to_time.nil?

    if from_time_utc >= to_time_utc
      raise ArgumentError, 'The end time must be more recent than start time.'
    end

    return from_time_utc, to_time_utc
  end

end
