#
# Copyright 2016 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

class Api::V4::ReportsController < Api::V3::ReportsController
  # Preprocess the query results.
  #
  # @param [DecoratedPluckyQuery] query_results Query results
  # @param [String] version API version string.
  # @return [Hash] Query results suitable for rendering as json.
  def processed_results(query_results, version)
    walk_and_transform(super)
  end

  def source_display_info(source)
    # Build basic source display information:
    source_data = {
        _type: source._type,
        _links: {
            type_icon: {
                href: icon_url(source),
                type: icon_type(source)
            },
        },
        name: source.display_name,
        background_color: icon_color(source),
        source_association_key: source.source_association_key
    }

    # Add the schema name, if applicable:
    source_data[:schema_display_name] = source.schema.name if source.is_a? CustomActivitySource


    if source.is_a? WorkItemSource
      # Inject information on the source icon, if present (i.e., Tracker sources):
      source_icon = source_icon_url(source)
      if source_icon
        source_data[:_links][:source_icon] = {
            href: source_icon,
            type: source_icon_type(source)
        }
      end
    end

    source_data
  end

  # Walk a data structure, performing any specified proccessing on nodes.
  #
  # @param [DecoratedPluckyQuery, Hash, Array, String, Object] results Node to walk.
  # @return [Hash, Array, String, Object] Node procesed by the transformations.
  def walk_and_transform(results)
    # Force execution of query
    results = results.respond_to?(:all) ? results.all : results

    results = interpolate_ctf_urls(results) if results.is_a?(String)
    results = translate_summary_to_decorated_hash(results) if results.is_a?(ActivitySummary)

    # Walk the structure (destructively):
    if results.is_a? Array
      results.map! { |item| walk_and_transform(item) }
    elsif results.is_a? Hash
      results.each { |k, v| results[k] = walk_and_transform(v) }
    end
    results
  end

  # Perform CTF %VAR% interpolations on the provided string.
  #
  # @param [String] text The string to interpolate
  # @return [String] Transformed string.
  def interpolate_ctf_urls(text)
    CtfInterpolations.interpolate_ctf_tokens(text)
  end

  # Translates an ActivitySummary to Hash and decorates with additional information.
  #
  # @param [ActivitySummary] summary_obj An activity summary object.
  # @return [Hash] Decorated and converted hash representation of input.
  def translate_summary_to_decorated_hash(summary_obj)
    current_source = summary_obj.source
    return_hash = {'source_association_key' => current_source.source_association_key}
    if summary_obj.is_a? CustomSummary
      return_hash['schema_display_name'] = current_source.schema.name
    end
    summary_obj.attributes.merge!( return_hash )
  end

end
