#
# Copyright 2016 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

# Controller for CRUD actions on Webhooks
class Api::V5::WebhooksController < Api::BaseController
  include RabbitServices
  include CtfApiAuthentication

  before_action :authenticate_with_teamforge_and_get_permissions
  before_action :get_project
  before_action :require_crud_permissions

  # Return all the active webhooks for the project
  def index
    webhooks = WebhookSummary.where(project_id: current_project.id).all
    render json: { webhooks: webhooks }, status: :ok

  end

  # This method contains steps to generate Webhook URL with source.
  def create
    # Steps Creating Webhook URL:
    # 1. create the source
    result = create_source_for_webhook(params[:source])
    return render json: result, status: :unprocessable_entity if result.key?(:errors)
    webhook_source = result[:source]

    # 2. construct the webhook URL
    webhook_url = construct_webhook_url(params[:source], webhook_source)

    # 3. save the document to WebhookSummary collection
    save_result = save_webhook(webhook_url, webhook_source.source_association_key, params[:source][:field_mapping])

    if save_result[:saved]
      render json: { webhook_source: webhook_source, webhook_info: save_result[:webhook_info] }, status: :ok
    else
      render json: { errors: save_result[:webhook_info] }, status: :unprocessable_entity
    end
  end

  private

  def use_existing_or_create_tool(tool_hash)
    if Tool.name_unique?(current_project, tool_hash[:name])
      new_tool = Tool.new(name: tool_hash[:name], service_name: tool_hash[:service_name], project: current_project)
      if new_tool.save && new_tool.errors.empty?
        return new_tool
      else
        render json: { errors: new_tool.errors.messages }, status: :unprocessable_entity
      end
    else
      existing_tool = Tool.where(name: tool_hash[:name]).first
      return existing_tool
    end
  end


  # Create a source
  def create_source_for_webhook(source_info={})
    source = ApiSourceAdapter.new(source_info, current_project)

    if source.save
      create_queue_account(source.native_source.queue_username, source.native_source.queue_password)
      return { source: source.native_source }
    end
    { errors: source }
  end

  # Construct Webhook URL
  def construct_webhook_url(src_payload = {}, webhook_source)

    orc_iaf_url = Rails.configuration.teamforge[:iaf][:orc_url]
    project_url = current_project.project_configuration.project_url
    field_mapping = src_payload[:field_mapping]
    sak = webhook_source.source_association_key

    webhook_url_endpoint ="/api/latest/projects/#{project_url}/webhook/packets/#{field_mapping}/#{sak}"
    return orc_iaf_url + webhook_url_endpoint
  end

  # def save_webhook(webhook_url, sak, tool_id, service_name)
  def save_webhook(webhook_url, sak, packet_map)
    project_id = current_project.id

    webhook = WebhookSummary.new(associated_sak: sak, webhook_url: webhook_url, project_id: project_id, packet_map: packet_map)
    save_result = {}

    if (save_result[:saved] = webhook.save)
      # Merge webhook_info to the payload to be returned
      webhook_info = {}
      webhook_info[:id] = webhook.id
      webhook_info[:webhook_url] = webhook_url
      webhook_info[:associated_sak] = sak
      webhook_info[:packet_map] = packet_map
      save_result[:webhook_info] = webhook_info
    else
      save_result[:webhook_info] = webhook
    end

    return save_result
  end

end