#
# Copyright 2016 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

class Api::V6::ActivitiesController < Api::V5::ActivitiesController

  # Return all the activities as per requested criteria
  def index

    # Get requested response limits or set a limit
    limit = params.has_key?(:limit) ? params[:limit].to_i : 25

    query_condition = generate_query(params, limit)

    activities = Activity.where(query_condition).sort(:time_created.desc).limit(limit).all
    latest_activity = Activity.where(query_condition).sort(:atomic.desc).limit(limit).first
    latest_atomic_id = latest_activity.try(:atomic)

    sources = get_info_for_activity_sources(activities)
    project = current_project.project_configuration
    processed_activities = []

    activities.each do |activity|
      processed_activities << walk_and_transform(activity)
    end

    render json: { activities: processed_activities, sources: sources, project: project, latest_atomic_id: latest_atomic_id }, status: :ok
  rescue ArgumentError, SyntaxError, JSON::ParserError, NoMethodError, TypeError => ex
    logger.error(ex)
    render json: { :error => ex.class.to_s + ': ' + ex.message }, status: 400
  end


  # Generate the query based on the parameters passed in to index
  #
  # @param [Hash] params The parameters sent with the request
  # @param [Integer] limit The number of activities requested
  # @return [Hash] query_condition The hash constructed from received query parameters
  def generate_query(params, limit)
    query_condition = {}

    # Check for requested sources in query params
    query_condition[:source_id] = params.has_key?(:sources) ? add_sources_constraint(params[:sources]) : add_sources_constraint

    # Check for requested types in query params
    query_condition[:_type] = add_type_constraint(params[:types]) if params[:types]

    # Check for requested users in query params. Restrict to activities from a particular project's sources
    query_condition[:created_by] = add_users_constraint(params[:created_by]) if params[:created_by] && query_condition.has_key?(:source_id)

    # Check for requested time range.
    if params[:start_date] && params[:end_date]

      from_time = format_and_validate_timestamp(params[:start_date])
      to_time = format_and_validate_timestamp(params[:end_date])
      gte_time, lt_time = add_time_constraint(from_time, to_time)
      query_condition[:time_created.gte] = gte_time
      query_condition[:time_created.lt] = lt_time

    elsif (params[:start_date] && !params[:end_date]) || (!params[:start_date] && params[:end_date])
      raise ArgumentError, 'Both start and end timestamps are required, if either is provided.'
    end

    # List the requested activities
    query_condition[:_id] = add_activities_constraint(params[:activities]) if params[:activities]

    # Since_atomic_id is used for fetching latest activities while prior_to_atomic_id
    # is used for fetching older activities for infinite scroll. Both the parameters
    # are based on the atomic id of the activity, and have contrast uses. Hence, only
    # one of them can be used at a time. Frontend code has to take care of sending in
    # only one of the parameters. Else, just due to sequential execution, prior_to_atomic_id
    # gets preference.
    #
    # since_atomic_id condition supports the requests for activity updates
    if params[:since_atomic_id]
      query_condition[:atomic.gt] = params[:since_atomic_id].to_i
    end

    # prior_to_atomic_id condition supports requests for activities previous to stated atomic_id.
    # Used for infinite scroll functionality of EventQ
    if params[:prior_to_atomic_id]
      query_condition[:atomic.lt] = params[:prior_to_atomic_id].to_i
    end

    query_condition
  end


  # Get the necessary information from sources for all the activities being returned
  #
  # @param [Array] all_activities All the activities to be returned for this request
  # @return [Array] all_sources Array of hash objects with source information
  def get_info_for_activity_sources(all_activities)

    source_ids_for_activities = []
    all_sources = []
    relative_path = '/orc/assets/icons/'

    all_activities.each { |each_activity| source_ids_for_activities << each_activity.source_id}

    sources = ActivitySource.where(id: source_ids_for_activities).all

    sources.each do |source|

      source_data = {}
      source_data[:id] = source.id
      source_data[:display_name] = source.display_name


      if source.is_a? WorkItemSource
        tracker = source.tracker
        statuses = {}
        source_data[:icon] = WorkItemSource.convert_to_absolute_url(tracker.icon.to_s)

        tracker.statuses.each do |status|
          statuses[status[:remote_id]] = status[:name]
        end

        source_data[:statuses] = statuses
        source_data[:work_item_source_url] = source.url

      elsif source.is_a? CustomActivitySource
        normalized_url = CustomActivitySource.normalize_orchestrate_icon_uri(source.icon_uri, true)
        source_data[:icon] = CustomActivitySource.convert_to_absolute_url(normalized_url)
        source_data[:background_color_value] = source.background_color_value

      elsif source.is_a? CtfRepository
        source_data[:icon] = CtfRepository.convert_to_absolute_url(relative_path + 'commit.png')

      elsif source.is_a? ExternalRepository
        source_data[:icon] = ExternalRepository.convert_to_absolute_url(relative_path + 'commit.png')

      elsif source.is_a? BuildServer
        source_data[:icon] = BuildServer.convert_to_absolute_url(relative_path + 'build.png')

      elsif source.is_a? ReviewServer
        source_data[:icon] = ReviewServer.convert_to_absolute_url(relative_path + 'review-request.png')
      end

      all_sources << source_data
    end

    return all_sources
  end


  # Walk a data structure, performing any specified processing on nodes.
  #
  # @param [DecoratedPluckyQuery, Hash, Array, String, Object] results Node to walk.
  # @return [Hash, Array, String, Object] Node processed by the transformations.
  def walk_and_transform(results)

    # Force execution of query
    results = results.respond_to?(:all) ? results.all : results

    results = interpolate_ctf_urls(results) if results.is_a?(String)

    results = transform_activity_to_hash(results) if results.is_a? Activity

    # Walk the structure (destructively):
    if results.is_a? Array
      results.map! { |item| walk_and_transform(item) }
    elsif results.is_a? Hash
      results.each { |k, v| results[k] = walk_and_transform(v) }
      results['link_to'] = flatten_link_to(results['link_to']) if results.key?('link_to')
    end
    results
  end

  # Perform CTF %VAR% interpolations on the provided string.
  #
  # @param [String] text The string to interpolate
  # @return [String] Transformed string.
  def interpolate_ctf_urls(text)
    CtfInterpolations.interpolate_ctf_tokens(text)
  end

  # Flatten the contents of 'link_to' to a single string, looking for the most
  # URL-like element (or failing that, the first element, or the thing itself).
  # See related: custom_helper.rb#linkify_custom_linkto
  #
  # @param [String, Array, Object] obj Thing accused of harboring a URI
  # @return [String] Our best guess as to what the link was meant to be.
  def flatten_link_to(obj)
    return obj if obj.is_a?(String) # Already a string.  Go home.
    if obj.kind_of? Array
      obj.each do |element|
        return element if (element.is_a?(String) && element =~ /^https?:\/\//i)
      end  
      return obj[0].to_s # Ok, it was an array, no good match.  Return the first.
    end  
    obj.to_s # I got no earthly idea what it was.  String it!
  end  
  
  # Transform the Activity object to a hash and interpolate the link_to field for XDS
  #
  # @param [Activity Object] The activity object to transform
  # @return [Hash] Transformed hash
  def transform_activity_to_hash(results)
    JSON.parse(results.to_json)
  end

end
