#
# Copyright 2016 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

class Api::V6::SourcesController < Api::V5::SourcesController

  include CtfApiAuthentication


  ####################################################################################
  #  push an array of  activity sources or a sub-array thereof to the trusted sources
  #  array in a XDS source
  #  returns: the array of sources inserted into the trusted sources list and
  #           status code 200
  #           status code 404 if the XDS source does not exist
  #           status code 403 if the user does not have permission in his
  #           project
  #
  #  POST   https://ctf_hostname/orc/api/:version/projects/:project/sources/:id/trusted_sources
  #         id - XDS source id
  #####################################################################################

  def trusted_sources
    trusted_sources = params['trusted_sources']
    target_source = params['id']

    verify_trusted_sources(target_source, trusted_sources)
    xds_source = get_xds_source(target_source)

    trusted_list = xds_source.insert_trusted_sources(trusted_sources)

    response     = {trusted_sources: trusted_list}
    render :json => response, status: :ok
  rescue ArgumentError => arg
    render json: { errors: arg.message }, status: :not_found
  rescue NoPermissionError => npe
    render json: { errors: npe.message }, status: :forbidden
  end


  ##################################################################################
  #  lists all activity sources that can be trusted and hence can be pushed into
  #  a XDS source's trusted sources array
  #
  #  Returns: status code 404 if the XDS source does not exist
  #           status code 200 and the array of all sources which can b inserted
  #           as trusted sources in the XDS trusted sources array otherwise
  #
  #  GET  https://ctf_hostname/orc/api/:version/projects/:project/sources/:id/trusted_list
  #       id - XDS source id
  ###################################################################################

  def list_trusted_sources
    # get the XDS source
    xds_source = get_xds_source(params['id'])

    target_source = params['id']

    verify_trusted_sources(target_source, [])

    trustable_sources = []
    projects = current_user.projects_with_permission :pipeline_trusted_association
    projects.each do |project|
      trustable_sources << project.sources
    end

    trustable_sources = trustable_sources.flatten.map(&:id).reject {|src_id| src_id == xds_source.id}
    trustable_sources = trustable_sources.uniq.compact

    response = {trustable_sources: trustable_sources}
    render :json => response, status: :ok
  rescue ArgumentError => arg
    render json: { errors: arg.message }, status: :not_found
  rescue NoPermissionError => npe
    render json: { errors: npe.message }, status: :forbidden
  end

  def delete_trusted_sources
    xds_source = get_xds_source(params['id'])

    verify_trusted_sources(params['id'], xds_source ? xds_source.trusted_sources : [])

    count = xds_source.trusted_sources.length

    xds_source.trusted_sources = []
    xds_source.save

    render :json => {count: count}, status: :ok
  rescue ArgumentError => arg
    render json: { errors: arg.message }, status: :not_found
  rescue NoPermissionError => npe
    render json: { errors: npe.message }, status: :forbidden
  end

  private

  def verify_trusted_sources(target_source, trusted_sources)
    xds_source = validate_target_source(target_source) unless params[:action] == 'create'

    sources = validate_trusted_sources(trusted_sources) unless params[:action] == 'list_trusted_sources'
    all_sources = sources || []
    all_sources << xds_source unless params[:action] == 'create'
    raise response_for_lack_of_permissions unless permitted_to_add_trusted_sources?(all_sources)
  end

  def validate_trusted_sources(trusted_sources)
    raise response_for_lack_of_sources if trusted_sources.blank?
    sources = ActivitySource.find(trusted_sources)
    raise response_for_lack_of_sources if sources.count < trusted_sources.count
    sources
  end

  def validate_target_source(target_source)
    xds_source = get_xds_source(target_source)
    raise response_for_nil_xds_source if xds_source.nil?
    xds_source
  end

  def get_xds_source(id)
    CustomActivitySource.find(id)
  end

  def response_for_nil_xds_source
    ArgumentError.new('xds source not found')
  end

  def permitted_to_add_trusted_sources?(sources)
    projects = sources.map(&:project) << current_project
    projects.uniq.all? {|project| create_trusted_associations?(project.id, current_user) }
  end
end

