#
# Copyright 2013 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

# Base controller for the Orchestrate application.
class ApplicationController < ActionController::Base
  include ApiWebCommon

  include ::NewRelic::Agent::MethodTracer
  rescue_from SystemCallError, with: :service_unavailable
  rescue_from MongoMapper::DocumentNotFound, with: :not_found
  rescue_from CanCan::AccessDenied, with: :unauthorized

  helper_method :current_project
  helper_method :path_options_with_project_context
  helper_method CtfUserAuthorization.instance_methods # Needed for access to permission check methods in views
  before_filter :save_path_for_devise_redirect

  # The user authentication must happen before we include the TeamforgeApplication module.
  before_filter :authenticate_user!

  before_filter :try_completing_project_info

  include TeamforgeApplication if Rails.configuration.teamforge[:enabled]
  include ApplicationHelper

  before_filter :body_id

  protect_from_forgery
  # DO NOT MOVE the overridden route because putting them in a module will require a chain of dependencies to move in the module
  # And cause some of the helpers and controller method to cache memoized object in 2 different places and therefore
  # not work correctly (specifically current_project)
  #

  private 
  # Get the current project context Hash to use in named path as querystring.
  # Will create a hash such as {:project => 'project_url'}
  #
  # @param [Hash] options the options to control where to pick or not the project context
  # @option  :ignore_project_context  when set to true, it generate an empty hash and basically ignore the project context
  # @return [Hash] a hash containing the project url and the key we use in our app to keep the context {:project => 'project_url'}
  def project_context_querystring_hash(options)
    ignore_project_context = !!options.delete(:ignore_project_context)

    #keep current project context, or if empty, use the current project
    project = options.delete(:project) || current_project
    !ignore_project_context && project ? {:project => project.project_configuration.project_url} : {}
  end

  # Attempt to ensure that the current project (if available) is complete in its configuration.  Practically speaking,
  # this means trying to make sure that the link_plug_id has been fetched and saved for the project.
  #
  # NOTE: this method *may* make a SOAP call into CTF (if the project has never had its link_plug_id set), but this is an
  # exceedingly rare occurrence - a superuser (avoiding permissions checks and thus project info evaluation) would have
  # to hit a page with a project context before a normal user (subject to the aforementioned checks) hit the page, and
  # even then, the superuser is paying the price of the hit once to the benefit of every subsequent user.
  def try_completing_project_info
    if current_project
      current_project.project_configuration.get_link_plug_id(session['ctf_soap_token'])
      logger.debug "link_plug_id for #{current_project.name} is #{current_project.project_configuration.link_plug_id}"
    end
  rescue
    logger.debug "Unable to complete project info (soap token:#{session['ctf_soap_token']})"
  end

  # Respond to the end user with unauthorized, rather than opaquely with 404
  def render_unauthorized
    render_exception(401)
  end

  # session["user_return_to"] is used by Devise to do a proper redirect.
  # If the request is not a GET then redirect to current project page (the save will be lost),
  # otherewise redirect to the root path (project index page).
  # We do not want to set the return-to path if it's an XHR request, because those will not be user-friendly paths.
  def remember_path
    if request.post? || request.put? || request.patch?
      session["user_return_to"] = project_path(current_project) if current_project
    else
      session["user_return_to"] = request.fullpath unless request.xhr?
    end
  end

  # we want to save the path only when the user is not already authenticated
  def save_path_for_devise_redirect
    remember_path unless user_signed_in?
  end

  # Sets a body id tag for all controllers, override when needed by individual controllers
  def body_id
    @body_id = [controller_name, action_name].join('-')
  end

  add_method_tracer :save_path_for_devise_redirect
  add_method_tracer :try_completing_project_info
  add_method_tracer :authenticate_user!
end
