#
# Copyright 2013 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

# Controller for accessing and manipulating build server data.
class BuildServersController < ActivitySourceController
  helper_method :available_commit_sources, :valid_associated_source?

  def index
    sources_index(params)
  end

  # Shows the form for creating a new build server.
  def new
    render template: 'shared/crud_form', locals: {source_server: BuildServer.new}
  end

  # Edits a build server.  Sets the deactivated message if the build source is inactive.
  def edit
    render_crud_form(@build_server)
  end

  # Updates a build server.
  def update

    if @build_server.update_attributes(params[:build_server]) && !deleted_commit_source?(@build_server)
      flash[:success] = t("ci_server.messages.server_updated") if @build_server.save
      redirect_to build_servers_path(@build_server.project)
    else
      render_crud_form(@build_server)
    end
  end

  # Creates a build server. If it's successfully saved, the server will be added to the project
  def create
    add_source_to_project(
        source_type: BuildServer,
        source_attributes: params[:build_server],
        saved_message: t('ci_server.messages.server_created'),
        not_created_message: t('ci_server.messages.server_not_created'))
  end

  # Sets the active status on a build server. The params[:new_status] must be one of "active" or "inactive". Anything
  # else does nothing.
  def active_status
    update_active_from_param(@build_server, params[:new_status])
    path_opts = path_options_with_project_context(action: :edit)
    redirect_to polymorphic_path(@build_server, path_opts)
  end

  # Ordered list of the available commit sources
  def available_commit_sources
    current_project.commit_sources.sort_by {|current_src| current_src.display_name.downcase}
  end

  def canonicalize_server_params
    @server_params = params[:build_server]
  end

  private

  # @return [Boolean] true if the associated commit source has been deleted
  def deleted_commit_source?(source)
    source.commit_server.present? && !valid_associated_source?(source)
  end

  # @return [Array][String] messages to be displayed
  def warning_messages(source)
    super << deleted_commit_source_msg(source) << missing_associations_msg(source)
  end

  # @return [String] warning message if the associated commit source (or step) has been deleted
  def deleted_commit_source_msg(source)
    if deleted_commit_source? source
      I18n.t('form.messages.deleted_association_source',
               source_type: I18n.t('entity.commit').downcase,
               commit_source: ERB::Util.html_escape(source.commit_server.display_name)).html_safe
    end
  end

  # @return [String] warning message if the selected commit source is not valid
  def missing_associations_msg(source)
    I18n.t('form.messages.missing_associations') unless valid_associated_source?(source)
  end

  # @return [Boolean] true if the associated commit source is a commit source in this project
  def valid_associated_source?(source)
    available_commit_sources.include?(source.commit_server)
  end

  # Finds the build server by its ID or raises an error if it's not found.
  def find_source
    @source_server = @build_server = BuildServer.where(id: params[:id], project_id: current_project.id).first
    raise MongoMapper::DocumentNotFound.new(t("ci_server.messages.not_found")) unless @build_server
  end

end
