#
# Copyright 2013 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

# Controller for SCM repositories

class CommitServersController < ActivitySourceController
  before_filter :find_ctf_repositories!, only: [:new, :create, :edit]
  before_filter :ctf_id_cannot_be_changed, only: [:update]
  before_filter :check_repository_type_is_valid, only: [:create]

  VALID_TYPES = ["CtfRepository", "ExternalRepository"]

  def index
    sources_index(params)
  end

  # Shows the form for creating a new repository.
  def new
    # the default type is CtfRepository
    commit_server = CtfRepository.new
    render template: 'shared/crud_form', locals: {source_server: commit_server}
  end

  # Shows the form for editing a repository.
  # @raise [MongoMapper::DocumentNotFound] if a repository with the given project doesn't exist
  def edit
    render_crud_form(@commit_server)
  end

  # Updates a scm server.
  def update
    if @commit_server.update_attributes(params[:commit_server])
      flash[:success] = t('commit_server.messages.repository_updated')
      redirect_to commit_servers_path(@commit_server.project)
    else
      render_crud_form(@commit_server)
    end
  end

  # Creates a scm server or tests that it is valid.
  def create
    add_source_to_project(
        source_type: params[:commit_server][:repository_type].constantize,
        source_attributes: params[:commit_server].merge(ActivitySource.generated_columns),
        saved_message: t('commit_server.messages.repository_created'),
        not_created_message: t('commit_server.messages.repository_not_created'))
  end

  # Sets the active status on a scm server. The params[:new_status] must be one of "active" or "inactive". Anything
  # else does nothing.
  def active_status
    update_active_from_param(@commit_server, params[:new_status])
    path_opts = path_options_with_project_context(action: :edit)
    redirect_to polymorphic_path(@commit_server, path_opts)
  end

  def canonicalize_server_params
    @server_params = params[:commit_server]
  end

  # @return [Array<CtfRepositoryData>] a list of ctf repositories for a project that ctf is aware about ant that are
  # not being used yet in the current project
  def find_ctf_repositories!
    @ctf_repositories = repository_list(@ctf_session, @commit_server) if @ctf_session && (@commit_server.blank? || @commit_server.is_a?(CtfRepository))
  end

  # List of information on repositories that aren't in use by any project
  # @see ScmService#repositories_for_ctf_project
  # @return [[Hash]] hash of name, url and ctf_id for this pipeline
  def repository_list(ctf, commit_server)
    repositories_reported_by_ctf = CtfRepository.project_repositories(ctf, current_project)
    if commit_server  # grab self from the list reported by CTF
      repositories_reported_by_ctf.select{|repo| repo.ctf_id == commit_server.ctf_id}
    else
      repositories_reported_by_ctf
    end
  end

  private

  # Finds the repository with the ID given by the params hash.
  # @raise [MongoMapper::DocumentNotFound] if there is no repository with that ID in the current pipeline
  def find_source
    @source_server = @commit_server = CommitServer.where(id: params[:id], project_id: current_project.id).first
    raise MongoMapper::DocumentNotFound.new(t("commit_server.messages.not_found")) unless @commit_server
  end

  # once set, the ctf_id should not be writable
  def ctf_id_cannot_be_changed
    requested_ctf_id = params[:commit_server][:ctf_id]
    params[:commit_server].delete(:ctf_id) if requested_ctf_id && @commit_server.ctf_id != requested_ctf_id # if the user decides to forge the form
  end

  # a guard against the malicious users in case they try to pass something like "User" as a repository type
  def check_repository_type_is_valid
    unless VALID_TYPES.include? params[:commit_server][:repository_type]
      flash[:alert] = t('commit_server.errors.invalid_repository_type')
      render_crud_form(CtfRepository.new)
    end
  end

end
