#
# Copyright 2013 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#
# Controller for resources under Projects
# Convenient point to hang authorization and project-specific functionality
class ProjectScopedController < ApplicationController
  helper_method :breadcrumbs

  ## using alias for auth..pipe..edit for accessing sources_overview
  before_filter :authorize_for_project!
  
  protected

  # Check to see if the user is authorized to access a given project, defaulting to current_project.
  # @param [project] project The project to check permission on. Optional, defaults to current_project.
  # @return [boolean] true if the user is allowed to access the project, or current_project if none provided
  # @raises CanCan::AccessDenied if the user cannot perform the operation
  def authorize_for_project!(project = current_project)
    authorize! :pipeline_read, project
  end

  # Check to see if the user is authorized to edit a given project, defaulting to current_project.
  # @param [project?] project The project to check permission on. Optional, defaults to current_project.
  # @return [boolean] true if the user is allowed to edit the project, or current_project if none provided
  # @raises CanCan::AccessDenied if the user cannot perform the operation
  def authorize_for_pipeline_edit!(project = current_project)
    authorize! :pipeline_edit, project
  end

  alias :authorize_for_sources_edit! :authorize_for_pipeline_edit!

  # Check to see if the user is authorized to create sources in projects
  # @return [boolean] true if the user is allowed to create a source for any project
  # @raises CanCan::AccessDenied if the user cannot perform the operation
  def authorize_for_pipeline_create!
    authorize! :pipeline_create, Project
  end

  alias :authorize_for_sources_create! :authorize_for_pipeline_create!

  def authorize_for_edit_create_permissions!(project = current_project)
    # TODO: Couldn't come with a better alternative to
    # authorize! :pipeline_edit, project || authorize! pipeline_create, project
    if has_permission_for_project_id?(:pipeline_edit, project.id)
      authorize! :pipeline_edit, project
    elsif has_permission_for_project_id?(:pipeline_create, project.id)
      authorize! :pipeline_create, project
    end
    authorize! :pipeline_edit, project
  end

  def add_project_crumbs
    Hash.new.tap do |breadcrumbs_hash|
      breadcrumbs_hash[current_project.name] = project_path(current_project) if current_project
    end
  end

  # breadcrumbs for the activity-browsing contexts
  # @return [Hash] of display text/href for the breadcrumbs
  def breadcrumbs_activities
    add_project_crumbs.tap do |breadcrumbs_hash|
      breadcrumbs_hash[current_project.name] = project_path(current_project) unless current_project
      breadcrumbs_hash[t("menubar.activity", {project_name: current_project.name})] = activity_index_path(current_project) if current_project
      breadcrumbs_hash[activity_display_name(@activity, true)] = activity_path(@activity) if @activity.is_a? ActivitySummary
    end
  end

  # breadcrumb navigation default
  # @return [Hash] of display text/href for the breadcrumbs
  def breadcrumbs
    breadcrumbs_activities
  end

  add_method_tracer :authorize_for_project!
end
