#
# Copyright 2013 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

# Handles responses to TeamForge for the IAF integration.
# Described more in our wiki: https://forge.collab.net/sf/wiki/do/viewPage/projects.orch/wiki/CreatingAnIntegratedApp
class Services::TeamForgeController < ActionController::Base
  include ::NewRelic::Agent::MethodTracer

  rescue_from TeamForge::ServiceError, with: :not_found # authentication failure in this context (can't verify that the request came from TeamForge)
  before_filter :debug
  # Do not authenticate pre-commit messages until we are gating/preventing commits,
  # so as not to incur the authentication overhead.
  before_filter :authenticate, except: [:pre_commit_message, :detailed_precommit_message]

  respond_to :xml

  # Handles GET ConfigurationParameters
  def get_configuration_parameters
    # passed null tf_project_id to get the global config
    # passed tf_project_id (ex. proj1008) when it's added to a project

    # When added to a project:
    # {"format"=>:xml, "controller"=>"services/team_forge", "action"=>"get_configuration_parameters", "token"=>"3315DFE7B419A9FBF40EA47E62A72EC6", "tf_project_id"=>"proj1008"}
    ctf_project_id = params['tf_project_id']
    if ctf_project_id != 'null'
      project_config = TeamForge::ProjectConfiguration.find_by_ctf_project_id(ctf_project_id)
      # TF also sends this request when Orchestrate is added to the project, but before sending a POST request.
      # Hence we might receive a request for a specific project before we actually created it inside Orchestrate.
      @tracker_source = project_config.try(:tracker_source)
      @rabbit_uri = rabbit_uri
    end
    render 'configuration_parameters'
  end

  # Handles POST ProjectConfiguration
  def create_project_configuration
    config_params = params["CreateProjectConfigurationRequest"]["Configuration"]
    # when added to the CTF site:
    # {"projectId"=>"proj1001", "projectName"=>"Site", "projectURL"=>".", "userId"=>"user1003", "template"=>"false"}
    # when added to a project:
    # {"projectId"=>"proj1008", "projectName"=>"Brokerage System (Sample)", "projectURL"=>"brokerage_system_sample", "userId"=>"user1003", "template"=>"false"}
    config_hash = Hash[config_params.map do |k, v|
      k = "ctf_project_id" if k == "projectId"
      [k.underscore, v]
    end]
    return head :ok if config_hash['project_url'] == '.'
    existing_configuration = TeamForge::ProjectConfiguration.find_by_ctf_project_id(config_hash['ctf_project_id'])
    if existing_configuration
      existing_configuration.update_attributes(active: true)
      Resque.enqueue_in(60.seconds, OrchestrateSyncListenerService, existing_configuration.ctf_project_id)
    else
      project = Project.find_or_initialize_by_name(config_hash["project_name"])
      # do this since in MongoMapper project.project_configuration = some_proj_configuration will try to save both the parent and the child on the fly
      project.build_project_configuration(config_hash)
      if project.save && project.project_configuration.save
        Resque.enqueue_in(60.seconds, OrchestrateSyncListenerService, project.project_configuration.ctf_project_id)
      else
        logger.error("Could not create a project and project_configuration from the following parameters #{config_params}")
      end
    end
    head :ok
  end

  # Handles PUT ProjectConfiguration
  def update_project_configuration
    head :ok
  end

  # Handles DELETE ProjectConfiguration
  def delete_project_configuration
    # After deleting from the site when Orchestrate is added as a project to proj1008:
    # {"format"=>:xml, "controller"=>"services/team_forge", "action"=>"get_configuration_parameters", "token"=>"2FF9FB4F39726A26EAFD6F0C8EAD1B33", "tf_project_id"=>"null"}
    # {"format"=>:xml, "controller"=>"services/team_forge", "action"=>"get_configuration_parameters", "token"=>"3DC668F69D081EABBCC1A4622777D487", "tf_project_id"=>"proj1008"}
    # {"format"=>:xml, "controller"=>"services/team_forge", "action"=>"delete_project_configuration", "token"=>"B66C4F0A8802E1B8ED371599BA669CB2", "tf_project_id"=>"proj1008"}
    # {"format"=>:xml, "controller"=>"services/team_forge", "action"=>"get_configuration_parameters", "token"=>"FEA6086A33933206F53D9D073E660638", "tf_project_id"=>"proj1001"}
    # {"format"=>:xml, "controller"=>"services/team_forge", "action"=>"delete_project_configuration", "token"=>"B4CA4D60A978DDF3792FB76B16A3FB55", "tf_project_id"=>"proj1001"}
    # we need to mark the configuration as inactive and clear the link_plug_id. If Orchestrate is added back later it will have a different plug_id.
    # Since we don't get this piece of data in the POST ProjectConfiguration request and query for link_plug_id on the first request
    # (see TeamForge::ProjectConfiguration.get_link_plug_id) we need to set link_plug_id to nil, so that CTF gets queried
    # for that.
    TeamForge::ProjectConfiguration.where(ctf_project_id: params[:tf_project_id]).each { |conf| conf.update_attributes(active: false, link_plug_id: nil) }
    head :ok
  end

  # Gets the CTF project ID for the given ID (params[:object_id]). Handles GET ProjectId.
  # Raises errors if the activity or ctf project id is not found.
  def ctf_project_id
    activity = ActivitySummary.first(opaque_id: params[:object_id])
    raise MongoMapper::DocumentNotFound.new(t("errors.activity.not_found")) unless activity
    project = activity.source.project
    @ctf_project_id = project.project_configuration.ctf_project_id unless (project.nil? || project.project_configuration.nil?)
    raise MongoMapper::DocumentNotFound.new(t("errors.activity.not_found")) if @ctf_project_id.nil?
  end

  # Gets the title for the object identified by the given ID (params[:object_id]). Handles GET Title.
  def title
    activity = ActivitySummary.first(opaque_id: params[:object_id])
    raise MongoMapper::DocumentNotFound.new(t("errors.activity.not_found")) unless activity
    @title = activity.ctf_description
  end

  # pre-commit notification by TeamForge
  def pre_commit_message
    head :ok
  end

  # pre-commit notification by TeamForge, renders a default template (XML that OKs the commit)
  def detailed_precommit_message
  end

  # post-commit notification
  def post_commit_message
    head :ok
  end

  # post-commit notification
  def detailed_post_commit_message
    PostCommitMessageHandler.new(params["DetailedSCMPostCommitRequest"]).process_message
    head :ok
  end

  # ctf search
  def search
    head :ok
  end

  private

  # Logs the parameters given by TeamForge with a prefix to make discovery and debugging easier.
  def debug
    logger.debug "TeamForge params: #{params}"
  end

  # verify that the request came from TeamForge. Use a provided one-time token to make a SOAP call. Raises TeamForge::ServiceError
  # if the call can't be processed
  def authenticate
    token = params[:token]
    session = CtfService.create_teamforge_session
    session.teamforge.get_session_id(session_id: token)
  end

  # VT not sure if and what we need to log something here (this is called when we can't verify that the request came from TeamForge)
  def not_found(exception)
    head :not_found
  end

  # fetches rabbit uri from the configuration and strips username and password from it
  def rabbit_uri
    uri = URI.parse(Rails.configuration.local_settings.load('rabbitmq')['uri'])
    uri.password = uri.user = nil
    sanitize_uri(uri)
    uri.to_s

  end

  # if the hostname is localhost replace with the orchestrate hostname.
  def sanitize_uri(uri)
    forge_config = Rails.configuration.local_settings.load('teamforge')
    orc_url = forge_config && forge_config['iaf'] && forge_config['iaf']['orc_url']
    if (uri.host == 'localhost' || uri.host == '127.0.0.1') && orc_url
      begin
        forge_uri = URI.parse(orc_url)
        uri.host = forge_uri.host
      rescue URI::InvalidURIError => uri_error
        logger.warn "Teamforge orchestrate URL is invalid"
      end
    end
  end


  add_method_tracer :debug
  add_method_tracer :authenticate
end
