#
# Copyright 2016 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

require 'lexi'

module ApplicationHelper
  include UriHelper

  DATE_TIME_FORMATS = {
      basic_time: '%H:%M',
      time_with_tz: '%-I:%0M %p %Z',
      basic_date: '%-m/%-d/%0y',
      date_padded: '%m/%d/%Y',
      date_time_with_tz: '%-I:%0M %p %Z %-m/%0d/%Y',
      date_long: '%^B %d, %Y'
  }

  # Constructs a page title from the host app name and subtitle.
  # @param [String] page subtitle
  # @return [String] the full title of the page
  def page_title(subtitle = nil)
    t('host_app_name') + # e.g. TeamForge
        (subtitle ? " : #{subtitle}" : '') + # add subtitle if provided
        (Rails.env.production? ? '' : " (#{Rails.env})") # add environment if not production
  end

  # Converts the given time to a common format.
  # using 24-hour time.
  # Takes the current_user.time_zone value
  #   and if it is valid TimeZone displays accordingly
  #   if the zone cannot be found, it uses getlocal
  # @param [Time] at the time to convert
  # @param [String] format the format to use
  # @return [String] a formatted time
  def common_date_time(at, format = :basic_time)
    tz = get_current_user_time_zone if current_user

    if Time.find_zone(tz)
      at.in_time_zone(tz).strftime DATE_TIME_FORMATS[format]
    else
      at.getlocal.strftime DATE_TIME_FORMATS[format]
    end
  end

  def get_current_user_time_zone
    current_user.time_zone
  end

  # Converts the given date to a common format.
  # @param [Date] date to convert
  # @param [String] format to use
  # @return [String] a formatted date, or nil if none provided
  def common_date(date, format = :basic_date)
    date ? date.strftime(DATE_TIME_FORMATS[format]) : nil
  end

  # Formatting for whitespace characters
  # Generate safe HTML for formatted text by html escaping the contents and then replacing
  # linefeeds with breaks and/or paragraphs
  def format_newlines(text)
    simple_format(h(text))
  end

  # Generate will_paginate toolbar for passed collection
  # @param [Array] paginated collection
  # @param [options] overrride default arguments to will_paginate
  # @return [String] pagination HTML or nil if no pagination
  def pagination_for(paginated_collection, options = {})
    left_arrow = '&#8592;'
    right_arrow = '&#8594;'
    defaults = { outer_window: 0, inner_window: 2, previous_label: left_arrow, next_label: right_arrow }
    will_paginate(paginated_collection, defaults.update(options))
  end

  # Creates the link to the CTF project header for the given project.
  #
  # The linkId parameter is required in order for the scrolling within drop down menus to be activated.
  # Supply a bogus linkId here when there is no project.
  #
  # @param [Hash] project the project information to create the header link for
  # @return [String] the URL
  def ctf_header_url(project)
    ctf = Rails.configuration.teamforge
    server_url = ctf[:server]
    slashless_server_url = normalize_repository_uri server_url
    if project && project.active?
      project_url = project.project_configuration.project_url
      link_id = project.project_configuration.link_plug_id
      URI.join(server_url, "/sf/sfmain/do/topInclude/projects.#{project_url}?linkId=#{link_id}&base=#{slashless_server_url}").to_s
    else
      URI.join(server_url, "/sf/sfmain/do/topInclude?linkId=noProject&base=#{slashless_server_url}").to_s
    end
  end

  # Creates the hash with the CTF project header info for the given project.
  #
  # @param [Hash] project the project information to create the header link for
  # @return [Hash] the hash with project information
  def ctf_header_info(project)
    ctf_info = {}
    ctf = Rails.configuration.teamforge
    provided_host = ctf[:server]
    ctf_host = (provided_host.end_with?('/')) ? provided_host[0,provided_host.size - 1] : provided_host
    ctf_info[:ctf_host] = ctf_host
    if project && project.active?
      ctf_info[:project_url] = project.project_configuration.project_url
      ctf_info[:iaf_app_id] = project.project_configuration.link_plug_id
    end
    ctf_info
  end

  # Creates a link to the Details page for an activity
  # @param [Activity] activity the activity to link to
  # @param [Hash] opts title and other options for the link
  # @param [Block] block Optionally provides the content of the <a> tag.
  # @return an anchor
  def link_to_activity(activity, opts = {}, &block)
    opts = opts.dup
    opts[:href] = activity_path(activity) unless opts.has_key? :href
    unless opts.has_key? :title
      opts[:title] = activity_display_name(activity)
      opts[:title] += "\n#{activity.title}" if activity.respond_to? :title
    end
    if block_given?
      content_tag(:a, opts, &block)
    else
      content_tag(:a, opts[:title], opts)
    end
  end

  # Custom path resolution for activities
  #
  # @return [String] path or url for the activity, or nil if there is none
  def activity_path(activity)
    if activity.respond_to? :target_url
      activity.target_url
    elsif activity.settled?
      polymorphic_path(activity, project: activity.source.project || current_project)
    end
  end

  # Creates a link to the edit page of a ProjectActivitySource
  # @param [Hash] title and other options for the link
  # @return an anchor
  def link_to_edit_source_from_project(source, opts = {}, &block)
    path_opts = path_options_with_project_context(project: source.project)
    opts[:href] = edit_polymorphic_path(source, path_opts)

    if block_given?
      content_tag(:a, opts, &block)
    else
      text = opts[:title] || source.display_name
      content_tag(:a, text, opts)
    end
  end

  # Creates a link to the new page for the source of the given type
  # @param project sources overview to get a source to link to
  # @param [Hash] title and other options for the link
  # @return an anchor
  def link_to_new_source_from_project(source_type, opts = {}, &block)
    path_opts = path_options_with_project_context(project: current_project)
    begin
      opts[:href] = new_polymorphic_path(source_type, path_opts)
    rescue => e
      return "<span class='unobtrusive'>#{t('project.errors.no_sources')}</span>".html_safe
    end
    if block_given?
      content_tag(:a, opts, &block)
    else
      content_tag(:a, opts[:title], opts)
    end
  end

  def link_to_sources_from_project(source_type, opts = {}, &block)
    begin
      opts[:href] = polymorphic_path(source_type.pluralize)
    rescue => e
      return "<span class='unobtrusive'>#{t('project.errors.no_sources')}</span>".html_safe
    end
    if block_given?
      content_tag(:a, opts, &block)
    else
      content_tag(:a, opts[:title], opts)
    end
  end

  def cache_if(condition, activity, &blk)
    if condition
      cache(activity, &blk)
    else
      yield
    end

    nil
  end

  def image_url(source)
    "#{root_url}#{image_path(source).gsub(/^\//, '')}"
  end

  # Return HTML string for the avatar for this activity.  Encapsulates hierarchy
  # of preferences for avatars do display.
  #
  # @param [Activity|ActivitySummary] activity to evaluate
  # @return [String] HTML string for the image.
  def avatar_for_activity(activity)
    if activity.is_a?(Activity)|| activity.is_a?(ActivitySummary)
      if activity.try(:updated_by)
        user_image(activity.updated_by)
      elsif activity.is_first? && activity.try(:created_by)
        user_image(activity.created_by)
      else
        (activity.is_a?(Custom)|| activity.is_a?(CustomSummary)) ? user_image('XDS') : user_image(activity._type.to_s)
      end
    else
      user_image('_') # Shows the profile-silhouette image
    end
  end

  # Return HTML fragment containing image tag for supplied user.
  #
  # @param [User|String] user User or username to render
  # @return [String] the html
  def user_image(user)
    user = (user.blank?) ? '_unknown' : user
    render partial: "shared/ctf_user_image", locals: { ctf_avatar: ctf_avatar_url(user) }
  end

  # Calculate the avatar url for a given user
  # @param [String] avatar_param
  # @return [String] the CTF avatar url for the user's profile
  def ctf_avatar_url(avatar_param)

    service = 'avatar'
    size = '64'

    # Remove all non-word characters to avoid incorrect initial to be passed for avatar URL
    parsed_avatar_param = avatar_param.gsub(/\W/, '')
    initial = (parsed_avatar_param.blank?) ? '_' : parsed_avatar_param.first.downcase

    # Remove all spaces and convert to lowercase before calculating MD5 hash
    md5_hash = Digest::MD5.hexdigest(avatar_param.gsub(/\s/, '').downcase)

    URI.join(Rails.configuration.teamforge[:server], (service + '/' + size + '/' + initial + '/' + md5_hash)).to_s
  end

  def crud_health_status_text(access_status)
    if access_status.last_access_status == :success
      I18n.t('access_status.successful')
    else
      I18n.t('access_status.failed')
    end
  end

  def crud_health_status_icon(access_status)
    if access_status.last_access_status == :success
      'icons-log-status-success'
    else
      'icons-log-status-failure'
    end
  end

  def crud_health_status_time(access_status)
    if access_status.last_access_status == :success
      access_status.last_success
    else
      access_status.last_error
    end
  end

  ICON_MAP = {
      BuildServer => 'build',
      ReviewServer => 'review-request',
      WorkItemSource => 'work-item',
      CtfRepository => 'commit',
      ExternalRepository => 'commit',
      CustomActivitySource => 'custom'
  }

  def source_icon(source)
    ICON_MAP[source.class]
  end

  # Returns text denoting the icon for the repository type.
  # @param [CommitServer] commit_server the repository
  # @return [String] text for the repository icon
  def repository_icon(commit_server)
    "icons-commit"
  end

  # Get the CSS classes for this activity
  #
  # @param [Activity] activity the activity
  # @return [String] value used for css class names for the activity's step
  def activity_css_classes(activity)
    "#{activity_css_class_name(activity)} #{ activity.settled? ? 'clickable' : 'disabled'}"
  end

  # Get the step type with dashes for use in CSS class names, selectors
  #
  # @param [Activity] activity the activity
  # @return [String] value used for css class names for the activity's step
  def activity_css_class_name(activity)
    activity.class.to_s.sub('Summary', '').underscore.dasherize
  end

  # Get the status icon class for a given status
  #
  # @param [Status] status to render
  # @param [Symbol] size :small (default) or :normal
  # @return [String] the css classes to apply for activity status icons
  def activity_status_css_class(status, size=:small)
    "icons-status-#{status}#{ '-small' if size == :small}"
  end

  # Renders the "display number" for an activity, optionally
  # using the full git revision number
  # @param [Activity] the activity whose display_number to render
  # @param [Boolean] display_long_number yes or no
  # @return [String] the display number
  def activity_display_number(activity, display_long_number=false)
    (display_long_number && activity.respond_to?(:revision_number)) ? activity.revision_number : activity.display_number
  end

  # Return the display name for the activity.  Tooltip will include the full revision number for commits.
  #
  # @param [Activity] activity the activity
  # @param [Boolean] source_first - (optional, default: false) Display the name of the activity as <source name> - <activity ID>
  # @return [String] the display name
  # TODO split this method into two (for heading and tooltip)
  def activity_display_name(activity, source_first=false)
    return activity.display_name if activity.respond_to?(:display_name)

    translation_key = 'activity'
    translation_key += '.phantom' unless activity.settled?
    translation_key += source_first ? '.heading' : '.tooltip'

    activity_number = source_first ? activity.display_number : activity_display_number(activity, true)
    I18n.t(translation_key, number: activity_number, name: activity.source.display_name)
  end


  # CSS class to add to the activity-comment-count display.  'hidden' if no
  # comments.
  # @param [Integer] the comment count
  # @return [String] the CSS class to apply
  def comment_count_css_class(count)
    "activity-comment-count icons-speech-bubble #{'hidden' unless count > 0}"
  end

  # Returns a count of comments on an activity, up to a threshold N where
  # it returns "N+".
  # @param [Integer] the comment count
  # @param [Integer] threshold of "a lot".
  # @return [String] comment count to display
  def activity_comment_indicator(count, threshold=99)
    (count > threshold) ? t('comments.count_over', number: threshold) : t('comments.count', number: count)
  end

  # # Return hash with project context added if context is set. The hash is compatible with namned_route method parameter
  # #
  # # @param [Hash] path_opts  the path options that you would path in the named route helper (like new_pipeline_path)
  # # @return [Hash] the options with a project merged if a context is set (but ignores the current pipeline project)
  # def path_options_with_project_context(path_opts)
  #   path_opts.merge!({project: current_project.project_configuration.project_url}) if current_project(use_pipeline_project: false)
  #   path_opts
  # end

  # Return hash with project context added if context is set. The hash is compatible with named_route method parameter
  #
  # @return [Hash] the options with a project merged if a context is set
  def path_options_with_project_context(path_opts)
    path_opts.merge!({ project: current_project.project_configuration.project_url }) if current_project
    path_opts
  end

  CONTENT_PATTERNS = { CtfGoLinks::ANCHORED_REGEX => :ctf,
                        CtfGoLinks::ALL_REGEX => :ctf,
                        ActionView::Helpers::TextHelper::AUTO_LINK_RE => :uri,
                        ActionView::Helpers::TextHelper::AUTO_EMAIL_RE => :uri,
                        /\r?\n/ => :newline
                     }

  # runs regex on text and replaces captures with links where href=url+capture
  # the rest of the text is HTML escaped and the whole output is marked as html safe (so that erb does not escape it again)
  # @param [String] text
  # @param [Regexp]regex, defaults to the TeamForge's regular expression
  # @param [String] url, defaults to the TeamForge's go link base url
  def linkify(text, url = CtfGoLinks.base_go_url)
    result = ""
    text = interpolate(text)
    start_point = 0
    Lexi.new(text, CONTENT_PATTERNS).parse.each do | match |
      end_point = start_point + match.from-1
      result += html_escape(text[start_point..end_point]) if match.from > 0
      start_point += match.to
      result += case match.key
                  when :ctf
                    go_url = URI.join(url, match.match_data[0]).to_s
                    content_tag(:a, match.match_data[0].to_s, href: go_url)
                  when :newline
                    "<br>"
                  when :uri
                    auto_link(match.match_data[0], :html => { :target => '_blank' })
                  else
                    match.match_data[0].to_s
                end
    end
    result += html_escape(text[start_point..text.size-1])
    result.html_safe
  end

  # Interpolate ctf tokens in text and mark html safe.
  #
  # @param [String] text The string to interpolate
  # @return [String] Transformed string, marked safe
  def interpolate(text)
    CtfInterpolations.interpolate_ctf_tokens(text).html_safe
  end

  def best_key
    if current_project
      current_project.to_param
    else
      "orc"
    end
  end

  def ui_friendly_timestamp(activity)
    t('activity.timestamp',
      date_created: common_date_time(activity.time_created, :date_padded),
      time_created: common_date_time(activity.time_created, :time_with_tz))
  end

  # model attributes that signal that we need to display a status icon
  STATUS_ATTRIBUTES = { Custom => [:status_type, :status_name], WorkItem => [:deleted, :closed, :status] }


  # decide if we need to display a status icon in the info bubble.
  # currently used for work_items and custom activities
  def display_status_icon_in_activity_stream?(activity)
    activity.is_first? || (activity.attribute_changes & STATUS_ATTRIBUTES[activity.class]).present?
  end

  # creates an updated attribute-change list for use in Event History view
  # @param [Activity] activity, currently used for ReviewRequest. Custom activities have their own method.
  # @return [Array<Symbol>] array of attribute symbols changed in this instance
  def activity_attribute_change_events(activity)
    activity.attribute_changes.tap do |change_list|
      change_list.unshift(change_list.delete(:status)) if change_list.include? :status
    end
  end

  def source_type_snake_case(source_type)
    source_type.to_s.underscore
  end

  def source_type_css_classes(source_type)
    source_type_snake_case(source_type).dasherize
  end

end
