#
# Copyright 2013 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

# Helpers for builds.
module BuildHelper

  # maps the build statuses to their css class representation
  TEST_TYPE_CLASSES = %w(passed failed ignored)

  # Converts seconds to hours:minutes:seconds format.
  # @param [integer] duration the number of seconds to convert
  # @return the formatted string
  def duration_in_hours(duration)
    hours = duration / 1.hour
    minutes = (duration % 1.hour) / 1.minute
    seconds = duration % 1.minute
    sprintf("%02d:%02d:%02d", hours, minutes, seconds)
  end

  # Returns the css class which indicates the build's status.
  # @param [Build] build the build
  # @return [String] the name of the css class to indicate the build's status
  def build_status_class(build)
    "build-#{build_status(build)}"
  end

  # Formats a duration provided in seconds in precise english terms
  # @param [integer] delta The difference in time in seconds to translate to text
  def duration(delta)
    return "0 seconds" if delta.to_i == 0

    places = [
        [ :day, delta / 1.day ],
        [ :hour, (delta % 1.day) / 1.hour ],
        [ :minute, (delta % 1.hour) / 1.minute ],
        [ :second, delta % 1.minute ]
    ]

    # TODO: (JDM) "5 minutes" probably should be "5&nbsp;minutes" for most uses of this method.
    # Probably have to gsub the response from pluralize() to accomplish that though.
    trimmed = places.reject {|h| h.last == 0}
    trimmed.to_a.map { |key, value| pluralize(value, key.to_s)}.join(", ")
  end

  # Class for the test type and count
  # @param [String] status the status name
  # @param [Integer] the count of tests in this status
  def status_class(status, status_count)
    count = status_count.to_i
    "status-type #{status} #{count > 0 ? 'active' : 'inactive'}"
  end

  # Class for the appropriate icon for the test type and count
  # @param [String] status the status name
  # @param [Integer] the count of tests in this status
  def status_icon(status, status_count)
    count = status_count.to_i
    "icons-status-#{status}-small#{count > 0 ? '' : '-inactive'}"
  end

  # Localizable text for the build status summary
  # @param [Build] build to display the status and time for
  # @return [String] containing text similar to "Build successful in X minutes and Y seconds."
  def status_line(build)
    text_key = 'build.message.duration'
    text_key = 'build.message.no_duration' if build.duration.to_i == 0
    t(text_key, status: friendly_status(build), time: duration(build.duration))
  end

  # Creates a div tag for displaying the test count for the given status
  # @param [Build] build the build to link to
  # @return [String] the string to display for the build status type
  def friendly_status(build)
    t("build.status.#{build_status(build)}")
  end

  private

  def build_status(build)
    if Build::VALID_STATUS_TYPES.include? build.status_type
      build.status_type.downcase
    else
      'unknown'
    end
  end
end
