#
# Copyright 2014 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

module CustomHelper

  # creates an updated attribute-change list for use in Event History view
  # Also hides changes to hidden fields from the History view
  # @param [Activity] custom custom activity
  # @return [Array<Symbol>] array of attribute symbols changed in this instance
  def custom_activity_attribute_change_events(custom)
    custom.attribute_changes.tap do |change_list|

      # do not show changes to :status_type and :associated_remote_id
      [:status_type, :associated_remote_id].each {|unwanted| change_list.delete(unwanted)}

      # move :status to the front of the list if present
      change_list.unshift(change_list.delete(:status_name)) if change_list.include? :status_name
    end
    .delete_if {|hidden_field| custom.custom_schema.hidden.member? hidden_field.to_s }
  end

  # Label all changes to attributes
  # @param [Activity] activity the activity
  # @param [Symbol] attribute the attribute of the activity
  # @return [String] The display
  def custom_activity_changed_attribute_description(activity, attribute)
    case attribute
      when :is_created then t('activity.created')
      else
        value = activity.fetch_attribute attribute
        if value
          t('activity.attribute_updated_to', attribute: attribute.to_s.humanize, value: value)
        else
          t('activity.attribute_removed', attribute: attribute.to_s.humanize)
        end
    end
  end

  def displayed_important_changes(activity)
    activity.important_changes - [:status_name, :status_type] # we are already going to display an icon and status name, no need to display 'Status name changed to foo'
  end

  # Adds created_by field, if present, to the list of key-value pairs for display
  # Also excludes hidden fields
  def displayed_custom_data(activity)
    Hash.new.tap do |h|
      h['created_by'] = activity.created_by if activity.created_by.present?
    end
    .merge(activity.custom_data)
    .delete_if do |k, v|
        activity.custom_schema.hidden.member? k
    end
  end

  # Linkify the given value if its key is included in the given activity's parsed set of fields.
  # @param activity the activity to look up parsed fields
  # @param key the key
  # @param value the value
  def linkify_custom_field(activity, key, value)
    activity.custom_schema.parsed.include?(key) ? linkify(value) : value
  end

  # Allow either a scalar (assume it's a link) or an array of upto two elements
  # If array, one is a URI and the other its anchor. Determine which
  # intelligently failing which assume first link, second the anchor
  def linkify_custom_linkto(link_to)
    href, anchor = nil, nil
    if link_to.kind_of? Array
      elems = link_to[0, 2].map do |e|
        val = interpolate(e);
        URI(val) rescue val
      end
      elems.reverse! unless elems[0].kind_of? URI
      href, anchor = elems.map(&:to_s)
    else
      href = interpolate(link_to)
    end
    anchor ||= t('custom.link') # Default anchor if not given
    link_to anchor, href, target: '_blank', class: 'link-back'
  end

  def custom_activity_history_icon_class(custom, attribute_change, size=:small)
    show_status = ([:status_name].include? attribute_change)
    "icons-status-#{(show_status) ? custom.status_icon_class : 'bullet'}#{ "-small" if size == :small}"
  end
end
