#
# Copyright 2013 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

module ReviewRequestHelper

  # Links to the original review request
  # @param [ReviewRequest] review_request the review request
  # @return the link to the original review request on the review server
  def link_to_original_review_request(review_request, &block)
    raise "You must provide a block" unless block_given?
    link_to review_request.link, target: "_blank", &block
  end

  # Links to a user on a review server
  # @param [Hash] user_hash the user to link to
  # @return the link to the user on the review's server
  def link_to_review_user(user_hash)
    present = user_hash['link'].present?
    var = user_hash['username']
    var2 = user_hash['link']

    link_to_if(present, var, var2, target: "_blank")
  end

  # Translates the review status_name into a string
  # @param [ReviewRequest] review the review request
  # @return [String] the string to display for the review status type
  def review_status_string(review)
    review.status_name
  end

  # Return an english past-tense term describing the action taken for this review
  # @param [ReviewRequest] review the review request
  def review_action_verb(review)
    review.is_first? ? t('review_request.created_by') : t('review_request.updated_by')
  end

  # Creates the summary string for a review request header including a date and time
  #
  # @param [ReviewRequest] review the review request
  # @return [String] the string to display as the header for the review comment
  def review_summary_header(review)
    t('review_request.summary_header',
      action_verb: review_action_verb(review),
      acting_user: (!review.is_first? && review.updated_by)? review.updated_by : review.created_by,
      timestamp: ui_friendly_timestamp(review))
  end

  # Summarize the changes between this review request, and its most recent ancestor
  # into a single sentence.
  # @param [ReviewRequest] review the review request
  # @param [Array] the attribute symbols to characterize (all by default, but this
  # can be used for iterating changes to produce multi-line descriptions)
  # @return the string to display as the description for the review comment
  def review_changed_description(review, attribute_changes = nil)
    attribute_changes = review.attribute_changes unless attribute_changes != nil
    new_reviewers = review.new_reviewers
    if review.is_first?
      t('review_request.review_created', reviewer: review.created_by)
    elsif new_reviewers.size > 0
      t('review_request.reviewer_added', reviewer: new_reviewers.to_sentence, count: new_reviewers.size)
    else
      t('review_request.attributes_updated',
        attributes_changed: attribute_changes.to_sentence.humanize)
    end
  end

  # Creates string containing classes for review history status icon, or a simple "bullet", depending on the change
  # event being rendered
  # @param [ReviewRequest] review the review request
  # @param [Symbol] the attribute change in question
  def review_history_icon_class(review, attribute_change, size=:small)
     show_status = attribute_change == :status
    "icons-status-#{(show_status) ? review.status_icon_class : 'bullet'}#{ "-small" if size == :small}"
  end

  # Creates string containing classes for review status
  # @param [ReviewRequest] review the review request
  def review_status_class(review)
    "review-#{review.status_type} review-status"
  end
end
