#
# Copyright 2013 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

# Helper module
module UriHelper
  def self.included(base)
    base.send(:extend, UriHelper::ClassMethods)
  end

  module ClassMethods
    def uri_extractor(uri_field)
      [:host, :port, :path, :ssl?, :host_parts].each do |method|
        define_method method do
          UriUtils.new(send(uri_field)).send(method)
        end
      end
    end

    # Local and self referential icon paths intended to be served from orchestrate should be normalized to match
    # the application expectations of paths starting relative to the rails root with /assets/...
    # Non-local paths are not modified, which would represent any icon served outside of orchestrate
    # @param [String, Boolean] icon_uri, non_rails_display to modify
    # @return [String] uri modified to be a local rails root path if necessary

    # NOTE: non_rails_display params will be used to determine whether the result should be returned
    # as proxied url so non_rails image link can find the image correctly.
    def normalize_orchestrate_icon_uri(icon_uri, non_rails_display=false)
      URI.parse(icon_uri).tap { |uri|
        if (uri.to_s.starts_with? Rails.configuration.teamforge[:iaf][:orc_url]) # self-referential uri
          uri.scheme = uri.host = uri.port = nil # converted to local path
        end

        if (!uri.scheme && !uri.host) # A local path
          path = '/assets' # default assets path

          if non_rails_display
            # if relative url root exist?
            if orc_root = Rails.application.config.try(:relative_url_root)
              path = "#{orc_root}" + path
            end
          end
          uri.path = uri.path.gsub(/^.*\/assets/, path) #normalized to serve icons from app location
        end
      }.to_s
    rescue
      # This is a pre-validation attempt to normalize uris, we will fail silently and return the original value
      # on any exception so that we can let the upstream validators catch and report them appropriately
      # Example exceptions include: URI parse failure or failure to fetch the orc_url value
      icon_uri
    end
  end

  # Normalize a repository URI so it can be compared to other URIs
  #
  # Strips user name and password, params, fragments and trailing slashes so that prefix matchers can work.
  # Invalid URIs a going to be returned unchanged.
  # @param [String] uri_string to modify
  # @return [String] uri minus the user:password@ bits, and trailing slash
  def normalize_repository_uri(uri_string)
    return "" if uri_string.blank?
    begin
      URI(uri_string).tap do | stripped |
        stripped.path = stripped.path.gsub(/\/+\Z/, '')
        stripped.user = nil
        stripped.password = nil
        stripped.query = nil
        stripped.fragment = nil
      end.to_s
    rescue URI::InvalidURIError
      uri_string
    end
  end


end