#
# Copyright 2013 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

# Common Activity functionality.
class Activity
  include MongoMapper::Document

  key :created_by, String
  key :time_created, Time
  key :atomic, Integer
  key :displayable, Boolean, default: true
  belongs_to :source, class: ActivitySource
 
  belongs_to :activity_summary
  delegate :opaque_id, :to => :activity_summary
  delegate :comments, :to => :activity_summary

  attr_accessible *(keys.keys.map(&:to_sym) + [:comments])

  before_create :set_atomic

  # An activity is always settled. So to be able to use Activity and ActivitySummary with similar API
  # we enforce here the call to return true so we can transparently pass either one to a helper or view
  # @return |Boolean| always true!
  def settled?
    true
  end

  # return the activity summary commentable since comments are stored in the summary
  # 
  # @returns [Commmentable] the activity summary commentable object
  def commentable
    activity_summary
  end

  # Returns a status string for any instance which has a :status attribute
  # This string is used to construct the appropriate CSS class for the status icon
  #
  # @return [String] the lower-case status if the class supports it, or "none"
  def status_icon_class
    #TODO CM: This would be a good candidate for a presenter pattern,
    # It is a little strange to be calling the model for CSS class specifics
    if self.respond_to?(:status_type) && !self.status_type.nil?
      self.status_type.to_s.downcase
    else
      "none"
    end
  end

  # Used by the Rails routing code to generate a URL pointing to this object. Delegating back to the
  # activity summary opaque_id
  #
  # @return [String] opaque id for this object
  def to_param
    activity_summary.opaque_id
  end

  # Used to determine if this activity is first in its sequence.  Compare with same
  # method in ActivitySummary and ChangeHistory
  #
  # TODO: Unify this method with the same in ChangeHistory!  This method is needed
  # here because UI methods that don't distinguish between Activity and ActivitySummary
  # have a need to be able to make uniform calls to such a method, but ChangeHistory
  # provides an implementation of is_first? which is suitable to some (but not all)
  # Activity types.
  #
  # @return [Boolean]
  def is_first?
    self == activity_summary.activities.first(order: :time_created.asc)
  end

  private

  # Associate this activity with a an 'activity source', and add some useful methods.
  #
  # @param [Symbol] parent_source Where this activity type comes from.
  def self.activity_source(parent_source)
    parent_class = parent_source.to_s.camelize.constantize

    define_singleton_method :source_type do
      parent_class
    end
  end

  def set_atomic
    activity_clock_counter = Counters.first_or_create(name: "AtomicActivityClock", count: 0)
    self.atomic = activity_clock_counter.next
  end
end
