#
# Copyright 2013 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

class ActivityFilter::Base

  include ActiveModel::Validations
  include ActiveModel::Conversion
  extend ActiveModel::Naming
    
  validate :parameter_name_validations, :date_validations
  
  # Inflection of the last part of the class name
  # Used to distinguish between filter types in the view
  # @return [String] underscore-style string of class name, stripping module prefixes
  def type
    self.class.name.demodulize.underscore
  end
  
  # Constructor for non-persistent activity filter model
  # Assigns each passed attribute to its writer method and records bad parameters for validation phase
  # @constructor
  # @param [Hash] attributes to assign to writer methods
  def initialize(attributes = nil)
    if attributes
      attributes.each do |name, value|
        begin
          send("#{name}=", value)
        rescue NoMethodError => e
          (@invalid_parameter_names ||= []) << name.to_s
        end
      end
    end
  end
  
  # Process string date argument into Date object
  # @param [String] string to be parsed
  # @return [Date or Symbol] date object, :invalid, or nil if input string is empty
  def parse_date(date_string)
    return nil unless date_string.present?
    begin
      date = Date.strptime(date_string, I18n.t('activity_filter.form.date.strptime_format'))
      date += 2000.years if date < Date.new(100, 1, 1)
      date
    rescue StandardError => e
      :invalid
    end
  end

  # Attribute writer for start date
  # @param [String] date string to be parsed
  def start=(date)
    @start = parse_date(date)
  end
  
  # Attribute writer for end date
  # @param [String] date string to be parsed
  def end=(date)
    @end = parse_date(date)
  end
  
  # Attribute reader for start date
  # @return [Date] date or nil if empty or has error condition
  def start
    (@start == :invalid) ? nil : @start
  end
  
  # Attribute reader for start date
  # @return [Date] date or nil if empty or has error condition
  def end
    if (@end == :invalid) || (@end && start && (@end < @start))
      nil
    else
      @end
    end
  end

  # Date field validations
  def date_validations
    valid = true
    
    if @start == :invalid
      valid = false
      errors.add(:start, :invalid)
    end  
    
    if @end == :invalid
      valid = false
      errors.add(:end, :invalid)
    end
    
    # check for (valid) end date before start date
    if valid && @start && @end && (@start > @end)
      errors.add(:end, :before_start_date)
    end
  end
  
  # Parameter name validations
  # Reports failed parameter assignments to validator
  def parameter_name_validations
    if @invalid_parameter_names.present?      
      errors.add(:base, :invalid_parameter_names, names: @invalid_parameter_names.join(', '))
    end
  end
  
  # Returns model validation error messages
  # @return [String] error messages on multiple lines, separated by line breaks
  def error_messages
    errors.messages.values.flatten.join('<br>').html_safe
  end
  
  # Flag to ActiveModel that this is not a database model
  def persisted?
    false
  end
  
end
