#
# Copyright 2013 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

# Base class for all activity sources.

class ActivitySource
  include MongoMapper::Document

  include ::NewRelic::Agent::MethodTracer

  MAX_NAME_LENGTH = 100

  key :active, Boolean, default: true
  key :source_association_key, String, default: -> { UUID.generate }
  validates :source_association_key, uniqueness: true
  key :trigger_status, Hash
  key :queue_username, String
  key :queue_password, String
  key :display_name, String
  one :access_status

  many :activity_summaries, foreign_key: :source_id
  belongs_to :project
  belongs_to :tool
  timestamps!
  SOURCE_TYPES = ['ExternalRepository','CtfRepository', 'WorkItemSource', 'ReviewServer', 'BuildServer', 'CustomActivitySource', 'CommitServer' ]

  before_validation :strip_whitespace

  validates_presence_of :display_name
  validates_presence_of :project
  validates_length_of :display_name, maximum: MAX_NAME_LENGTH
  validates :active, inclusion: {in: [true, false]}
  validate :presence_of_valid_type
  validate :validate_existence_of_tool
  validate :validate_existence_of_project

  scope :types, ->(type) { where(:_type => type) }
  scope :by_type, ->(type) { where(:_type => type) }
  scope :by_project, ->(project_id) {where( project_id: project_id) }
  scope :by_sak, ->(sak) {where( source_association_key: sak) }
  scope :by_tool, ->(tool_id) {where( tool_id: tool_id ) }

  def self.permitted_attrs
    ['display_name', 'active', 'project_id', 'tool_id']
  end

  def self.nativized_keys
    {}
  end

  def self.nativized_key(key)
    nativized_keys.keys.include?(key.to_s) ?  nativized_keys[key.to_s] : key
  end

  # returns icon URL for the activity source
  # supposed to be overridden by the specific activity sources types
  def icon
    nil
  end

  def presence_of_valid_type
    unless SOURCE_TYPES.include?(self._type)
      errors.add(:_type, "is not valid")
    end
  end
  private :presence_of_valid_type

  def self.generated_columns
    sak = UUID.generate 
    { 'source_association_key' => sak,
      'queue_username' => sak,
      'queue_password' => SecureRandom.hex(32).to_s
    }
  end

  # Strips the leading and trailing whitespace from the display name.
  def strip_whitespace
    self.display_name.strip! unless self.display_name.blank?
  end

  # Get all of the sources in the project
  # @param [Project] project you want the sources of
  # @return [Array] an array of source ids for the project
  def self.in_project(project)
    source_ids = project.sources.flat_map(&:id)
    where(id: source_ids).map(&:id)
  end

  # All activities associated (directly) with this source
  #
  # @return [[Activity]] activities that belong to this source
  def activities
    Activity.where(source_id: self.id)
  end

  # Whether or not activate/deactivate buttons show up in the CRUD edit display.
  def deactivatable?
    true
  end


  # An interface method for how activity_sources should handle_messages
  # Handles the parsed json message hash to create an activity
  # @param [Symbol] data_key - the data_key this message was included inside of
  # @param [Hash] message - the message to be created as an activity
  def handle_message(data_key, message)
    raise NotImplementedError, "Please implement the handle_message method on your activity source"
  end
  add_method_tracer :handle_message

  # Check to see if a given date is valid in RFC3339 format.
  def validate_event_time(test_date)
    raise MissingFieldError, 'event_time' unless test_date
    DateTime.rfc3339(test_date)
  rescue ArgumentError
    raise MessageProcessingError, I18n.t('api.time_format')
  end


  def flatten_hash_and_append_key_name(hash, delim = '_')
    hash.each_with_object({}) do |(key, value), hash|
      if value.is_a? Hash
        flatten_hash_and_append_key_name(value).map do |hash_key, hash_value|
          hash["#{key}#{delim}#{hash_key}"] = hash_value
        end
      else
        hash[key] = value
      end
    end
  end

  def self.source_types
    source_type_hash = Hash.new
    Project::KNOWN_TYPES.each { | type, source_type | source_type_hash[type.to_s] = source_type.to_s }
    source_type_hash
  end

  def is_member_of_any_tool?
    Tool.all.map {|tool| tool.sources.map(&:id)}.flatten.include? self.id
  end

  private

  def validate_existence_of_tool
    return if tool_id.blank?
    #flag = Tool.all.any?{ |tool| tool.id.to_s == tool_id.to_s }
     flag = Tool.all.find { |tool| tool.id.to_s == tool_id.to_s  }
     errors.add(:tool_id, 'invalid reference to a tool') unless flag
  end

  def validate_existence_of_project
    flag = Project.all.find { |project| project.id.to_s == project_id.to_s }
    errors.add(:project_id, 'invalid reference to a project') unless flag
  end

  # A commit server must be associated with a project
  def validate_project_association
    if project.blank?
      errors.add(:project, "activity source must belong to a project")
    end
  end

  # Convert relative URLs to full URLs with a CTF domain, if they are not full URLs already
  #
  # @param [String] Relative or Full URL
  # @return [String] FUll URL
  def self.convert_to_absolute_url(url)

    # Regex to check whether an URL is a full URL
    # To render assets served by EventQ in activity stream displayed in integrated applications and other possible
    # places, the relative URLs of images are converted to full URLs so that the assets do not break.
    # Validates URIs starting with http://, https://, ftp:// and protocol-independent URIs like //cdn.example.com
    # as valid. Case insensitive.

    is_full_url = url[/^(?:[a-z]+:){0,1}\/\//i]

    # If it is a relative URL, convert to full URL
    if is_full_url.nil?
      return Rails.configuration.teamforge[:server] + url
    else
      return url
    end
  end

end
