#
# Copyright 2013 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

class ActivitySummary
  include MongoMapper::Document
  include CtfAssociationPublisher

  include ::NewRelic::Agent::MethodTracer

  embedded_callbacks_off

  before_create :create_opaque_id

  key :opaque_id, String
  key :time_created, Time
  key :created_by, String
  key :status_type, String
  key :status_name, String
  key :settled, Boolean, default: true

  timestamps!

  belongs_to :source, class: ActivitySource

  many :comments, :as => :commentable, touch: true

  validates_associated :comments

  many :activities, foreign_key: :activity_summary_id

  key :associated_activity_summary_ids, Set
  many :activity_summaries, in: :associated_activity_summary_ids
 
  attr_accessible *(keys.keys.map(&:to_sym) + [:comments, :source])

  alias :first_order_associations :activity_summaries

  def commentable
    self
  end

  # @return [String] Username of the user who performed the last update (if known)
  def updated_by
    most_recent = activities.order(:time_created.desc).first
    most_recent.try(:updated_by) if most_recent
  end

  # Used by the Rails routing code to generate a URL pointing to this object.
  #
  # @return [String] opaque id for this object
  def to_param
    opaque_id
  end

  # Returns a icon status for any instance which has a :status attribute present
  # @return [String] the lower-case status or "none"
  def status_icon_class
    if self.status_type.present?
      status_type.to_s.downcase
    else
      "none"
    end
  end

  def title
    ''
  end

  # adds self as an association to all passed activity summaries
  # Note that this method will write to the database objects, but not the objects in memory! In general, to get the database
  # and application in sync, you will need to call #reload on each passed activity summary.
  # @param [Array<ActivitySummary>] activity_summaries
  def associate_to_activity_summaries!(activity_summaries)
    act_sums = activity_summaries.map {|a| a._id}
    ActivitySummary.add_to_set({:id => { :$in => act_sums}}, associated_activity_summary_ids: id)
  end
  add_method_tracer :associate_to_activity_summaries!

  # Fetch first order associations for a list of ActivitySummaries
  # Does a bulk fetch, faster than .each(&:activities)
  #
  # @param [[Activity]] A list of activities to fetch associations for
  # @return [[Activity]] unique list of Activities associated with the given Activities
  def self.activity_summaries_for(activity_summaries)
    ids = Set.new(activity_summaries.map(&:associated_activity_summary_ids)).flatten

    (ids.length > 0) ? ActivitySummary.where(id: ids).all : []
  end

  # returns first and second order associations for builds, review requests and custom and first order associations for commits
  # sorted by time_created and created at.
  # @param [Boolean] include_self. Include self into returned activities if true.
  # @return [Array<Activity>] an array of activities
  def all_activity_summary_associations(include_self=false)
    associations = if self.is_a? CommitSummary
                    first_order_associations
                   else
                    first_order_associations + ActivitySummary.activity_summaries_for(first_order_associations)
                   end
    # At this point self may be in the set of associations (if self is a build/review_request and
    # is associated with any commit, so that its in the second order associations) or not (if self is a commit)
    include_self ? associations << self : associations.delete(self)
    associations.uniq.sort_by! {|activity_summary|  [activity_summary.time_created, activity_summary.created_at]}.reverse!
  end

  # if called on the summary, we respond to first as true if we only had 1 activity related
  # Otherwise we are not a in the state of being the first request
  def is_first?
    activities.length <= 1
  end

  # @return [Array<Activity>] The series of activity instances sharing this remote and server ids,
  #in reverse chrono order.
  def full_history
    activities.order(:time_created.desc)
  end

  # returns the attribute if the method for it is defined, else returns nil
  # @param [String|Symbol] the attribute name
  def fetch_attribute(attr)
    self.send(attr) if self.respond_to? attr
  end

  # Returns the existing summary or creates a phantom one except Work Item Summary
  # For Work Item Summary, returns the existing summary or nothing.
  #
  # For other summaries, We only create one phantom if needed.
  # WorkItemSummary does not create a phantom.
  def self.find_or_create(remote_id, sources)
    sources = sources.class == Array ? sources : [sources]
    summary_class = summary_type_from_source(sources.first)
    remote_id_key = if sources.first.is_a? CommitServer then :revision_number else :remote_id end

    # We take the first summary from the list, sorted by settled (which may or may not be settled) descending order.
    found = ActivitySummary.where(source_id: sources.map(&:id), remote_id_key => remote_id).sort(:settled.desc).first

    unless sources.first.is_a? WorkItemSource
      # FIXME - Jan 21, 2016
      # We need to revisit the logic of skipping the validations when summary is saved here
      # so there won't be any one off case.
      found ||= summary_class.new(remote_id_key => remote_id, source_id: sources.first.id, time_created: Time.now, settled: false).tap {|summary| summary.save(validate: false)}
    end
    found
  end

  # infers the activity summary type from the source
  def self.summary_type_from_source(source)
    case source
      when BuildServer then BuildSummary
      when CommitServer then CommitSummary
      when ReviewServer then ReviewRequestSummary
      when WorkItemSource then WorkItemSummary
      when CustomActivitySource then CustomSummary
    end
  end

  private

  # Creates the next opaque_id.
  # @return [String] a new opaque id
  def create_opaque_id
    activity_log_counter = Counters.first_or_create(name: Activity.to_s)
    self.opaque_id = OpaqueID.convert(activity_log_counter.next)
  end
end