#
# Copyright 2013 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

require 'http/http_connection_validator'

class BuildServer < ActivitySource
  belongs_to :commit_server
  many :builds, foreign_key: :source_id

  validate :validate_commit_server_id

  def self.permitted_attrs
    super.concat(['associated_commit_source_id']) 
  end

  def self.nativized_keys
    super.merge({'associated_commit_source_id' => 'commit_server_id'})
  end

  # creates a build and saves it in the database
  # @param [Hash] message_hash parsed JSON message from Rabbit
  # @return [Build] generated build
  def handle_message(message_type, message_hash)
    raise ProjectNotActive, I18n.t('project.messages.not_active', project_name: self.project.name) unless self.project.active?
    build_build(flatten_hash_and_append_key_name message_hash).tap do |build|
      raise MessageProcessingError, I18n.t("api.duplicate", type:"Build") if duplicated? build
      if build.save
        build_summary = BuildSummary.create_or_update(build)
        build_summary.publish_associations_to_ctf
        AccessLogUpdater.update_log(self)
      else
        error_message = "#{I18n.t('errors.could_not_save',
                                  type: I18n.t('build.description'))}: #{build.errors.messages.values.to_sentence}"
        AccessLogUpdater.update_log(self, error_message)
        raise MessageProcessingError, error_message
      end
    end
  end

  # validates that the build is unique in time and in remote_id
  # @param [Build] build the build to look up
  # @return [Boolean] return true if a duplicate build exists
  def duplicated?(build)
    builds.exists?(remote_id: build.remote_id, time_created: build.time_created)
  end

  def as_json(options={})
    # Get the API version
    #
    # Format:
    # {:prefixes=>["api/v1/trackers", "api/base"]}
    #
    api_version = 'v1'
    api_version = options[:prefixes][0].split('/')[1] if options[:prefixes].present?
    "#{api_version.camelize}::BuildSourceSerializer".constantize.new(self).serialize
  end

  # makes a build (without saving)
  # @param [Hash] build_hash parsed JSON message from Rabbit
  # @return [Build] generated build
  def build_build(build_hash)
    validate_event_time build_hash['event_time']

    build_data = translate_to_orchestrate(build_hash)

    #don't want revision data to go into the mass assignment when creating a build, revisions are processed separately
    revisions_info = build_data.delete('revisions')
    builds.build(build_data).tap do |build|
      build.provided_revisions = generate_revisions(revisions_info.to_a)
    end
  end

  # parses the revision info and generates revisions
  # @param [Array] revisions array of revision hashes
  # @return [Array] array of IncludedRevisions
  def generate_revisions(revisions)
    revisions.map { |revision_hash| IncludedRevision.new(revision_hash) }
  end

  # Converts data to the Orchestrate format.
  # @param [Hash] build_data request_data hash of incoming request to convert
  def translate_to_orchestrate(build_data)
    Hash[build_data].tap do |orchestrate_format|
      orchestrate_format['time_created'] = orchestrate_format.delete('event_time')
      orchestrate_format['revisions'].to_a.each do |revision_info|
        revision_info['repository'] = revision_info.delete('repository_url')
      end
    end
  end

  private

  def validate_commit_server_id
    return if commit_server_id.blank?
    flag = CommitServer.all.any? { |server| server.id == commit_server_id}
    errors.add(:commit_server_id, 'cannot be associated with a fictitious commit server' ) unless flag
  end
end
