#
# Copyright 2013 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

# NOTE: DEVELOPMENT environments lazy load SCI classes!
# When you create a child of this class, add it to 'config/environments/development.rb'
# or you will end up with _weird_ behaviour in dev mode.
class CommitServer < ActivitySource
  include UriHelper

  uri_extractor :repository_uri

  key :repository_uri, String
  # does not appear to be used
  key :uuid, String
  key :commit_gate, Boolean, default: false     # For pre-commit hook

  many :commits, foreign_key: :source_id

  before_validation :alter_incoming_uris

  validates :repository_uri, presence: true
  validates :repository_uri, format: { with: /\A(svn:|http:|https:|git:|ssh:)?\/\/[a-zA-Z0-9_.-\/]+/i   }
  validates :commit_gate,  inclusion: { in: [true, false]  }
  validates_with HttpConnectionValidator, attribute: :repository_uri
  validate :validate_project_association

  # Builds a new commit object
  def handle_message(data_type, message_hash)
    raise ProjectNotActive, I18n.t('project.messages.not_active', project_name: self.project.name) unless self.project.active?
    create_or_update_from_queue(message_hash)
  end

  def self.permitted_attrs
    super.concat(["repository_uri"])
  end

  # Simply returns the class' name (which is also its repo type)
  def repository_type
    self.class.name
  end

  # Create PathChange objects from the given commit data
  # @param [Hash] commit_data commit info hash from a parsed json
  # @return [[PathChange]] generated change set
  def path_changes(commit_data)

    # allow for the changes key to be missing entirely
    return [] unless commit_data[:changes]
    commit_data[:changes].map do |change|
      PathChange.new(path: change[:path], type: change[:action])
    end
  end

  def as_json(options={})
    # Get the API version
    #
    # Format:
    # {:prefixes=>["api/v1/trackers", "api/base"]}
    #
    api_version = 'v1'
    api_version = options[:prefixes][0].split('/')[1] if options[:prefixes].present?
    "#{api_version.camelize}::CommitSourceSerializer".constantize.new(self).serialize
  end

  private

  # TODO: doesn't actually update anything
  def create_or_update_from_queue(commit_data)
    validate_commit_data(commit_data)

    path_changes = path_changes(commit_data)
    commit = Commit.find(revision_number: commit_data[:revision_id], source_id: id)
    unless commit
      commit = Commit.create(
          created_by: commit_data[:created_by],
          time_created: commit_data[:event_time],
          revision_number: commit_data[:revision_id],
          branch_name: commit_data[:branch_name],
          message: commit_data[:message],
          source_id: id,
          ctf_item: commit_data[:ctf_item],
          path_changes: path_changes,
      )
    end
    CommitSummary.create_or_update(commit)
    commit
  end

  # Strings the trailing slashes from the user-entered URIs
  def alter_incoming_uris
    self.repository_uri.strip! unless self.repository_uri.blank?
    self.repository_uri = normalize_repository_uri(self.repository_uri)
  end

  # verifies required fields are present
  def validate_commit_data(commit_data)
    Array(%w{revision_id type event_time created_by}).each do |field|
      raise MessageProcessingError, I18n.t("api.missing_field", field: field) if missing_field?(commit_data, field)
    end
  end

  # returns true if the given field is not present within the given hash
  def missing_field?(hash, field)
    !(hash.key?(field) && hash[field].present?)
  end

end
