#
# Copyright 2013 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#


class CommitSummary < ActivitySummary
  key :message, String
  key :branch_name
  many :path_changes
  key :revision_number
  key :ctf_commit_url, String

  one :ctf_item

  alias :title :message

  attr_accessible *(keys.keys.map(&:to_sym) + [:ctf_item])

  embedded_callbacks_off

  # Returns the PathChanges in sorted order
  # @return [Array<PathChange>] the sorted list of path changes
  def path_changes_sorted
    # TODO: (JDM) This should have already been sorted in the backend
    path_changes.sort_by(&:path).reverse
  end


  # Returns the associated work items, if any.
  # @return [Array<CtfAssociation>] the work items or an empty array if this is not a CTF item
  def associated_work_items
    ctf_item.try(:sorted_associations) || []
  end

  # Returns commit_summaries for a CTF Item
  # @param [CtfAssociation] ctf_association the ctf association to find commit summaries for
  # @param [Project] project the project to find commits in
  # @return [Array<CommitSummary>] the commit_summary in the project that are associated to the ctf_association
  def self.find_commit_summaries_by_ctf_association(ctf_association, project)
    project_sources_map = project.sources.collect(&:id)
    CommitSummary.where('ctf_item.ctf_associations.item_id' => ctf_association.item_id,
        :source_id.in => project_sources_map).order(:time_created).all
  end

  # Override for CtfAssociationPublisher.publish_associations_to_ctf
  # which, in this case, enqueues only the self CommitSummary
  def publish_associations_to_ctf
    Resque.enqueue(CtfAssociationJob, id)
  rescue
    Rails.logger.error "Unable to create item associations in TeamForge using #{Resque.inspect}"
    raise
  end

  # This method is used when an underlying  commit arrives via a queue or CTF post-commit notification.
  # Check for an existing and update if found.
  # Otherwise create a new CommitSummary object with the given data.
  # @param [Commit] commit the underlying commit.
  # @return [CommitSummary] A persisted CommitSummary object, either the one found or a newly created one.
  def self.create_or_update(commit)
    commit_summary = find_or_initialize_by_revision_number_and_source_id(commit.revision_number, commit.source_id)
    commit_summary.populate_fields(commit)
    activity_summaries_from_message = associations_from_commit_message(commit)
    commit_summary.activity_summaries << activity_summaries_from_message

    # write to associated review_request summaries
    if commit_summary.save
      commit_summary.associate_to_activity_summaries!(activity_summaries_from_message)
      commit_summary.publish_associations_to_ctf
      commit_summary
    else
      LogUtils.log_error_and_info(commit_summary.errors.messages, commit_summary.inspect)
      nil
    end
  end
  add_method_tracer :create_or_update

  # Find matching work item / review request from the commit message
  # @param [Commit] a commit object
  # @return [Array<ReviewRequestSummary><WorkItemSummary>]
  def self.associations_from_commit_message(commit)
    message = commit.message.to_s
    association_regexes = commit.source.project.combined_association_regexes

    associated_summaries = []
    Lexi.new(message, association_regexes).parse.map do | association |
      remote_id = association.keys[0]
      sources = association[remote_id]

      associated_summaries << ActivitySummary.find_or_create(remote_id, sources)
    end

    associated_summaries.compact
  end
  add_method_tracer :associations_from_commit_message

  # Return the display 'number' for this commit_summary
  # @return [String] the display number
  def display_number
    revision_number.size > 20 ? revision_number[0, 7] : revision_number
  end

  # copy the fields from the commit object to commit summary
  # @param [Commit]
  def populate_fields(commit)
    self.time_created = commit.time_created
    self.created_by = commit.created_by
    self.message = commit.message
    self.branch_name = commit.branch_name
    self.ctf_item = commit.ctf_item
    self.path_changes = commit.path_changes
    self.settled = true
    self.activities << commit
    self.ctf_commit_url = URI.join(CtfGoLinks.base_go_url, commit.ctf_item[:ctf_id]).to_s unless commit.ctf_item.nil?
  end
  add_method_tracer :populate_fields

end