#
# Copyright 2013 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

class Custom < Activity
  include ChangeHistory

  STOCK_FIELDS = CustomActivitySource::STOCK_FIELDS.map &:to_sym
  VALID_STATUS_TYPES = %w(open success failed)
  (STOCK_FIELDS - [:event_time, :associated_remote_id, :link_to, :opaque_ids]).each do |field|
    key field, String
  end

  key :associated_remote_id, Array

  key :custom_data, Hash

  key :link_to

  alias :display_number :remote_id

  key :attribute_changes, Array
  key :ancestor_id, String

  # optional array of opaque ids
  key :opaque_ids, Array

  belongs_to :custom_schema

  before_create :calculate_attribute_changes, :set_displayable


  # Override of Activity.status_icon_class, since work item uses booleans to map to an Orchestrate status
  #
  # @return |String| the status, based on status type, or 'open' if the status type is not recognized
  def status_icon_class
    if status_type.blank?
      ''
    else
      (VALID_STATUS_TYPES.include? status_type) ? status_type : 'open'
    end
  end

  def important_changes
    attribute_changes & custom_schema.important.map(&:to_sym)
  end

  # Fill out the activity with data from the queue message.  Stock fields get
  # filled in directly, everything else is stuffed in a #custom_data hash field.
  # We rely on prior levels to have validated out all the data, its conformance,
  # and that it's a desirable message.  We also rely on the caller to save us.
  def handle_message(message)
    self.custom_schema_id = message.delete(:custom_schema_id) # it is not in the original message, but gets merged into the message when parsing
    self.associated_remote_id = get_array_from_message(:associated_remote_id, message)
    self.opaque_ids = get_array_from_message(:opaque_ids, message)

    stock = message.slice(*STOCK_FIELDS)
    optional = message.except(*STOCK_FIELDS)

    self.time_created = DateTime.rfc3339(stock.delete(:event_time))

    stock.each {|k,v| send("#{k}=", v)}

    self.custom_data = optional
  end

  # deletes and returns an expected array value from the message
  # returns an array of values if the key points to an array
  # returns an array of a single value if the key points to a single value
  # returns an empty array otherwise
  def get_array_from_message(key, message)
    if message[key].present? # handles string value or array value
      Array(message.delete(key))
    else
      [] # handles "" and nil
    end.select { |id| id.present? }#removes empty ids if passed in associated_remote_id array
  end


  # builds attribute_changes array against previous Activity and stores
  # result in the :attribute_changes field.
  # @return |Array| the collected attribute changes
  def calculate_attribute_changes
    self.attribute_changes = []
    current_instance_time = time_created ? time_created : Time.now
    previous_instance = Custom.most_recent_activity(remote_id, source_id, current_instance_time)
    if previous_instance
      self.ancestor_id = previous_instance.id
      (custom_schema.fields - ['event_time']).each do |attr|
        attribute_changes << attr.to_sym unless self.fetch_attribute(attr) == previous_instance.fetch_attribute(attr)
      end
    else
      attribute_changes << :is_created
    end
    attribute_changes
  end

  def set_displayable
    self.displayable = custom_schema.important.blank? ||
        if self.is_first?
          custom_schema.important.any? {|attr| fetch_attribute(attr).present?}
        else
          important_changes.present?
        end
    nil # The deal here is that if a before create filter evaluates to the boolean then the filter chain will be stopped if the return value is false
  end

  def fetch_attribute(attr)
    if attributes.has_key?(attr) then attributes[attr] else custom_data[attr] end
  end
end
