#
# Copyright 2014 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

require 'uri'

class CustomActivitySource < ActivitySource
  include MongoMapper::Document
  include UriHelper

  # remote_id is a required field, associated_remote_id is optional
  REQUIRED_STOCK_FIELDS = %w(event_time remote_id schema_id schema_version)
  OPTIONAL_STOCK_FIELDS = %w(associated_remote_id created_by summary description status_name status_type link_to updated_by opaque_ids)
  STOCK_FIELDS = REQUIRED_STOCK_FIELDS + OPTIONAL_STOCK_FIELDS

  # the custom activity record contains the foreign key: activity_source_id
  belongs_to :activity_source

  # the custom activity record contains the foreign key: custom_schema_store_id
  belongs_to :custom_schema_store

  # custom_activity_source has many custom records
  many :customs, foreign_key: :source_id

  key :activity_source_tag_ids, Set

  many :activity_source_tags, in: :activity_source_tag_ids

  key :background_color_value, String, :default => "#61c0e5"

  key :icon_uri, String, :default => "/assets/icons/custom_source/XDS_Icon_01.png"

  alias :schema :custom_schema_store

  # associated_source is an alias for activity_source
  alias :associated_source :activity_source

  # the array of the ids of trusted activity source objects for this
  # custom activity source.
  # it will be used to build associations for inter-project as well as
  # intra-project events.
  key :trusted_sources, Array


  attr_accessor :activity_source_tag_names


  before_validation :normalize_icon_uri
  before_validation :normalize_trusted_sources

  validates :icon_uri, presence: true
  validates :background_color_value, presence: true
  validates :background_color_value, format: { with: /\A(#[a-fA-F0-9]{6}|[a-zA-Z]*)\z/i, message: 'invalid RGB value' }
  validate :valid_icon_uri
  validate :valid_custom_schema_store_id
  validate :valid_activity_source_id

  # nativized keys in the base class is {}
  def self.nativized_keys
    super.merge({'associated_activity_source_id' => 'activity_source_id'})
  end

  before_save :find_or_create_tags

  def self.permitted_attrs
    super.concat(["associated_activity_source_id", 
                  "activity_source_tag_names",
                  "background_color_value",
                  "icon_uri",
                  "custom_schema_store_id",
                  "trusted_sources"])
  end

  def find_or_create_tags
    if self.activity_source_tag_names
      self.activity_source_tag_ids = self.activity_source_tag_names.map do |tag|
        ActivitySourceTag.first_or_create(name: tag.downcase)
      end.map(&:id)
    end
  end

  # called from custom queue parser
  # Processes a message that has a new activity. Creates a new activity record.
  # Creates or updates an existing CustomSummary record..
  def create_activity(message)
    activity = customs.build.tap {|activity| activity.handle_message(message) }
    activity.save
    CustomSummary.create_or_update(activity)
  end

  def can_be_associated?
    !activity_source
  end

  def as_json(options={})
    # Get the API version
    #
    # Format:
    # {:prefixes=>["api/v1/trackers", "api/base"]}
    #
    api_version = 'v1'
    api_version = options[:prefixes][0].split('/')[1] if options[:prefixes].present?
    "#{api_version.camelize}::CustomActivitySourceSerializer".constantize.new(self).serialize
  end

  # update the custom_schema_store_id key with the schema store id
  def associate_with_schema_store(schema_store)
    update_attributes(custom_schema_store_id: schema_store.id) unless custom_schema_store.present?
  end

  # Get the schema with specific ID and version from message.
  # Validate message, schema and project before creating a new activity
  def handle_message(message)
    schema = schema_from_message(message)
    associate_with_schema_store(schema.custom_schema_store)
    validate_message_and_schema(schema, message)
    validate_project_active
    create_activity(message.merge(custom_schema_id: schema.id))
  end

  # Get the schema from the schema ID and version in the message
  def schema_from_message(message)
    schema = CustomSchema.where(:schema_id => message["schema_id"], :schema_version => message["schema_version"]).first
    raise MessageProcessingError, I18n.t('custom_activity_source.messages.schema_not_found', schema_id: message["schema_id"],
                                         schema_version: message["schema_version"] ) if schema.blank?
    schema
  end

  # Ensure that the message is valid
  def validate_message_and_schema(schema, message)
    validate_event_time message['event_time']
    raise MessageProcessingError, I18n.t('custom_activity_source.messages.nonconforming',
                                         schema_name: schema.name) unless schema.valid_message?(message)
  end

  # Ensure the pipeline project for the source is active
  def validate_project_active
    raise ProjectNotActive, I18n.t('project.messages.not_active', project_name: project.name) unless self.project.active?
  end

  ##########################################################################
  # insert_trusted_source
  #
  # Receives: (i)  the user
  #           (ii) an array of source  ids off sources sought to be inserted
  #                into the trusted sources array
  # Returns: the trusted sources actually inserted (as an array)
  #
  # FIXME: this doc is no longer an accurate depiction of how things work.
  # Algorithm: inserts the id of a trusted source into the trusted sources
  # array if the user has application admin permission in the project to
  # which the source belongs and the project is active
  ###########################################################################

  def insert_trusted_sources(source_ids)
    # return an empty array if the source_ids array is empty
    return [] if source_ids.blank?
    sources_inserted = []

    source_ids.each do |source_id|
      source  = ActivitySource.find(source_id)
      next if source.nil?
      next if associated_source && associated_source.id == source_id # do not add the associated source as a trusted source

      sources_inserted << source.id
    end

    if sources_inserted.present?
      self.trusted_sources = (trusted_sources + sources_inserted).uniq
      self.save
    end

    sources_inserted
  end

  private

  def normalize_trusted_sources
    return if self.trusted_sources.blank?
    dirty_sources = self.trusted_sources
    clean_sources = []

    dirty_sources.each do |source|
      case source
        when String
          found_source = ActivitySource.find(source)
          clean_sources << found_source.id if found_source.present?
        when BSON::ObjectId
          clean_sources << source
      end
    end

    self.trusted_sources = clean_sources.compact.uniq
  end

  def valid_icon_uri
    !!URI.parse(icon_uri)
  rescue URI::InvalidURIError
    errors.add(:icon_uri, "invalid uri")
  end

  def valid_custom_schema_store_id
    return if custom_schema_store_id == nil
    # If custom_schema_store_id is not nil, validate for existence of SchemaStore for the given schema_id
    schemas = CustomSchema.where(custom_schema_store_id: custom_schema_store_id).to_a.map(&:schema_id)
    flag = CustomSchemaStore.where(schema_id: schemas).first
    errors.add(:custom_schema_store_id, 'reference to an invalid schema') unless flag
  end

  def valid_activity_source_id
    return if activity_source_id.blank?
    flag = ActivitySource.find(activity_source_id)
    errors.add(:activity_source_id, 'reference to an invalid activity_source') unless flag
  end

  def normalize_icon_uri
    self.icon_uri = CustomActivitySource.normalize_orchestrate_icon_uri self.icon_uri if self.icon_uri
  end

end
