#
# Copyright 2014 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

class CustomSummary < ActivitySummary
  key :remote_id, String
  key :associated_remote_id, Array
  key :created_by, String
  key :summary, String
  key :description, String
  key :status_name, String
  key :status_type, String
  key :link_to # Was String, now can be String or Array and, therefore, untyped
  key :custom_data, Hash
  VALID_STATUS_TYPES = %w(open success failed)

  belongs_to :custom_schema

  #widget aliases
  alias :display_number :remote_id
  alias :title :summary
  delegate :background_color_value, :to => :source


  attr_accessible *(keys.keys.map(&:to_sym))

  # @return |String| the status, based on status type, or 'open' if the status type is not recognized
  def status_icon_class
    if latest_activity then latest_activity.status_icon_class else '' end
  end

  # Adds an activity to the activities associated with this summary.
  # puts the activity_summary_id in the activity
  def add_activity(activity)
    activities << activity
  end
  add_method_tracer :add_activity

  # In CustomSummary find a custom summary by source_id and remote_id. If the custom summary is not found
  # add the activity to the activities collection with the custom_summary_id set in the activity record
  # then update the custom-summary record from the activity. If the custom-summary record was found simply
  # update the custom summary record

  def self.create_or_update(activity)
    CustomSummary.find_or_initialize_by_source_id_and_remote_id(activity.source_id, activity.remote_id).tap do |summary|
      summary.add_activity(activity)
      summary.update_from_activity(activity)
      summary.associate_to_activity_summaries!(summary.activity_summaries)
      summary.publish_associations_to_ctf
    end
  end
  add_method_tracer :create_or_update

  def update_from_activity(activity)
    activity_associations = associations_from_activity(activity)

    self.status_name = activity.status_name
    self.status_type = activity.status_type
    self.time_created = activity.time_created
    self.custom_schema_id = activity.custom_schema_id unless self.custom_schema_id.present?
    self.activity_summaries << activity_associations if activity_associations.present?
    self.associated_remote_id << activity.associated_remote_id
    self.created_by = activity.created_by
    self.summary = activity.summary
    self.description = activity.description
    self.link_to = activity.link_to
    self.custom_data = activity.custom_data #HashWithIndifferentAccess so cannot <<
    self.settled = true
    save
  end
  add_method_tracer :update_from_activity

  def latest_activity
    activities.order(:time_created.desc).first
  end

  # return the array of custom summaries which are commit summaries
  def associated_commit_summaries
    activity_summaries.select { |x| x.is_a? CommitSummary}
  end

  # Algorithm: associate the incoming activity with the custom summary iff:
  # (a) The activity source of the incoming summary exists in the trusted sources array
  #     of the activity source pertaining to the activity summary
  def associations_from_activity(activity)
    trusted_sources = self.source.trusted_sources

    parsed_associations = associations_from_parsed_fields(activity)
    explicit_associations = explicit_associations_from_activity(activity)
    opaque_associations = []

    valid_opaque_ids = activity.opaque_ids.uniq.select { |opaque_id| can_associate_with_opaque_id?(trusted_sources, opaque_id) }
    opaque_associations << ActivitySummary.where('opaque_id' => {:$in => valid_opaque_ids}).to_a

    parsed_associations + explicit_associations + opaque_associations
  end
  add_method_tracer :associations_from_activity

  # receives: [1] the trusted sources list for the custom activity source
  #               of this custom summary
  #           [2] an opaque id
  #
  # Algorithm: get the activity source pertaining to the opaque
  # id. see whether this source exists in the trusted source array of
  # the activity summary source.
  def can_associate_with_opaque_id?(trusted_sources, opaque_id)
    src = get_source_from_opaque_id(opaque_id)
    return false if src.nil?
    trusted_sources.include?(src.id)
  end

  # for the activity source in the custom_summary record get the associated_source , this is the activity_source record
  # get the activity's associated-remote_id array. return an empty array if the associated_source or the array is empty
  # get the array of remote_ids
  # return the array of ActivitySummaries for each remote_d and associated_source

  def explicit_associations_from_activity(activity)
    associated_source = source.associated_source
    return [] unless activity.associated_remote_id.present? && associated_source.present?
    remote_ids = if associated_source.is_a? ReviewServer then activity.associated_remote_id.map(&:upcase) else activity.associated_remote_id end
    remote_ids.map do |remote_id|
      ActivitySummary.find_or_create(remote_id, [associated_source]) if remote_id.present?
    end
  end
  add_method_tracer :explicit_associations_from_activity

  def associations_from_parsed_fields(activity)
    work_item_sources = source.project.work_item_sources
    parseable_fields = activity.custom_schema.parsed.map { |attr| activity.custom_data[attr.to_s]}
    text_to_parse = (parseable_fields + [activity.description, activity.summary]).compact.join(' ')
    artf_references = Lexi.new(text_to_parse, {WorkItemSummary::CTF_ARTF_REGEX => :ctf}).parse.map{| artf | artf.match_data[0] }.compact
    WorkItemSummary.where(source_id: work_item_sources.map(&:id), remote_id: artf_references).all
  end
  add_method_tracer :associations_from_parsed_fields

  # Only show activities that have attribute changes
  #@see ActivitySummary#full_history
  def full_history
    super.select {|activity| activity.attribute_changes.present?}
  end

  # @Override
  def publish_associations_to_ctf
    associations_to_enqueue.each do |summary|
      Resque.enqueue(CtfAssociationJob, summary.id)
    end
  end
  add_method_tracer :publish_associations_to_ctf

  # we need to enqueue only commit summaries and work items that we suspect(based on their remote id) are ctf artifacts
  def associations_to_enqueue
    activity_summaries.select do |summary|
      summary.is_a?(CommitSummary) || (summary.is_a?(WorkItemSummary) && summary.remote_id =~ WorkItemSummary::CTF_ARTF_REGEX)
    end
  end

  # returns the attribute if the method for it is defined, else returns nil
  # @param [String|Symbol] the attribute name
  def fetch_attribute(attr)
    if attributes.has_key?(attr) then attributes[attr] else custom_data[attr] end
  end

  private

  # gets the activity source pertaining to an opaque id
  def get_source_from_opaque_id(opaque_id)
    summary = ActivitySummary.first('opaque_id' => opaque_id)
    summary.source if summary
  end


end
