#
# Copyright 2013 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

# Projects have sources. Permissions define the user's access to project.
class Project
  include MongoMapper::Document

  key :name, String
  key :active, Boolean, default: true

  one :project_configuration, class_name: "TeamForge::ProjectConfiguration", :conditions => {active: true}

  has_many :sources, class: ActivitySource
  has_many :tools

  validates :name, presence: true
  timestamps!

  KNOWN_TYPES = { WorkItem => WorkItemSource,
                  Commit => CommitServer,
                  ReviewRequest => ReviewServer,
                  Build => BuildServer,
                  Custom => CustomActivitySource
                }.with_indifferent_access

  # Intercept the default query, and attach the 'active' condition to it.  Forces only active records to be
  # returned in queries.  To access inactive records, pass 'active: false'.
  #
  # TODO: in this implementation, passing 'active: false' results in returning records which are either true
  # OR false (i.e., all records) since criteria are successively merged.
  #
  # NOTE: This override affects result sets which are produced by query, and NOT those produced from the
  # collection as a whole.  E.g.:
  #
  # Class.all(active: false) - Only the false records; 'all' is acting like a filter on the collection.  No query is created.
  # Class.where(active: false).all - 'where' is merging the criteria; 'all' is acting as a filter on the query results.
  def self.query(options={})
    super.tap { |query| query[:active] = true unless options.key?(:active) }
  end

  def to_param
    project_configuration.present? ? project_configuration.project_url : super
  end

  # the project is active if it has a project configuration
  # @return [Boolean] true if the project configuration present, false otherwise
  def active?
    self.active && project_configuration.present? && project_configuration.active?
  end

  # Used for sorting because MongoMapper (and Mongo underneath) doesn't have case insensitive sort from DB.
  def name_key
    name.downcase
  end

  # Retrieve a list of active projects.
  #
  # @return [Array[Project]] The list of projects that are active.
  def self.active_projects
    Project.all.select(&:active?).sort_by(&:name_key)
  end

  # Retrieve a list of active projects.
  #
  # @param [ids] A list of BSON ids for the projects we care about.
  # @return [Array[Project]] The list of projects that are active and whose ids were present in the ids list
  def self.active_projects_by_id(ids)
    Project.where(id: ids).all.select(&:active?).sort_by(&:name_key)
  end

  # Returns the ctf ids for the active projects.
  # @return [Array] the ctd ids
  def self.active_projects_ctf_ids
    active_projects.map{|project| project.project_configuration[:ctf_project_id]}
  end

  # returns a list of schemas used in the project
  def schemas
    KNOWN_TYPES[Custom].where(project_id: id).select(&:active?).map(&:custom_schema_store)
  end

  # Generates methods to list all the types of sources
  KNOWN_TYPES.each do |type, source_type|
    activity_type = type.to_s.underscore.to_sym

    define_method "#{activity_type}_sources" do
      KNOWN_TYPES[type].where(project_id: id).all
    end
  end

  # return hash of regexes of work item / review request sources and array of corresponding sources
  # returns an array of [Regexp => [Sources]]
  # NOTE: downstream matchers are restricted to whatever the regex is defined as here
  #       make the regex case insensitive
  def combined_association_regexes
    association_regexes = Hash.new

    valid_work_item_sources.each{ |wis|
      regex = wis.regex
      association_regexes[regex] = work_item_sources_with_regex(regex) unless association_regexes.has_key?(regex)
    }

    review_request_sources.each{ |rrs| association_regexes[rrs.regex] = review_request_sources unless association_regexes.has_key?(rrs.regex) }
    association_regexes
  end

  # filter deactivated and unconfigured work item sources from the list of all Work item sources
  def valid_work_item_sources
    work_item_sources
        .select{ |source| source.active? && source.tracker.present?}
  end

  # returns the valid work item sources that match the given regex string
  # @string [String] the regex string to find in sources
  # @return [Array[WorkItemSource]] Array of matching sources
  def work_item_sources_with_regex(string)
    valid_work_item_sources
        .select{ |source| source.regex == string }
  end


  # return the activity types in a given project
  # @param [Project] project
  def activity_types
    stock_types = ["Build", "Commit", "ReviewRequest", "WorkItem"]
    custom_types = CustomActivitySource.where(project_id: id).map(&:custom_schema_store).compact.map(&:name)
    (stock_types + custom_types).sort
  end

  # Return the most recent activities from all activity sources in this project,
  # offsetting the results with the given page number.
  # @param [Numeric] page_number The page number to show
  # @param [number] count The number of activities to return.  If this is -1 all results will be returned.
  # @param [Hash] options (:start, :end, :chosen_activity_type, :include_external_items) An optional hash for :start and :end timestamps
  # @return [WillPaginate::Collection[..activities...] | Array] A Collection of all the different kinds of activities
  #         most recently created, paginated if count was not -1.
  def most_recent(page_number, count, options = {})
    selected_sources = options[:activity_types] ? sources_for_activity_types(options[:activity_types]) : all_sources
    source_ids = selected_sources.map(&:id)

    query = Activity.where(source_id: {:$in => source_ids}).order('time_created desc')
    query = query.where(:time_created.gte => options[:start]) if options[:start] && !options[:count]
    query = query.where(:time_created.lt => options[:end]) if options[:end]
    query = query.where(:displayable.ne => false)

    if options[:count]
      query.where(:time_created.gt => options[:start]).count
    else
      (count == -1) ? query.all : query.paginate(page: page_number, per_page: count)
    end
  end

  def sources_for_activity_types(activity_types)
    app_stock_types = KNOWN_TYPES.keys.map(&:to_s)
    selected_stock_types = activity_types & activity_types
    stock_type_sources = ActivitySource.where(project_id: id, _type: KNOWN_TYPES.values_at(selected_stock_types)).all
    selected_custom_types = activity_types - app_stock_types
    custom_type_sources = CustomActivitySource.where(project_id: id).all.select do |source|
      source.custom_schema_store.present? && selected_custom_types.include?(source.custom_schema_store.name)
    end
    stock_type_sources + custom_type_sources
  end

  def all_sources
    ActivitySource.where(project_id: id).all
  end

  # Get the list of source ids for this
  def all_source_ids
    Set.new(self.sources.map(&:id)).flatten
  end

  # Finds activities in the project by their unique opaque ids.
  # @param [String] id the opaque ID of an activity
  # @return [Activity] the activity with the given opaque id
  # @raise [MongoMapper::DocumentNotFound] if an activity is not found
  def find_activity_summary_by_opaque_id(id)
    activity = ActivitySummary.first(opaque_id: id)
    if activity && activity.source.project.id == self.id
      activity
    else
      raise MongoMapper::DocumentNotFound.new(I18n.t("errors.activity.not_found"))
    end
  end

  def count_newer(atomic_id)
    source_ids = all_source_ids
    Activity.where(source_id: {:$in => source_ids}).where(:atomic.gt => atomic_id.to_i).where(:displayable.ne => false).count
  end

  ######################################################################
  # get_ctf_project_id: gets the symbolic CTF project ID for a project
  ######################################################################
  def get_ctf_project_id
    self.project_configuration[:ctf_project_id]
  end

end
