#
# Copyright 2013 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

# Data model for code reviews
class ReviewRequest < Activity
  include ChangeHistory

  embedded_callbacks_off

  # The length to truncate the comment to when showing it on the activity view.
  MAX_SUMMARY_LENGTH = 100

  VALID_STATUS_TYPES = %w(open submitted rejected discarded other)
  SIGNIFICANT_ATTRIBUTES = [:summary, :description, :status_type, :status_name, :reviewers, :reviewed_files]
  REVIEW_MATCHER = /\[(?<key>\w+):(?<id>.*?)\]/i # a pattern to match a review request from a message

  key :remote_id, String
  key :link, String
  key :summary, String
  key :description, String

  key :status_type, String
  key :status_name, String
  key :updated_by, String

  key :reviewers, Set # Array of {username: 'username', link: 'http://somelink'}
  # Array of attribute keys describing significant changes since the last request for this remote_id
  key :attribute_changes, Array
  # Id of preceding instance in the series, or nil if first
  key :ancestor_id

  activity_source :review_server
  many :reviewed_files

  ##### Deprecated RB-only things that will be removed when we are MQ-API only
  key :change_id, Integer # the ID of the change in Review Board this object is associated with
  #################

  attr_accessible *(keys.keys.map(&:to_sym) + [:reviewed_files])

  alias :display_number :remote_id

  timestamps!

  validates_presence_of :remote_id
  validates_presence_of :status_type, :status_name
  validates_associated :reviewed_files
  before_create :calculate_attribute_changes, :validate_significant_delta

  # returns a list of usernames new to this review instance or an empty list
  # @return |Array| list of username, empty list
  def new_reviewers
    if attribute_changes.include? :reviewers
      current_usernames = reviewers.collect { |elem| elem['username'] }
      previous_usernames = previous_instance.reviewers.collect { |elem| elem['username'] }
      current_usernames - previous_usernames
    else
      []
    end
  end

  # activity type name that gets displayed in CTF -- overrides method of CtfAssociationPublisher
  def ctf_activity_type_name
    I18n.t('review_request.ctf_activity_type_name')
  end

  # builds attribute_changes array against previous Activity and stores
  # result in the :attribute_changes field.
  # @return |Array| the collected attribute changes
  def calculate_attribute_changes
    self.attribute_changes = []
    current_instance_time = time_created ? time_created : Time.now
    previous_instance = self.class.most_recent_activity(remote_id, source_id, current_instance_time)
    if previous_instance
      self.ancestor_id = previous_instance.id
      self.class::SIGNIFICANT_ATTRIBUTES.each do |attr|
        attribute_changes << attr unless send(attr) == previous_instance.send(attr)
      end
    else
      attribute_changes << :status # so that we can display a status icon in front of 'User foo requested code review' in the history of changes
    end
    if (attribute_changes & [:status_name, :status_type]).present?
      attribute_changes.delete(:status_name)
      attribute_changes.delete(:status_type)
      attribute_changes.unshift(:status)
    end
    attribute_changes
  end

end
