#
# Copyright 2013 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#


class ReviewRequestSummary < ActivitySummary
  key :remote_id, String
  key :link, String
  key :summary, String
  key :created_by, String

  key :description, String

  key :reviewers, Set # Array of {username: 'username', link: 'http://somelink'}

  many :reviewed_files

  embedded_callbacks_off

  alias :display_number :remote_id

  alias :associated_commit_summaries :activity_summaries

  alias :title :summary

  attr_accessible *(keys.keys.map(&:to_sym) + [:reviewed_files])

  REVIEW_MATCHER = /\[(?<key>\w+):(?<id>.*?)\]/i # a pattern to match a review request_summary from a message

  ##### Deprecated RB-only things that will be removed when we are MQ-API only
  key :change_id, Integer # the ID of the change in Review Board this object is associated with
  #################

  validates_presence_of :remote_id
  validates_presence_of :status_type, :status_name
  validates_associated :reviewed_files


  # This method is used when an underlying review arrives via a queue 
  # Check for an existing and update if found.
  # Otherwise create a new ReviewRequestSummary object with the given data.
  # @param [ReviewRequest] review_request the underlying review.
  # @return [ReviewRequestSummary] A persisted ReviewRequestSummary object, either the one found or a newly created one.
  def self.create_or_update(review_request)
    review_summary = ReviewRequestSummary.find_or_initialize_by_remote_id_and_source_id(review_request.remote_id, review_request.source_id)
    review_summary.time_created = review_request.time_created if (review_summary.new? || !review_summary.settled)
    review_summary.status_type = review_request.status_type
    review_summary.status_name = review_request.status_name
    review_summary.activities << review_request
    review_summary.reviewed_files = review_request.reviewed_files
    review_summary.reviewers = review_request.reviewers
    review_summary.description = review_request.description
    review_summary.link = review_request.link
    review_summary.created_by = review_request.created_by
    review_summary.summary = review_request.summary
    review_summary.settled = true

    associated_commit_summaries =  review_summary.resolve_commits(review_request)
    review_summary.activity_summaries << associated_commit_summaries

    if review_summary.save
      review_summary.associate_to_activity_summaries!(associated_commit_summaries)
      review_summary.publish_associations_to_ctf
      review_summary
    else
      LogUtils.log_error_and_info(review_summary.errors.messages, review_summary.inspect)
      nil
    end
  end
  add_method_tracer :create_or_update

  # adds commits matched from the reviewed files to the first request in the series, creates phantom commits if necessary
  def resolve_commits(review_request)
    files = review_request.reviewed_files.select{ |file| file[:repository_url] && file[:revision]}
    if files.present?
      commits_from_reviewed_files(files)
    else
      []
    end
  end
  add_method_tracer :resolve_commits

  # finds the commit source by matching the repository uri
  # @params [Array<CommitServer>] commit_sources an array of active commit sources for the project
  # @params [String] repository_uri
  # @return [CommitServer]
  # the comments below have been moved from another file so that they didn't get lost
  # since the bits of related code have been moved here as well
  # FIXME: there's a larger problem of comparing what a ReviewRequest says to what an
  # CommitServer knows about itself than can be resolved here.  This is due to the fact
  # that on one side, the repository URL is a user-supplied (to orchestrate) field which
  # is not validated for actually being a repository, and on the other (crucible) side, it
  # is only the repository root URL (no "trunk", "branch", etc.), but IS validated as being
  # a repository.  This comparison is unfortunate, as the URL field of the CommitServer is
  # otherwise completely unnecessary - you only enter it as a step in order to get the
  # association key that you supply to a hook script (which is one-per-repo).  The best we
  # can do here is find and CommitServer that starts with the URI as reported by the
  # review server.
  #
  # There are a number of things that could be done to address this problem, depending upon
  # what we are actually trying to accomplish - we could remove the user-supplied field, and
  # pull repo data from the first commit we see fly by from that source; we could implement
  # something that pings the repo and confirms before moving on; we could modify the crucible
  # plugin to supply info (like the repo UUID, in the case of svn) which makes it easier to
  # lookup.  Or we could support path-matching, which means rework of this code, as well as
  # reworking how the post commit hooks work.
  def find_commit_source_by_repository_url(commit_sources, repository_uri)
    commit_sources.find{|source| source.repository_uri =~ /^#{repository_uri}(\/|\Z)/}
  end

  private

  # returns an array of existing or created phantom commits matched from reviewed files
  # @param [Array<ReviewedFile>] files reviewed files attribute
  def commits_from_reviewed_files(files)
    commit_sources = source.project.commit_sources.select {|source| source.active? }

    files_for_sources = files.inject(Hash.new([])) do |memo, file|
      current_revisions = memo[file[:repository_url]]
      current_revisions << file[:revision]
      memo[file[:repository_url]] = current_revisions
      memo
    end

    commits_reviewed = []
    files_for_sources.each do |repo, revisions|
      if commit_source = find_commit_source_by_repository_url(commit_sources, repo)
        commits_reviewed << match_commits_for_source(commit_source, revisions)
      end
    end
    commits_reviewed.flatten
  end

  # Returns all CommitSummary for the given source and revision numbers
  # @param [CommitSource] commit_source The source to add/collect revisions with
  # @param [ArrayOfStrings] revisions The revisions to add/collect
  # @return [CommitSummary] matched existing summaries and added phantom summaries
  def match_commits_for_source(commit_source, revisions)
      existing_revisions = commit_source.activity_summaries.where(:revision_number => revisions).map(&:revision_number)
      new_revisions = revisions - existing_revisions

      new_revisions.each do |revision|
        CommitSummary.find_or_create(revision, commit_source)
      end

      commit_source.activity_summaries.where(:revision_number => revisions).to_a
  end
end
