#
# Copyright 2013 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

# Data model for specific Review Board configurations
class ReviewServer < ActivitySource

  include Verifiable
  include UriHelper

  MAX_KEY_LENGTH  = 20

  key :association_key, String
  before_validation :clean_association_key

  validates :association_key,
            presence: true,
            format: {with: /\A\w+\z/i},
            length: {maximum: MAX_KEY_LENGTH},
            uniqueness: {case_sensitive: false}

  many :review_requests, foreign_key: :source_id

  # Builds a new review request object
  def handle_message(data_type, message)
    raise ProjectNotActive, I18n.t('project.messages.not_active', project_name: project.name) unless self.project.active?
    create_review(flatten_hash_and_append_key_name message)
  end

  def self.permitted_attrs
    super.concat(['commit_association_prefix']) 
  end

  def self.nativized_keys
    super.merge({'commit_association_prefix' => 'association_key'})
  end


  def as_json(options={})
    # Get the API version
    #
    # Format:
    # {:prefixes=>["api/v1/trackers", "api/base"]}
    #
    api_version = 'v1'
    api_version = options[:prefixes][0].split('/')[1] if options[:prefixes].present?
    "#{api_version.camelize}::ReviewSourceSerializer".constantize.new(self).serialize
  end

  # Get the matching group from commit message
  # ex: message = [rv:MD-1]
  #     remote_id = 'MD-1'
  def regex
    Regexp.new("(#{association_key}:([\\d\\w][-_\\d\\w]*))", true)
  end


  private

  # Builds a new ReviewRequest object
  # @param [Hash] review_hash a hash of the review data
  def create_review(review_hash)
    validate_event_time(review_hash['event_time'])
    review_data = translate_to_orchestrate(review_hash)
    validate_status_type(review_data)
    validate_status_name(review_data)
    validate_reviewers(review_data)

    raise InconsistentStateError, I18n.t('api.duplicate', type: ReviewRequest.to_s) if is_duplicate?(review_data)

    review_requests.build(review_data).tap do |review_request|
      if review_request.save
        AccessLogUpdater.update_log(self)
        ReviewRequestSummary.create_or_update(review_request)
      else
        model_errors = review_request.errors.messages
        error_message = "#{I18n.t('errors.could_not_save', type: I18n.t('associations.review_request'))}: #{model_errors}"
        full_error = "#{error_message} #{review_hash}"
        AccessLogUpdater.update_log(self, full_error)
        LogUtils.log_error_and_info(error_message, review_hash)
      end
    end
  end

  # Returns true if the review is a duplicate
  def is_duplicate?(review_data)
    ReviewRequest.where(source_id: id,
                        remote_id: review_data['remote_id'].to_s.upcase,
                        time_created: DateTime.rfc3339(review_data['time_created']).to_time,
                        created_by: review_data['created_by']).count > 0
  end

  # Strips the leading and trailing whitespace from the display name and association key.
  def clean_association_key
    self.association_key.strip! unless self.association_key.blank?
  end

  # Converts incoming request to the format we're going to parse.
  # @param [Hash] request_data hash of incoming request to convert
  #
  def translate_to_orchestrate(request_data)
    HashWithIndifferentAccess[request_data].tap do |orchestrate_format|
      orchestrate_format['remote_id'] = orchestrate_format.delete('remote_id').to_s.upcase
      orchestrate_format['reviewed_files'] = normalize_reviewed_files(orchestrate_format.delete('files'))
      orchestrate_format['time_created'] = orchestrate_format.delete('event_time')
    end
  end

  # remove user, password. trailing slashes etc. from the repository urls
  # @param [Array<Hash>] files Array of reviewed files hashes
  def normalize_reviewed_files(files)
    files.to_a.each do |file|
      file['repository_url'] = normalize_repository_uri(file['repository_url'])
    end
  end

  # Validates the (optional) reviewer information.
  def validate_reviewers(request_data)
    request_data['reviewers'].to_a.each do |reviewer|
      raise MessageProcessingError, I18n.t('api.missing_fields') unless check_data(reviewer, [:username])
    end
  end

  # Validates the status type
  # @param [Hash] hash the hash to validate
  def validate_status_type(hash)
    # status_type data is required
    raise MessageProcessingError, I18n.t('api.missing_field', field: 'status_type') unless hash.has_key?('status_type') && !hash['status_type'].nil?

    status = hash['status_type']
    raise MessageProcessingError,
          I18n.t('api.invalid.type', type_name: 'status',
                 valid_types: ReviewRequest::VALID_STATUS_TYPES.join(', ')) unless ReviewRequest::VALID_STATUS_TYPES.include?(status)
  end

  # Validates the status type
  # @param [Hash] hash the hash to validate
  def validate_status_name(hash)
    # status_name data is required, this is the title displayed from the review
    raise MessageProcessingError, I18n.t('api.missing_field', field: 'status_name') unless hash.has_key?('status_name') && !hash['status_name'].nil?
  end

end
