#
# Copyright 2013 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

# Represents the users
class User
  include CtfUserAuthorization
  include MongoMapper::Document

  DEVISE_PARAMS = :database_authenticatable, :trackable, :omniauthable

  devise *DEVISE_PARAMS

  attr_accessible :uid, :provider, :name, :email, :password, :password_confirmation, :authentication_token

  # Omniauthable
  key :uid, String, required: true
  key :provider, String, required: true
  key :name, String, required: true
  validates_uniqueness_of :uid, :scope => :provider

  # Token Authenticable
  key :authentication_token

  ## Database authenticatable
  key :email, String, :required => true
  key :encrypted_password, String, :required => true

  ## Trackable
  key :sign_in_count, Integer, :default => 0
  key :current_sign_in_at, Time
  key :last_sign_in_at, Time
  key :current_sign_in_ip, String
  key :last_sign_in_ip, String

  #Store the user's specified timezone, default to Pacific zone until set at users first login
  key :time_zone, String, default: 'America/Los_Angeles'

  has_many :user_permissions
  key :refreshed_permissions_at, default: Time.at(0)

  key :super_user, Boolean, default: false

  validates_associated :user_permissions

  embedded_callbacks_off

  # Retrieve a list of active projects that the user has the provided permissions on.
  #
  # @param [Symbol] permission The permission to check against.
  # @return [Array[Project]] The list of projects that are active and the user has the provided permission in.
  def projects_with_permission(permission)
    if @super_user
      Project.active_projects
    else
      Project.active_projects_by_id(project_ids_with_permission(permission))
    end
  end


  # Refreshes permissions and sets the next update time to 20 hours from the current time
  # @param [TeamForge::Session || Hash] black_box, right now is either a CTFSession object (for CTF authorization) or a hash (for local auth)
  def refresh_permissions(black_box = nil)
    permission_method = "#{provider}_refresh_permissions"
    send(permission_method, black_box) if respond_to? permission_method
    update_attribute :refreshed_permissions_at, Time.now.utc
  end

  # Finds a user with the given omniauth authentication information or creates them if one does not already exist.
  # 'auth' is an AuthHash provided by OmniAuth; provider is the current OmniAuth
  # strategy in use.
  def self.find_or_create_user(auth)
    provider = auth.provider
    uid = auth.uid
    user = User.first(provider: provider, uid: uid)
    user ||= User.create(email: auth.info.email, name: auth.info.name,
                         uid: uid, provider: provider, password: Devise.friendly_token[0, 20])
    user.tap { |u| u.ensure_authentication_token }
  end

  #TODO migrate to the recommended devise implementation (https://gist.github.com/josevalim/fb706b1e933ef01e4fb6)
  # devise no longer supports this type of authentication so adding this to allows us to move forward
  def ensure_authentication_token
    if authentication_token.blank?
      self.authentication_token = generate_authentication_token
    end
  end

  # Updates the user name and email, since it may have changed in teamforge.
  def after_login(auth)
    self.name = auth.info.name
    self.email = auth.info.email
    save!
  end

  # Load a set of permissions from the provided file (in YAML format) and apply them to all the users; if a user isn't
  # specifically mentioned in the file, use the 'default' user permissions.
  # @param [String] permissions_file The path to the YAML file containing permissions
  def self.apply_permissions(permissions_file)
    stored_permissions = YAML.load_file(permissions_file)

    # Now we grant permissions to users according to the user_auth_store.yml file.
    User.all.each do |user|
      auth = stored_permissions[user.uid] || stored_permissions['default']
      user.refresh_permissions auth
    end
  end

  # Given a set of permissions, clear the user's permissions, and add all provided project permissions.
  #
  # @param [Hash] auth A collection of permissions in the form of { 'project_name' => 'permission'}
  def localauth_refresh_permissions(auth = nil)
    return unless auth
    self.user_permissions = []
    auth['project_permissions'].each do |project_name, permissions|
      project = Project.find_by_name(project_name)
      permissions.each do |permit|
        self.user_permissions << UserPermission.new(name: permit, project: project)
      end if project
    end
    save
  end

  private
  def generate_authentication_token
    loop do
      token = Devise.friendly_token
      break token unless User.where(authentication_token: token).first
    end
  end
end
