#
# Copyright 2013 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

class WorkItemSource < ActivitySource
  include Verifiable
  include Comparable

  key :event_time, Time
  key :url, String

  WORKITEM_SETTINGS_REQUIRED = [:url]

  belongs_to :project
  has_one :tracker
  many :work_items, foreign_key: :source_id

  validate :validate_url, allow_blank: true

  # case insensitive tracker regex
  def regex
    Regexp.new("(#{tracker.regex})", true)
  end

  def as_json(options={})
    # Get the API version
    #
    # Format:
    # {:prefixes=>["api/v1/trackers", "api/base"]}
    #
    api_version = 'v1'
    api_version = options[:prefixes][0].split('/')[1] if options[:prefixes].present?
    "#{api_version.camelize}::WorkItemSourceSerializer".constantize.new(self).serialize
  end

  def icon
    tracker.icon.to_s
  end

  def deactivatable?
    true
  end

  def handle_message(data_type, message)
    raise ProjectNotActive, I18n.t('project.messages.not_active', project_name: project.name) unless project.active?
    if data_type == :work_item
      handle_work_item_message(message)
    elsif data_type == :workitem_settings
      handle_tracker_state_message(message)
    end
  end

  # No longer need redirection through tracker since there is only one tracker per work item source
  def handle_work_item_message(work_item_hash)
    raise MessageProcessingError, I18n.t('work_item_source.errors.tracker_state_not_received', source_association_key: source_association_key) unless tracker.present?
    validate_event_time(work_item_hash['event_time'])
    validate_work_item_status(work_item_hash[:status_id])
    work_item_data = translate_to_orchestrate work_item_hash
    work_item = work_items.build(work_item_data)
    if work_item.save
      WorkItemSummary.create_or_update(work_item)
    else
      raise MessageProcessingError, "#{I18n.t('work_item_source.errors.could_not_save', source: display_name)}: #{work_item.errors.messages}"
    end
  end

  # processes the tracker state message and adds or replaces the tracker to this work item source
  def handle_tracker_state_message(tracker_state_hash)
    raise MessageProcessingError, I18n.t('api.missing_fields') unless check_data(tracker_state_hash, WORKITEM_SETTINGS_REQUIRED)

    validate_event_time(tracker_state_hash[:event_time])
    event_time = DateTime.rfc3339(tracker_state_hash[:event_time]).utc.to_time
    #TODO: think if there is a reason for the first check, seems redundant since self's event_time presumably
    #TODO: would have been run through the validate_event_time method before being set
    return if !self.event_time.nil? && event_time <= self.event_time

    translate_tracker_state_to_orchestrate(tracker_state_hash)
    self.event_time = event_time
    self.url = tracker_state_hash[:url]
    tracker_hash = tracker_state_hash[:tracker]

    add_or_replace_work_item_tracker(tracker_hash)

    if save
      AccessLogUpdater.update_log(self)
    else
      error_message = "#{I18n.t('work_item_source.errors.could_not_update')}: #{errors.messages}"
      full_error = "#{error_message} #{tracker_state_hash}"
      AccessLogUpdater.update_log(self, full_error)
      LogUtils.log_error_and_info(error_message, tracker_state_hash)
    end
  end

  # Sets tracker's :remote_id key to the value of :id in the received hash argument
  # arguments: receives a hash.
  # tracker_state_hash[:tracker] is a hash
  def translate_tracker_state_to_orchestrate(tracker_state_hash)
    tracker = tracker_state_hash[:tracker]
    tracker['remote_id'] = tracker.delete 'id'
    tracker['statuses'].each { |status| status['remote_id'] = status.delete 'id' }
  end

  # Converts hash to the format we're going to parse.
  # @param [Hash] work_item_data hash of incoming message to convert
  def translate_to_orchestrate(work_item_data)
    Hash[work_item_data].tap do |orchestrate_format|
      orchestrate_format['remote_id'] = orchestrate_format.delete 'id'
      orchestrate_format['time_created'] = orchestrate_format.delete 'event_time'
      orchestrate_format['remote_status_id'] = orchestrate_format.delete 'status_id'
    end
  end

  # Removes any existing and creates a new tracker from the tracker_state_message
  # @param [Hash] tracker_from_message, hash containing a single Tracker's attributes
  def add_or_replace_work_item_tracker(work_item_tracker_hash)
    #Delete the associated tracker since these no longer can be reused
    self.tracker.try(:delete)
    self.tracker = Tracker.new({work_item_source: self, active: true}.merge!(work_item_tracker_hash))
    raise MessageProcessingError, build_comprehensive_error(tracker) unless self.tracker.save
  end

  # Provide an error for each tracker that was (or became) invalid.
  def build_comprehensive_error(tracker)
    error_message = ""
    unless tracker.valid?
      error_message += "(#{tracker.remote_id}, #{tracker.name}, #{tracker.regex}) - "
      error_message += tracker.errors.messages.values.flatten.to_sentence + "\n"
    end
    error_message
  end

  def validate_work_item_status(work_item_status_id)
    status_ids = tracker.statuses.map(&:remote_id)
    raise MessageProcessingError, "The provided status_id (#{work_item_status_id}) was not found." unless status_ids.include? work_item_status_id
  end

  private
  def validate_url
    return if url.blank?
    !!URI.parse(url)
  rescue URI::InvalidURIError
    errors.add(:url, "invalid url")
  end

  # must be associated with no commit server or one commit servers
  def validate_commit_server_id
    return if commit_server_id.blank?
    flag = CommitServer.all.any? { |server| server.id == commit_server_id}
    errors.add(:commit_server_id, 'cannot be associated with a fictitious commit server' ) unless flag
  end

end
