class WorkItemSummary < ActivitySummary
  # Implements display_status to display the Work item trackers status
  include WorkItemStatus
  include WorkItemHelper

  key :remote_id, String
  key :remote_status_id, String

  delegate :attribute_changes, to: :latest_work_item

  key :summary, String
  key :description, String
  key :closed, Boolean
  key :deleted, Boolean
  key :priority, String
  key :assigned_to, String
  key :creation_time, Time
  key :tags, Array

  attr_accessible *(keys.keys.map(&:to_sym))

  alias :display_number :remote_id
  alias :associated_commit_summaries :activity_summaries
  alias :title :summary

  validates_presence_of :remote_id


  CTF_ARTF_REGEX = /artf\d{4,}/

  # Could not add created_by to the delegate list above because created_by is already defined in ActivitySummary
  def created_by
    latest_work_item.created_by if latest_work_item
  end

  def self.create_or_update(work_item)
    source_id = work_item.source_id
    work_item_summary = WorkItemSummary.where(source_id: source_id, remote_id: /^#{work_item.remote_id}/i).first
    
    # Check whether the work item summary already exists for this work_item and
    # moved to other tracker (Work Item Source)
    # if so, we need to re-link all existing associations with new work item summary.
    # assume that we only keep one work item summary for one remote_id
    work_item_summary ||= WorkItemSummary.where(remote_id: /^#{work_item.remote_id}/i).first
    work_item_summary ||= WorkItemSummary.new(source_id: source_id, remote_id: work_item.remote_id)

    unless work_item_summary.persisted? && work_item_summary.settled?
      work_item_summary.created_by = work_item.created_by
      work_item_summary.time_created = work_item.time_created
    end

    work_item_summary.source_id = source_id unless work_item_summary.source_id == source_id
    work_item_summary.activities << work_item
    work_item_summary.settled = true
    work_item_summary.remote_status_id= work_item.remote_status_id
    work_item_summary.summary = work_item.summary
    work_item_summary.description = work_item.description
    work_item_summary.closed = work_item.closed
    work_item_summary.deleted = work_item.deleted
    work_item_summary.priority = work_item.priority
    work_item_summary.assigned_to = work_item.assigned_to
    work_item_summary.creation_time = work_item.creation_time
    work_item_summary.tags = work_item.tags

    if work_item_summary.save
      # Activity needs to re-link with new tracker as well.
      Activity.set({:activity_summary_id => work_item_summary.id}, :source_id => source_id)
      work_item_summary
    else
      LogUtils.log_error_and_info(work_item_summary.errors.messages, work_item_summary.inspect)
      nil
    end
  end
  add_method_tracer :create_or_update

  # used when matching work item summary from the commit message
  def self.find_or_create(sources, remote_id)
    WorkItemSummary.where(source_id: sources.map(&:id), remote_id: /^#{remote_id}/i).first
  end

  def latest_work_item
    activities.order(:time_created.desc).first
  end

  # Returns the link-back url of this work item
  # @return [String] the link-back url
  def url
    URI.join(source.url + '/', remote_id).to_s
  end

  def status_icon_class
    if latest_work_item then latest_work_item.send(:status_icon_class) else '' end
  end

end