############################################################################################
# coupler service : This service couples the scheduler service to the legacy
# system
#            |                |                     |                  |
#            |                |   COUPLER SERVICE   |        LEGACY    |
#            | SCHEDULER      |                     |        SYSTEM    |
#            |                |                     |                  |
#            |                |                     |                  |
#
# Design Notes:
#
# If the variable Rails.application.config.BIGLOCK is less than or equal to 1 then the record/document
# insertion goes from the scheduler to the coupler service. The coupler service makes
# whatever modifications to the packet that are necessary to persist the document into
# the legacy system.
# The Interpreter and Scheduler Service have no knowledge at all about the legacy
# system. All they know is that if Rails.application.config.BIGLOCK is <= 1 then the data packet
# must be handed over to the coupler service after the scheduler has processed the packet
#
# If BIGLOCK == 2 then the packet is only persisted into the new system
# If BIGLOCK == 0 then the packet is only persisted into the legacy system
# If BIGLOCK == 1 then the packet is persisted into both systems
#
# Motivation: The stability of systems increases as executable code becomes less
# volatile. We want the executor and scheduler to be invariant code systems. Changes
# should be confined only to the coupling service
#
#
#  ALGORITHM NOTES ON STATUSES AND REMOTE IDS
#
#  [1] invoke the PacketInjection service to populate the incoming pattern in
#      conformity with the expectations of the legacy activities collection
#  [2] push the injected packet onto RabbitMQ or insert directly into the
#      activities collection
#
# copyright collabnet, 2015
############################################################################################

require 'singleton'

class CouplerService

  include Singleton

  ###################################################################################
  # interface: interface into the coupler service.
  # Receives (i)   a data packet (ruby object)
  #          (ii)  the packet type
  #          (ii)  a packet_mapper (hash)
  # returns the packet if the packet is pushed onto RabbitMQ; false otherwise
  ###################################################################################

  def interface(data_packet, packet_type, packet_mapper)

    ServiceLog.stack_trace('CouplerService::interface') if Rails.application.config.STACK_TRACE_COUPLER

    # basic validation on the data packet
    return false unless sanity_test?(data_packet, packet_type, packet_mapper)


    ##############################################
    # call the packet injection service to make
    # a conforming legacy packet for insertion
    # into the Activities collection. This
    # requires an applicable activity map
    # ret is false or [wis, packet]
    ##############################################

    ret = PacketInjector.instance.interface(data_packet, packet_type, packet_mapper)
    return false if ret.blank?

    #####################################################
    # replace dots in namespaced keys with the namespace
    # separator
    #####################################################
    ret[1] = map_namespace_separator(ret[1])


    ################################################
    # log in to RabbitMQ and push the packet into
    # the legacy system or do a direct insertion
    ################################################

    if [ 'ExternalRepository', 'CommitServer', 'CtfRepository' ].include?(ret[0]._type)
      queue_type = 'commits'
    elsif ret[0]._type == 'CustomActivitySource'
      queue_type = 'custom'
    else
      queue_type = ActivitySource.source_types.key(ret[0]._type).snakecase.pluralize
    end

    push_packets_to_legacy_system(ret[0].queue_username, ret[0].queue_password, queue_type, ret[1])

    # return the packet which was placed on Rabbit's Queue
    ret[1]

  end


  private

  ##################################################################################
  # sanity_test? tests whether the data packet passes basic validation
  ##################################################################################

  def sanity_test?(data_packet, packet_type, packet_mapper)
    ServiceLog.stack_trace('CouplerService::sanity_test?') if Rails.application.config.STACK_TRACE_COUPLER

    if data_packet.blank?
      ServiceLog.log('CouplerService::sanity_test? : data packet is blank')
      return false
    end

    if packet_type.blank?
      ServiceLog.log('CouplerService::sanity_test? : packet type is blank')
      return false
    end

    if packet_mapper.is_a?(Hash).blank?
      ServiceLog.log('CouplerService::sanity_test? : packet mapper is not an non-empty hash')
      return false
    end

    if data_packet.is_a?(Hash).blank?
      ServiceLog.log('CouplerService::sanity_test? : data packet is not of hash type')
      return false
    end

    # data packet must contain the source association key
    if data_packet['source_association_key'].blank?
      ServiceLog.log('CouplerService::sanity_test? : source association key not found for the data packet :' + data_packet.to_json)
      return false
    end

    true
  end


  #################################################################
  # direct_insertion: inserts the packet directly into the
  # activity collection without using RabbitMQ
  # NOTE: right now only works for work_items
  #################################################################

  def direct_insertion(packet_hash)
    ServiceLog.stack_trace('CouplerService::direct_insertion') if Rails.application.config.STACK_TRACE_COUPLER


    #----------------------------------------------
    # also insert the packet into the
    # legacy Activity Collection
    #----------------------------------------------
    begin
      if (wis = ActivitySource.find_by_source_association_key(packet_hash['sak']))
        if wis.tracker
          wis.handle_message(:work_item, packet_hash)
        end
      end
    rescue => e
      ServiceLog.log('CouplerService::direct_insertion : : ' + e.message)
      raise e.message
    end
  end

  #######################################################################
  # push_packets_to_legacy_system: pushes a data packet to rabbitMQ
  #######################################################################

  def push_packets_to_legacy_system(queue_username, queue_password, queue_type, data_packet)
    ServiceLog.stack_trace('CouplerService::push_packets_to_legacy_system') if Rails.application.config.STACK_TRACE_COUPLER


    if Rails.application.config.RABBITMQ_SWITCH
      RabbitConnector.instance.interface(queue_username, queue_password, queue_type, data_packet)
    else
      direct_insertion(data_packet)
    end

  end


  ############################################################
  # map_namespace_separator
  # mongoDB does not permit a document key to have dots in
  # the name. We namespace packet keys as (for example):
  #      fields.status.name
  # The namespace separator is used to substitute dots in a
  # namespaced key (i.e attribute)
  # returns the transformed packet
  ############################################################

  def map_namespace_separator(packet)
    ServiceLog.stack_trace('CouplerService::namespace_separator') if Rails.application.config.STACK_TRACE_COUPLER

    data = {}

    packet.map do  |key, value|
      if key.is_a? Symbol
        data[key] = value
        next
      end
      tmp = key.gsub('.', '_')
      data[tmp] = value
    end

    data
  end

end
