############################################################################################
# rabbit_connector : manages connections to RabbitMQ, pushes packets onto the rabbit queue
# Uses the singleton pattern
# copyright collabnet,2015
############################################################################################

require 'singleton'
require 'uri'
require 'bunny'


class RabbitConnector

  include Singleton

  ################################################################
  # interface for inter-operating with RabbitMQ
  # returns true if the packet is pushed onto Rabbit, otherwise
  # false
  ################################################################

  def interface(user, password, queue_type, data_packet)

    ServiceLog.stack_trace('RabbitConnector:interface') if  Rails.application.config.STACK_TRACE_RABBIT_CONNECTOR

    # Make the connection string
    rabbit_uri = URI.parse(Rails.configuration.rabbitmq[:uri])
    host = rabbit_uri.host
    port = rabbit_uri.port

    connection_string = 'amqp://' + user + ':'+ password + '@' + host + ':' + port.to_s

    # instantiate a connection
    return false unless open_connection(connection_string, queue_type)

    # push packet and close the connection
    ret = push_packet(data_packet)
    close_connection

    ret
  end


  private

  ################################################################
  # open_connection:
  # connect to RabbitMQ return false if a connection cannot
  # be established
  ################################################################

  def open_connection(connection_string, queue_type)
    ServiceLog.stack_trace('RabbitConnector:open_connection') if  Rails.application.config.STACK_TRACE_RABBIT_CONNECTOR

    begin
      @connection = Bunny.new(connection_string)
      @connection.start
    rescue => e
      ServiceLog.log("RabbitConnector::open_connection : failed to connect to rabbit: " + e.message)
      return false
    end

    begin
      @channel  = @connection.create_channel
      @exchange = @channel.topic(EVENT_HUB_NAME, durable: true)
      @queue = @channel.queue(@queue_name, durable: true).bind(@exchange, :routing_key => @queue_name)

    rescue => e
      ServiceLog.log("RabbitConnector::open_connection : connection or queue setup failed: " + e.message)
      return false
    end

    true
  end


  ################################################################
  # push_packet: pushes the data packet onto the RabbitMQ
  # receives a data packet(ruby hash object)
  # Return false if the packet is not pushed
  ################################################################

  def push_packet(data_packet)

    ServiceLog.stack_trace('RabbitConnector:push_packet') if  Rails.application.config.STACK_TRACE_RABBIT_CONNECTOR

    begin
      @exchange.publish(data_packet.to_json, :routing_key => @queue.name)
    rescue => e
      ServiceLog.log('RabbitConnector::push_packet : failed to publish packet: ' + e.message)
      return false
    end

    ServiceLog.log("RabbitConnector::push_packet : SENT a packet to Rabbit")

    true
  end


  ################################################################
  # close_connection: close the connection to RabbitMQ
  # return true or false
  ################################################################

  def close_connection
    ServiceLog.stack_trace('RabbitConnector:close_connection') if  Rails.application.config.STACK_TRACE_RABBIT_CONNECTOR
    begin
      @channel.close
      @connection.close
    rescue => e
      ServiceLog.log('RabbitConnector::push_packet : failed to close connection to RabbitMQ: ' + e.message)
      return false
    end

    true
  end

  end



