#################################################################################################
# datadump: Persists data packets into the DataDump with minimal processing
#           Only HTTP POST operations are permissible
# This service receives (i) the data packet, (ii) URL of the external data source
#
# POST /api/3/projects/:project_id/webhook/packets/:packet_type/:sak/:user_param
#
# Algorithm
#     [1] Receives a Ruby Hash Object from the REST interface, receives the URL of the incoming
#         external source from the  REST interface
#     [3] Deconstructs the received URL and adds the following keys to the Ruby Hash object:
#         sak, packet_type: project_id, packet time, machine_code, cryptographic_hash to the object.
#     [3] Stringifies the object to JSON and then persists it to the data dump.
#     [4] Sends the packet to the scheduler for further processing
#     [4] Returns false if there are errors in processing the packet
#
# copyright collabnet,2015
#################################################################################################


require 'uri'
require 'digest'


class DataDump

  include MongoMapper::Document


  ###############################################################################
  # interface
  # receives (i)   packet of data destined for persistence (ruby hash object)
  #          (ii)  url (string) of the external source from which the data
  #                packet was received
  #
  # returns false if the data packet is not persisted; otherwise the array
  #    [views_inserted, views_queued,  projected_packet]

  # Valid data packet urls have form:
  #     POST  http://www.teamforge.com/version/:version/project/:project_id/socket/packets/:packet_type/:sak/:user_param
  #
  # packet_type is completely user specified and not activity_type
  # NOTE: packet_type is also the name of the default pattern for the
  # incoming packet. Each incoming packet must have a default pattern associated
  # with it
  ###############################################################################

  def self.interface(payload, url)

    ServiceLog.stack_trace('DataDump::interface') if  Rails.application.config.STACK_TRACE_DATADUMP

    return false unless validate_payload?(payload)

    #-------------------------------------
    # persist the packet
    #-------------------------------------

    url_object = self.make_url_object(url)
    return false if url_object.blank?

    return false unless self.packet_operations(payload, url_object)
    return self.persist_data_packet(payload, url_object)

  end


  private

  ##########################################################################################
  # validate_url: examines the URL object to determine if the URL is valid
  # Receives the external source URL as a deconstructed ruby hash object
  # Returns false if the URL is not in a recognized format. true otherwise
  #
  ###########################################################################################

  def self.validate_url?(url_object)

    ServiceLog.stack_trace('DataDump::validate_url?') if  Rails.application.config.STACK_TRACE_DATADUMP

    # initiate URL validation
    if url_object["scheme"].blank?
      ServiceLog.log("datadump : validate_url?: scheme not specified")
      return false
    end

    if url_object["host"].blank?
      ServiceLog.log("datadump : validate_url?: host not specified")
      return false
    end

    # the scheme must be http or https
    if url_object["scheme"] != 'http' && url_object["scheme"] != 'https'
      ServiceLog.log("datadump : validate_url?: scheme is not http or https")
      return false
    end

    # return false if the url contains a fragment
    if url_object["fragment"].present?
      ServiceLog.log("datadump : validate_url?: there is a fragment in the url")
      return false
    end

    # return false if there is no path specified in the URL
    if url_object["path_array"].empty?
      ServiceLog.log("datadump : validate_url?: path is empty")
      return false
    end

    return true

  end


  #################################################################
  # make_url_object
  # receives a URL string
  # returns: a ruby hash object comprising of the parts of the URL
  ##################################################################

  def self.make_url_object(url)

    ServiceLog.stack_trace('DataDump::make_url_object') if  Rails.application.config.STACK_TRACE_DATADUMP

    scheme, host, query, fragment = nil

    uri        = URI(url)
    scheme     = uri.scheme.strip.downcase if uri.scheme
    host       = uri.host.strip.downcase if uri.host
    path_array = uri.path.split('/')
    query      = uri.query.strip.downcase if uri.query
    fragment   = uri.fragment.strip.downcase if uri.fragment

    if path_array.length - 1 == '/'
      path_array = path_array.delete("path_array".length - 1)
    end

   url_object = {
      'scheme' => scheme,
      'host' =>   host,
      'path_array' => path_array,
      'query' => query,
      'fragment' => fragment
  }

    url_object
  end




  ##############################################################################
  # validate_payload: receives a data packet object and verifies that it can be
  # converted to a json string object. Catches errant packets
  # receives: (i)  a ruby object
  ##############################################################################

  def self.validate_payload?(payload)

    ServiceLog.stack_trace('DataDump::validate_payload?') if  Rails.application.config.STACK_TRACE_DATADUMP

    begin
      payload.to_json
    rescue => e
      ServiceLog.log("datadump : validate_payload? : exception converting data packet to json string" + e.message)
      return false
    end

    true
  end



  ##########################################################################################
  # packet_operations: validate and persist packets to the datadump
  # receives the data packet (ruby object) and the external source URL converted into a
  # Ruby Hash object
  # return true or false
  #########################################################################################

  def self.packet_operations(packet, url_object)

    ServiceLog.stack_trace('DataDump ::packet_operations') if  Rails.application.config.STACK_TRACE_DATADUMP

    # validate the url that was received
    return false unless self.validate_url?(url_object)

    true
  end


  ########################################################################
  # get_default_pattern
  # returns the default pattern pertaining to a packet. The name
  # of the default pattern is the packet type of the packet, it is
  # at index position 7 of the url array
  # return false if the default pattern cannot be extracted
  ########################################################################

  def self.get_default_pattern(url_array)

    ServiceLog.stack_trace('DataDump::get_default_pattern') if  Rails.application.config.STACK_TRACE_DATADUMP

    return false if url_array.length < 7 || url_array[7].blank?
    return url_array[7]
  end


  ##############################################################################
  # persist_data_packet:
  # Adds additional attributes to the packet from the path components of
  # the URL received by the interface of this service object. Persists
  # the packet as a JSON string to the data dump
  # invokes the scheduler for further processing of the packet (if necessary)
  # by appending to a view or creating a new view
  #
  # receives:  the data packet
  #            the URL Object
  #
  # Returns false if the packet is not persisted, an array if the packet is
  # persisted or true if there is no error but the packet is not persisited
  #
  # Attributes Added:
  #  [1] SAK
  #  [2] packet_type
  #  [3] project_id
  #  [4] default pattern name
  #  [5] id of packet
  #  [5] machine_code - for machine learning, clustering, neural networks etc.
  #  [6] time packet received in epoch time
  #  [7] cryptographic_hash - message digest of the received packet
  #  [8] the query string
  #
  # NOTE: The source type semantics are set by the user of the system: example:
  #       jira. trac, github etc. It does not correspond to the activity type in the
  #       legacy system
  #
  #  canonical url for external data
  #  POST  http://www.teamforge.com/api/:version/:project_id/:type/:sak/:string
  #
  # Note: the legacy system requires an event_time attribute transformable into
  # rfc3339 format
  #
  #############################################################################

  def self.persist_data_packet(packet, url_object)
    ServiceLog.stack_trace('DataDump::persist_data_packet') if  Rails.application.config.STACK_TRACE_DATADUMP

    path_array = url_object["path_array"]

    #-------------------------------------------------------
    # get the default pattern pertaining to the source type
    #-------------------------------------------------------
    path_array[6].downcase!
    return false if get_default_pattern(path_array).blank?

    #----------------------------------------------
    # transform nil values in the data packet
    # to empty strings
    #----------------------------------------------

    packet = nil_transformer(packet)
    return false if packet.blank?


    #----------------------------------------------
    # add these fields only after computing the
    # message digest
    # path_array are extractions from the URL
    #----------------------------------------------

  #  http://www.teamforge.com/version/:version/project/:project_id/socket/packets/:packet_type/:sak/

    packet['project_id']   = path_array[path_array.length - 5]
    packet['packet_type']  = path_array[path_array.length - 2]
    packet['query_string'] = url_object["query"]
    packet['source_association_key'] = path_array[path_array.length - 1]

    # for analytics engine, value is some differentiating
    # evaluable string
    packet['machine_code'] = ""

    #------------------------------------------------------
    # each packet has a default pattern associated with it
    # this default pattern is called packet['packet_type']
    #-----------------------------------------------------
    packet['pattern_name'] = packet['packet_type']

    # always timestamp the packet
    packet['packet_time']  = Time.now.to_i

    #-------------------------
    # Persist to the DataDump
    #-------------------------
    begin
      DataDump.create(packet);
    rescue => e
      ServiceLog.log('datadump::persist_packet : failed to insert packet into the datadump : ' + e.message)
      return false
    end

    #-------------------------------------------------------------
    # invoke the Scheduler to further process the packet
    # the scheduler will return either true, false or
    # the array [views_inserted, views_queued,  projected_packet]
    #-------------------------------------------------------------
    ret = Scheduler.interface("run", nil, nil, packet)

    ret

  end


  ##########################################################################
  # nil_transformer: converts nil values in a data packet to empty strings
  # receives a data packet (Hash and possibly a nested hash)
  # returns the mutated object  or false if there is an error
  # Algorithm:
  #  [0] note only values can be nil, a key cannot be nil
  #  [1] convert the data packet to a json string object
  #  [2] use regex to convert nil values to empty string
  ##########################################################################

  def self.nil_transformer(packet)
    ServiceLog.stack_trace('DataDump::nil_transformer') if Rails.application.config.STACK_TRACE_DATADUMP

    # note: to_json changes nil to null
    begin
      json_string = packet.to_json
    rescue => e
      ServiceLog.log('datadump::nil_transformer:  packet object cannot be cannot changed to a json: ' + e.message)
      return false
    end

    return packet if json_string.include?('null').blank?

    json_string = json_string.gsub(/:\s*null/, ":\"\"")

    # convert the string to a ruby hash
    begin
      packet = JSON.parse(json_string)
    rescue => e
      ServiceLog.log('datadump::nil_transformer: cannot parse the json string to a ruby object: ' + e.message)
      false
    end

    packet
  end

end
