########################################################################################
# Emulator Service: performs  dynamic translation and adaptor emulation services prior to
# persisting a packet into a view or a legacy collection. Also enables CRUD operations.
# Functionality:
#      (i)   CRUD operations on emulation tables
#      (ii)  transform data packets with emulation tables
# Receives a emulation packet from the data dump as a ruby hash object.
# Transforms the  keys of the object in accordance with an emulation table.
# Designed to emulate some of the functionality of legacy adaptors
##
#   Emulation Hash Object Structure:
#     #       "packet type" =>  'some packet type',
#
#       'emulation' => [
#               { "old_key"   => 'old_namespaced_key',
#                 'new_key'   => 'new_namespaced_key',
#               },
#                   ....
#           ]         .
#                           .
#        'parser' => [ {
#                       'input_key' => { 'name' => value,
#                                         "metadata" => 'static_string or namespaced_value'
#                                        }
#                                          ...
#
#                       'output_keys'  => [ { 'name' => some_output_key name,',
#                                             'value'=> "some_value",
    #                                         "metadata" => 'static_string or parsed_key_value',
#                                             "head_key" => 'x,y,z',
#                                             'container' => 'hash or array,'
#                                    }
    #                                          ...
    #                                 ]
    #                             }
    #                      ...
    #               ]
    #    }

    #
# Explanation
#
#
#   (i)    emulator type: A new key will be created. Its value is the value of
#          some.other.namespaced.key If the transform is dynamic. If the transform is  then the
#          namespaced new key is resolved to its value - which becomes the new key (name)
#          if the transform is static the actual value is used as the new key name
#   (ii)   the parser  type creates a hash or array data structure which is the value
#          of the head_key
#
#
# some_packet_type is the packet type (which is extractable from the URL hitting the REST interface)#
#
# If the packet_type is '*' the key in the emulator table will be applied to all
# received packets
#
# NOTE: The Emulator must be called BEFORE the Activity Map has operated upon the data packet
# copyright: collabnet, 2015
#########################################################################################


class Emulator

  include MongoMapper::Document

  ##################################################################
  # Interface To The Emulator
  # Receives: [i]   an emulation reference which is either an optional
  #                 ruby hash object,
  #           [ii]  the type of the packet
  #           [ii]  a verb which indicates the type of operation
  #                 to be performed on the emulation table
  #
  # verbs:  GET, DELETE, PUT, POST, RUN
  #
  # [1]  The RUN verb runs the emulator against a data packet which is
  #      received at the interface
  #
  # return false if there is an error otherwise returns the
  # transformed data packet as a ruby hash object
  #
  ###################################################################


  def self.interface(emulation_reference = nil, packet_type = nil,  verb = nil)

    ServiceLog.stack_trace('emulator::interface') if Rails.application.config.STACK_TRACE_EMULATOR

    if verb.blank?
      ServiceLog.log('Emulator::interface : the emulator must receive a verb parameter')
      return false
    end

    verb.upcase!
    verb.strip!


    #########################################
    # sanity test
    #########################################

    if verb != 'RUN'
      return false unless self.sanity_test?(emulation_reference, packet_type, verb)
    else
      self.run_tests?(emulation_reference, packet_type)
    end


    #############################################
    # validate any emulation packets received
    # on the interface
    #############################################


    if (verb == 'POST' || verb == 'GET')  && emulation_reference.present?
      return false unless self.valid_emulation_packet?(emulation_reference)
    end


    ###################################
    # get all of the emulation tables
    ###################################
    if verb == 'GET'
      return self.get_emulation_tables


    ###################################
    # get a  emulation table
    ###################################
    elsif verb == 'GET' && packet_type.present?
      return self.get_emulation_table(packet_type)


    ###############################
    # create an emulation table
    ###############################

    elsif verb == 'POST'
      return self.create_emulation_table(emulation_reference)

    ###############################
    # update an emulation table
    ###############################

    elsif verb == 'PUT'
      return self.update_emulation_table(emulation_reference, packet_type)


    ###############################
    # delete an emulation table
    ###############################

    elsif verb == 'DELETE'
      return self.delete_emulation_table(packet_type)


    ##################################################
    # perform the transformations of the data packet
    ##################################################

    elsif verb == 'RUN'
      if packet_type.blank?
        ServiceLog.log('Emulator::interface : packet type not specified for the RUN verb')
        return false
      end

      return self.transform(emulation_reference, packet_type)
    end

    ServiceLog.log("Emulator::interface : unknown verb:  #{verb}" )
    false
  end


  private


  #############################################################################
  # sanity_test? - tests the validity of parameters received on the interface
  # Only executed for putative emulation packet
  #############################################################################

  def self.sanity_test?(emulation_reference, packet_type, verb)

    ServiceLog.stack_trace('emulator::sanity_test?') if Rails.application.config.STACK_TRACE_EMULATOR

    #---------------------
    # POST Verb
    #---------------------
    if verb == 'POST'

      if emulation_reference.blank?
        ServiceLog.log('Emulator::sanity_test? : an emulator hash was not specified along with POST verb')
        return false
      end

      if emulation_reference.is_a?(Hash).blank?
        ServiceLog.log('Emulator::sanity_test? : emulation table is not a hash object; a POST verb was received')
        return false
      end

      if emulation_reference == {}
        ServiceLog.log('Emulator::sanity_test? : empty emulation table specified along with POST verb')
        return false
      end

      if emulation_reference['emulation'].blank? || emulation_reference['parser'].blank?
        ServiceLog.log('Emulator::sanity_test? :  one of the emulation or parser keys is required in the emulation object')
        return false
      end

      if emulation_reference['emulation'].is_a?(Array).blank?
        ServiceLog.log('Emulator::sanity_test? :  the emulation key value is not of type array')
        return false
      end

      if emulation_reference['parser'].is_a?(Array).blank?
        ServiceLog.log('Emulator::sanity_test? :  the parser key value is not of type array')
        return false
      end

    end

    #-------------------------
    # Delete Verb
    #-------------------------
    if verb == 'DELETE'
      if packet_type.class != String
        ServiceLog.log('Emulator::sanity_test? : a search key of type string not received for a DELETE verb parameter')
        return false
      end

      if packet_type.blank?
        ServiceLog.log('Emulator::sanity_test? : search key  not received for a DELETE verb parameter')
        return false
      end
    end

    #------------------------
    # PUT Verb
    #------------------------
    if verb == 'PUT'

      if emulation_reference.blank? || packet_type.blank?
        ServiceLog.log('Emulator::sanity_test? : missing parameters for PUT verb')
        return false
      end


      if emulation_reference.is_a?(Hash).blank?
        ServiceLog.log('Emulator::sanity_test? : emulator table is not a Hash type for the PUT verb parameter')
        return false
      end

      if packet_type.is_a?(String).blank?
        ServiceLog.log('Emulator::sanity_test? : search key  not a String type for the PUT verb parameter')
        return false
      end
    end

    #-----------------------
    # GET Verb
    #-----------------------
    if verb == 'GET'
      if packet_type.is_a?(String).blank? && packet_type.present?
        ServiceLog.log('Emulator::sanity_test? : invalid parameter in interface for the GET verb')
        return false
      end
    end

    true
  end


  ##############################################################################
  # run_tests?: tests performed when the RUN verb is received on the interface
  ##############################################################################

  def self.run_tests?(data_packet, packet_type )

    ServiceLog.stack_trace('emulator::run_tests?') if Rails.application.config.STACK_TRACE_EMULATOR

    if data_packet.blank?
      ServiceLog.log('Emulator::run_tests? : missing data packet for RUN verb')
      return false
    end

    if packet_type.blank?
      ServiceLog.log('Emulator::run_tests? : missing packet_type for RUN verb')
      return false
    end

    true

  end


  ##############################################################################
  # validate_emulation_packet?: verifies that an emulation packet (received
  # through the REST interface) has the proper format.
  # A valid emulation packet has format:
  #
  #
=begin
     #   Emulation Hash Object Structure:
     #   {
     #       "packet type" =>  'some packet type',
     #
     #       'emulation' => [
     #               { "old_key"   => 'old_namespaced_key',
     #                 'new_key'   => 'new_namespaced_key'
     #               },
     #                   ....
     #           ]         .
     #                           .
     #        'parser' => [ {
     #                       'base_node' => 'x,y,z',
     #                       'head_key'  => 'a.namespaced.key',
     #                       'container' => 'hash or array,'
     #                       'input_key' => { 'name' => value,
     #                                         "metadata" => 'static_string or namespaced_value'
     #                                        }
     #                                          ...
     #
     #                       'output_keys'  => [ { 'name' => some_output_key_not_namespaced',
                                                   'value'=> "some_value",
     #                                             "metadata" => 'static_string or parsed_key_value'
     #                                    }
     #                                          ...
     #                                 ]
     #                             }
     #                      ...
     #               ]
     #    }
     #
=end
     #
  ##############################################################################

  def self.valid_emulation_packet?(emulation_packet)

    ServiceLog.stack_trace('emulator::valid_emulation_packet?') if Rails.application.config.STACK_TRACE_EMULATOR

    #----------------------
    # test the packet type
    #----------------------
    if emulation_packet['packet_type'].blank?
      ServiceLog.log('Emulator::valid_emulation_packet : emulation packet does not contain a packet_type key')
      return false
    end

    return false unless self.validate_substitutions?(emulation_packet['emulation'])
    

    #----------------------------------
    # validate the head node - it is
    # a comma delimited string that
    # has at least one value
    #----------------------------------

    if emulation_packet['parser'].present?
      emulation_packet['parser'].each do |obj|
          if obj['input_key']["head_key"].strip.blank?
            ServiceLog.log('Emulator::valid_emulation_packet : the head key string is empty in the parser')
            return false
          end
       end
      end

    return true
    
  end


  #############################################################################
  # validate_substitutions: validates the substitutions that are to be done
  # by the emulator. These substitutions are specified under the emulator
  # key
  # receives the empulation_packet["emulation"] array
  #############################################################################

  def self.validate_substitutions?(emulator_array)

    ServiceLog.stack_trace('emulator::validate_substitutions') if Rails.application.config.STACK_TRACE_EMULATOR

    #-----------------------------------------
    # iterate through the emulation key array
    #-----------------------------------------

    if emulator_array.present?
      if emulator_array.is_a?(Array).blank?
        ServiceLog.log('Emulator::validate_substitutions? : emulation key does not have an array value')
        return false
      end

      emulator_array.each do |el|
        el.keys.each do |key|
          if key.blank?
            ServiceLog.log('Emulator::validate_substitutions?: emulation type array contains a blank key')
            return false
          end
        end

      end
    end

    true
  end


  #############################################################################
  # validate_parser_object
  # parses the emulation parser array object
  # receives the emulator parser array - emulation_packet['parser']
  #############################################################################

  def self.validate_parser_object?(parser_hash)

    ServiceLog.stack_trace('emulator::validate_parser_object') if Rails.application.config.STACK_TRACE_EMULATOR

    #-----------------------------------
    # iterate through the parser array
    #-----------------------------------
    parser_hash.each do |el|

        #----------------------------------------------
        # iterate through and validate the output keys
        #----------------------------------------------
        el['output_keys'].each do | output_key |

          if output_key['head_key'].blank?
            ServiceLog.log('Emulator::validate_parser_object?: emulation parser contains a blank head key')
            return false
          end
          if ['array', 'hash'].include?(output_key['container'].downcase).blank?
            ServiceLog.log('Emulator::validate_parser_object?: emulation parser does not have an array or hash container')
            return false
          end

          if output_key['metadata'].blank? ||
              ['namespaced_value', 'static_string', 'parsed_key_value'].include?(output_key['metadata'].downcase)
            ServiceLog.log('Emulator::validate_parser_object?: emulation parser input_key does not have valid metadata')
            return false
          end

          if output_key['name'].blank?
            ServiceLog.log('Emulator::validate_parser_object?: emulation parser output_key is blank')
            return false
          end
        end

        # validate the input name key
        if el['input_key']['name'].blank?
          ServiceLog.log('Emulator::validate_parser_object?: emulation parser input_key does not have a name key value')
          return false
        end

        # validate the input metadata key
        if el['input_key']['metadata'].blank? ||
            ['namespaced_value', 'static_string'].include?(el['input_key']['metadata'].downcase)
          ServiceLog.log('Emulator::validate_parser_object?: emulation parser input_key does not have valid metadata')
          return false
        end
    end
  end


  #############################################################################
  # create_emulation_table: creates and returns the new emulation table
  # receives a emulation table as a ruby hash object
  # persists the object to the Emulator collection
  # returns false if the object cannot be persisted otherwise returns the new
  # table
  #
  ###############################################################################

  def self.create_emulation_table(new_emulation_table)

    ServiceLog.stack_trace('emulator::create_emulation_table') if Rails.application.config.STACK_TRACE_EMULATOR

    # delete an existing matching emulation table
    self.delete_emulation_table(new_emulation_table['packet_type'])

    #----------------------------------------
    # Persist to the Emulator Collection
    #----------------------------------------
    begin
      Emulator.create(new_emulation_table);
    rescue => e
      ServiceLog.log('Emulator::create_emulation_table : emulation table was not persisted to the Emulator collection' + e.message)
      return false
    end

    new_emulation_table

  end


  ##############################################################################
  # delete_emulation_table: deletes an emulation table from the Emulator
  # collection.
  # Receives a packet_type string which is the search key
  # returns true or false
  ##############################################################################

  def self.delete_emulation_table(packet_type)

    ServiceLog.stack_trace('Emulator::delete_emulation_table') if Rails.application.config.STACK_TRACE_EMULATOR

    return true if Emulator.count == 0

    array = Emulator.all(:packet_type => packet_type).to_a
    return false if array.count == 0

    begin
      Emulator.destroy_all(:packet_type => packet_type)
    rescue => e
      ServiceLog.log('Emulator: destroy_emulation_table : failed to delete emulation : ' + e.message)
      return false
    end

    true
  end


  ##############################################################################
  # get_emulation_tables: gets the entire emulation table from the Emulator
  # collection.
  # returns false or an array of emulation tables
  ##############################################################################

  def self.get_emulation_tables()

    ServiceLog.stack_trace('Emulator::get_emulation_tables') if Rails.application.config.STACK_TRACE_EMULATOR

    # return the empty array if there are no emulation tables
    return [] if Emulator.count == 0

    begin
      tables = Emulator.all.to_a
    rescue => e
      ServiceLog.log('Emulator: get_emulation_tables : failed to get all emulation tables: ' + e.message)
      return false
    end

    tables
  end


  ##############################################################################
  # get_emulation_table: gets an emulation table from the Emulator
  # collection based on a search key
  # receives a search key
  # returns false or an array of emulation tables
  ##############################################################################

  def self.get_emulation_table(search_key)

    ServiceLog.stack_trace('Emulator::get_emulation_table') if Rails.application.config.STACK_TRACE_EMULATOR

    return [] if Emulator.count == 0

    begin
      table = Emulator.where( 'packet_type' => search_key ).to_a

    rescue => e
      ServiceLog.log("Emulator: get_emulation_table : failed to get the emulation table, search_key : #{search_key} - " + e.message)
      return false
    end

    return false if table.length == 0
    table
  end


  ##############################################################################
  # update_emulation_table: updates an emulation table
  # receives an emulation_table and a search key
  # returns false or the updated emulation table as an array
  ##############################################################################

  def self.update_emulation_table(updated_emulation_table, search_key)

    ServiceLog.stack_trace('Emulator::update_emulation_table') if Rails.application.config.STACK_TRACE_EMULATOR

    emulator_array = Emulator.all('packet_type' => search_key)

    if emulator_array == []
      ServiceLog.log('Emulator: update_emulation_table : failed to get the emulator for the search_key : ' + search_key)
      return false
    end

    self.delete_emulation_table(emulator_array[0].packet_type)
    self.create_emulation_table(updated_emulation_table)

  end


  ##############################################################################
  # transform: transforms attribute names (keys) in the data packet by reference
  # to the emulation table. Also creates new data structures for inclusion
  # in a data packet by reference to the parser key in an emulation structure
  # if there is no emulator for the packet type, the data packet is returned
  #
  # Receives [1] a  data packet (ruby hash object)
  #          [2] the type of packet
  # returns the transformed data packet. If an emulation record does not
  # exist for the data packet returns the untransformed data packet
  #############################################################################$

  def self.transform(data_packet, packet_type)

    ServiceLog.stack_trace('emulator::transform') if Rails.application.config.STACK_TRACE_EMULATOR

    #------------------------------------------------
    # get the emulator for the type of packet
    # that has been received
    #-----------------------------------------------

    emulator = self.get_emulation_table(packet_type)
    return data_packet if emulator.blank?
    emulator = emulator[0]

    #--------------------------------------------
    # do the key transformations for the packet
    #--------------------------------------------
     data_packet = self.key_transformations(data_packet, emulator)

    #--------------------------------------------
    # create a new data structure with the parser
    #--------------------------------------------

    if emulator["parser"].present?
      data_packet = self.create_data_object(data_packet, emulator['parser'])
    end

    data_packet

  end


  #########################################################################
  # key_transformations: performs key transformations on a data packet
  # though an emulator
  # returns the transformed data packet
  #
  # uses the following portion of the emulator data structure
  #       {
  #        "packet type" =>  'some packet type',
  #
  #        'emulation' => [
  #                        {
  #                          "old_key"   => 'old_namespaced_key',
  #                          'new_key'   => 'new_namespaced_key'
  #                         },
  #                          ...
  #                       ]
  #       }
  #
  #########################################################################

  def self.key_transformations(data_packet, emulator)

    ServiceLog.stack_trace('emulator::key_transformations') if Rails.application.config.STACK_TRACE_EMULATOR

    begin
      emulator["emulation"].each do |element|

        #--------------------------------------------
        # insert the new key into the data packet if
        # the old_key exists in the data packet
        #--------------------------------------------

        if data_packet[element["old_key"]].present?
            data_packet[element["new_key"]] = data_packet[element['old_key']]
          # delete the old_key from the data packet
          data_packet = data_packet.tap { |hash| hash.delete(element["old_key"]) }
        end
      end
    rescue => ex
      ServiceLog.log('Emulator: key_transformations : exception caught: ' + ex.message)
      return false
    end

    data_packet

  end

  #########################################################################
  # create_data_object: creates a data object for insertion into a data
  # packet with the aid of the parser in the emulator object
  # Returns the data packet with the inserted object
  #
  #  Uses this portion of the emulator object
  #
  #        'parser' => [ {
  #
  #                       'head_key'  => 'a.namespaced.key',
  #                       'container' => 'hash or array,'
  #                       'input_key' => { 'name' => value,
  #                                         "metadata" => 'static_string or namespaced_value'
  #                                        }
  #                                          ...
  #
  #                       'output_keys'  => [ { 'name' => some_output_key_not_namespaced',
  #                                             'value'=> "some_value",
  #                                             "metadata" => 'static_string or namespaced_value or parsed_key_value'
  #                                    }
  #                                          ...
  #                                 ]
  #                             }
  #                      ...
  #               ]
  #########################################################################

  def self.create_data_object(data_packet, parser)

    ServiceLog.stack_trace('emulator::create_data_object') if Rails.application.config.STACK_TRACE_EMULATOR

    container = nil

      #--------------------------------------
      # iterate through the parser creating
      # data objects
      #--------------------------------------
      parser.each do |el|

        #---------------------------------
        # iterate through the output keys
        #---------------------------------

        output_keys_array = []

        el['output_keys'].each_with_index do |output_key, index|

          # persist the  output_key
          output_keys_array << output_key

        end

        #--------------------------------------------
        # synthetic processing
        #-------------------------------------------

        tmp = self.synthetic_transformer(data_packet, parser)
        return false if tmp.blank?

        # construct container
        container = self.construct_container(tmp, el["input_key"]["head_key"] )
        return false if container.blank?

        container.merge!(data_packet)
      end

    container
  end


  ##########################################################################
  # construct_container: constructs the container in which the
  # custom data structure will be built,
  #
  # receives: (1) the data object to be inserted into the container
  #           (2) parse object
  # returns the container with the inserted data object
  ##########################################################################

  def self.construct_container(data_packet,  head_key)

    ServiceLog.stack_trace('emulator::construct_containert') if Rails.application.config.STACK_TRACE_EMULATOR


    key_array = head_key.split(',').compact

    #------------------------------------------------------
    # build a nested hash key container from the head_key
    # comma delimited string
    #-----------------------------------------------------
    key_array.reverse.inject(data_packet) { |a, n| { n => a } }

  end



  ################################################################################
  # synthetic_transformer
  # Receives : (i)    the data packet to be transformed
  #            (ii)   the output_keys array in the emulation object
  #            (iii)  the input key from the emulation object
  #
  # Motivation: So we can build different transformer methods for an emulation
  # object and these transformers then generate a data structure object for each
  # data packet received. The transformer is just a mathematical function such as
  #
  #          transformedPacket = synthetic_transformer(data_packet, emulator)
  #
  #################################################################################

  def self.synthetic_transformer(data_packet, parser_array)

    ServiceLog.stack_trace('emulator::synthetic_transformer') if Rails.application.config.STACK_TRACE_EMULATOR

    result_object = []


    #-------------------------------------------------------
    # iterate through the parser array
    #-------------------------------------------------------

    begin

      parser_array.each do |parser|

        values_array  = []


        #------------------------------------------------------
        #  create a abstract transformation based on the
        #  output_key object objects in the emulation object
        #------------------------------------------------------

        parser["output_keys"].each do |obj|
          if obj['metadata'] == 'namespaced_value'
            tmp = data_packet[parser["input_key"]["name"]]
            if tmp.is_a?(String)
              values_array << tmp
            elsif tmp.is_a?(Array)
              values_array = values_array + tmp
            else
              next
            end
          end
        end

        values_array.uniq!


        #-----------------------------------------------------------
        # process input_key and get the parsed leaf of the name key
        #----------------------------------------------------------
        input_name = parser["input_key"]['name']
        parts      = input_name.split('.')
        leaf_key  = parts[parts.length - 1]



        #---------------------------------------------
        # iterate through the values array
        #--------------------------------------------

        values_array.each do |element|

          tmp_hash = {}
            parser["output_keys"].each  do |output_key|

              # generate a key
              key = output_key['name']

              #-------------------------------------------
              # assign a value corresponding to the key
              #-------------------------------------------

              if output_key["metadata"] == 'parsed_key_value'
                 value = leaf_key

              elsif output_key["metadata"] == 'static_string'
                value = output_key['value']

              elsif output_key["metadata"] == 'namespaced_value'
                if data_packet[parser["input_key"]["name"]].include?(element)
                  value = element
                end
              end

              #---------------------------------------------
              # push the key value into the tmp_hash
              #---------------------------------------------
              tmp_hash[key] = value if value.present?

              result_object << tmp_hash

            end
          end
        end

    rescue => e
      ServiceLog.log('Emulator: synthetic_transformer : exception caught: ' + e.message)
    end

    # Return the result object array
    result_object.uniq!
  end


end
