###################################################################################
# Emitter: emits a sequence of synthetic DML, SQL and MongoDB DML statements
# receives a DML object
#
# Design Note: Uses the singleton pattern
###################################################################################

class Emitter
  include MongoMapper::Document


  #############################################################################
  # interface into the emitter
  # receives a ruby object
  # returns the DML object that has been augmented with DML statements
  #############################################################################

  def self.interface(dml_object = nil)
    ServiceLog.stack_trace('emitter::interface') if Rails.application.config.STACK_TRACE_EMITTER

    return false unless self.sanity_test?(dml_object)

    #------------------------------------------------
    # add required keys for selection into the view
    #------------------------------------------------
    self.add_keys(dml_object)

    dml_walker(dml_object)

  end


  private

  ############################################################################
  # sanity_test?: simple validation of arguments received on the interface
  ############################################################################

  def self.sanity_test?(dml_object)
    ServiceLog.stack_trace('emitter::sanity_test') if Rails.application.config.STACK_TRACE_EMITTER

    return false if dml_object.blank?
    return false if dml_object.is_a?(Hash).blank?
    return false if dml_object["required_attributes"].blank?

    true
  end


  #############################################################################
  # dml_walker: walks through the DML object adding synthetic DML, SQL
  # and Mongodb commands
  # returns the mutated DML object
  #############################################################################

  def self.dml_walker(dml_object)
    ServiceLog.stack_trace('emitter::dml_walker') if Rails.application.config.STACK_TRACE_EMITTER

    #add a create Table or create Collection dml statement
    self.create_collection_dml(dml_object)

    # select target documents in the Datadump collection
    self.make_command_string(dml_object)

    dml_object
  end


  #############################################################################
  # adds_keys to the DML object. These will be populated with command strings
  # by the emitter
  #############################################################################

  def self.add_keys(dml_object)
    ServiceLog.stack_trace('emitter::self.add_keys') if Rails.application.config.STACK_TRACE_EMITTER

    dml_object["synthetic_dml"] = []
    dml_object["sql"] = []
    dml_object["mongodb"] = []
  end


  ##############################################################################
  # add_self.create_collection_dml: add create a collection or table DML to the
  # dml object
  ##############################################################################

  def self.create_collection_dml(dml_object)
    ServiceLog.stack_trace('emitter::self.create_collection_dml') if Rails.application.config.STACK_TRACE_EMITTER

    dml_object["synthetic_dml"].push("CREATE " + dml_object["view_name"])
    dml_object["sql"].push("CREATE " + dml_object["view_name"])

    # not required for mongodb as the collection will be automatically
    # created on the first insert if the collection does not exist
  end


  #############################################################################
  # self.make_command_string: makes a command string to select documents
  # which are candidates for inclusion in the view and forwarding to the
  # legacy system
  ###############################################################################


  def self.make_command_string(dml_object)
    ServiceLog.stack_trace('emitter::self.make_command_string') if Rails.application.config.STACK_TRACE_EMITTER

    #-----------------------------------
    # create the command string for
    # the candidate documents
    #-----------------------------------

    command_string = 'DataDump.where('

    #----------------------------------------------
    # start adding segments to the command string
    #----------------------------------------------

    command_string = self.required_attributes_segment(dml_object, command_string)

    if dml_object["sak"].present?
      command_string = self.sak_segment(dml_object, command_string)
    end

    if dml_object["selection"].present?
      command_string = self.selection_segment(dml_object, command_string)
    end

    #----------------------------
    # close the command
    #----------------------------

    command_string += ')'

    #----------------------------
    # add the projection
    #----------------------------
    if dml_object["projection"].present?
      command_string = self.projection_segment(dml_object, command_string)
    end


    #-------------------------------------
    # push the command string onto the
    # dml_object's mongodb array
    #-------------------------------------
    dml_object["mongodb"].push(command_string)

  end


  ##############################################################################
  # self.required_attributes_segment: determines the documents in the datadump which
  # are candidates for inclusion in the view. A document must have all of the
  # required attributes to be a candidate
  ##############################################################################

  def self.required_attributes_segment(dml_object, command_string)

    ServiceLog.stack_trace('emitter::self.required_attributes_segment') if Rails.application.config.STACK_TRACE_EMITTER

    #----------------------------------------------
    # the fields in the required attributes
    # attribute specify which documents are
    # candidates for inclusion in the view
    #----------------------------------------------

    dml_object['required_attributes'].split(',').each do |el|
      command_string += ":#{el} => " + "{ '$exists' => true },"
    end

    #----------------------------------------------
    # if the projection attribute is present in
    # the DML object then a candidate document
    # must contain all the projection fields
    #----------------------------------------------

    if dml_object["projection"].present?
      dml_object['projection'].split(',').each do |el|
        command_string += ":#{el} => " + "{ '$exists' => true },"
      end
    end

    #----------------------------------------------
    # if the sak attribute is present in the DML
    # then a candidate document must also contain
    # a SAK field
    #----------------------------------------------

    if dml_object["sak"].present?
        command_string += ":sak => " + "{ '$exists' => true },"
    end

    #---------------------------------------------
    # if the selection attribute is present in
    # the DML then add the attributes to the
    # required attributes. Loop through the
    # selection expression adding the variable
    # conditions
    #---------------------------------------------

    if dml_object["selection"].present?

      ctr = 0
      array = dml_object["selection"].split(" ")
      array.reject! { |element| element.blank? }

      while ctr < array.length do
        sub_command = ":#{array[ctr]} => " + "{ '$exists' => true },"

        if command_string.include?(sub_command).blank?
          command_string += sub_command
        end
        ctr += 4
      end

    end

    command_string.chomp(',')

  end


  ###############################################################################
  # self.projection_segment: augments the command string by adding a segment that
  # projects the datadump onto a view
  #
  # Design: specifies the projection attributes that will be projected for a
  # view
  ################################################################################

  def self.projection_segment(dml_object, command_string)

    ServiceLog.stack_trace('emitter::self.projection_segment') if Rails.application.config.STACK_TRACE_EMITTER

    #----------------------------------------------
    # the fields in the projection attribute
    # specify the document fields which will be
    # present in the view
    #----------------------------------------------

    command_string += ".fields("
    dml_object['projection'].split(',').each do |el|
      command_string += ":#{el},"
    end

    command_string.chomp!(',')
    command_string += ')'

  end


  ###############################################################################
  # self.selection_segment: augments the command string by adding a segment that
  # adds sql like selection to the candidate documents in the datadump
  #
  #  ARITHMETIC OPERATORS SUPPORTED:    gt,  lt, gte, lte, eq, ne
  #  RELATIONAL OPERATORS SUPPORTED:   AND, OR
  #
  # A selection expression must have the expansible form:
  #          variable ARITHMETIC OPERATOR constant RELATIONAL OPERATOR  ....
  ##############################################################################

  def self.selection_segment(dml_object, command_string)

    ServiceLog.stack_trace('emitter:: selection segment') if Rails.application.config.STACK_TRACE_EMITTER

    if command_string.present? && command_string[-1] == ','
      command_string.chomp!(',')
    end

    ctr = 0
    or_flag = false
    sub_command = ", "

    array = dml_object["selection"].split(" ")
    array.reject! { |element| element.blank? }

    #--------------------------------------
    # loop and add selection conditions
    # to the command string
    #-------------------------------------

    while ctr < array.length do

      #--------------------------------------
      # add the arithmetic operator clause
      # special expression required for the
      # equality operator
      #--------------------------------------
      if array[ctr + 1] != 'eq'
        sub_command += " :#{array[ctr]}.#{array[ctr + 1]} => " + "#{array[ctr + 2]},"
      else
        sub_command += " :#{array[ctr]} => { '" + "$#{array[ctr + 1]}" +  "' => #{array[ctr + 2]} },"
      end


      #-----------------------------------------
      # close OR sub-expression
      #-----------------------------------------
      if (ctr + 3) >= array.length
        if or_flag
          sub_command.chomp!(',')
          sub_command += ' }]'
          break
        end
      end

      #-----------------------------------------
      # relational OR Operator
      #-----------------------------------------

      if array[ctr + 3] == 'or'

        if or_flag.blank?
          sub_command += "'$or' => [{ "
          or_flag = true
        end
        ctr += 1
      end

      #----------------------------------------
      # relational AND operator: by default
      # sub-clauses in the expression are
      # ANDED so we skip over the and operator
      # we close the relational OR operator
      # sub-clause if necessary
      #---------------------------------------

      if array[ctr + 3 ] == 'and'
        if or_flag
          or_flag = false
          sub_command.chomp!(',')
          sub_command += '}], '
        end

        ctr += 1
      end

      #------------------------------------
      # go the beginning of the next
      # sub-expression
      #------------------------------------
       ctr += 3

      #-----------------------------------------
      # close OR sub-expression
      #-----------------------------------------

      if ctr >= array.length
        if or_flag
          sub_command.chomp!(',')
          sub_command += ' }]'
        end
        break
      end

    end

    # remove the trailing comma
    sub_command.chomp!(',')
    command_string += sub_command

  end


  ###############################################################################
  # self.sak_segment: augments the command string by adding segments containing
  # the sak constraint
  # Design: The sak attribute in the DML object is a comma separated list of
  # sak values. For a document from the datadump to be included in the view
  # it must contain the sak key and the field value must equal a value in
  # the comma separated string
  ################################################################################

  def self.sak_segment(dml_object, command_string)
    ServiceLog.stack_trace('emitter::self.sak_segment') if Rails.application.config.STACK_TRACE_EMITTER


    #----------------------------------------------
    # a candidate document in the data dump can be
    # included in the view if its sak key has a
    # value included in dml['sak']
    #----------------------------------------------

    command_string += ", \'$or\' => [ "

    dml_object['sak'].split(',').each do |sak_value|
      command_string += "{'sak'=> \'#{sak_value}\'},"
    end


    command_string.chomp!(",")
    command_string += ']'
  end


  ###############################################################################
  # self.exec_mongo:  executes a mongodb command
  # receives the dml object and a command string
  # returns false if the view collection is not created, true otherwise
  ###############################################################################

  def self.exec_mongo(dml_object, command_string)
    ServiceLog.stack_trace('emitter::self.exec_mongo') if Rails.application.config.STACK_TRACE_EMITTER

    # ---------------------------------------
    # dynamically create a view class with
    # the MongoMapper module mixed in
    #-----------------------------------------

    Object.const_set dml_object["view_name"].camelize, Class.new { include MongoMapper::Document  }

    #-----------------------------------------------
    # get the documents selection from the datadump
    # by executing the command string
    #-----------------------------------------------

    datadump_selection = instance_eval(command_string).to_a

    if datadump_selection == []
      return false
    end

    #-------------------------------------------------
    # add all of the selected documents into the
    # view collection create the view collection if
    # it does not exist
    #-------------------------------------------------

    view =  dml_object["view_name"].camelize.constantize

    datadump_selection.each do |doc|
      doc_hash = JSON.parse(doc.to_json)
      doc_hash = doc_hash.reject { |key, value| key == "id" }
      begin
        view.create(doc_hash)
      rescue
        return false
      end
    end

    true
  end

end


