###################################################################################
# interpreter: interprets pattern objects
# receives a pattern as a ruby object
#
# Algorithm
# (1) Invoke the tokenizer input on the Pattern Object
# (2) The tokenizer examines the pattern object and creates a State Object and
#     a DML Object
# (3) Emitter walks through the DML object and for each token:
#     (a) Emitter iteratively creates  a SQL statement and for MongoDB DML mutatis
#     mutandis
# (4) When emitter is finished the interpreter takes the State object constructed
#     by the tokenizer and the DML object created by the emitter and passes these
#     objects to the scheduler
#     The scheduler adds the node object to its traversable node list
# (5) Scheduler wakes up
# (6) For each node the scheduler executes the dml statements in the node
# (7) scheduler sleeps
# (7) scheduler goes to step 5
#
# Note: Syntax checking is not required since this function is performed by the
# pattern service which validates the pattern before persisting it to the
# pattern collection
#
# Design Note: Uses the singleton pattern
###################################################################################


require 'singleton'

class Interpreter
  include Singleton

  ##############################################################################
  # interface into the interpreter
  # Receives: pattern object as a ruby object
  # returns a ruby hash object of the form:
  #               { "state" => ..., "dml" => ... } or false
  ###############################################################################

  def interface(pattern)

    ServiceLog.stack_trace('interpreter::interface') if Rails.application.config.STACK_TRACE_INTERPRETER

    return false unless sanity_check?(pattern)

    #---------------------------------------------------
    # invoke tokenizer: returns false or the hash
    # { "state" => state_object, "dml" => dml_object}
    #---------------------------------------------------
    tokenized_pattern = Tokenizer.instance.interface(pattern)

    #-------------------------------
    # construct the DML object
    #-------------------------------

    if tokenized_pattern.present?
       dml_object = Emitter.interface(tokenized_pattern["dml"])
    else
       return false
    end

    #---------------------------------
    # add the node to the scheduler
    #---------------------------------

    Scheduler.interface("create", tokenized_pattern["state"], dml_object)

    { "state" => tokenized_pattern["state"],  "dml" => dml_object }

  end

  private

  #############################################################################
  # basic validation on the parameters received by the interpreter interface
  #############################################################################

  def sanity_check?(pattern)
    ServiceLog.stack_trace('interpreter::sanity_check?') if Rails.application.config.STACK_TRACE_INTERPRETER

    # type check
    return false unless pattern.is_a?(Hash)

    true
  end

end
