####################################################################################################
# Tokenizer: receives a Pattern (Ruby Hash Object)
#            Walks through the pattern object and constructs a state object and a DML object
# returns false if there is an error
#         or the array [ state_object, dml_object ] otherwise
# Algorithm note: the DML object is the sequence of "SQL" or "momgodb" statements for execution
# The state object specifies when and in what conditions the "sql" or 'mongodb' statements
# will be executed
# the state object is complete, the dml object will be further processed by the parser
#
# Implemented as a singleton object
####################################################################################################

require 'singleton'

class Tokenizer
  include Singleton

  #########################################################################
  # tokenizer interface
  # receives a valid pattern ruby object
  # returns false if there is an error
  #         or : the hash object  { "state" => state_object,
  #                                 "dml"   => dml_object
  #                               }
  #########################################################################

  def interface(pattern)
    ServiceLog.stack_trace('tokenizer::interface') if Rails.application.config.STACK_TRACE_TOKENIZER

    # empty state object and dml objects
    @state_object = {}
    @dml_object   = {}

    return false unless walk_pattern(pattern)

    mutate_projection

    { "state" => @state_object, "dml" => @dml_object}
  end

  private

  ##########################################################################
  # walks through a received pattern
  # tries to recognize each key and upon recognition adds the key and the
  # value to the state object or the dml object
  # returns false if the key cannot be recognized
  # note: adds the view_name to both the state object and the dml object
  ##########################################################################

  def walk_pattern(pattern)
    ServiceLog.stack_trace('tokenizer::walk_pattern') if Rails.application.config.STACK_TRACE_TOKENIZER

    pattern.each_pair do |key, value|

      ret = recognize_attributes(key)
      return false if ret.blank?

      add_dml_attribute(key, value)   if ret == 'dml'
      add_state_attribute(key, value) if ret == 'state'

      #-------------------------------------------------
      # the view_name key must be added to both objects
      #-------------------------------------------------
      if key == "view_name"
        add_dml_attribute(key, value)
        add_state_attribute(key, value)
      end

      # add the pattern id to the state object
      add_state_attribute('pattern_id', pattern['pattern_id'])
      add_dml_attribute('pattern_id', pattern['pattern_id'])
    end

    true
  end


  ##########################################################################
  # add state_attribute: adds a key and corresponding value to the state
  # object
  # receives an attribute and and an attribute value
  ##########################################################################

  def add_state_attribute(attribute, value)
    ServiceLog.stack_trace('tokenizer::add_state_attribute') if Rails.application.config.STACK_TRACE_TOKENIZER

    @state_object.merge!({ attribute => value })
  end


  ##########################################################################
  # add dml_attribute: adds a key and corresponding value to the DML
  # object
  # receives an attribute and and an attribute value
  ##########################################################################

  def add_dml_attribute(attribute, value)
    ServiceLog.stack_trace('tokenizer::add_dml_attribute') if Rails.application.config.STACK_TRACE_TOKENIZER

    @dml_object.merge!({  attribute => value })
  end


  ##########################################################################
  # mutate_projection: If the projection attribute is  empty then all of
  # the required_attributes will be projected. In the latter case the
  # required attributes are copies to the projection attribute.
  ##########################################################################

  def mutate_projection()
    ServiceLog.stack_trace('tokenizer::mutate_projection') if Rails.application.config.STACK_TRACE_TOKENIZER

    return if @dml_object["projection"].present?

    @dml_object["projection"] = @dml_object["required_attributes"]

  end


  ##########################################################################
  # recognize_attributes: determines whether an attribute in a pattern
  # belongs in the state object or the DML object
  #
  # receives an attribute
  # returns: false if the attribute is not recognized
  #          otherwise returns: 'state' or 'dml'
  #
  # Note: the view_name attribute will be added to both objects
  ##########################################################################

  def recognize_attributes(attribute)

    ret = case attribute
            when 'pattern_name' then 'state'
            when 'pattern_id' then 'state'
            when 'view_name' then 'state'
            when 'required_attributes' then 'dml'
            when 'sak' then 'dml'
            when 'projection' then 'dml'
            when 'selection' then 'dml'
            when 'join' then 'dml'
            when 'time_interval' then 'dml'
            when 'generate_on_query' then 'state'
            when 'generate_once' then 'state'
            when 'append_interval' then 'state'
            when 'create_at' then 'state'
            when 'append_always' then 'state'
            when 'indexes' then 'dml'
            when 'reset_view' then 'state'
            else
              ServiceLog.log('tokenizer::recognize_attributes : pattern attribute ' + attribute + ' not recognized' )
              return false
    end
  end

end
