############################################################################################
# PacketMapService : maps data packet fields to fields required by the legacy activity
# collection as well as doing other defined mappings.
# This mapping enables the packet injection service to remap the incoming data packet so
# that it contains document fields which are required by the legacy activity collection.
#
# Group_id Motivation
# A packet_type may have different associated group_ids. This is used to do different
# types of mappings for a given type of packet. Each such mapping will spawn a new
# mapped data packet which will then be used to populate a view and be sent to the
# legacy system. This is packet-type "jira-type" has five group_ids: group_a,... group_e
# THIS WILL spawn five packets destined for persistence
#
# Structure of a good packet map:
#
#  #  Example
#
#      { {   packet_type: 'jira',
#            group_id: 'group_a'
#
#           'jira' => {
#                     'status_name' => namespaced key (such as fields.issue.name),
#                     'status'      => namespaced_key (such as fields.issue.latest.outcome)
#                     },
#
#      presentation_layer_category =>
#      } }
#
# group id is the primary key
#
# NOTE: THE PRESENTATION_LAYER_CATEGORY FIELD IS FOR THE CONVENIENCE OF THE PRESENTATION
# LAYER ONLY. THE NEW SYSTEM DOES NOT USE OR CARE ABOUT THIS FIELD. THIS IS AN IMPOSED
# KLUDGE TO RECTIFY DEFICIENCIES IN PRESENTATION LAYER UI DESIGN ONLY.IT CAN BE BLANK
# THIS FIELD NEED NOT BE PRESENT IN AN PACKET MAP
#
# The first level key (jira in this example) specifies the type of data packet to which
# the map is to be applied); the value of this key is a hash which shows the mappings
# required in order to successfully insert a value into the legacy table.
# So the status_name field in a data packet to be inserted into the legacy activity
# collection is mapped as follows:  "status_name" => data_packet[namespaced key]
#
# Note that a packet type need not have a corresponding packet map. In the latter
# case no mappings are performed for a packet of this type which is destined for the
# legacy activities collection
#
# NOTE: mongomapper does NOT work with singletons
#
# copyright collabnet,2015- 2016
############################################################################################



  class PacketMaps

    include MongoMapper::Document

    ##################################################################
    # Interface To The PacketMap Service
    # Receives: [i]   a ruby hash object (a packet map) or a string
    #                 as the second parameter
    # Receives: [i]   a http verb which indicates the type of operation
    #                 to be performed: GET, POST,PUT, DELETE
    #           [ii]   a ruby hash object (a packet map) or a string
    #                  as the second parameter
    #           [iii]  a string for the key parameter (optional)
    #
    # return false if there is an error otherwise true
    #
    # Note: the second interface parameter is a ruby hash if the verb
    #       is POST. It is a string if the verb is DELETE
    ###################################################################


    def self.interface(http_verb, packet_map = nil, group_id = nil, packet_type = nil)

      ServiceLog.stack_trace('PacketMap::interface') if Rails.application.config.STACK_TRACE_PACKET_MAP

      if http_verb.is_a?(String) == false
        ServiceLog.log('PacketMaps::interface? : verb is not of type string')
        return false
      end

      http_verb.upcase!
      http_verb = http_verb.strip

      if packet_type.present? && packet_type.is_a?(String)
        packet_type = packet_type.strip
      end

      if group_id.present? && group_id.is_a?(String)
        group_id = group_id.strip
      end

      #--------------------------------------------------
      # parameters sanity test since they are overloaded
      #--------------------------------------------------
      return false unless sanity_test?(http_verb, packet_map, group_id, packet_type)

      if http_verb == 'POST'
        return false unless self.map_valid?(packet_map)
        ret = persist_packet_map(packet_map)

      elsif http_verb == 'DELETE'
        ret = self.destroy_packet_map(group_id)

      elsif http_verb == 'GET' && packet_type.blank?
        ret = self.get_packet_maps

      elsif http_verb == 'GET' && packet_type.present?
        ret = self.get_packet_map(packet_type)

      elsif http_verb == 'PUT' && packet_map.present? && group_id.present?
        ret = self.put_packet_map(packet_map, group_id)
      else
        ServiceLog.log('PacketMaps::interface? : invalid interface parameters')
        return false
      end

      ret
    end


    private

    ####################################################################
    # sanity_test?: basic integrity tests
    # receives a packet hash object
    ####################################################################

    def self.sanity_test?(http_verb, packet_map, group_id, packet_type)
      ServiceLog.stack_trace('PacketMaps::sanity_tests') if Rails.application.config.STACK_TRACE_PACKET_MAP

      #------------------------------
      # test the HTTP verb received
      #------------------------------
      if ['GET', 'POST', 'PUT', 'DELETE'].include?(http_verb).blank?
        ServiceLog.log('PacketMaps::sanity_test? : interface did not receive a recognized verb')
        return false
      end

      # test key match
      if packet_map.is_a?(Hash) && packet_map["packet_type"].blank?
        ServiceLog.log('PacketMaps::sanity_test? : packet map does not have a packet_type key')
        return false
      end

      # test matching keys
      if packet_map.is_a?(Hash)
        if packet_map.keys.include?(packet_map["packet_type"]).blank?
          ServiceLog.log('PacketMaps::sanity_test? : packet map does not have a key matching the value of packet_type')
          return false
        end
      end

      #---------------------------
      # post verb
      #---------------------------

      if http_verb == 'POST'
        if packet_map.is_a?(Hash).blank?
          ServiceLog.log('PacketMaps::sanity_test? : packet map is not a Hash')
          return false
        end

        if packet_map.empty? && http_verb == 'POST'
          ServiceLog.log('PacketMaps::sanity_test? : packet map is empty for a post request')
          return false
        end

        # test the packet_map keys
        if packet_map['packet_type'].blank?
          ServiceLog.log('PacketMaps::sanity_test?: post verb: packet_type is not specified')
          return false
        end
      end

      #---------------------------
      # delete verb
      #---------------------------

      if http_verb == 'DELETE'
          if group_id.blank?
            ServiceLog.log('PacketMaps::sanity_test? delete verb: search key is empty')
            return false
          end

        if group_id.class != String
          ServiceLog.log('PacketMaps::sanity_test?: delete verb: group id is not a non-empty string')
          return false
        end

      end

      #---------------------------
      # get verb
      #---------------------------

      if http_verb == 'GET'
        if packet_type.present? && packet_type.is_a?(String).blank?
          ServiceLog.log('PacketMaps::sanity_test?: get verb: packet type is not of type string')
          return false
        end
      end

      #---------------------------
      # put verb
      #---------------------------

      if http_verb == 'PUT'
        if  group_id.blank?
        ServiceLog.log('PacketMaps::sanity_test?: get verb: a packet map update requires a group id')
        return false
        end
      end

      true
    end


    ########################################################################
    # validate_map?: validates that the packet map received has the
    # proper format:
    #      { {   packet_type: 'jira',
    #            group_id: 'group_alpha',
    #
    #           'jira' => {
    #                     'status_name' => namespaced key (such as fields.reporter.name),
    #                     'status'  =>  namespaced_key (such as fields.report.outcome)
    #                     }
    #      } }
    #
    # NOTE: The two 'jira' attributes they MUST be the same string
    #
    #
    #########################################################################

    def self.map_valid?(packet_map)
      ServiceLog.stack_trace('PacketMap::validate_map?') if Rails.application.config.STACK_TRACE_PACKET_MAP

      key = packet_map['packet_type']

      # boundary case: the packet map is an empty hash
      # just return true

      map_subset = packet_map[key]

      return true if map_subset == {}

      # verify that each key in the map has a value of type hash

      if map_subset.is_a?(Hash).blank?
        ServiceLog.log('PacketMap::valid_map? : packet_map[key] must be a key-value hash')
        return false
      end

      # iterate through the keys validating each key value
      map_subset.values.each do |value|

        if value.class != String || value.length == 0
          ServiceLog.log('PacketMap::valid_map? : key value in map is not a string or has length zero')
          return false
        end

      end

    true
    end


    ######################################################################
    # persist_packet_map: persists a packet map to the PacketMaps
    # collection
    # NOTE: can persist the multiple activity maps with the same
    # packet-type key
    #
    # receives an packet map (ruby hash object)
    # returns the persisted packet map or false
    #######################################################################

    def self.persist_packet_map(packet_map)
      ServiceLog.stack_trace('PacketMap::persist_packet_map') if Rails.application.config.STACK_TRACE_PACKET_MAP

      #----------------------------------------
      # Persist to the PacketMaps Collection
      #----------------------------------------
      begin
        PacketMaps.create(packet_map);
      rescue => e
        ServiceLog.log('PacketMap::persist_packet_map : packet map not persisted to the PacketMaps collection: ' + e.message)
        return false
      end

      packet_map
    end

    #######################################################################
    # get_packet_maps: get all packet maps
    # returns an array of packet maps or false if there is an error
    #######################################################################

    def self.get_packet_maps
      ServiceLog.stack_trace('PacketMap::get_packet_maps') if Rails.application.config.STACK_TRACE_PACKET_MAP

      begin
        maps = PacketMaps.all.to_a
      rescue => e
        ServiceLog.log('PacketMap::get_packet_maps : failed to get all packet maps : ' + e.message)
        return false
      end

      maps
    end


    #######################################################################
    # get_packet_map: returns the packet map pertaining to a search
    # key
    # receives the group id and the packet_type
    #######################################################################

    def self.get_packet_map(packet_type)
      ServiceLog.stack_trace('PacketMap::get_packet_map') if Rails.application.config.STACK_TRACE_PACKET_MAP

      begin
        map = PacketMaps.where(:packet_type => packet_type).to_a
      rescue => e
        ServiceLog.log('PacketMap::get_packet_map: failed to get the packet map : ' + e.message)
        return false
      end

      map
    end

    #######################################################################
    # put_packet_map: updates a packet map
    # receives the new packet map and the group id
    # returns the updated packet map or false if the update fails
    #######################################################################

    def self.put_packet_map(packet_map, group_id)

      ServiceLog.stack_trace('PacketMap::put_packet_map') if Rails.application.config.STACK_TRACE_PACKET_MAP
      begin
        packet_array = PacketMaps.where(group_id: group_id).to_a
      rescue => e
        ServiceLog.log('PacketMap::put_packet_map : failed to get the packet map : ' + e.message)
        return false
      end

      if packet_array.size == 0
        ServiceLog.log('PacketMap::put_packet_map : cannot update an empty packet map collection')
        return false
      end

      unless self.destroy_packet_map(group_id)
        ServiceLog.log('PacketMap::put_packet_map : failed to find the packet map')
        return false
      end

      self.persist_packet_map(packet_map)
    end



    #######################################################################
    # destroy_packet_map: deletes an packet map that has been persisted
    # into the PacketMaps collection
    # receives: a group_id
    # returns true if the record is deleted; false otherwise
    # returns true if the PacketMaps collection is empty
    #######################################################################

    def self.destroy_packet_map(group_id)

      ServiceLog.stack_trace('PacketMap::destroy_packet_map') if Rails.application.config.STACK_TRACE_PACKET_MAP

      return true if PacketMaps.count == 0

      begin
        PacketMaps.destroy_all(:group_id => group_id)

      rescue => e
        ServiceLog.log('PacketMap::destroy_packet_map : failed to delete the packet map : ' + e.message)
        return false
      end

      true
    end

  end

