################################################################################################
# Validates patterns received from the REST interface. Called by the REST interface which
# receives a pattern or a search key. Performs HTTP operations
# Algorithm
#    [0]   receives: a pattern that has been converted into a Ruby Hash Object or a plain string
#                  and a HTTP verb
#    [1] invokes various syntax checking tests
#    [2] invokes various problem domain  logic tests
#    [3] returns false if a test fails
#    [4] otherwise performs the operation implied by the HTTP verb
#
# Structure Of The Pattern JSON Object
# The CTF UI will send this stringified JSON Pattern object to the REST interface:
#
#  {
#       "pattern_name" :         required, stringValue
#       "view_name" :            required, stringValue
#       "required_attribute" :   optional, comma separated attribute names
#       "sak" :                  optional, comma separated SAKS
#       "required_columns" :     optional, comma separated column_names
#       "projection" :           optional,  comma separated column_names
#       "selection" :            optional, comma separated selection values such as column_value >= 5 AND c < p
#       "join" :                 optional comma separated column_names
#       "required_columns" :     optional comma separated column_names
#       "time_interval":         time_1, time_2
#       "append_interval":       optional, string seconds
#       "create_at":             optional time string
#       "index_on":              optional comma separated column_names
#       "reset_view" :           optional date string
#       "append_always":         optional boolean flag
#       "generate_once":         optional boolean flag
#
#      }
#
# Note: All attribute names must be exactly as shown; in particular attributes are case sensitive
#
# Validations Performed:
#
#   (1) The pattern name must not be blank and must be unique. Required pattern attribute
#
#   (2) The view name must be present and unique. Required
#
#   (3) required_attributes: a datadump document must have these attributes present in order
#             for the document to be considered for selection into a view (optional pattern attribute)
#
#   (4) sak : a datadump document must have ANY of the sak keys specified in order for the
#             document to be considered for inclusion in a view. The SAK attribute need not
#             however be present in any view document
#
#   (5) projection: the document attributes which must be present in a view. Optional pattern
#             attribute
#
#   (6) indexes: Optional pattern attribute. The view will be indexed on these fields. all
#             indexes that are created are optional indexes
#
#   (7) generate_once: optional boolean attribute in a pattern. If true specifies that the
#       view will be generated only once. Cannot be blank
#
#   (8) generate_on_query:  optional boolean attribute in the pattern
#             If true, it specifies that the view will be updated every time a query is received.
#             This gives a dynamic view. If False the  view is updated periodically by the scheduler
#             Attribute value must be true or false and never blank
#
#   (9) append_always: optional boolean flag. When data pertaining to the view comes in then
#       immediately append to the view
#
#  (10)  A join expression is a colon separated string of the form: type_of_join:col_1,col2...
#       At least two attributes must be joined. This is an equi-join on the data dump
#
#  DESIGN NOTE: all times received must be epoch times cast as strings
#
#
#  ARITHMETIC RELATION SUPPORTED: gt,  lt, gte, lte, eq, ne
#  LOGICAL OPERATORS SUPPORTED:    AND, OR
#
# A selection expression must have the expansible form:
#                          variable ARITHMETIC OPERATOR constant RELATIONAL OPERATOR  ....
#
#  # Valid pattern URL format:
#   POST  http://www.teamforge.com/version/:version_id/project/:project_id/webhook/patterns
#
################################################################################################


class Pattern

  include MongoMapper::Document

  ##############################################################################
  # Interface To The Pattern Validator
  # receives (i)   the pattern (ruby hash object) that is to operated on OR a
  #                simple string object (which will then be interpreted as a
  #                search key to GET or DELETE pattern operations)
  #          (ii)  the HTTP verb which designates the operation to be performed
  #          (iii) an optional pattern search key
  ##############################################################################

  def self.interface(payload, http_verb = nil, search_key = nil)

    ServiceLog.stack_trace('pattern::interface') if Rails.application.config.STACK_TRACE_PATTERN

    # check parameters for no operation
    if http_verb.blank? || http_verb.is_a?(String).blank?
      ServiceLog.log('Pattern: sanity_check? : blank http verb or verb is not a string')
      return false
    end

    http_verb.upcase!
    http_verb = http_verb.strip

    # sanity check for the HTTP Verb
    return false if self.sanity_check?(payload, http_verb, search_key).blank?

    # perform the validation  tests on the received pattern
    if payload.kind_of?(Hash)
      return false if self.valid_pattern?(payload).blank?
    end


    #---------------------------------
    # sanitize the received pattern
    #---------------------------------
    if payload.present?
      payload = self.sanitize_pattern(payload)
    end

    #---------------------
    # Pattern processing
    #---------------------

    # persist the pattern
    if http_verb == 'POST'
      payload['view_name'] = clean_view_name(payload['view_name'])
      return self.persist_pattern(payload)

    # test for pattern deletion
    elsif http_verb == 'DELETE'
      return self.destroy_pattern(search_key)

    # test for pattern update
    elsif http_verb == 'PUT'
      payload['view_name'] = clean_view_name(payload['view_name'])
      return self.update_pattern(payload, search_key)

    # GET#show (DO NOT INVERT ORDER OF THE GET STATEMENTS)
    elsif http_verb == 'GET' && search_key.present?
      return self.get_pattern(search_key)

    # test for GET#index
    elsif http_verb == 'GET'
      return self.get_all_patterns
    else
      ServiceLog.log('Pattern: interface: unrecognized http verb: ' + http_verb)
      return false
    end

  end


  private


  ############################################################
  # sanity_check?: validates parameters received on
  # the interface
  ###########################################################

  def self.sanity_check?(payload, http_verb, search_key)

    ServiceLog.stack_trace('pattern::sanity_check?') if Rails.application.config.STACK_TRACE_PATTERN

    if ['GET', 'POST', 'DELETE', 'PUT'].include?(http_verb).blank?
      ServiceLog.log('Pattern: sanity_check? : unrecognized http verb ')
      return false
    end

    if http_verb == 'POST' and payload.is_a?(Hash).blank?
      ServiceLog.log('Pattern: sanity_check? : payload not received for POST verb ')
      return false
    end

    if http_verb == 'PUT' and (payload.is_a?(Hash).blank? || search_key.kind_of?(String).blank?)
      ServiceLog.log('Pattern: sanity_check? : payload or search key type is improper')
      return false
    end

    if http_verb == 'DELETE' && search_key.is_a?(String).blank?
      ServiceLog.log('Pattern: sanity_check? : invalid search key type of DELETE verb')
      return false
    end

    true
  end

  ##############################################################
  # valid_pattern?: validates a pattern
  # returns true or false
  ###############################################################

  def self.valid_pattern?(payload)

    return false unless self.validate_pattern_name?(payload)
    return false unless self.validate_pattern_id?(payload)
    return false unless self.validate_view_name?(payload)
    return false unless self.validate_required_attributes?(payload)
    return false unless self.validate_sak?(payload)
    return false unless self.validate_projection?(payload)
    return false unless self.validate_indexes?(payload)
    return false unless self.validate_generate_once?(payload)
    return false unless self.validate_generate_on_query(payload)
    return false unless self.validate_append_always?(payload)
    return false unless self.validate_join?(payload)
    return false unless self.validate_reset_view?(payload)
    return false unless self.validate_append_interval?(payload)
    return false unless self.validate_create_at?(payload)
    return false unless self.consistent_attributes?(payload)
    return false unless self.validate_selection?(payload)
    
    true
  end





  ############################################################
  # validate_pattern_name: ensure that the pattern name
  # is not blank and is unique
  ############################################################

  def self.validate_pattern_name?(pattern)

    ServiceLog.stack_trace('pattern::validate_pattern_name?') if Rails.application.config.STACK_TRACE_PATTERN

    if pattern["pattern_name"].blank?
      ServiceLog.log('Pattern: validate)pattern_name? : blank pattern name')
      return false
    end

    if pattern["pattern_name"][0] == '$'
      ServiceLog.log('Pattern: validate_pattern_name? : pattern name has an invalid leading char')
      return false
    end

    true
  end



  ############################################################
  # validate_pattern_name: ensure that the pattern name
  # is not blank and is unique
  ############################################################

  def self.validate_pattern_id?(pattern)

    ServiceLog.stack_trace('pattern::validate_pattern_id?') if Rails.application.config.STACK_TRACE_PATTERN

    if pattern["pattern_id"].blank?
      ServiceLog.log('Pattern: validate)pattern_id? : blank pattern id')
      return false
    end

    true
  end



  ##########################################################
  # validate_view_name: ensure that the view name is not
  # blank and is unique
  ##########################################################

  def self.validate_view_name?(pattern)

    ServiceLog.stack_trace('pattern::validate_view_name?') if Rails.application.config.STACK_TRACE_PATTERN

    if pattern['view_name'].blank?
      ServiceLog.log('Pattern: validate_view_name? : the view name was not specified')
      return false
    end

    if pattern["view_name"][0] == '$'
      ServiceLog.log('Pattern: validate_view_name? : the view name has an invalid leading char')
      return false
    end


    # view name must be unique
    array = Pattern.find('view_name' => pattern["view_name"]).to_a
    if array.size > 0
      ServiceLog.log('Pattern: validate_view_name? : the view already exists')
      return false
    end

    true
  end


  #########################################################################
  # validate_required_attributes: this attribute in the pattern
  # specifies the document attributes which must be present in order
  # for the document to be included in a view. These required
  # attributes need not however be present in the view documents.
  # this attribute cannot be empty in order to preclude the inclusion
  # of documents in the datadump in a view
  #########################################################################

  def self.validate_required_attributes?(pattern)

    ServiceLog.stack_trace('pattern::validate_required_attributes?') if Rails.application.config.STACK_TRACE_PATTERN

    if pattern['required_attributes'].blank?
      ServiceLog.log('Pattern: validate_required_attributes? : required attributes not specified')
      return false
    end

    # the required attributes must be string data type
    if pattern['required_attributes'].class != String
      ServiceLog.log('Pattern: validate_required_attributes? : required attributes are not of string type')
      return false
    end

    if pattern["required_attributes"][0] == '$'
      ServiceLog.log('Pattern: validate_required_attributes? : required attributes has an invalid leading char')
      return false
    end

    true
  end


  ###################################################################
  # validate_sak: this attribute in the pattern specifies the
  # keys, any of which,  must be present in a document in the datadump
  # in order for the document to be included in a view. The SAK
  # attribute need not however be in the view documents
  # Note: The sak attribute must be lowercase
  ####################################################################

  def self.validate_sak?(pattern)

    ServiceLog.stack_trace('pattern::validate_sak?') if Rails.application.config.STACK_TRACE_PATTERN

    return true if  pattern['sak'].blank?

    # the sak fields must be string data type
    if pattern['sak'].class != String
      ServiceLog.log('Pattern: validate_sak? : sak is not of type string')
      return false
    end

    if pattern["sak"][0] == '$'
      ServiceLog.log('Pattern: validate_sak? : the view name has an invalid leading char')
      return false
    end


    true
  end


  ###################################################################
  # validate_projection: this attribute in the pattern specifies the
  # attributes which must be present in a persisted datadump document
  #  in order for the document to be included in a view. These
  # attributes are all included as attributes in each view document
  ####################################################################

  def self.validate_projection?(pattern)

    ServiceLog.stack_trace('pattern::validate_required_projection?') if Rails.application.config.STACK_TRACE_PATTERN

    return true if pattern['projection'].blank?

    # the project field  must be string data type
    if pattern['projection'].class != String
      ServiceLog.log('Pattern: validate_projection? : projection is not of type string')
      return false
    end

    if pattern["projection"][0] == '$'
      ServiceLog.log('Pattern: validate_projection? : the projection attribute has an invalid leading char')
      return false
    end


    true
  end


  ########################################################################
  # validate_generate_once: this attribute in the pattern is a boolean
  # cast as a string
  # If 'true', it specifies that the view will be generated only once
  # this gives a static view
  # Attribute value must be 'true' or 'false' and never blank
  #######################################################################

  def self.validate_generate_once?(pattern)

    ServiceLog.stack_trace('pattern::validate_generate_once?') if Rails.application.config.STACK_TRACE_PATTERN

    return true if pattern['generate_once'].blank?

    return false if pattern['generate_once'] != 'true' && pattern['generate_once'] != 'false'
    true
  end


  ###################################################################
  # validate_join: this attribute in the pattern specifies the
  # columns on which a join will occur. The view is the result of multi-
  # self-joins on the datadump
  ####################################################################

  def self.validate_join?(pattern)

    ServiceLog.stack_trace('pattern::validate_join?') if Rails.application.config.STACK_TRACE_PATTERN

    return true if pattern['join'].blank?

    # sanitize the join field
    pattern['join'] = sanitize_join(pattern['join'])

    str = pattern['join']
    array = str.split(':')

    # the first element must be the join type. second element is a
    # comma separated  join attributes string
    # only self equi-joins on the datadump are supported right now
    # a join expression must specify at least three columns for the join
    return false if array.size < 2

    # test for the join type
    return false if array[0].strip != 'equi-join'

    # test the join arguments
    return false unless self.valid_join_arg?(array[1])

    return true
  end


  ###################################################################
  # validate_generate_on_query: this attribute in a pattern is a
  # boolean represented as the string 'true' or 'false'
  # If true, it specifies that the view will be updated every time a
  # query is received. This gives a dynamic view. If false the
  # view is updated periodically by the scheduler or not updated
  # at all
  ####################################################################

  def self.validate_generate_on_query(pattern)

    ServiceLog.stack_trace('pattern::validate_generate_on_query?') if Rails.application.config.STACK_TRACE_PATTERN

    return true if pattern['generate_on_query'].blank?

    return false if pattern['generate_on_query'] != 'true' && pattern['generate_on_query'] != 'false'
    true
  end


  #######################################################################
  # validate_indexes: this attribute in the pattern specifies
  # the keys which will be used to index the view.
  # index cannot contain a dot operator
  ########################################################################

  def self.validate_indexes?(pattern)

    ServiceLog.stack_trace('pattern::validate_indexes?') if Rails.application.config.STACK_TRACE_PATTERN

    return true if pattern['indexes'].blank?
    return false if pattern['indexes'].include?('.')

    # the indexes field must be string data type
    return false if pattern['indexes'].class != String

    # sanitize the indexes field
    pattern['indexes'] = self.sanitize_comma_separated_string(pattern['indexes'])

    return self.valid_comma_separated_format?(pattern['indexes'])
  end


  ###########################################################################
  # validate_append_always_flag: this attribute in the pattern is an
  # optional boolean attribute.
  # If true, it specifies that when data relevant to a view comes in then
  # it must be immediately appended to the view.
  # Attribute value must be the string 'true' or 'false' or empty
  # empty is interpreted as append_always = false
  ############################################################################

  def self.validate_append_always?(pattern)

    ServiceLog.stack_trace('pattern::validate_append_always?') if Rails.application.config.STACK_TRACE_PATTERN

    return true if pattern['append_always'].blank?

    if pattern['append_always'] != 'true' && pattern['append_always'] != 'false'
      ServiceLog.log('Pattern: append-always? : attributes is not string : true or false')
      return false

    end

    return true
  end


  ##########################################################################
  # validate_reset_view: this attribute in the pattern specifies
  # the view is to be deleted and all documents after the reset time to
  # be henceforth included in the view
  ###########################################################################

  def self.validate_reset_view?(pattern)

    ServiceLog.stack_trace('pattern::validate_reset_view?') if Rails.application.config.STACK_TRACE_PATTERN

    return true if pattern['reset_view'].blank?
    return self.is_epoch_time?(pattern['reset_view'])
  end


  #######################################################################
  # validate_append_interval: this attribute in the pattern specifies
  # the view is to updated every x seconds
  #######################################################################

  def self.validate_append_interval?(pattern)

    ServiceLog.stack_trace('pattern::validate_append_interval?') if Rails.application.config.STACK_TRACE_PATTERN

    return true if pattern['append_interval'].blank?
    return self.is_epoch_time?(pattern['append_interval'])
  end


  #######################################################################
  # validate_create_at: this attribute in the pattern specifies
  # the epoch time at which the view is to be created
  #######################################################################

  def self.validate_create_at?(pattern)

    ServiceLog.stack_trace('pattern::validate_create_at?') if Rails.application.config.STACK_TRACE_PATTERN

    return true if pattern['create_at'].blank?
    return self.is_epoch_time?(pattern['create_at'])
  end


  #######################################################################
  # inconsistent_attributes? verifies whether various combinations of
  # attributes are consistent
  #######################################################################

  def self.consistent_attributes?(pattern)
    ServiceLog.stack_trace('pattern::consistent_attributes?') if Rails.application.config.STACK_TRACE_PATTERN

    return false if pattern['generate_once'] == 'true' &&  pattern['append_always'] == 'true'
    return false if pattern['generate_once'] == 'true' &&  pattern['generate_on_query'] == 'true'
    return false if pattern['generate_once'] == 'true' &&  pattern['append_interval']

    true
  end


  #########################################################
  # sanitize_pattern: clean up the received pattern
  ########################################################

  def self.sanitize_pattern(pattern)
    return if pattern.blank?

    # sanitize the required attributes field
    if pattern['required_attributes'].present?
      pattern['required_attributes'] = self.sanitize_comma_separated_string(pattern["required_attributes"])

    # sanitize the Projection field

    elsif pattern['projection'].present?
      pattern['projection'] = self.sanitize_comma_separated_string(pattern['projection'])
    end

    pattern
  end


  ####################################################################
  # A FEW HELPER FUNCTIONS TO ASSIST VALIDATIONS
  ####################################################################


  ####################################################################
  # valid_comma_separated_format?
  # The calling function must sanitize the string so that it does
  # not contain string sub-sequences such as ,,
  # tests that a comma separated attribute string has a proper format
  # cannot include prohibited characters: '/' or $ (as the first char)
  # SEE NOTE IN service_environment.rb
  ####################################################################

  def self.valid_comma_separated_format?(comma_separated_string)

    ServiceLog.stack_trace('pattern::valid_comma_separated_format?') if Rails.application.config.STACK_TRACE_PATTERN

    # boundary cases
    return true  if comma_separated_string.blank?
    return false if comma_separated_string.class != String

     array = comma_separated_string.split(',')

     # prevent simple injection and buffer overrun attacks
     array.each {|column| return false if column.include?("/'") || column.include?("\/") || column[0] == '$'}

     true
  end


  ####################################################################
  # sanitize_comma_separated_string
  # remove ,, sub-sequences in the string, remove whitespace
  # returns the sanitized string
  ####################################################################

  def self.sanitize_comma_separated_string(str)

    ServiceLog.stack_trace('pattern::sanitize_comma_seperated_string') if Rails.application.config.STACK_TRACE_PATTERN

    array = str.split(',').to_a
    array.reject! { |element| element.empty? }

    # remove leading and trailing whitespace
    tmp = array.map! { |el|  el.strip }

    tmp.join(',')
  end


  ####################################################################
  # sanitize_join
  # remove :: sub-sequences in the join string
  # returns the sanitized string
  ####################################################################

  def self.sanitize_join(str)
    ServiceLog.stack_trace('pattern::sanitize_join') if Rails.application.config.STACK_TRACE_PATTERN

    array = str.split(':').to_a
    array.reject! { |element| element.empty? }
    array.join(':')
  end


  ###################################################################
  # valid_join_arg
  # tests for simple sql injection, buffer overrun patterns
  # validates that a string has the format for a valid mongoDB key
  ###################################################################

  def self.valid_join_arg?(join_argument)
    ServiceLog.stack_trace('pattern::valid_join_args?') if Rails.application.config.STACK_TRACE_PATTERN

    str = join_argument.strip

    return false if str == '' || str.include?("'") || str.include?("'")  ||
        str.include?('\'') || str.include?(".") || str[0] == "$"

    true
  end

  ####################################################################
  # validate that a time is a valid epoch time
  # receives the epoch time as a string
  ####################################################################

  def self.is_epoch_time?(epoch_time)
    ServiceLog.stack_trace('pattern::is_epoch_time?') if Rails.application.config.STACK_TRACE_PATTERN

    return true if epoch_time =~ /\A\d+\Z/
    true if Float(epoch_time) rescue false
  end


  ######################################################################
  # validate_selection: validates a selection expression
  ######################################################################

  def self.validate_selection?(pattern)
    ServiceLog.stack_trace('pattern::validate_selection?') if Rails.application.config.STACK_TRACE_PATTERN

    return true if pattern["selection"].blank?

    operator_array   = [ 'gt', 'lt', 'gte', 'lte','eq', 'ne' ]
    relational_array = [ 'and', 'or' ]

    #--------------------------------------
    #  normalize selection expression
    #--------------------------------------
    self.normalize_selection(pattern)


    # validate that the variables and operators
    # are where they are supposed to be
    array = pattern["selection"].split(' ')
    array = array.reject { |el| el.blank? }

    ctr = 0

    while ctr < array.length do

      # evaluate the variable operand - variable cannot be an operator
      if operator_array.include?(array[ctr])
        ServiceLog.log('Pattern: validate_selection? : an operand is an operator')
        return false
      end


        if relational_array.include?(array[ctr])
          ServiceLog.log('Pattern: validate_selection? : an operand is an predicate operator')
          return false
        end

      # variable operand must be of type string
      if array[ctr].class != String
        ServiceLog.log('Pattern: validate_selection? : operand is not of type string')
        return false
      end

      if array[ctr].to_f != 0
        ServiceLog.log('Pattern: validate_selection? : operand cannot be a number')
        return false
      end

      ctr += 1
      return true if ctr >= array.length

      # check for a valid arithmetic operator
      if operator_array.include?(array[ctr]).blank?
        ServiceLog.log('Pattern: validate_selection? : an unknown arithmetic operator was encountered')
        return false
      end

      # expression terminates with the arithmetic operation
      ctr += 1
      if ctr >= array.length
        ServiceLog.log('Pattern: validate_selection? : selection expression terminates with an arithmetic operator')
        return false
      end

      # check for a valid string or numeric constant on the right hand side
      if array[ctr].class != String && array[ctr].class != Fixnum
        ServiceLog.log('Pattern: validate_selection? : right hand side of an expression is not a string or a number')
        return false
      end

      # expression validly terminates with a string constant or a numeric
      ctr += 1
      return true if ctr >= array.length

      # check for a valid relational operator
      if relational_array.include?(array[ctr]).blank?
        ServiceLog.log('Pattern: validate_selection? : an unknown boolean operator was encountered')
        return false
      end

      # expression terminates with the relational operator
      ctr += 1
      if ctr >= array.length
        ServiceLog.log('Pattern: validate_selection? : selection expression terminated with a boolean operator')
        return false
      end

    end

    true
  end



  ##############################################################################
  # normalize_selection: makes the selection expression into a canonical
  # form. Capitalizes the relation operators, converts the arithmetic
  # operators into a form suitable for mongodb
  ##############################################################################

  def self.normalize_selection(pattern)
    ServiceLog.stack_trace('pattern::normalize_selection') if Rails.application.config.STACK_TRACE_PATTERN

    # uniform spaces in the expression
    tmp = pattern["selection"].split(' ')
    tmp.reject!{ |element| element.blank? }
    tmp.map! { |element| ' ' + element + ' ' }
    pattern["selection"] = tmp.join

    # normalize the relational operators
    pattern["selection"].gsub!( /\bor\b/i, 'or')
    pattern["selection"].gsub!( /\band\b/i, 'and')

    # normalize the arithmetic operators
    pattern["selection"].gsub!( /\blt\b/i, 'lt')
    pattern["selection"].gsub!( /\blte\b/i, 'lte')

    pattern["selection"].gsub!( /\bgt\b/i, 'gt')
    pattern["selection"].gsub!( /\bgte\b/i, 'gte')

    pattern["selection"].gsub!( /\beq\b/i, 'eq')
    pattern["selection"].gsub!( /\bne\b/i, 'ne')

  end


  ####################################################################
  # clean_view_name
  # remove spaces from the view name.
  # returns the sanitized view name string in snakecase format
  ####################################################################

  def self.clean_view_name(view_name)
    ServiceLog.stack_trace('pattern::clean_view_name') if Rails.application.config.STACK_TRACE_PATTERN

    array = view_name.split(' ').to_a
    array.reject! { |element| element.empty? }
    array.join('_')
  end


  ########################################################################
  # get_all_patterns:
  # returns all patterns as a Array of hash objects
  ########################################################################

  def self.get_all_patterns

    begin
      docs = Pattern.all.to_a
    rescue => e
      ServiceLog.log('Pattern: get_all_patterns : failed to get all patterns')
      return false
    end
    docs
  end


  ########################################################################
  # search_for_pattern:
  # returns a pattern (as a hash object)
  # argument : pattern name which is the search key
  ########################################################################

  def self.get_pattern(search_key)

    begin
      docs = Pattern.all("pattern_name" => search_key).to_a
    rescue
      ServiceLog.log('Pattern: get_pattern : pattern search failed')
      return false
    end
    docs
  end


  #######################################################################
  # persist_pattern: receives a pattern ruby object to be persisted into
  # the  pattern collection. In addition invokes the Interpreter Service
  # in order to create a node to be added to the scheduler
  # return true or false
  # returns true if the pattern with the sae pattern_id already exists
  # in the Pattern collection
  #######################################################################

  def self.persist_pattern(pattern)
    ServiceLog.stack_trace('pattern::persist_pattern') if Rails.application.config.STACK_TRACE_PATTERN

    # do not persist trivial patterns
    if pattern == {}
      ServiceLog.log('Pattern: persist_pattern : cannot persist an empty pattern')
      return false
    end

    #----------------------------------------------------------
    # ascertain whether the pattern has already been persisted
    # uniqueness constraint on pattern id
    #----------------------------------------------------------
    patterns_array = Pattern.all
    patterns_array.each do |pat|
      return false if pat["pattern_id"] == pattern["pattern_id"]
    end

    # persist to the patterns collection
    begin
      Pattern.create(pattern);
    rescue => e
      ServiceLog.log('persist_pattern : failed to insert pattern into the pattern collection : ' + e.message)
      return false
    end

    #----------------------------------------------------
    # invoke the interpreter to interpret the pattern
    # and create state and dml objects for the Scheduler
    #-----------------------------------------------------

    ret = Interpreter.instance.interface(pattern) ? true : false
    ret
  end


  ########################################################################
  # update_pattern : update a pattern
  # receives the replacement pattern
  # a search key
  # returns true if the operation succeeds
  ########################################################################

  def self.update_pattern(pattern, search_key)
    ServiceLog.stack_trace('pattern::update_pattern') if Rails.application.config.STACK_TRACE_PATTERN

    Pattern.destroy_all(pattern_name: search_key)

    begin
      self.persist_pattern(pattern)
      return true
    rescue => e
      ServiceLog.log('update_pattern : failed to update pattern : ' + e.message)
      false
    end
  end


  ########################################################################
  # destroy_pattern : deletes a pattern
  # receives the pattern name of the pattern to be deleted
  # removes the pattern from the pattern collection and from the
  # the Scheduler's node array
  # returns true if the operation succeeds, false otherwise
  ########################################################################

  def self.destroy_pattern(search_key)

    ServiceLog.stack_trace('pattern::destroy_pattern') if Rails.application.config.STACK_TRACE_PATTERN

    begin
      Pattern.destroy_all(:pattern_name => search_key)
    rescue => e
      ServiceLog.log('destroy_packet : failed to delete pattern from the Pattern collection : ' + e.message)
      return false
    end

    # delete the node in the scheduler corresponding to the pattern
    Scheduler.interface('delete', search_key)

    true

  end

end
