############################################################################################
# The executor traverses the nodes in the scheduler and executes the DML if so warranted
# by the state object in the node
# Receives the node array
#
# Data Structure: linked list of nodes
# Each node contains
#      [1] a state object (as created by the interpreter's tokenizer)
#      [2] a DML object (as created by the interpreter's DML emitter)
#
# uses the singleton pattern
#
# ALGORITHM NOTE: A pattern is not EXPLICITLY linked to any type of data packet.
# For each node in the node list of the scheduler, the Executor checks whether the
# data_packet satisfies all of the constraints that are set out in the nodes State Object,
# if so the data_packet will be further processed and if all goes well, the packet shall be
# inserted into a view and forwarded to the legacy system.
# Note that a single data_packet can be in placed in multiple views (for example, multiple
# report views). A data_packet qualifies for further processing if it satisfies all of the
# topology constraints in the pattern object.ß
# copyright: collabnet, 2015-2016
############################################################################################


class Executor

  include MongoMapper::Document


  #############################################################################
  # interface: entry point into the Executor service
  # Receives:  the array of scheduler nodes
  #            the data packet (a hash) to be persisted
  #
  # Algorithm : traverses the set of nodes and execute the DML in a node if
  # warranted by the node's state object
  #
  # returns false if the data_packet was not inserted into a view
  # otherwise true - meaning that the data_packet met the constraints of a pattern
  # and was thus insertable and was actually inserted into one or more
  # views and/or  sent to RabbitMQ for insertion into the legacy activity
  # collection
  #
  # NOTE: data_packet can be inserted into one or more views (say for example, report
  # views) but only once into the legacy Activity collection
  ##############################################################################

  def self.interface(nodes, data_packet)

    ServiceLog.stack_trace('executor::interface') if Rails.application.config.STACK_TRACE_EXECUTOR

    # Test that scheduler nodes exist. If no nodes exist
    # the data packet is only persisted into the datadump
    if nodes.blank?
      ServiceLog.log('executor::interface - The scheduler has no nodes, only persisting the data_packet to the data dump')
      return true
    end

    ServiceLog.log('executor::interface - scheduler node count :' + nodes.size.to_s)

    return false if self.sanity_test?(nodes, data_packet).blank?

    self.initialize

    #-----------------------------------
    # save the type of the data_packet
    #-----------------------------------

    @data_packet_type = data_packet['packet_type']

    #-----------------------------------
    # create a flat namespaced key hash
    # from the data_packet
    #-----------------------------------

    self.make_flat_namespaced_data_packet('', data_packet)

    if @flat_namespaced_data_packet == {}
      ServiceLog.log('executor::interface the flat namespaced data_packet is an empty hash')
      return false
    end

    #-----------------------------------------------------
    # scheduler node processing loop
    # iterate through the nodes testing
    # for data_packet insertion
    # every time the data_packet is inserted in
    # a view the view_insertions counter will increment
    # @queued_ctr is: data_packets queued by the scheduler
    #-------------------------------------------------------
    nodes.each do |node|

      ret = self.process_scheduler_node(node, data_packet)

      #----------------------------------
      # increment the insertions counter
      #----------------------------------
      (@insertions_ctr += 1) if ret

    end

    [@insertions_ctr, @queued_ctr, @projection_object]

  end


  private


  ############################################################################
  # sanity_test? :  validation of interface parameters
  # receives: array of scheduler nodes
  #           the data data_packet
  ############################################################################

  def self.sanity_test?(nodes, data_packet)

    ServiceLog.stack_trace('executor::sanity_test?') if Rails.application.config.STACK_TRACE_EXECUTOR

    #---------------------
    # packet verification
    #---------------------
    if data_packet.blank?
      ServiceLog.log('Executor::sanity_test?: data_packet not received')
      return false
    end

    if data_packet['packet_type'].blank?
      ServiceLog.log('executor::interface: the packet does have a packet-type ')
      return false
    end

    #-------------------
    # node verification
    #-------------------
    nodes.each do |node|
      if node['dml_object'].blank? || node['state_object'].blank?
        ServiceLog.log('Executor::sanity_test?: missing state object or dml object in scheduler node')
        return false
      end

      if node['dml_object'].is_a?(Hash).blank? || node['state_object'].is_a?(Hash).blank?
        ServiceLog.log('Executor::sanity_test?: scheduler node has invalid type')
        return false
      end
    end

    true
  end



  ############################################################################
  # initialize the singleton
  ############################################################################

  def self.initialize

    ServiceLog.stack_trace('executor::initialize') if Rails.application.config.STACK_TRACE_EXECUTOR

    @projection_object = {}
    @view_object = {}
    @flat_namespaced_data_packet = {}
    @namespaced_key = ''
    @emulation_flag = true
    @insertions_ctr = 0
    @queued_ctr     = 0


  end

  #############################################################################
  # self.process_scheduler_node: processes a node in the scheduler
  # receives: the node
  #           the data_packet
  # returns false if the node is not processed, true otherwise
  #############################################################################

  def self.process_scheduler_node(node, data_packet)
    ServiceLog.stack_trace('executor::self.process_scheduler_node') if Rails.application.config.STACK_TRACE_EXECUTOR

    #---------------------------------------------
    # test that the data_packet has the required,
    # projection and selection attributes set
    # forth by the pattern; these are the
    # packet topology constraints
    #----------------------------------------------

    return false unless self.insertable_packet?(node["dml_object"], data_packet)


    #--------------------------------------------
    # it is an insertable data_packet, so make a
    # projection object
    #--------------------------------------------

    return false unless self.make_projection_object(node["dml_object"])


    #-------------------------------------------------
    # transform the projection object in accordance
    # with the emulator table
    # this function must be called after
    # self.make_self.projection_object
    #--------------------------------------------------

    if @emulation_flag
      @projection_object = Emulator.interface(@projection_object, @data_packet_type, 'RUN')
      @emulation_flag = false

      if @projection_object.blank?
        ServiceLog.log('Executor::self.process_scheduler_node: emulation failed - the projection object is blank')
        return false
      end
    end

    #-----------------------------------------------------------------
    # get the packet maps applicable to this data packet.
    # There can be more than one and apply the
    # packet mappings to the data packet. One data_packet
    # can spawn n resultant data_packets destined for persistence
    # Algorithm note:
    # This is a n by n relationship. one packet map can be associated
    # with n packet maps and one packet map can be associated with n
    # patterns
    #------------------------------------------------------------------

    mappings_array = self.get_packet_maps(node['pattern_name'])
    return false if mappings_array.blank?

    #----------------------------------------------------
    # spawn packets for persistence
    # persist one packet for each mappings_array element
    #----------------------------------------------------
    mappings_array.each do |packet_map|
      next if self.packet_persistence_processor(packet_map, node).blank?
    end

  end


  #############################################################################
  # self.make_flat_namespaced_data_packet - recursive function
  # receives: the current namespaced key
  #           the data data_packet (ruby hash object)
  # traverses the data_packet hash and converts keys into namespaced keys (such
  # as fields.name.address.tel)
  #
  # converts the (possibly nested) data_packet hash object:
  #   {
  #     'sak' => 'abc'
  #     'fields' => { 'status' => { 'name' => 'abc] } },
  #     etc.
  #   }
  # 
  # and makes it into a flat
  # key value hash such as:
  #   {
  #       "fields.status.name" => 'abc',
  #       "sak" => "abc"
  #       "name.address.area.code" => "abcdefg"
  #   }
  #############################################################################

  def self.make_flat_namespaced_data_packet(current_namespaced_key, data_packet)

    ServiceLog.stack_trace('executor::self.make_flat_namespaced_data_packet') if Rails.application.config.STACK_TRACE_EXECUTOR

    data_packet.each do |key, value|

      # set the namespaced key
      namespaced_key = current_namespaced_key.blank? ? key : current_namespaced_key + '.' + key

      #-----------------------------------------
      # map the key traversal to the service log
      # for debugging
      #-----------------------------------------

       if Rails.application.config.SHOW_NAMESPACE_TRAVERSAL
         ServiceLog.log("key : #{namespaced_key}")
       end

      #-----------------------------------------
      # either save the key value to the flat
      # hash or traverse into the value hash
      #-----------------------------------------

      if value.kind_of?(Hash) && value != {}
        self.make_flat_namespaced_data_packet(namespaced_key, value)
      else
        @flat_namespaced_data_packet[namespaced_key] = value
      end

    end

  end



  ######################################################################################
  # self.insertable_packet?: tests whether a data_packet satisfies the constraints
  # in a DML object
  # receives a DML object as well as a data_packet
  #
  # Algorithm:  Check whether the data_packet satisfies the constraints set out in the
  # pattern's required_attributes, selection and projection fields
  #
  # Returns true or false
  ######################################################################################

  def self.insertable_packet?(dml, data_packet)

    ServiceLog.stack_trace('executor::self.insertable_packet?') if Rails.application.config.STACK_TRACE_EXECUTOR

    #---------------------------------------------------
    # does the data_packet have all of the fields required
    # for selection as a candidate document
    #---------------------------------------------------

    if dml['required_attributes'].present?

      required_attributes = dml["required_attributes"].split(',')

      required_attributes.each do |field|

        if @flat_namespaced_data_packet.has_key?(field).blank?
          ServiceLog.log("executor::insertable_data_packet - data data_packet does not contain the required attribute: #{field}")
          return false
        end

      end
    end

    #--------------------------------------------------------
    # does the data_packet have all of the projection fields
    #--------------------------------------------------------

    if dml["projection"].present?
      projection_attributes = dml["projection"].split(',')

      projection_attributes.each do |field|

        if @flat_namespaced_data_packet.has_key?(field).blank?
          ServiceLog.log("executor::insertable_data_packet - data data_packet does not contain projection attribute: #{field}")
          return false
        end
      end
    end


    #-----------------------------------------------------
    # does the data_packet have all of the fields required by
    # the selection attribute in the pattern and are the
    # selection attribute constraints met by the data_packet
    #------------------------------------------------------

    if dml["selection"].present?
      @namespaced_key = ''

      return false if self.selection_evaluation(dml["selection"], data_packet).blank?
    end

    true
  end


  ###########################################################################
  # self.make_projection_object: receives the DML object of the node being
  # processed
  #
  # mutates the projection object. This object will be inserted into a
  # legacy view or a new system view (ruby hash object)
  #
  # Algorithm: traverses a single level or nested data_packet (ruby hash object)
  # searching for target keys in the data_packet (as required by the DML).
  # Inserts the key and and the value of the target key will be inserted into a
  # projection object.
  #
  # Algorithm Notes (i) if the projection attribute is empty then the required
  # attributes will be projected
  # (ii) the entire data data_packet CANNOT be projected into a view
  #
  ############################################################################

  def self.make_projection_object(dml_object)

    ServiceLog.stack_trace('executor::self.make_self.projection_object') if Rails.application.config.STACK_TRACE_EXECUTOR

    #----------------------------------------------------
    # if the DML does not have a projection attribute
    # then the fields in the projection object will be
    # the required attributes
    #-----------------------------------------------------

    @projection_object = {}

    #-------------------------------------
    # iterate over the projection fields
    #-------------------------------------

    if dml_object['projection'].present?
      array = dml_object['projection'].split(',')

      array.each do |attribute|
        @namespaced_key = ''
        
        if @flat_namespaced_data_packet.has_key?(attribute).blank?
          ServiceLog.log("executor::self.make_self.projection_object - data_packet does not contain projection attribute #{attribute}")
          return false
        else
          @projection_object[attribute] = @flat_namespaced_data_packet[attribute]
        end
      end


      #----------------------------------------------
      # projection attribute was not present
      # therefore iterate over the required attributes
      #----------------------------------------------
    else
      # iterate over the required_attributes field
      array = dml_object['required_attributes'].split(',')

      array.each do |attribute|
        @namespaced_key = ''

        if @flat_namespaced_data_packet.has_key?(attribute).blank?
          ServiceLog.log("executor::insertable_data_packet - data_packet does not contain required attribute #{attribute}")
          return false
        else
          @projection_object[attribute] = @flat_namespaced_data_packet[attribute]
        end

      end
    end

    #------------------------------------
    # projection object cannot be blank
    #------------------------------------
    if @projection_object.blank? || @projection_object == {}
      ServiceLog.log("Executor::self.make_self.projection_object : error: the projection object is empty (some attributes must be projected")
      return false
    end

    true
  end


  ######################################################################################
  # self.selection_evaluation: receives
  #                        [1] the selection attribute in the DML (a string)
  #                        [2] data data_packet (ruby hash)
  #
  # evaluates whether the data_packet contains all of the fields required by
  # the selection attribute of the DML  and then evaluates the possibly compound
  # selection attribute given the data in the data_packet
  # returns true or false
  #
  # selection expression format:
  #     [fully qualified namespaced key] ARITHMETIC OPERATOR
  #                  [string or numeric contsant] [RELATIONAL OPERATOR: AND, OR ] ...
  #
  #  The exact parameter with name 'data_packet' is used by self.map_namespaced_operands-
  #  do not change.
  ######################################################################################

  def self.selection_evaluation(selection_attribute, data_packet)

    ServiceLog.stack_trace('executor::self.selection_evaluation') if Rails.application.config.STACK_TRACE_EXECUTOR

    # constraints are trivially met
    return true if selection_attribute.blank?

    selection_array = selection_attribute.split(' ')


    #--------------------------------------------
    # test that all of the required namespaced
    # attributes exist
    #--------------------------------------------
    ctr = 0
    while ctr < selection_array.length do
      return false if selection_array[ctr].blank?
      ctr += 4
    end

    #------------------------------------------
    # modify the selection operators to create
    # an evaluatable boolean predicate
    #------------------------------------------
    ctr = 0

    while ctr < selection_array.length do

      #------------------------------------
      # modify operands: enclose operands
      # in quotes for instance_eval
      #------------------------------------
      if selection_array[ctr].class == String

        operand = self.map_namespaced_operands(selection_array[ctr])
        return false if operand.blank?

        selection_array[ctr] =  operand
      else
        ServiceLog.log('Executor::selection_expression?: fully qualified key name operand is not off type string')
        return false
      end

      ctr += 4
    end

    #-----------------------------------------
    # modify the operators in the expression
    # so we have an evaluatable predicate
    # The operator transformation table is:
    #     'lt'   to   '<'
    #     'lte'  to   '<='
    #     'eq'   to   ' =='
    #     'gt'   to   '>'
    #     'gte'  to   '>='
    #     'ne'   to   '!='
    #-----------------------------------------

    ctr = 1

    while ctr < selection_array.length do
      if selection_array[ctr] == 'lt'
        selection_array[ctr] = '<'

      elsif selection_array[ctr] == 'lte'
        selection_array[ctr] = '<='

      elsif selection_array[ctr] == 'eq'
        selection_array[ctr] = '=='

      elsif selection_array[ctr] == 'gt'
        selection_array[ctr] = '>'

      elsif selection_array[ctr] == 'gte'
        selection_array[ctr] = '>='

      elsif selection_array[ctr] == 'ne'
        selection_array[ctr] = '!='
      end

      ctr += 4
    end

    #----------------------------------------
    # do relational operator substitutions
    #----------------------------------------

    ctr = 3

    while ctr < selection_array.length do

      operator = selection_array[ctr].upcase

      if operator == 'OR'
        selection_array[ctr] = '||'
      elsif operator == 'AND'
        selection_array[ctr] = '&&'
      else
        ServiceLog.log('Executor::selection_expression?: unknown operator symbol in selection attribute')
        return false
      end

      ctr += 4

    end

    # re-assemble the selection expression
    select_expression = selection_array.join(' ')

    #---------------------------------------
    # now evaluate the predicate selection
    # predicate expression
    #---------------------------------------
    begin
      return false if instance_eval(select_expression).blank?
    rescue => e
      ServiceLog.log("Executor::selection_expression?: selection evaluation error #{e.message}")
      return false
    end

    true
  end


  ######################################################################################
  #  self.map_namespaced_operands
  #  Receives a namespaced operand such as 'issue.yield.usd'
  #  Algorithm: [1] prefix the namespaced operand with 'data_packet['
  #             [2] replace each occurrence of '.' with '][]'
  #             [3] suffix the operand with ']'
  #  This converts the namespaced operand into a key in the data packet
  #  returns true if the key has a value; false otherwise
  ######################################################################################

  def self.map_namespaced_operands(operand)

    ServiceLog.stack_trace('executor::self.map_namespaced_operands') if Rails.application.config.STACK_TRACE_EXECUTOR

    tmp = operand.gsub('.', "']['")
    if tmp.blank?
      ServiceLog.log("Executor::self.map_namespaced_operands: empty operand ")
      return false
    end

    "data_packet['#{tmp}']"

  end


  #####################################################################################
  # self.packet_persistence_processor : preliminary processing to persist the projection
  # object
  # Verifies that the packet contains all of the fields required by the packet map
  #
  # receives (i)   packet map (hash) that will be used to transform the data_packet
  #          (iii) the scheduler node that is being processed
  #
  # returns false if the data_packet was not persisted into the legacy system
  #         otherwise returns the data_packet pushed onto RabbitMQ
  #####################################################################################

  def self.packet_persistence_processor(packet_map, node)
    ServiceLog.stack_trace('executor::self.packet_persistence_processor') if Rails.application.config.STACK_TRACE_EXECUTOR

    #----------------------------------------
    # verify that the mapping fields exists
    #----------------------------------------
    return false if self.valid_mapping?(@projection_object, packet_map).blank?

    persisted_data_packet = false
    
    #---------------------------------------------------
    # persist the data_packet into the legacy Activity
    # collection
    #---------------------------------------------------
    persisted_data_packet =  self.execute_for_legacy_system(packet_map) if Rails.application.config.BIGLOCK <= 1

    return false if persisted_data_packet.blank? && Rails.application.config.PERFECT_LEGACY_EMULATION


    #---------------------------------------------------
    # persist the data_packet into the new system
    #---------------------------------------------------
    if Rails.application.config.PERFECT_LEGACY_EMULATION
      self.persist_for_interpreter_system(node, persisted_data_packet) if Rails.application.config.BIGLOCK >= 1
    else
      self.persist_for_interpreter_system(node, @projection_object) if Rails.application.config.BIGLOCK >= 1
    end

  end


  ######################################################################################
  # self.persist_for_interpreter_system: persists a data data_packet into a view
  # collection
  #####################################################################################

  def self.persist_for_interpreter_system(node, data_data_packet)

    ServiceLog.stack_trace('executor::self.persist_for_interpreter_system') if Rails.application.config.STACK_TRACE_EXECUTOR

    #----------------------------------------------
    # execute the DML statements with respect to
    # the new interpreter system or the legacy
    # system or both
    #----------------------------------------------
   self.persist_into_view(node["dml_object"], data_data_packet)


  end


  ###############################################################################
  # self.persist_into_view  executes a mongodb command.
  # receives the dml object and a ruby hash object for persistence
  # returns false if the view collection is not created, true otherwise
  ###############################################################################

  def self.persist_into_view(dml_object, data_packet)
    ServiceLog.stack_trace('executor::self.persist_into_view') if Rails.application.config.STACK_TRACE_EXECUTOR

    # ---------------------------------------
    # dynamically create a view class with
    # the MongoMapper module mixed in
    #----------------------------------------

    Object.const_set dml_object["view_name"].camelize, Class.new { include MongoMapper::Document  }

    #-------------------------------------------------
    # add all of the selected documents into the the
    # view collection create the view collection if
    # it does not exist
    #-------------------------------------------------

    view =  dml_object["view_name"].camelize.constantize

    #-------------------------------------------------------
    # IMPORTANT
    # get rid of the id field in the data_packet since MongoDB
    # uses it's own id as the primary key
    #-------------------------------------------------------

    data_packet = data_packet.reject { |key, value| key == "id" }


    #-------------------------------------
    # Required because mongoDB keys cannot
    # contain dots
    #-------------------------------------
    data_packet = self.map_namespace_separator(data_packet)

    #-----------------------
    # persist to a view
    #-----------------------
    begin
      view.create(data_packet)
    rescue => e
      ServiceLog.log("Executor::self.persist_into_view: failed to persist data_packet into a view: #{e.message}")
      return false
    end

    true
  end


  ################################################################################
  # execute the mongoDB DML to maintain the legacy system
  # Receives: the packet map
  # returns the data packet if persisted into the legacy system else false
  #
  # NOTE: TO PERSIST INTO THE LEGACY SYSTEM data_packet HASH MUST HAVE SAK AND
  # EVENT_TIME ATTRIBUTES THE SAK IS PICKED UP FROM THE URL
  ###############################################################################

  def self.execute_for_legacy_system(packet_map)
    ServiceLog.stack_trace('executor::self.execute_for_legacy_system') if Rails.application.config.STACK_TRACE_EXECUTOR

    begin
      packet_map = JSON.parse(packet_map.to_json)
    rescue => e
      ServiceLog.log('self.execute_for_legacy_system: json parse error: '  + e.message)
      return false
    end

    CouplerService.instance.interface(@projection_object, @data_packet_type, packet_map)

  end


  ##########################################################################
  # self.get_packet_maps: gets the packet maps which will be used to transform
  # data packets which are destined for the views and legacy activities
  # collection
  # receives the data_packet_type (through the received data_packet)
  # NOTE: there can be more than one packet map record with the same
  # packet_type key
  # return [] if there is no matching packet mapping  records or otherwise
  # the array of data_packet maps corresponding to this type of data_packet.
  # return false if there is an error
  ###########################################################################

  def self.get_packet_maps(packet_type)
    ServiceLog.stack_trace('Executor::self.get_packet_maps') if Rails.application.config.STACK_TRACE_EXECUTOR

    if packet_type.blank?
      ServiceLog.log('executor:self.get_packet_maps : empty packet type received')
      return false
    end

    begin
      mappings = PacketMaps.get_packet_map(packet_type)
      if mappings.size == 0
        ServiceLog.log('executor:get_data_packet_maps : there is no packet map for the packet type: ' + packet_type)
        return []
      end
      rescue => e
        ServiceLog.log('executor:get_data_packet_maps : error getting packet maps for: ' + packet_type +" : " + e.message)
        return false
      end

    mappings.to_a
  end


  ############################################################
  # valid_mapping: determines whether all of the fields in
  # a packet mapping exist in the data packet
  # rejects the mapping if the fields do not exist in the
  # packet
  # returns true or false
  ############################################################

  def self.valid_mapping?(data_packet, map)
    ServiceLog.stack_trace('Executor::self.valid_mapping?') if Rails.application.config.STACK_TRACE_EXECUTOR

    root = map["packet_type"]

    if root.blank?
      ServiceLog.log('executor:valid_mapping: the packet map is invalid')
      return false
    end

    # verify that all of the fields required by the packet map actually
    # exist in the data packet
    map[root].each do  |key, value|

      begin
        return false if data_packet[value].nil?
      rescue
        ServiceLog.log('executor:valid_mapping: exception: the packet map is invalid')
        return false
      end
    end

  end


  ############################################################
  # self.map_namespace_separator
  # mongoDB does not permit a document key to have
  # dots in a key's name. We namespace data_packet keys as
  # (for example): fields.status.name
  # The namespace separator in service_environment.rb is used
  # to substitute dots in the namespaced keys
  # returns the re-keyed data_packet
  ############################################################

  def self.map_namespace_separator(data_packet)

    ServiceLog.stack_trace('Executor::self.map_namespace_separator') if Rails.application.config.STACK_TRACE_EXECUTOR

    separator = Rails.application.config.NAMESPACE_SEPARATOR
    data = {}


    data_packet.map do  |key, value|
      if key.is_a? Symbol
        data[key] = value
        next
      end

      tmp = key.gsub('.', separator)
      data[tmp] = value
    end

    data
  end


  ############################################################
  # projection object accessor
  ############################################################

  def self.projection_object
    @projection_object
  end

end
