############################################################################################
# This scheduler determines whether an arriving data packet (or a packet retrieved from the
# scheduler's queue) needs to be persisted. It the packet has to be persisted it CREATE the 
# view and sends the packet onwards to the to the Coupler Service. 
#
# Data Structure: nodes which are CREATEd from MongoDB persisted nodes
# Each node contains
#      [1] a state object (as CREATEd by the interpreter's tokenizer)
#      [2] a DML (data manipulation object) object (as CREATEd by the interpreter's DMLMapper)
#
# Algorithm:
#   [1] sleep
#   [2] receive data packet
#   [2] wake up
#   [3] start traversing the node linked list starting at the head node
#   [4] For each node, examine the state object and determine if the data packet satisfies 
#       the conditions in the nodes state object
#   [5] If the state conditions are satisfied, it:
#       [a] Use the DML Object to CREATE the view
#   [6] Go to the next node - repeat 4 and 5
#   [7] after last node processed go to sleep
#
# copyright, collabnet, 2016
# copyright, collabnet, 2016
############################################################################################



class Scheduler

  include MongoMapper::Document


  #############################################################################
  # interface: entry point into the Scheduler Service
  # Receives: A Verb : 'RUN', 'STOP',  'CREATE' or 'view'
  #           An optional  contextual parameter: either a ruby hash or a
  #           pattern name(for the DELETE verb only)
  #           An optional data manipulation language object
  #           The data packet

  # The state object embodies the information required to CREATE a view,
  # in other words it states the conditions under which a view is to be
  # CREATEd
  # The DML object embodies the symbolic SQL that will be used to CREATE the
  # view
  # returns true or false
  #
  # Verbs: 'RUN':  RUN the scheduler (iterate through the scheduler nodes)
  #        'STOP': STOP the scheduler
  #        'CREATE': add a new scheduler node embodying the received state_object
  #        'DELETE'  remove a node pertaining to the received pattern name
  ##############################################################################


  def self.interface(verb, state_object = nil, dml_object = nil, packet = nil)

    ServiceLog.stack_trace('Scheduler:interface') if Rails.application.config.STACK_TRACE_SCHEDULER

    if verb.blank?
      ServiceLog.log('Scheduler::interface : the scheduler has not received a verb')
      return false
    end

    verb = verb.upcase.strip if verb.is_a?(String)

    self.load_scheduler

    return false unless self.sanity_test?(verb, state_object, dml_object, packet)


    state_object

    #-----------------------
    # execute the verb
    #-----------------------
    case verb
      when 'CREATE' then return true if self.create_node(state_object, dml_object)
      when 'DELETE' then return true if self.delete_node(state_object)
      when 'RESET'  then return self.reset
      when 'RUN'
            ret = Executor.interface(@nodes, packet)
            return ret
      else
        ServiceLog.log('scheduler::instance : scheduler references an unknown verb ')
        return false
    end

    false

  end

  private

  ############################################################################
  # sanity_test?: RUN some basic tests on the arguments received on the
  # interface
  # returns a boolean true or false value
  ############################################################################

  def self.sanity_test?(verb, state_object, dml_object, packet)

    ServiceLog.stack_trace('Scheduler:sanity_test?') if Rails.application.config.STACK_TRACE_SCHEDULER

    #--------------------
    # verb verification
    #--------------------
    if verb.class != String
      ServiceLog.log('sanity_test? : the scheduler interface has a verb which is not of type string')
      return false
    end

    if ['RUN', 'STOP', 'CREATE', 'DELETE', 'RESET'].include?(verb).blank?
      ServiceLog.log('sanity_test? : the scheduler interface received an unknown verb:' + verb)
      return false
    end

    #---------------------------
    # DML Object verification
    #---------------------------
    if dml_object != nil && dml_object.is_a?(Hash).blank?
      ServiceLog.log('sanity_test? : the scheduler interface has received a DML object which is not of type Hash')
      return false
    end

    if verb == 'CREATE' && dml_object.class   != Hash
      ServiceLog.log('sanity_test? : the scheduler interface has the CREATE verb but the interface has not received a DML object of type Hash')
      return false
    end

    #-----------------------------
    # State Object Verification
    #-----------------------------

    # missing state object for CREATE verb
    if verb == 'CREATE' && state_object.class != Hash
      ServiceLog.log('sanity_test? : the scheduler interface has the CREATE verb but the interface did not receive a state object of type Hash')
      return false
    end

    # pattern name missing in the state object
    if verb == 'CREATE'
      if state_object["pattern_name"].nil?
        ServiceLog.log('sanity_test? : the scheduler interface has the CREATE verb but the interface received a state object with no pattern name')
        return false
      end

      if verb == 'DELETE' && state_object.blank?
        ServiceLog.log('sanity_test? : the scheduler interface has received the DELETE verb but the interface did not receive a DELETE key')
        return false
      end

      if verb == 'DELETE' && state_object.is_a?(String).blank?
        ServiceLog.log('sanity_test? : the scheduler interface has received the DELETE verb but did not receive a search key of type string')
        return false
      end

    end

    #------------------------------------------------------
    # RUN verb  cannot be associated with an empty packet
    #------------------------------------------------------
    if verb == 'RUN' && packet.blank?
      ServiceLog.log('sanity_test? : the scheduler interface has the RUN verb but the interface received no packet')
      return false
    end

    true

  end


  ############################################################################
  # load_scheduler : loads the scheduler nodes into memory from the
  # scheduler collection
  ############################################################################

  def self.load_scheduler

    ServiceLog.stack_trace('Scheduler:load_scheduler') if Rails.application.config.STACK_TRACE_SCHEDULER

    begin
      @nodes = Scheduler.all
    rescue => e
      ServiceLog.log('scheduler::load_scheduler : failed to get all the scheduler nodes from the collection : ' + e.message)
      return false
    end

    @nodes = @nodes || []
  end


  ##############################################################################
  # create_node : creates a new Scheduler node on the and persists the new node
  # into the database
  # receives a state object and the DML object
  # returns true or false
  ##############################################################################

  def self.create_node(state_object, dml_object)

    ServiceLog.stack_trace('Scheduler:create_node') if Rails.application.config.STACK_TRACE_SCHEDULER

    #----------------------------------------------------
    # do not persist the node if the collection already
    # contains a node with the pattern id
    #----------------------------------------------------

    begin
      ret = Scheduler.all(:pattern_id => state_object["pattern_id"] )
    rescue => e
      ServiceLog.log('scheduler::create_node : exception - failed to get the scheduler node from the collection: ' + e.message)
      return false
    end

    return true if ret.length > 0

    #--------------------------------------
    # Persist to the Scheduler Collection
    #--------------------------------------

    node = { "pattern_name" => state_object["pattern_name"], 'pattern_id' => state_object['pattern_id'],
             "state_object" => state_object, "dml_object" => dml_object }

    begin
      Scheduler.create(node);
    rescue => e
      ServiceLog.log('scheduler::create_node : failed to CREATE a scheduler node for the Scheduler collection : ' + e.message)
      return false
    end

    # Add the node
    @nodes << node

    true
  end


  ###################################################################################
  # delete_node: deletes a scheduler node
  # receives the pattern name for the node to be DELETEd
  ###################################################################################

  def self.delete_node(pattern_name)

    ServiceLog.stack_trace('Scheduler:delete_node') if Rails.application.config.STACK_TRACE_SCHEDULER

    return true if @nodes == []

    counter = @nodes.length

    begin
      Scheduler.delete_all(:pattern_name => pattern_name)
    rescue => e
      ServiceLog.log('delete_node : Exception deleting scheduler node: ' + e.message)
      return false
    end

    @nodes = @nodes.delete_if { |x| x["pattern_name"] == pattern_name }

    if counter == @nodes.length
      ServiceLog.log('delete_node : failed to a node from the scheduler collection')
      return false
    end

    true
  end


  ###################################################################################
  # reset_scheduler: DELETEs all of the scheduler nodes
  ###################################################################################

  def self.reset
    ServiceLog.stack_trace('Scheduler:reset') if Rails.application.config.STACK_TRACE_SCHEDULER

    begin
      Scheduler.delete_all
    rescue => e
      ServiceLog.log('reset_node : destroy all of the nodes in the scheduler collection : ' + e.message)
      return false
    end

    @nodes = []
    true

  end


  ##################################################################################
  # accessor
  ##################################################################################

  def self.accessor
    @nodes
  end

  ##################################################################################
  # mutator
  ##################################################################################

  def self.mutator
    @nodes = []
  end

end
