#!/opt/collabnet/ruby/bin/ruby
#
# Copyright 2016 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

require 'erb'
require 'getoptlong'

require File.expand_path('../../lib/application_constants.rb',  __FILE__)
require File.expand_path('../../lib/site_options.rb',  __FILE__)

# TODO: honor existing file permissions

class ConfigurationManager

  def initialize(options)
    @script_options = options
    opts_file = @script_options[:options_file]

    # NOTE: @site_options is a symbol-keyed hash, but with string values, so 
    # booleans need to be interpreted from strings.
    @site_options = SiteOptions.new()
    begin
      @site_options.load(APP_SITE_OPTIONS_FILE) if loadable?(APP_SITE_OPTIONS_FILE)
      @site_options.load(opts_file) if loadable?(opts_file)
    rescue Exception => e
      warn "Error in loading site options: #{e}"
    end

    if @site_options.empty?
      warn "No site options file was loaded; starting inferrence"
      infer_app_options
    end

    raise "No site options were loaded or inferred" if @site_options.empty?
  end

  def infer_app_options
    if File.file?(NGINX_CONFIG_FILE)
      File.open(NGINX_CONFIG_FILE) do |file|

        file.find { |line| line =~ /^\s*(server_name)\s+([^;\s]+)\s*;/ }
        file.rewind
        @site_options.set("#{APP_NAME}.server", $2) if $2 && !$2.empty?

        file.find { |line| line =~ /^\s*(listen)\s+(\d+)\s*;/ }
        file.rewind
        @site_options.set("#{APP_NAME}.port",   $2) if $2 && !$2.empty?

        file.find { |line| line =~ /^\s*(ssl)\s+(on)\s*;/ }
        file.rewind
        @site_options.set("#{APP_NAME}.use_ssl", $2 == 'on' ? 'true' : 'false')

        debug @site_options.inspect
      end
    end
    # TODO: Load defaults.yml file, and infer the balance of the settings from there
  end
  
  def update_site_options
    if @site_options.empty?
      warn "Skipping site options update, as no options are loaded."
      return 
    end    
    begin
      File.open(APP_SITE_OPTIONS_FILE, 'w') { |file| file.write(@site_options.to_options_string) }
    rescue Exception => e
      warn "Unable to update site options: #{e}"
    else 
      info "Site options file updated: #{APP_SITE_OPTIONS_FILE}"
    end
  end

  # Perform all the nginx-related updates
  def update_nginx
    update_nginx_config
  end

  # Update the nginx server config file based on current site-options settings
  def update_nginx_config
    begin
      @raw = File.read(NGINX_CONFIG_TEMPLATE)
      @template = ERB.new(@raw)

      # Vars for template substitution
      # NOTE: app_server is assumed to be true, until/unless behind a load 
      # balancer.  The installer's nginx plugin asserts this via merge.
      # TODO: or defaults?

      app_server         = true
      use_ssl            = booleanish(get_app_option(:use_ssl))
      app_server_name    = get_app_option(:server)
      app_server_port    = get_app_option(:port)

      debug "Nginix template vars:"
      debug "app_server      = #{app_server}"
      debug "use_ssl         = #{use_ssl}"
      debug "app_server_name = #{app_server_name}"
      debug "app_server_port = #{app_server_port}"
    
      # FIXME: support --dry-run
      debug @template.result(binding)
      File.open(NGINX_CONFIG_FILE, 'w') { |file| file.write(@template.result(binding)) }
    rescue Exception => e
      warn "Unable to update nginx configuration: #{e}"
    else 
      info "Updated nginx configuration: #{NGINX_CONFIG_FILE}"
    end
  end
  
  # Teases out an option which might have been under different product names
  def get_app_option(key)
    [:eventq, :orchestrate].each do |s| 
      return @site_options.options[s][key] if (@site_options.options[s] && @site_options.options[s].key?(key))    
    end
    raise "No options for #{key} found"
  end
  
  # TODO: move to SiteOptions
  def loadable?(file)
    file && !file.empty? && File.file?(file)
  end
  
  # NOTE: this has a strong relationship with the definition in Asky#truth and 
  # Asky#truthy
  def booleanish(v)
    if v.is_a? String
      return true  if %w(true yes y 1).include?(v.downcase)
      return false if %w(false no n 0).include?(v.downcase)
    end
    return v
  end

  def debug(m)
    if @script_options[:debug]
      m.split("\n").each {|l| puts "DEBUG: #{l}" }
    end
  end

  def warn(m)
    puts "WARN: #{m}"
  end

  def info(m)
    puts "INFO: #{m}"
  end
end

def main
  opts = get_options
  abort("Script must be run as root!") unless is_root?

  begin
    cf_mgr = ConfigurationManager.new(opts)
  rescue RuntimeError => e 
    puts "INFO: Unable to initialize configuration manager: #{e}"
    puts "INFO: Possibly a new installation; exiting"
    exit(0)
  end

  cf_mgr.update_site_options       if opts[:options_file]  
  # TODO: handle 'all'
  cf_mgr.update_nginx              if opts[:update_nginx]  
end

def is_root?
  Process::uid == 0 ? true : false
end

def get_options(options={})
  begin
    getopts = GetoptLong.new(
      ['-f', '--options_file',GetoptLong::REQUIRED_ARGUMENT],
      ['-d', '--debug', GetoptLong::NO_ARGUMENT],
      ['-h', '--help', GetoptLong::NO_ARGUMENT],
      ['-u', '--update', GetoptLong::REQUIRED_ARGUMENT],
    )
    getopts.each do |opt, arg|
      case opt
        when '-f'
          options[:options_file] = arg
        when '-d'
          options[:debug] = true  
        when '-h'
          help
        when '-u'
          options["update_#{arg}".to_sym] = true
      end
    end
  rescue Exception => e
    help e  
  end
  puts "DEBUG: options=#{options}" if options[:debug]
  help if options.empty?
  options
end

def help(msg=nil)
  script = File.basename($0)
  
  puts
  if msg
    puts msg
    puts 
  end 

  puts "Script to update site options, and to update supporting components with"
  puts "the current site options."
  puts 
  puts "Usage: #{script} [options....] "
  puts
  puts "Options:"
  puts "  -d, --debug                 Turn on debugging"
  puts "  -h, --help                  Display this message"
  puts "  -u, --update=<component>    Update the configuration of the component based on site options"
  puts 
  puts "Examples:"
  puts "  Update nginx settings"
  puts "  $ #{script} -u nginx"
  puts

  exit 
end

main()
