#!/bin/sh
#
# Copyright 2016 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

#
# Application installer wrapper
#
set -e

PRODUCT_NAME=EventQ
APP_NAME=eventq

RUBY_ROOT="/opt/collabnet/ruby"
APP_ROOT="/opt/collabnet/$APP_NAME"

APP_PACKAGE=CN-$APP_NAME

# Prints basic usage information
usage()
{
cat << EOF
usage: $0 [options]

This script starts the TeamForge $PRODUCT_NAME installer.

OPTIONS:
   -h,--help                        Show this message
   -f,--file <config>               Specify a site-options answer file
   -l,--local                       Assume all components should be installed on localhost
   --install-version <version>      Install a specific version or revision of the $PRODUCT_NAME RPM, e.g. '16.7' or '16.7-64'
EOF
}

# Rolls back the install, generally called after a problem occurs
rollback()
{
  # Removing libyaml actually removes everything related to the application 
  yum -y remove $APP_PACKAGE collabnet-libyaml collabnet-redis
}

# Check for the existence of an answers file, and report no file found and exit if it's provided but not there
nofile()
{
  if [ ! -z "$1" -a ! -r "$1" ]; then
    echo "The answers file you specified ($1) does not exist.  Either provide the proper path to the answers file or run the installer in interactive mode." >&2
    exit 1
  fi
}

# Parse the parameters for options
TEMP_OPTS=`getopt -o hf:l -l help,file:,install-version:,local -n "$0" -- "$@"`

# Exit if parsing failed
if [ $? != 0 ] ; then
  exit 1
fi

#Set $@ to the reordered parameters from getopt 
eval set -- "$TEMP_OPTS"

# Parse any options
while true 
  do
    case "$1" in
      -h|--help)
          usage
          exit
          ;;
      -f|--file)
          ANSWERS_FILE=$2
          shift 2
          ;;
      -l|--local)
          LOCAL_OPTION='--local'
          shift
          ;;
      --install-version)
         INSTALL_VERSION=$2
         shift 2
         ;;
      --)
          shift
          break
          ;;
      *)
          echo "Parse error! Aborting"
          exit 1
          ;;
    esac
  done

if [ $(id -u) -ne 0 ]; then
  echo "You must be root to run the $PRODUCT_NAME installer!" >&2
  exit 1
fi

# Install the Application RPM
echo "Downloading and extracting $PRODUCT_NAME ..."
if [ -n "$INSTALL_VERSION" ]; then
  APP_PACKAGE=$APP_PACKAGE-$INSTALL_VERSION
fi

# NOTE: `set -e' should stop this in its tracks if the package cannot be found
# in any installed repo; we assume that the repo file was installed as part of
# the main product installation:
yum -y install $APP_PACKAGE

# Ensure any linking issues are resolved (artf250601)
ldconfig /opt/collabnet/yaml/lib/


# Find the full path for the answers file
if [ ! -z "$ANSWERS_FILE" ]; then
  export ANSWERS_FILE
  CANONICALIZED_ANSWERS=`$RUBY_ROOT/bin/ruby -e "print File.expand_path(ENV['ANSWERS_FILE'])"`
  nofile $CANONICALIZED_ANSWERS
  ANSWERS_OPTION="-f $CANONICALIZED_ANSWERS"
fi

echo -e "\nStarting the installer...\n"

cd $APP_ROOT
env PATH=$RUBY_ROOT/bin:/sbin:$PATH bundle exec $APP_ROOT/bin/install.rb $ANSWERS_OPTION $LOCAL_OPTION
