#!/bin/bash
#
# Upgrade script for application 
#

PRODUCT_NAME=EventQ
APP_NAME=eventq

SERVICE_NAMES=($APP_NAME mongodb rabbitmq)

APP_INIT_SCRIPT=/etc/init.d/$APP_NAME
APP_PACKAGE=CN-$APP_NAME
APP_ROOT=/opt/collabnet/$APP_NAME

RABBITMQ_INIT=/etc/init.d/collabnet-rabbitmq-server
RABBITMQ_PACKAGE=collabnet-rabbitmq
ERLANG_PACKAGE=collabnet-erlang

MONGODB_INIT=/etc/init.d/collabnet-mongod
MONGODB_PACKAGE=collabnet-mongodb

DEFAULTS_YAML=${APP_ROOT}/config/defaults.yml

RUBY_ROOT=/opt/collabnet/ruby

REDIS_INIT=/etc/init.d/collabnet-redis

OLD_APP_NAME=orchestrate
OLD_PRODUCT_NAME=Orchestrate
OLD_APP_INIT_SCRIPT=/etc/init.d/$OLD_APP_NAME

# TODO: this might be under the old name; move this whole thing to the 
# configurer.
if [ ! -f $DEFAULTS_YAML ]; then
  DEFAULTS_YAML=/opt/collabnet/${OLD_APP_NAME}/config/defaults.yml
fi

# Ensure that PATH will have chkconfig, etc.
export PATH=${PATH}:/sbin:/usr/sbin:/usr/local/sbin

# Usage statement
usage() {
  cat << EOF | sed -e 's/^ //g'

  usage: $0 $APP_NAME|mongodb|rabbitmq|all
   or
  usage: $0 --help

  The script will upgrade the specified application.

  OPTIONS:
     -h,--help             Show this message
EOF
}

ensure_linking() {
  # Ensure any linking issues are resolved (artf250601)
  ldconfig /opt/collabnet/yaml/lib/
}

upgrade_application() {
  ensure_linking

  set -e
  echo -e "\nShutting down existing $PRODUCT_NAME instance..."
  # Work around a conflict betweeen old-fashined init.d scripts and
  # systemd-managed services by skipping redirect while shutting down:
  export SYSTEMCTL_SKIP_REDIRECT=1
  # Shut down, regardless of name:
  [ -f ${APP_INIT_SCRIPT} ] && ${APP_INIT_SCRIPT} stop
  [ -f ${OLD_APP_INIT_SCRIPT} ] && ${OLD_APP_INIT_SCRIPT} stop
  unset SYSTEMCTL_SKIP_REDIRECT

  echo -e "\nUpgrading  $PRODUCT_NAME ..."
  #these rpm yum commands work around an upgrade bug in libyaml and ruby installed for versions 1.0.*
  set +e
  rpm -qa | grep collabnet-libyaml-0.1.4   && rpm -e --nodeps --nopostun collabnet-libyaml-0.1.4
  rpm -qa | grep collabnet-ruby-1.9.3_p392 && rpm -e --nodeps --nopostun collabnet-ruby-1.9.3_p392
  set -e 

  yum clean all
  yum -y install "collabnet-libyaml"
  yum -y install "collabnet-ruby"
  
  ensure_linking

  yum -y update $APP_PACKAGE

  pushd ${APP_ROOT} > /dev/null

  # start redis before calling rails
  [ -f ${REDIS_INIT} ] && ${REDIS_INIT} start

  echo -e "\nCompiling $PRODUCT_NAME assets... (this may take a while)"
  su -c "source ~/.bashrc && bundle exec rake assets:clean" - $APP_NAME
  su -c "source ~/.bashrc && RAILS_RELATIVE_URL_ROOT=/orc RAILS_ENV=production bundle exec rake assets:precompile" - $APP_NAME
  echo -e "\nRunning database migrations"
  su -c "source ~/.bashrc && bundle exec rake db:migrate" - $APP_NAME
  su -c "source ~/.bashrc && bundle exec rake db:index" - $APP_NAME

  popd > /dev/null

  # Call restart (rather than start) to ensure that services have been completely
  # stopped under their current service definitions, as those definitions may have
  # changed during the course of the upgrade: 
  echo -e "\nUpgrade complete, restarting $PRODUCT_NAME ..."
  ${APP_INIT_SCRIPT} restart
  set +e

  echo -e "\nUpdating IAF configuration ..."
  su -c "source ~/.bashrc && bundle exec rake $APP_NAME:update_iaf_configuration" - $APP_NAME
  echo -e "\nIAF configuration updated ..."
  set +e
}

upgrade_rabbitmq() {
  set -e
  echo -e "\nShutting down rabbitmq service ..."
  ${RABBITMQ_INIT} stop

  echo -e "\nUpgrading rabbitmq ..."
  yum clean all
  # FIXME: due to changes in erlang's versioning numbers, it's not always clear
  # that rabbit will know to update erlang, so force it here:
  yum -y update ${ERLANG_PACKAGE}
  yum -y update ${RABBITMQ_PACKAGE}

  echo -e "\nUpgrade complete, starting rabbitmq service ..."
  ${RABBITMQ_INIT} start
  set +e
}

upgrade_mongodb() {
  set -e
  echo -e "\nShutting down mongodb service ..."
  ${MONGODB_INIT} stop

  echo -e "\nUpgrading mongodb ..."
  yum clean all
  yum -y update ${MONGODB_PACKAGE}

  echo -e "\nUpgrade complete, starting mongodb service ..."
  ${MONGODB_INIT} start
  set +e
}

check_prerequisites() {
  check_required_commands
  check_defaults_yml    
}

check_required_commands() {
  echo -e "\nTesting required commands in environment ..."

  for cmd in rpm yum chkconfig; do 
    which $cmd > /dev/null;
    if [ $? -ne 0 ]; then
      echo "FATAL: command $cmd not found in PATH.  Please install $cmd or ensure that it is in PATH."
    exit 1
    fi
  done
}

check_defaults_yml() {
  echo -e "\nChecking defaults.yml configuration ..."

  flags=''

  if [ ! -f "$DEFAULTS_YAML" ]; then
    echo "WARNING: no $DEFAULTS_YAML file found; assuming this is infrastructure node." 
    return
  fi

  # Does it look like Rabbit is advertising localhost?:
  egrep '^\s*uri:\s*("|'"'"')?amqps?:.*(127\.0\.0\.1|localhost)' $DEFAULTS_YAML >/dev/null 2>&1

  if [ $? -eq 0 ]; then
    flags="${flags} --update_rabbit_host"
  fi

  # Does it look like specific Rabbit user (permission_publisher) exist?:
  egrep '^\s*permission_publisher:' $DEFAULTS_YAML >/dev/null 2>&1

  if [ $? -ne 0 ]; then
    flags="${flags} --create_rabbit_publisher_user"
  fi

  if [ "$flags" != '' ]; then
    now=`date +%Y%m%d%H%M%S`
    cp -p $DEFAULTS_YAML ${DEFAULTS_YAML}.${now}
    $APP_ROOT/utils/defaults_surgeon.rb ${flags}
  fi
}


### Live code below:

# Complain and exit if not root
if [ $(id -u) -ne 0 ]; then
  echo "You must be root to run the $PRODUCT_NAME upgrader." >&2
  exit 1
fi

# Complain if SELINUX is enforcing
getenforce | grep -qi enforcing
if [ $? -eq 0 ]; then
  echo "SELINUX is set to enforcing; it should be permissive or disabled for upgrade." >&2
  exit 1
fi

# Show help message if -h or --help is passed
while true; do
  case "$1" in
    -h|--help|"")
        usage
        exit 1
        ;;
    *)
        SERVICE_NAME=$1
        break
        ;;
  esac
done

# Copy STDOUT and STDIN to the upgrade.log
TIMESTAMP=$(date +"%Y%m%d%H%M%S")
LOG_FILE="`dirname $0`/upgrade_${TIMESTAMP}.log"
exec > >(tee $LOG_FILE)
exec 2>&1

# make sure that the correct service name is passed as an argument
[[ $SERVICE_NAME =~ ^($APP_NAME|mongodb|rabbitmq|all)$ ]]
if [ $? -ne 0 ]; then
  usage
  exit 1
fi

check_prerequisites

# Upgrade the service
case $SERVICE_NAME in
  "all")
    upgrade_rabbitmq
    upgrade_mongodb
    upgrade_application
    ;;
  "$APP_NAME")
    upgrade_application
    ;;
  "rabbitmq")
    upgrade_rabbitmq
    ;;
  "mongodb")
    upgrade_mongodb
    ;;
esac
