Orchestrate::Application.routes.draw do


  devise_for :users, skip: [:session], controllers: { omniauth_callbacks: "users/omniauth_callbacks" }
  devise_scope :user do

    # TODO: What we would like to do here is remove the conditional, and have the auth provider
    # specify what to do with the routes.  More specifically, have the routes be guaranteed to
    # exist, with the provider determining where they should go, or what they should do.

    if Rails.configuration.teamforge[:enabled]
      active_auth_strategy = 'teamforge'
    else
      active_auth_strategy = 'localauth'
    end

    get "users/sign_in" => redirect("#{ActionController::Base.config.relative_url_root}/users/auth/#{active_auth_strategy}"), as: :new_user_session
    #get "users/sign_in", to: "devise/sessions#new", as: :new_user_session
    post "users/sign_out", to: "devise/sessions#create", as: :user_session
    get "users/sign_out", to: "devise/sessions#destroy", as: :destroy_user_session

  end

  root to: "project#index"

  namespace :commit_gate do
    match "/", action: :can_commit, as: "can_commit", via: [:get, :post]
  end

  resources :project, only: [:index]

  scope "projects/:project" do

    get '/', to: 'project#show', as: :project
    get '/trace', to: 'association_trace#index', as: :trace
    match "/comments/:id", to: "comments#create", via: :post, as: :comment

    resources :activity, only: [:index] do
      get :count
    end

    get '/sources_overview', to: 'project#sources_overview', as: :sources_overview
    scope "sources_overview" do
      [:commit_servers, :build_servers, :review_servers, :work_item_sources, :custom_activity_sources].each do |type|
        resources type do
          member do
            post :active_status
          end

          collection do
            get :poll
          end
        end
      end
    end

    [:commits, :builds, :review_requests, :work_items, :customs].each do |type|
      resources :activity, only: [:show], as: type, path: type.to_s
    end

    resources :commit_summaries, path: 'commits'
    resources :work_item_summaries, path: 'work_items'
    resources :build_summaries, path: 'builds'
    resources :review_request_summaries, path: 'review_requests'
    resources :custom_summaries, path: 'customs'
  end

  namespace :api do
    version_numbers = ('1'..'6').to_a
    version_symbols = ['stable', 'latest']

    (version_numbers + version_symbols).each do |v|
      v_module_name = version_symbols.include?(v) ? version_numbers.last : v

      api_version(:module => "V#{v_module_name}", :path => {:value => "#{v}"}) do
        resources :tags, only: :index

        scope "projects/:project" do
          resources :sources, only: [:index, :show, :new, :create, :update] do
            member do
              put :activate
              put :deactivate
              post  'trusted_sources', to: 'sources#trusted_sources'
              delete 'trusted_sources', to: 'sources#delete_trusted_sources'
              get   'trusted_list',    to: 'sources#list_trusted_sources'

            end
            collection do
              get :work_item_sources_by_tracker_ids
            end
          end

          resources :tools do
            member do
              put :activate
              put :deactivate
            end
            collection do
              get :is_name_unique
            end
            resources :sources, only: [:index]
          end


          #####################################################
          # Restful routing for socket resources:
          # resources are:
          # (i)   data packets from an external source to be
          #       persisted
          # (ii)  CRUD on patterns
          # (iii) CRUD on emulation
          # (iv)  CRUD on Packet maps
          # POST#create is the only method method on packets
          ######################################################

          namespace :webhook do
            post 'packets/:packet_type/:sak/', to: 'packets#create'

            resources :emulation, except: [:patch]

            get    'patterns/',                          to: 'patterns#index'
            get    'patterns/:pattern_id',               to: 'patterns#show'
            post   'patterns/',                          to: 'patterns#create'
            put    'patterns/:pattern_id',               to: 'patterns#update'
            delete 'patterns/:pattern_id',               to: 'patterns#destroy'

            get    'packet_maps/',                       to: 'packet_maps#index'
            get    'packet_maps/:packet_type/',          to: 'packet_maps#show'
            post   'packet_maps/',                       to: 'packet_maps#create'
            put    'packet_maps/:group_id',              to: 'packet_maps#update'
            delete 'packet_maps/:group_id',              to: 'packet_maps#destroy'
          end


          # Webhook CRUD endpoints
          resources :webhooks, only: [:index, :create, :update, :destroy]

          get "trackers", to: "trackers#index"
          get 'trackers/:tracker_id' => 'trackers#by_tracker_id', :as => :tracker_by_remote_id
          get 'activities', to: 'activities#index'
        end

        #create and view schemas
        post "schemas/", to: "schemas#create"
        get "schemas", to: "schemas#index", format: :js

        get 'schemas(/:schema_id)' => 'schemas#show', :as => :schema_id
        get 'schemas(/:schema_id)(/:schema_version)' => 'schemas#show', :as => :schema_id_by_version

        # get permission_queue_info
        get "permission_queue_info", to: "permission_queue_info#index", format: :js

        post "reporting/", to: "reports#index"

        get "docs", to: "docs#index", format: :html
        get "docs/index", to: "docs#index", format: :html
        get "docs/:page", to: "docs#show", format: :html

        # Route all preflight requests to here (CORS)
        match '*all', :to => '/api/base#cors_set_access_control_headers', :via => :options
      end
    end
  end

  namespace :services do
    namespace :team_forge, defaults: { format: :xml } do
      get "/ConfigurationParameters/:token/:tf_project_id", to: :get_configuration_parameters

      post "/ProjectConfiguration/:token", to: :create_project_configuration
      put "/ProjectConfiguration/:token", to: :update_project_configuration
      delete "/ProjectConfiguration/:token/:tf_project_id", to: :delete_project_configuration

      get "/ProjectId/:token/:object_id", to: :ctf_project_id
      get "/Title/:token/:object_id", to: :title

      get "/ScmPreCommit/:token", to: :pre_commit_message
      post "/DetailedScmPreCommit/:token", to: :detailed_precommit_message

      get "/SCMPostCommit/:token", to: :post_commit_message
      post "/DetailedScmPostCommit/:token", to: :detailed_post_commit_message
      post "/Search/:token", to: :search

    end
  end

  get "/go/:id", controller: :go, to: :redirect


end
