#
# Copyright 2013 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#
require 'migrations/exceptions'
class CreateActivitySummaries < Migration
  extend MigrationHelper

  DOCUMENT_BLOCK_SIZE = 100
  ASCENDING = 1
  DESCENDING = -1


  FIELDS_TO_COPY = {commit: ['time_created', 'settled', 'message', 'branch_name', 'path_changes', 'comments', 'ctf_item', 'created_by', 'opaque_id'],
                    build: ['time_created', 'status', 'test_results', 'duration', 'build_url', 'associated_ctf_ids', 'comments', 'opaque_id'],
                    review_request: ['time_created', 'status', 'link', 'summary', 'description', 'reviewers', 'reviewed_files', 'associated_ctf_ids' ,'created_by']}


  def self.up
    if ActivitySummary.count == 0
      [:commit, :build, :review_request].each do |activity_type|
        create_activity_summaries(activity_type)
      end
      update_reviews_and_review_summaries
      associate_summaries
      remove_phantom_commits
      clean_fields
    end
  end

  # creates an ActivitySummary for each Activity of the given type and copies the necessary fields over
  def self.create_activity_summaries(type)
    activity_type = type.to_s.camelize
    summary_type = activity_type + 'Summary'
    (0..document_count(:activities, _type: activity_type)).step(DOCUMENT_BLOCK_SIZE) do |records_to_skip|
      iterate_over(:activities, {_type: activity_type}, {timeout: false, sort: [['_id', ASCENDING]], limit: DOCUMENT_BLOCK_SIZE, skip: records_to_skip}) do |activity|
        data_to_insert = activity.slice(*FIELDS_TO_COPY[type])
        remote_id_key = (type == :commit ? 'revision_number' : 'remote_id')
        search_query_parameters = {_type: summary_type, source_id: activity['source_id'], remote_id_key => activity[remote_id_key]}
        if ['BuildSummary', 'CommitSummary'].include?(summary_type)
          search_query_parameters.merge!(time_created: activity['time_created'])
        end
        activity_summary = db[:activity_summaries].find_one(search_query_parameters) || search_query_parameters
        summary_id = db[:activity_summaries].save(activity_summary.merge(data_to_insert))
        db[:activities].save(activity.merge(activity_summary_id: summary_id))
      end
    end
  end

  # we need to ensure that each ReviewRequestSummary has a correct opaque_id (an opaque_id of the first review request
  # associated with the ActivitySummary). The reason we care about opaque_ids is that we want the links to the Orchestrate
  # activities that were created inside CTF before the migration work as before. We also need to copy comments that are stored
  # on the first review request. And we need to add :created to the attribute changes for the first review request (because of the after 1.0.2 model changes)
  def self.update_reviews_and_review_summaries
    (0..document_count(:activity_summaries, _type: 'ReviewRequestSummary')).step(DOCUMENT_BLOCK_SIZE) do |records_to_skip|
      iterate_over(:activity_summaries, {'_type' => 'ReviewRequestSummary'}, {timeout: false, sort: [['_id', ASCENDING]], limit: DOCUMENT_BLOCK_SIZE, skip: records_to_skip, fields: []}) do |activity_summary|
        first_review_request= within(:activities) { |activities| activities.find_one({activity_summary_id: activity_summary['_id']},
                                                                                     {sort: [['time_created', ASCENDING]], limit: 1, fields: ['opaque_id', 'comments']}) }
        opaque_id = first_review_request['opaque_id']
        comments = first_review_request['comments']
        db[:activity_summaries].update({'_id' => activity_summary['_id']}, {'$set' => {'opaque_id' => opaque_id, 'comments' => comments}})
        db[:activities].update({'_id' => first_review_request['_id']}, {'$set' => {'attribute_changes' => [:status]}})
      end
    end
  end


  # Associating summaries. For each activity (say x), find all its associations (say activities y,z) and then associate
  # ActivitySummary of x with ActivitySummaries of y and z.
  def self.associate_summaries
    activity_summaries = db[:activity_summaries]
    (0..document_count(:activities)).step(DOCUMENT_BLOCK_SIZE) do |records_to_skip|
      iterate_over(:activities, {}, {timeout: false, limit: DOCUMENT_BLOCK_SIZE, skip: records_to_skip, sort: [['_id', ASCENDING]]}) do |activity|
        next if activity['_type'] == 'External'
        summary_for_activity = activity_summaries.find_one(activity["activity_summary_id"],
                                                                      {fields: ['associated_activity_summary_ids']})
        existing_summary_associations = summary_for_activity['associated_activity_summary_ids'] || []
        existing_associations_for_activity = activity['associated_activity_ids'].to_a
        existing_associations_for_activity.each do |id|
          summary = db['activities'].find_one(id, {fields: ['activity_summary_id']})['activity_summary_id']
          existing_summary_associations << summary if summary
        end
        existing_summary_associations.uniq!
        activity_summaries.update({'_id' => summary_for_activity['_id']},
                                     {'$set' => {'associated_activity_summary_ids' => existing_summary_associations}})
      end
    end
  end

  #remove phantom commits (this info is already captured in the corresponding summary)
  def self.remove_phantom_commits
    db['activities'].remove('settled' => false)
  end

  # renames the fields from 'foo' to '__foo__'
  def self.clean_fields
    remove_fields(:activities, ['settled', 'opaque_id', 'associated_activity_ids', 'associated_ctf_ids'])
  end


  def self.down
    raise IrreversibleMigration.new
  end

end