#
# Copyright 2015 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#
require 'migrations/exceptions'
include RabbitServices

# This migration does the following data updates:
# 1. Converts all XDS from Orchestrate 1.4.0 to the XDS in Orchestrate 8.1
#    While doing this conversion:
#     - Background Color and Icon URI are being reset to default values for all XDS
#     - RabbitMQ username and password are generated for the source and created on RabbitMQ server
#     - XDS are migrated from Pipeline to Project (in case they are not already)
#     - Activity Source Tags for XDS are copied over from CustomSource objects (XDS Collection in Orchestrate 1.4.0)
# 2. All XDS Collections (Type: CustomSource) are deactivated and marked for deletion. Their association to the
#    project is broken by setting the project_id as nil. This is not reversible.
# 3. Orphaned sources (the ones that cannot be associated to a project) are deactivated and marked for deletion. In
#    addition to doing this, orphan XDS sources are converted to Orchestrate 8.1 format.

class Xds140ToXds81 < Migration
  extend MigrationHelper

  # Migrate
  def self.up
    self.create_individual_sources_for_existing_custom_activity_sources
    self.deactivate_xds_collections_and_mark_for_deletion
    self.deactivate_orphaned_sources_and_mark_for_deletion
  end

  # Migration methods

  # This method converts the first generation XDS sources (Created in Orchestrate 1.4.0) to current version.
  # Existing custom activity_sources are updated with the necessary fields and relationships.
  # For successfully updated sources, RabbitMQ credentials are created.
  def self.create_individual_sources_for_existing_custom_activity_sources

    # A hash of custom_source_ids (XDS collection IDs) to tags used in that collection.
    xds_collections_to_tags = {}

    # A hash of schema_store_ids to schema_store objects
    schema_store_ids_to_objects = {}

    # A hash of custom_source_ids to custom_source objects first generation (1.4.0)
    custom_source_ids_to_objects = {}

    # A hash of schema_store_ids to arrays of tags to be associated to sources that use that schema_store
    schema_store_id_to_tags = {}

    # Find all XDS collections and map their IDs to used Tags
    iterate_over(:activity_sources, {'_type' => 'CustomSource'}) do |xds_collection|
      xds_collections_to_tags[xds_collection['_id']] = xds_collection['activity_source_tag_ids']
    end

    # Filter records where custom_source_ids exist (This case only exists for XDS collections that were created in
    # Orchestrate 1.4.0) and populate the schema store map
    iterate_over(:custom_schema_stores, {custom_source_id: {'$exists' => true}}) do |schema_store|
      schema_store_ids_to_objects[schema_store['_id']] = schema_store
      schema_store_id_to_tags[schema_store['_id']] = xds_collections_to_tags[schema_store['custom_source_id']]
    end

    # Filter records and populate the custom source map. Filtration is targeted such that first generation XDS objects
    # that would not have queue_username and queue_password are only retrieved.
    iterate_over(:activity_sources, {_type: 'CustomActivitySource', queue_username: {'$exists' => false}, queue_password: {'$exists' => false}}) do |custom_src|
      custom_source_ids_to_objects[custom_src['_id']] = custom_src
    end

    # Default values to be used for assignment
    background_color_value = '#61c0e5'
    icon_uri = '/assets/icons/custom_source/XDS_Icon_01.png'

    # Variables for logging success and failure of the query for updating custom activity sources with proper fields
    db_query_successes = []
    db_query_failures = []
    # Variables for logging success and failure in creation of RabbitMQ credentials.
    queue_creds_successes = []
    queue_creds_failures = []

    iterate_over(:pipelines) do |pipeline|
      this_project = pipeline['project_id']
      steps = pipeline['pipeline_steps']

      # Operate only on Custom-Steps
      steps.each { |current_step|
        if current_step['type'] == 'custom'
          source_ids = current_step['source_ids']
          source_ids.each { |current_source_id|
            if custom_source_ids_to_objects.has_key?(current_source_id) && schema_store_ids_to_objects.has_key?(custom_source_ids_to_objects[current_source_id]['custom_schema_store_id'])
              current_source = custom_source_ids_to_objects[current_source_id]
              # Values to update in the activity source
              queue_username = current_source['source_association_key']
              queue_password = SecureRandom.hex(32).to_s
              activity_source_tag_ids = schema_store_id_to_tags[current_source['custom_schema_store_id']]

              # Update query
              query_status = db[:activity_sources].update({'_id' => current_source_id}, {'$set' => {'project_id' => this_project, 'queue_username' => queue_username, 'queue_password' => queue_password, 'background_color_value' => background_color_value, 'icon_uri' => icon_uri, 'activity_source_tag_ids' => activity_source_tag_ids}})

              # Log query success and create RabbitMQ credentials if successful.
              if query_status['updatedExisting'] && (query_status['err'] == nil)

                db_query_successes << "- source: '#{current_source['display_name']}', ID: #{current_source['_id']}, Type: #{current_source['_type']}, SAK: #{current_source['source_association_key']}"

                # Create Queue Credentials
                queue_cred_creation_successful = create_queue_account(queue_username, queue_password)

                if queue_cred_creation_successful
                  queue_creds_successes << "- source: '#{current_source['display_name']}', ID: #{current_source['_id']}, username: #{queue_username}"
                else
                  queue_creds_failures << "- source: '#{current_source['display_name']}', ID: #{current_source['_id']}, username: #{queue_username}"
                end
              else # Log query failure
                db_query_failures << "- source: '#{current_source['display_name']}', ID: #{current_source['_id']}, SAK: #{current_source['source_association_key']}"
                db_query_failures << "The DB query returned with: #{query_status}"
              end
            end
          }
        end
      }
    end

    # Logging status and result
    puts '********************************************* Results *********************************************'
    puts "The number of CustomActivitySources found -  #{custom_source_ids_to_objects.size}"

    log_title = "Successfully migrated: #{db_query_successes.size} Sources"
    self.log_records(log_title, db_query_successes)

    log_title = "Migration failed for: #{db_query_failures.size} Sources"
    self.log_records(log_title, db_query_failures)

    log_title = "Successfully created RabbitMQ credentials for: #{queue_creds_successes.size} users"
    self.log_records(log_title, queue_creds_successes)

    log_title = "Failed to create RabbitMQ Credentials for: #{queue_creds_failures.size} users"
    self.log_records(log_title, queue_creds_failures)

  end

  # The 'CustomSource' type objects (also called XDS Collections) which no longer exist in current version of
  # Orchestrate are disabled and marked for deletion. The association between the project and CustomSource objects is
  # explicitly broken and project_id for the sources is set to nil
  def self.deactivate_xds_collections_and_mark_for_deletion

    # Variables for logging query results.
    xds_deactivation_statuses = []
    xds_collections_to_projects = {}

    # Limited query for 'CustomSource' type objects
    iterate_over(:activity_sources, {_type: 'CustomSource'}) do |xds_collection|
      xds_collections_to_projects[xds_collection['_id']] = "This XDS collection was mapped to project: #{xds_collection['project_id']}"
      deactivation_query_status = db[:activity_sources].update({'_id' => xds_collection['_id']}, {'$set' => {'active' => false, 'to_delete' => true, 'project_id' => nil}})

      xds_collection_info = "Name: #{xds_collection['display_name']}, ID: #{xds_collection['_id']}, Type: #{xds_collection['_type']}, SAK: #{xds_collection['source_association_key']}"
      if deactivation_query_status['updatedExisting'] && (deactivation_query_status['err'] == nil)
        xds_deactivation_statuses << "Success: #{xds_collection_info}"
      else
        xds_deactivation_statuses << "Failure: #{xds_collection_info}"
      end
    end

    log_title = "XDS Collection to Project Mappings (now broken): Total of #{xds_collections_to_projects.size}"
    self.log_records(log_title, xds_collections_to_projects)

    log_title = "XDS Collection deactivation status: #{xds_deactivation_statuses.size} XDS Collections deactivated"
    self.log_records(log_title, xds_deactivation_statuses)

  end

  # Orphaned Sources are the activity sources that cannot be associated to a project as the connecting pipeline-step
  # has been deleted. This method marks all orphaned sources as inactive and marks them for deletion. For custom
  # activity sources, it additionally modifies their data so that they conform to the Orchestrate 8.1 XDS structure
  def self.deactivate_orphaned_sources_and_mark_for_deletion

    # A hash for logging type of source acted upon to the number of sources of that type
    type_to_count = {}

    # Variables to log results of migration queries.
    deactivated_orphan_sources_count = 0
    orphan_sources_deactivation_failure = []

    # Default values for Custom Actvity Source fields
    background_color_value = '#61c0e5'
    icon_uri = '/assets/icons/custom_source/XDS_Icon_01.png'
    activity_source_tag_ids = Set.new.to_a

    # Filter the orphan sources - these are the sources that do not belong to a project and thus do not have the
    # project_id field.
    iterate_over(:activity_sources, {'project_id' => {'$exists' => false}}) do |each_orphan_source|
      if type_to_count.has_key?(each_orphan_source['_type'])
        type_to_count[each_orphan_source['_type']] = type_to_count[each_orphan_source['_type']] + 1
      else
        type_to_count[each_orphan_source['_type']] = 1
      end

      # Update XDS objects and also deactivate them
      if each_orphan_source.is_a? CustomActivitySource
        queue_username = each_orphan_source['source_association_key'] || UUID.generate
        queue_password = SecureRandom.hex(32).to_s
        deactivation_query_status = db[:activity_sources].update({'_id' => each_orphan_source['_id']}, {'$set' => {'queue_username' => queue_username, 'queue_password' => queue_password, 'background_color_value' => background_color_value, 'icon_uri' => icon_uri, 'activity_source_tag_ids' => activity_source_tag_ids, 'active' => false, 'to_delete' => true}})
      else
      # Only deactivate other types of sources
        deactivation_query_status = db[:activity_sources].update({'_id' => each_orphan_source['_id']}, {'$set' => {'active' => false, 'to_delete' => true}})
      end

      # Log the result to the variables
      if deactivation_query_status['updatedExisting'] && (deactivation_query_status['err'] == nil)
        deactivated_orphan_sources_count += 1
      else
        orphan_sources_deactivation_failure << "Name: #{each_orphan_source['display_name']}, ID: #{each_orphan_source['_id']}, Type: #{each_orphan_source['_type']}, SAK: #{each_orphan_source['source_association_key']}"
      end
    end

    log_title = "Orphan Sources deactivated and marked for deletion: #{deactivated_orphan_sources_count + orphan_sources_deactivation_failure.size}"
    self.log_records(log_title, type_to_count)

    log_title = "Failed to deactivate and mark for deletion: #{orphan_sources_deactivation_failure.size}"
    self.log_records(log_title, orphan_sources_deactivation_failure)

  end

  # Helper methods

  # Method to print out the header information before logging the migration status data
  def self.print_log_header(title)
    puts '-'*95
    puts title
    puts '-'*95
  end

  # The primary logging method.
  # Takes a String and an Array/Hash to log the migration status data
  def self.log_records(title, migrated_records)

    unless migrated_records.empty?

      self.print_log_header(title)

      if migrated_records.is_a? Array
        migrated_records.each { |record| puts record}
      elsif migrated_records.is_a? Hash
        migrated_records.each {|key, value| puts "#{key}: #{value}" }
      end
    end
  end
end