#
# Copyright 2014 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

module ChangeHistory

  module ClassMethods

    # Finds the most recent revision of a review
    # @param [Integer] remote_id the review request's ID
    # @param [BSON::ObjectId] source_id the review board ID
    # @param [Time] created_before set the upper limit for the most recent item
    # @return [ReviewRequest] The most recent review from the review board server with the given review request ID
    def most_recent_activity(remote_id, source_id, created_before = nil)
      query = where(source_id: source_id, remote_id: remote_id)
      query = query.where(:time_created.lte => created_before) if created_before
      query.first(order: :time_created.desc)
    end

  end

  module InstanceMethods

    # Does this review request have changes from the previous request in this
    # review?
    # @return [Boolean] True if the request has attribute changes
    def has_changes?
      attribute_changes.present?
    end

    # Is this the first ReviewRequest instance in the series?
    # @return |Boolean| whether this is the first for remote_id and server_id
    def is_first?
      self.ancestor_id.nil?
    end

    # Finds the earliest request in the "review" that this request belongs to
    # (i.e., the creation of the review)
    # @return |ReviewRequest| the first request in the series
    def first_activity
      is_first? ? self : self.class.where(source_id: source_id,
                                          remote_id: remote_id).
          first(order: :time_created.asc)
    end

    # Quickly finds the preceding ReviewRequest instance, if it exists,
    # based on the ancestor_id recorded at creation
    # @return |ReviewRequest| the immediate ancestor, or nil
    def previous_instance
      (ancestor_id) ? self.class.find(ancestor_id) : nil
    end

    # builds attribute_changes array against previous Activity and stores
    # result in the :attribute_changes field.
    # @return |Array| the collected attribute changes
    def calculate_attribute_changes
      self.attribute_changes = []
      current_instance_time = time_created ? time_created : Time.now
      previous_activity = self.class == WorkItem ? WorkItemSummary.find_by_remote_id(/^#{remote_id}/i) : nil
      previous_instance = self.class.most_recent_activity(remote_id, previous_activity.try(:source_id) || source_id, current_instance_time)
      if previous_instance
        self.ancestor_id = previous_instance.id
        self.class::SIGNIFICANT_ATTRIBUTES.each do |attr|
          attribute_changes << attr unless send(attr) == previous_instance.send(attr)
        end
      else
        attribute_changes << :status
      end

      # This is because when the ticket (artifact) is moved to other tracker, the status id will be changed due to
      # different tracker will have different status ids. So, status change with this scenerio is false.
      attribute_changes.delete(:status) if attribute_changes.include?(:source_id)
      attribute_changes
    end

    # custom validator to discard insignificant delta from a previous instance
    def validate_significant_delta
      raise "Insignificant delta from preceding, discarding" unless attribute_changes.length > 0
    end

  end

  def self.included(base)
    base.extend ClassMethods
    base.send :include, InstanceMethods
  end

end
