module ApiWebCommon
  # Responds to SystemCallError (in particular when CTF can not be contacted)
  def service_unavailable(exception)
    logger.error "SystemCallError: #{exception.message}"
    render_exception(503, exception)
  end

  # This is called when an unauthorized action is attempted; we log the uid attempting the action, and return a 404,
  # under the theory that 404 exposes the least information, generally.
  # @param [Exception] The exception that triggered the unauthorized event;
  def unauthorized(exception)
    logger.warn "Unauthorized access attempt to #{request.url} by user #{current_user.try(:uid)} failed."
    render_exception(403, exception, :forbidden)
  end

  # Responds to not_found exceptions
  def not_found(exception)
    render_exception(404, exception, :not_found)
  end

  # Render the http exception page, serving html from the assets directory
  # @param [Integer] http_code - The http response code
  # @param [Exception] exception - The exception that triggered the event
  # @param [Integer or Symbol] status -Response status code, if different than the http code
  def render_exception(http_code, exception=nil, status=nil)
    status ||= http_code
    respond_to do |format|
      format.html do
        render text: Orchestrate::Application.assets.find_asset("#{http_code}.html").body,
               layout: false,
               formats: [:html],
               status: status
      end
      format.any(:js, :json) do
        render json: (exception ? { text: exception.message, status: status } : { status: status }),
               status: http_code
      end
    end
  end

  # Gets the current project if a context is found. See  extract_project_context
  # on how context is being parsed
  #
  # @return current project object
  def current_project(options={})
    unless defined?(@_current_project)
      @_current_project ||= begin
        project_url = extract_project_context
        project_conf = TeamForge::ProjectConfiguration.where(project_url: project_url).first
        project_conf.project if project_conf
      end
    end
    @_current_project
  end

  private

  # Extract the project context in the following order
  #
  # 1) querystring param project which the way to specify the context in our application for a project
  # 2) querystring param sfProj is the context sent via a project from CTF when using the Project toolbar
  # 3) session param project_url in the global menu so we need to remove it once read
  #
  # @return [String|nil] The project context
  def extract_project_context
    # TODO: better document the conditions under which we delete the project URL from the session (and why).
    if request.params[:project]
      request.params[:project]
    elsif request.params[:sfProj]
      session['ctf_project'].try(:delete, :project_url)
      request.params[:sfProj]
    else
      session['ctf_project'].try(:delete, :project_url)
    end
  end
end

module SharedTeamforgeApplication
  include CtfUserAuthorization
  # Performs a permission check, refreshing the permissions if necessary.
  def ctf_permissions_check

    if ctf_user[:super_user] # TODO is this still necessary?
      # If the user's permissions were refreshed, ever, we want to set that to never, so that if they ever lose
      # super-user access, their actual permissions will be reloaded immediately.
      if current_user.refreshed_permissions_at != Time.at(0)
        current_user.refreshed_permissions_at = Time.at(0)
      end
      current_user.super_user = true
      current_user.save
      current_ability.can :manage, :all
    else
      current_user.update_attribute :super_user, false unless !current_user.super_user? # update only if the flag is actually changing
      cache_permissions_for_user(current_user) if permissions_not_cached?(current_user)
    end
  end

  # Update the current_user's timezone from CTF get_current_user_data call
  def refresh_timezone_for_user
    tz = ctf_user[:time_zone]
    set_user_timezone(tz) if (current_user.time_zone != tz)
  end

  def set_user_timezone(time_zone)
    if (Time.find_zone(time_zone))
      current_user.update_attribute :time_zone, ctf_user[:time_zone]
    else
      logger.error(t('errors.user.time_zone_invalid', usr_id: "#{current_user.uid}", tz_passed: "#{ctf_user[:time_zone]}", tz_stored: "#{current_user.time_zone}"))
    end
  end

  # Determine if the user permissions need to be refreshed
  def refresh_permissions?
    permissions_timeout = Rails.configuration.security[:permissions_timeout].to_i
    (current_user.refreshed_permissions_at + permissions_timeout) <= Time.now # The refresh time has passed
  end

  def ctf_user
    @ctf_user
  end

end
