#
# Copyright 2013 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

require 'uri'
require 'readline'
require 'io/console'
$: << File.join(File.dirname(File.absolute_path(__FILE__)))
require 'site_options'

# A few utility functions, so that we can use #blank? easier outside of Rails.
unless String.instance_methods.include? :blank?
  class String
    def blank?; empty? end
  end

  class NilClass
    def blank?; true end
  end
end

class YAMLSafeHash < ActiveSupport::HashWithIndifferentAccess
  def deep_stringify_keys
    inject({}) { |result, (key, value)|
      value = value.deep_stringify_keys if value.is_a?(Hash)
      result[(key.to_s rescue key) || key] = value
      result
    }
  end unless Hash.method_defined?(:deep_stringify_keys)

  def to_yaml(opts = {})
    deep_stringify_keys.to_yaml(opts)
  end
end

# Asky provides a simple questioning DSL for creating scripts.  You can ask questions,
# they will be filled out into an answers hash, and you can provide a block at the
# end to do work with the resultant hash.
class Asky
  AFFIRMATIVE_RESPONSES = %w(true yes y 1)
  NEGATIVE_RESPONSES = %w(false no n 0)
  LOCALHOST = %w(127.0.0.1 localhost)

  attr_reader :answers

  # Validators
  attr_reader :truthy, :uri_parseable, :numeric, :present, :exists, :exists_if_specified, :not_localhost

  # Transformations
  attr_reader :boolean, :downcase, :number, :list, :truename

  ### PERSISTENT HOST AUTHENTICATION CODE
  class << self; attr_accessor :hostmap end
  # Default entry acts as a template for subsequent entries:
  @hostmap = {:ssh_hosts => {ssh_user: nil, key_auth: nil, ssh_key: nil, ssh_password: nil}}

  # Return the hash of ssh auth information that has been gathered for a particular host
  # @return Hash of ssh info, or nil if none defined.
  def auth_info
    hostname = @answers[:ssh_hosts]
    Asky.hostmap.key?(hostname) ? Asky.hostmap[hostname] : nil
  end

  # Update the internal ssh auth information map with data from the current answers.
  def update_auth_info
    auth_hash = {}
    hostname = @answers[:ssh_hosts]
    Asky.hostmap[:ssh_hosts].keys.each do |key|
      auth_hash[key] = @answers[key] if @answers[key]
    end
    Asky.hostmap[hostname] = auth_hash
  end
  ### END PERSISTENT HOST AUTHENTICATION CODE


  # Create and initialize the core methods of an Asky block, mainly called from Asky.script.
  # @param [Hash] args Not used yet; this would allow configuration, but isn't needed right now.
  def initialize(prefill = nil)
    @prefill = if prefill.nil?
      unless ARGV.empty?
        @prefill = SiteOptions.new(ARGV[0]).options
      end
    else
      prefill
    end || {}

    @answers = YAMLSafeHash.new #@prefill?

    # Validations
    @exists = Asky.method(:exists)
    @exists_if_specified = Asky.method(:exists_if_specified)
    @truthy = Asky.method(:truthy)
    @uri_parseable = Asky.method(:uri_parseable)
    @not_localhost = Asky.method(:not_localhost)
    @numeric = ->(response) { (response =~ /^\d+$/) ? true : $stderr.puts("'#{response}' must be a number")}
    @present = Asky.method(:present)

    # Transformations
    @number = ->(response) { response.to_i }
    @downcase = ->(response) { response.downcase }
    @truename = Asky.method(:truename)
    @boolean = Asky.method(:truth)
    @list = Asky.method(:csv_to_list_of_lists)
  end

  # Run a question/answer script either on a string, or a provided block, or both.
  def self.script(script = nil, prefill = nil, &block)
    ask = Asky.new(prefill)
    if script
      ask.instance_eval(script)
    end

    if block_given?
      ask.instance_eval(&block)
    end

    ask.answers
  end

  # Allow arbitrary methods to be treated as answer keys, if they are followed by a string for a question.
  # e.g.
  #   name "What is your name"
  # translates into
  #   ask "What is your name", :name
  # and will put the users response into answers[:name]
  def method_missing(meth, *args, &block)
    if args.first.is_a? String
      ask args.first, meth.to_sym, *args.drop(1), &block
    else
      super
    end
  end

  # Allow an answer that has no user-facing question, but still validate and
  # transform it.  These answers can only come from the site options configuration.
  # @param [Symbol] key The key to look up in the configuration.
  # @param [Array<Proc>] args A list of validators to run against the provided value.
  # @param [Proc] block An extra validator, definable in-line, to call before accepting the value.
  def allow(key, *args, &block)
    args, options = parse_options(args, block)
    answer = @prefill.delete(key)

    if answer
      # Validate, even from the configuration.  The value simply won't be set if the validation fails.
      successful = args.inject(true) {|accum, block| accum && block.call(answer)}
      if successful
        @answers[key] = options[:transform] ? options[:transform].call(answer) : answer
      end
    end
  end

  # Ask a question, with optional validators, and an options hash that can set transformations and defaults.
  # Available options for the last argument are:
  #   default [String]: A default value
  #   transform [Proc]: Code to call to convert the string response into some other object.
  #   password [boolean]: If true, hides input.
  #
  # @param [Symbol] key The key to associate the answer with.
  # @param [Array<Proc>] args A list of validators to run against the provided value. The last element can be an options hash.
  # @param [Proc] block An extra validator, definable in-line, to call before accepting the value.
  # @return The answer given, after transformation.
  def ask(question, key, *args, &block)
    successful = false
    answer = ''

    args, options = parse_options(args, block)

    default_string = " [#{options[:default]}]" if options[:default]
    until successful
      answer = @prefill.delete(key)
      ask_question = answer.nil?
      answer = prompt_user(question, default_string, options) if ask_question

      successful = args.inject(true) {|accum, block| accum && block.call(answer)}
      fail_out if !ask_question && !successful
    end

    @answers[key] = options[:transform] ? options[:transform].call(answer) : answer
  end

  def fail_out
    $stderr.puts "Configuration file did not successfully provide all the necessary inputs"
    exit 1
  end

  # Typically issued at the end of the script, it can be issued multiple times and do things at each stage with the
  # answers given so far.  Expects a block.
  def done
    yield @answers
  end

  # Create a child set of values in an inner hash the next level down.
  # If the validator fails, it will re-ask all child questions again.
  # @param [symbol] key The key that will refer to the inner hash
  # @param [Proc] validator A validator that will be called with the child answers to make sure they are consistent.
  # @return [Hash] The inner hash that was returned
  def child(key, validator=nil, &block)
    done = false
    child_answers = nil
    child_prefill = (@prefill || {})[key]
    until done do
      child_answers = Asky.script(nil, child_prefill, &block)
      done = validator.nil? || validator.call(key, child_answers)
      fail_out if child_prefill && !done
      $stderr.puts "Some data was not consistent or correct, please re-enter" unless done
    end
    answers[key] = child_answers
  end

  # Validations

  # Verify that the response contains a valid path to a file.  Prints out a warning message if the file does not exist or the path is in error.
  #
  # @param [String] response The file path and name to check for existence.
  # @return [boolean] true if the file exists, false if it does not or there is some other error.
  def self.exists(response)
    File.file?(File.expand_path(response)).tap {|valid| $stderr.puts 'You must provide a valid path to a file.' unless valid}
  rescue => e
    $stderr.puts "There are problems with the file name given: #{e.message}"
    return false
  end

  # Verify that the response contains a valid path to a file iff a response was specified.
  #
  # @param [String] response The file path and name to check for existence.
  # @return [boolean] true if the file exists or no response was given, false if it does not or there is some other error.
  def self.exists_if_specified(response)
    response.blank? ? true : exists(response)
  end

  # Verify that a response was given.  Prints out a warning message if no response was provided.
  #
  # @param [String] response The string to check for blankness.
  # @return [boolean] true if the string is non-nil and non-blank.
  def self.present(response)
    (response != nil && !response.blank?).tap {|valid| $stderr.puts 'You must provide an answer' unless valid }
  end

  # @param [String] response The string to check for a truth-like value
  # @return [boolean] Whether the string represents a truth-y value (true, yes, 1, false, no, 0)
  def self.truthy(response)
    (AFFIRMATIVE_RESPONSES + NEGATIVE_RESPONSES).include?(response.downcase).tap do |truthy|
      $stderr.puts "#{response} does not appear to be true or false" unless truthy
    end
  end

  # Is this input string parseable as a URI?
  # @param [String] response The string to check for a URI-like value
  # @return [boolean] Whether the string can be parsed out as a URI
  def self.uri_parseable(response)
    uri = URI.parse(response)
    (uri.scheme && uri.host).tap do |valid|
      $stderr.puts "#{uri} is not valid, URI requires a schema and host" unless valid
    end
  rescue URI::InvalidURIError
    $stderr.puts "Expected #{response} to look like a URI"
    false
  end

  # Require that the value not be localhost.
  # @param [String] response The string to check
  # @return [boolean] Whether or not value contained localhost.
  def self.not_localhost(response)
    if LOCALHOST.include?(response.downcase)
      $stderr.puts "Host must be the FQDN or IP"
      return false
    end
    true
  end

  # Transformations

  # Convert a (presumably validated) pathname into its canonical form.  Avoids expansion if no response provided
  # (for use in conjunction with exists_if_specified).
  #
  # @param [String] response The pathname to convert
  # @return [String] The expanded path, with relative references and '~' converted to appropriate paths.
  def self.truename(response)
    response.blank? ? response : File.expand_path(response)
  end

  # Convert a truth-like value into true/false.
  # @param [String] response The string to recognize as a truth-like value
  # @return [boolean] true if any of 'true', 'yes', '1', false if 'false', 'no', '0', and nil otherwise.
  def self.truth(response)
    if truthy(response)
      AFFIRMATIVE_RESPONSES.include? response.downcase
    end
  end

  # Converts as CSV list to a list of lists (mainly for host lists)
  # @param [String] response The string to parse into a list of lists
  # @return [List] List of lists, Ex: [[entry1], [entry2]]
  def self.csv_to_list_of_lists(response)
    response.split(',').collect {|host| [host.strip()]}
  end

  private
  def parse_options(args, block)
    args ||= []
    options = {}
    if args.last.is_a? Hash
      options = args.last
      args.pop
    end
    args += [block] if block
    return args, options
  end

  def prompt_user(question, default_string, options)
    # Questions always fall under some heading; add leading space to indicate subordinance:
    prompt = "  > #{question}#{default_string}: "
    answer = if options[:password]
               print prompt
               STDIN.noecho(&:gets).tap { STDOUT.puts }
             else
               Readline.readline(prompt, true)
             end.strip

    (answer.blank? && options[:default]) ? options[:default] : answer
  end
end
