#
# Copyright 2013 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

# Represents a node on the traceability graph
class Node

  include ApplicationHelper

  MIN_DEPTH = 1
  MAX_DEPTH = 5
  DEFAULT_DEPTH = 3
  DEPTH_OPTION = (MIN_DEPTH..MAX_DEPTH)

  FILTER_COMMIT_DEPTH = 1

  attr_reader :project, :activity, :max_depth, :filter_commits

  # Creates a new graph node (build, work_item, commit, etc.)
  # @param [Object] activity the build, work_item, commit, or review activity (type is either [ActivitySummary] or [CtfAssociation])
  # @param [Hash] activities stores each activity keyed by opaque id or item id
  # @param [Hash] filter_options - the optional filters to be applied to the search
  #
  # filter_options include:
  # :max_depth the maximum recursion depth
  # :filter_commits true if we want to prevent unrelated commits from being found (exclude commits with a relationship depth > 1)
  # :start limit activities to those created on or after this date
  # :end limit activities to those created on or before this date
  def initialize(project, activity, activities, filter_options = {})
    @project = project
    @activity = activity
    @activities = activities
    @filter_commits = filter_options[:filter_commits]

    @max_depth = filter_options[:association_depth] ? filter_options[:association_depth].to_i : DEFAULT_DEPTH
    @max_depth = [MIN_DEPTH, [MAX_DEPTH, @max_depth].min].max

    @start_date = filter_options[:start_date] if filter_options[:start_date]
    @end_date =  filter_options[:end_date] if filter_options[:end_date]

    if activity.is_a? CtfAssociation
      @activities[activity.item_id] = activity
    else
      @activities[activity.opaque_id] = activity
    end

  end

  # Generates the graph_model
  # @param [Integer] depth - the current recursion depth
  # @param [Node] parent - the parent node of this node
  # @return [Hash] the graph model data
  def generate_graph_model(depth = 0, parent = nil)
    get_node_data.tap do |model|
      if depth < max_depth
        model[:children] = find_children(depth).
            reject { |child_node| parent.present? && child_node.activity == parent.activity }.
            reject { |child_node| child_node.activity.try(:time_created) < @start_date if @start_date }.
            reject { |child_node| child_node.activity.try(:time_created) > @end_date if @end_date }.
            map { |child_node| child_node.generate_graph_model(depth + 1, self) }
      end
    end
  end

  private

  # Finds children for the given parent node
  # @return [Array<Node>] the child nodes
  def find_children(current_depth)
    associations = if filter_commits && current_depth >= FILTER_COMMIT_DEPTH # need to filter out commit summaries from the first_order_associations
                     activity.first_order_associations.reject { |x| x.is_a? CommitSummary}
                   else
                     activity.first_order_associations
                   end
    make_activity_nodes(project, associations)
  end


  # Creates [Node]s for the given activities
  # @param [project] project the project for the new node
  # @param [Array<Object>] items the activity items ([Activity] or [CtfAssociation]) to make into [Node]s
  def make_activity_nodes(project, items=[])
    items.map do |item|
      Node.new(project, item, @activities, {association_depth: max_depth, filter_commits: filter_commits})
    end
  end

  # Creates the model data for a node
  # @return [Hash] returns a hash of the model data in the node format expected by the graph
  def get_node_data
      {
          id: activity.opaque_id,
          name: activity.class.to_s + activity.display_number,
          data: {
              class: activity_css_class_name(activity),
              status: activity_status_css_class(activity.status_icon_class, :small),
              settled: activity.settled?
          },
          children: []
      }
  end

  # Finds an existing activity or creates a new ctf association.
  # @param [String] id an opaque id or a ctf item id
  def self.find_activity_by_id(id)
    ActivitySummary.first(opaque_id: id)
  end
end
