#
# Copyright 2016 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

# installer.rb
# Library for supporting the application installer
#
require 'yaml'
require 'etc'
require 'uri'
require 'net/ssh'
require 'fileutils'
require 'socket'
require_relative 'config'

# Checks to see if the user running the script is root
def is_root
  return false unless Process::uid == 0
  true
end

# Hijack Capistrano's `sudo` command to avoid sudo if running as root.
#
# FIXME: using *args may not be correct; there are two forms to this command, though the block
# form is deprecated.  See https://github.com/capistrano/capistrano/wiki/2.x-DSL-Action-Invocation-Sudo
def sudo(*args)
  if fetch(:user) != 'root'
    super
  else
    puts "WARNING: got unexpected arguments running sudo as root: #{args}" unless args.empty?
    ''
  end
end

### STATE MANAGEMENT METHODS

# Return a hash of marker paths the installer for a given Capistrano operation.
#
# @param [Array] cap_params An array of [command, arg, arg, ...] produced by the installer to be passed to Capistrano.
# @return [Hash] Hash of paths for the marker dirs and files associated with the command.
def get_marker_paths(cap_params)
  {
      marker_dir: File.join(INSTALLER_MARKER_DIR, cap_params[0]),
      answers_file: File.join(INSTALLER_MARKER_DIR, cap_params[0], 'answers'),
      state_file: File.join(INSTALLER_MARKER_DIR, cap_params[0], 'state')
  }
end

# Create a marker directory for the given Capistrano operation.  Optionally store marker state, and the answers used.
#
# @param [Array] cap_params An array of [command, arg, arg, ...] produced by the installer to be passed to Capistrano.
# @param [Hash] options Hash of current answers, state, etc.
# @return None
def set_marker(cap_params, options={answers: nil, state: nil})
  paths = get_marker_paths(cap_params)

  FileUtils.mkdir_p(INSTALLER_MARKER_DIR, mode: 0600) unless File.directory? INSTALLER_MARKER_DIR
  FileUtils.mkdir_p(paths[:marker_dir], mode: 0600)

  # TODO: if it ever becomes interesting or useful to store current state of the answers at the time a capistrano task
  # was being run, we can spit out the answers here (and strip SSH password data).  If not, refactor and remove this
  # altogether.
  #
  # if options[:answers]
  #   open(paths[:answers_file], 'w') { |file| file.write(options[:answers]) }
  #   File.chmod(0600, paths[:answers_file])
  # end

  if options[:state]
    open(paths[:state_file], 'w') { |file| file.write(options[:state].to_s) }
    File.chmod(0600, paths[:state_file])
  end
end

# Check the state marker for the given Capistrano operation.  State should be one of nil, :started, :completed
#
# @param [Array] cap_params An array of [command, arg, arg, ...] produced by the installer to be passed to Capistrano.
# @return [Symbol|Nil] Symbol of current state or nil.
def check_marker(cap_params)
  paths = get_marker_paths(cap_params)
  return nil unless File.exist? paths[:state_file]
  File.read(paths[:state_file]).strip.to_sym
end

# Clear all the markers.
def clear_markers
  FileUtils.rmtree INSTALLER_MARKER_DIR
end

### END STATE MANAGEMENT METHODS

### FIND/PROCESS/REINSERT EXISTING SETTINGS METHODS

# Search for any existing data on rabbitmq, etc. settings, either from settings in the installed application
# e.g. .rabbitmq-passwords, the installer.yml file, or in an existing defaults.yml file (in that order of preference).
#
# @param [Hash] answers The hash of answers from Asky
# @return [Hash] Hash of any existing settings found, suitable for merging into the answers hash.
def get_existing_settings(answers)
  existing_answers = {}

  # TODO: rename to 'applications'?  Refactor to static class map?
  roles = {
      'rabbitmq' => { password_file: RABBITMQ_PASSWORDS_FILE },
      'mongodb' => { password_file: MONGODB_PASSWORDS_FILE }
  }

  roles.keys.each do |role|
    unless answers[role]['existing_system']
      current_answers = get_existing_settings_from_passwords_file(role, roles[role][:password_file], answers)
      current_answers ||= get_existing_settings_from_yaml(role, INSTALLER_YAML_CONFIG_FILE)
      current_answers ||= get_existing_settings_from_yaml(role, APP_DEFAULTS_YAML_FILE)
      existing_answers[role] = current_answers if current_answers
    end
  end
  existing_answers
end

# Attempt to retrieve a hash of existing credentials, etc. from an application's password file
# (e.g., mongodb, rabbitmq, etc.)
#
# @param [String] role Role in question (e.g., 'mongodb', 'rabbitmq')
# @param [String] password_file File to check for password data
# @param [Hash] answers The hash of answers from Asky
# @return [Hash|Nil] Hash of settings found
def get_existing_settings_from_passwords_file(role, password_file, answers)
  host = answers[role]['ssh_config']['ssh_hosts']
  user = answers[role]['ssh_config']['ssh_user']
  use_keyauth = answers[role]['ssh_config']['key_auth']
  auth_args = use_keyauth ? {keys: [answers[role]['ssh_config']['ssh_key']]} : {password: answers[role]['ssh_config']['ssh_password']}

  result = nil
  result_hash = nil

  begin
    if [Socket.gethostname, 'localhost'].include?(host) and File.exist?(password_file)
      result = File.read(password_file).strip.split.map! { |r| r.chomp(':') }
    else
      Net::SSH.start(host, user, auth_args.merge!(timeout: 5)) do |ssh|
        result = ssh.exec!("cat #{password_file}").strip.split.map! { |r| r.chomp(':') }
      end
    end
    if result and result[0] != "cat" # This indicates start of 'no such file or dir' error
      result_hash = Hash[*result]
      result_hash = _answer_settings_from_passwords(role, result_hash)
    end
  rescue ArgumentError => ex
    puts "WARNING: couldn't convert password data for #{role} to hash - got: #{result}"
    puts "Exception was: #{ex.to_s}"
  rescue SocketError, Net::SSH::AuthenticationFailed => ex
    puts "WARNING: an SSH/network exception encountered attempting to obtain existing data for #{role} - got: #{result}"
    puts "Exception was: #{ex.to_s}"
  rescue => ex
    puts "WARNING: an uexpected exception occurred when attempting to obtain existing data for #{role} - got: #{result}"
    puts "Exception was: #{ex.to_s}"
  end
  result_hash
end

# Attempt to retrieve a hash of existing credentials, etc. from a yaml file (either defaults.yml or installer.yml).
#
# @param [String] role Role in question (e.g., 'mongodb', 'rabbitmq')
# @return [Hash|Nil] Hash of settings found
def get_existing_settings_from_yaml(role, filename)
  result_hash = nil
  begin
    if File.exist? filename
      existing_yaml = YAML::load_file(filename)
      result_hash = _answer_settings_from_yaml(role, existing_yaml)
    end
  rescue => ex
    puts "WARNING: unable to parse role #{role} data from existing #{filename}"
    puts "Exception was: #{ex.to_s}"
  end
  result_hash
end

# Convert a role hash (presumably from an application passwords file) into a hash suitable for merging with
# installer answers.
#
# @param [String] role Role in question (e.g., 'mongodb', 'rabbitmq')
# @return [Hash] Hash of transformed settings
def _answer_settings_from_passwords(role, result_hash)
  if role == 'rabbitmq'
    { 'admin_user'      => 'guest', # TODO: constant? RABBITMQ_ADMIN
      'admin_password'  => result_hash['guest'],
      'username'        => APP_NAME, # TODO: constant?
      'password'        => result_hash[APP_NAME] }
  elsif role == 'mongodb'
    { 'admin_password'  => result_hash['admin'],
      'backup_password' => result_hash['backup'],
      'password'        => result_hash[APP_NAME] }
  end
end

# Convert settings in a yaml document (presumably defaults.yml or installer.yml) into a hash suitable for merging with
# installer answers.
#
# @param [String] role Role in question (e.g., 'mongodb', 'rabbitmq')
# @param [Hash] existing_yaml Hash of yaml data
# @return [Hash] Hash of transformed settings
def _answer_settings_from_yaml(role, existing_yaml)
  # If 'production' found, this is defaults.yml, otherwise installer.yml:
  if role == 'rabbitmq'
    yaml_section = existing_yaml[role].key?('production') ? existing_yaml[role]['production'] : existing_yaml[role]
    if existing_yaml[role].key? 'production' # Ah, defaults.yml
      uri = URI.parse yaml_section['uri']
      { 'admin_user' => yaml_section['admin_user'] || RABBITMQ_ADMIN,
        'admin_password' => yaml_section['admin_password'],
        'username' => uri.user,
        'password' => uri.password }
    else # Aha! installer.yml
      { 'admin_user' => RABBITMQ_ADMIN,
        'admin_password' => yaml_section['guest_password'],
        'username' => yaml_section['app_user'],
        'password' => yaml_section['app_password'] }
    end
  elsif role == 'mongodb'
    alt_role_key = 'mongo' # Ug.  Mongo has a different name in defaults.yml
    is_defaults_yml = (existing_yaml.key?(alt_role_key)  && existing_yaml[alt_role_key].key?('production'))
    yaml_section = is_defaults_yml ? existing_yaml[alt_role_key]['production'] : existing_yaml[role]

    if is_defaults_yml # Ah, defaults.yml
      # TODO: no other data is stored in defaults!
      { 'password' => yaml_section['password'] }
    else # Aha! installer.yml
      { 'admin_password' => yaml_section['admin_password'],
        'backup_password' => yaml_section['backup_password'],
        'password' => yaml_section['app_password'] }
    end
  end
end

### END FIND/PROCESS/REINSERT EXISTING SETTINGS METHODS

# Creates the yaml that the backend reads to configure itself
# @param [Hash] answers The hash of answers from Asky
# @return [String] Yaml text
def create_installer_yaml(answers)
  installer_yaml = {}
  defaults_answers = answers.clone

  # password generation
  defaults_answers['mongodb']['password'] = generate_password unless defaults_answers['mongodb']['password']
  defaults_answers['mongodb']['admin_password'] = generate_password unless defaults_answers['mongodb']['admin_password']
  defaults_answers['mongodb']['backup_password'] = generate_password unless defaults_answers['mongodb']['backup_password']

  defaults_answers['rabbitmq']['password'] = generate_password unless defaults_answers['rabbitmq']['password']
  if defaults_answers['rabbitmq']['existing_system']
    defaults_answers['rabbitmq']['admin_password'] = defaults_answers['rabbitmq']['existing_system']['admin_password']
    defaults_answers['rabbitmq']['admin_user'] = defaults_answers['rabbitmq']['existing_system']['admin_user']
  else
    defaults_answers['rabbitmq']['admin_user'] = RABBITMQ_ADMIN
    defaults_answers['rabbitmq']['admin_password'] = generate_password unless defaults_answers['rabbitmq']['admin_password']
  end
  defaults_answers['rabbitmq']['permission_publisher'] = RABBITMQ_PERMISSION_PUBLISHER
  defaults_answers['rabbitmq']['permission_publisher_password'] = generate_password

  # RabbitMQ
  # app_user, app_password, guest_password
  installer_yaml['rabbitmq'] = {'app_user' => APP_SERVICES_USER,
                                'app_password' => defaults_answers['rabbitmq']['password'],
                                'guest_password' => defaults_answers['rabbitmq']['admin_password'],
                                'host' => defaults_answers['rabbitmq']['host'],
                                'permission_publisher' => RABBITMQ_PERMISSION_PUBLISHER,
                                'permission_publisher_password' => defaults_answers['rabbitmq']['permission_publisher_password']}.
                                merge(answers['rabbitmq']['ssh_config'])

  # MongoDB
  # admin_user, admin_password, app_user, app_password
  installer_yaml['mongodb'] = {'admin_user' => MONGODB_ADMIN_DB_USER,
                               'admin_password' => defaults_answers['mongodb']['admin_password'],
                               'backup_user' => MONGODB_BACKUP_DB_USER,
                               'backup_password' => defaults_answers['mongodb']['backup_password'],
                               'app_user' => APP_SERVICES_USER,
                               'app_password' => defaults_answers['mongodb']['password']}.
                              merge(answers['mongodb']['ssh_config'])

  # nginx
  # server, port, ssl_key_filename, ssl_crt_filename, ssl_ca_filename
  installer_yaml['nginx'] = {
      "app_server"         => answers[APP_NAME]['server'],
      "app_port"           => answers[APP_NAME]['port'],
      'use_ssl'            => answers[APP_NAME]['use_ssl'],
      'ssl_key_filename'   => answers[APP_NAME]['ssl_key_path'],
      'ssl_crt_filename'   => answers[APP_NAME]['ssl_cert_path'],
      'ssl_ca_filename'    => answers[APP_NAME]['ssl_ca_path'],
      'ssh_hosts'          => 'localhost',
      'ssh_user' => answers[APP_NAME]['ssh_config']['ssh_user'],
  }

  common_auth = {
      'ssh_key' => answers[APP_NAME]['ssh_config']['ssh_key'],
      'ssh_password' => answers[APP_NAME]['ssh_config']['ssh_password']
  }

  auth_type = {
      'nginx' => common_auth,
      APP_NAME => common_auth
  }

  # Merge ssh auth type answers
  installer_yaml.deep_merge!(auth_type)

  # Create the appropriate hash structure for defaults.yml
  installer_yaml.deep_merge!({APP_NAME => {'defaults_yml' => create_app_defaults(defaults_answers),
                                                'ssh_hosts' => 'localhost',
                                                'ssh_user' => answers[APP_NAME]['ssh_config']['ssh_user']}})
  installer_yaml.to_yaml
end

# Creates the structure for defaults.yml from the answers
# @param [Hash] answers The answers provided by Asky
def create_app_defaults(answers)
  #defaults = {'rabbitmq' => nil, 'mongodb' => nil, 'teamforge' => nil, 'security' => nil}
  defaults = {}

  # RabbitMQ
  # uri
  #defaults['rabbitmq']['production'] = {'uri' => create_rabbit_uri(answers['rabbitmq'])}
  defaults.merge!({'rabbitmq' => {'production' => {
      'uri' => create_rabbit_uri(answers['rabbitmq']),
      'admin_user' => answers['rabbitmq']['admin_user'],
      'admin_password' => answers['rabbitmq']['admin_password'],
      'permission_publisher' => answers['rabbitmq']['permission_publisher'],
      'permission_publisher_password' => answers['rabbitmq']['permission_publisher_password'],
  }}})

  # MongoDB
  # replset, host(s), database, username, password
  #defaults['mongodb']['production'] = create_mongo_config(answers['mongodb'])
  defaults.merge!({'mongo' => {'production' => create_mongo_config(answers['mongodb'])}})

  # Teamforge
  # server, enabled, iaf:orc_url, iaf:admin_user, iaf:admin_pass
  #default['teamforge']['production'] = create_teamforge_config(answers)
  defaults.merge!({'teamforge' => {'production' => create_teamforge_config(answers)}})

  # Misc options which are required
  # security:production
  #defaults['security']['production'] = {'permissions_timeout' => 7200}
  defaults.merge({'security' => {'production' => {'permissions_timeout' => 7200}}})
end

# Figures out if the application has been installed already
# @return [Boolean]
def existing_install
  File.exists? APP_DEFAULTS_YAML_FILE
end

# Ensure that ssh keys exist for localhost, and that they are authorized for passwordless ssh.  Allows the installer
# to run an all-local mode without pre-configured keys.
#
# @return [Nil]
def ensure_localhost_keys
  root_key_file = APP_INSTALLER_ROOT_SSH_KEY_FILE
  root_pubkey_file = APP_INSTALLER_ROOT_PUBKEY_FILE
  authorized_keys_file = APP_INSTALLER_AUTHORIZED_KEYS_FILE

  kegen_cmd = "ssh-keygen -t rsa -f #{root_key_file} -N '' -q"

  unless File.exist? root_key_file
    `#{kegen_cmd}`
  end

  root_pubkey = File.read(root_pubkey_file)

  FileUtils.touch(authorized_keys_file) unless File.exist? authorized_keys_file
  File.chmod(0600, authorized_keys_file)

  File.open(authorized_keys_file, 'r+') do |file|
    file.find { |line| line == root_pubkey } or file.write(root_pubkey)
  end
  nil
end

# Simply extracts the hostname from a URI
# @param [String] uri The URI to extract the host from
# @return [String] The host
def host_from_uri(uri)
  URI.parse(uri).host
end

# Uses the components provided by the user to construct a URI and replace the rabbit config with it
# @param [Hash] rabbit_config The rabbitmq answers from the Asky prompt system
# @return [Hash] The converted configuration
def create_rabbit_uri(rabbit_config)
  host = rabbit_config['host']
  if rabbit_config['ssh_config']['existing_system']
    system_config = rabbit_config['existing_system']
    # If user has specified / as the vhost, we need to empty it so the AMQP library parses it correctly
    vhost = system_config['vhost'] == '/' ? '' : system_config['vhost']
    credentials = [system_config['username'], system_config['password']]
    port = system_config['port'].to_i
    rabbit_uri = URI::Generic.build(scheme: 'amqp', host: host, port: port, userinfo: credentials).to_s
    vhost.present? ? URI.join(rabbit_uri, '/', CGI.escape("#{vhost}")).to_s : rabbit_uri
  else
    port = 5672
    credentials = [APP_SERVICES_USER, rabbit_config['password']]
    URI::Generic.build(scheme: 'amqp', host: host, port: port, userinfo: credentials).to_s
  end
end

# Creates the proper configuration for mongo (defaults.yml)
# @param [Hash] mongo_config The mongodb answers from the Asky prompt system
def create_mongo_config(mongo_config)
  unless mongo_config['ssh_config']['existing_system']
    {
        'host' => mongo_config['ssh_config']['ssh_hosts'],
        'port' => 27017,
        'database' => MONGODB_APP_DB_NAME,
        'username' => APP_SERVICES_USER,
        'password' => mongo_config['password']
    }
  else
    mongo_config['existing_system']
  end
end

# Creates the proper configuration for CTF (defaults.yml)
# @param [Hash] answers The full answer hash from the Asky prompt system
def create_teamforge_config(answers)
  teamforge_defaults = {}

  # These get rid of the 'YAMLSafeHash' object types (or they show up in the YAML)
  teamforge_defaults.merge!(answers['teamforge'])
  teamforge_defaults['iaf'] = {}.merge(answers['teamforge']['iaf'])

  # This is the URL by which the application will be accessed.  Note that the answer to teamforge
  # contains protocol, but to the application does not.  The URL could have been derived in 2 ways:
  #
  # 1. URI.join("https://#{answers[APP_NAME]['server']}", #{APP_RELATIVE_URL_ROOT}) - this is the old default,
  # which assumes that the application is accessed as a free-standing box
  #
  # 2. URI.join(answers['teamforge']['server'], #{APP_RELATIVE_URL_ROOT}).to_s - this derived from the answer to the
  # teamforge question, which included protocol, and presumes that CTF is set up to proxy traffic
  # to the application.
  teamforge_defaults['iaf']['orc_url'] = URI.join(answers['teamforge']['server'], APP_RELATIVE_URL_ROOT).to_s
  teamforge_defaults['enabled'] = true

  teamforge_defaults
end

# Kills the installer nicely
def diaf(exit_code)
  system 'stty echo'  # Just in case they cancel during a password request
  puts "\nExiting the #{FULL_PRODUCT_NAME} installer at user's request\n\n"
  exit(exit_code)
end

# Generates a random password for use in an auto-installed service
# @param [Integer] num_bytes Number of bytes of entropy for the password
# @return [String]
def generate_password(num_bytes=GENERATED_PASSWORD_LENGTH)
  SecureRandom.hex(num_bytes).to_s
end

# Uses the components provided by the user to construct a URI and replace the rabbit config with it
# @param [Hash] rabbit_config The rabbitmq answers
# @return [Hash] The converted configuration
def update_rabbit_uri(rabbit_config, new_host)
  exiting_str_uri = rabbit_config['uri']

  port = 5672
  uri = URI.parse(exiting_str_uri)
  credentials = [uri.user, uri.password]
  URI::Generic.build(scheme: 'amqp', host: new_host, port: port, userinfo: credentials).to_s

end
