#
# Copyright 2016 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

require 'socket'

module Installer::Plugins::Base

  LOCALHOST = %w(127.0.0.1 localhost)

  # Check if a remote file exists on a server
  # 
  # @param [String] path the full path of the file.
  # @return [Boolean] true if the file exists
  def remote_file_exists?(path)
    results = @config.capture("if [ -e '#{path}' ]; then echo -n 'true'; fi")
    results == 'true'
  end

  # load a yaml file and set 'installer_config' capistrano key 
  # to the loaded yaml content
  #
  # @param [String] filename  the file to load
  def load_installer_yaml_config(filename)
    if File.exists?(filename)
      set(:installer_config, YAML.load_file(filename).with_indifferent_access)
    end
  rescue Psych::SyntaxError
    logger.error "Internal Error during installation. Bad yaml config file #{filename}"
  end

  # Fetch a config param for a specific role from the 'installer_config' capistrano key 
  #
  # @param [Symbol] role  the role
  # @param [Symbol, String] key the key in that role
  # @return the value of the key or nil if not found (role or key)
  def fetch_installer_config(role, key)
    config = fetch(:installer_config, nil)
    config[role][key] if config && config[role]
  end

  # validate that a role was set and valid
  def validate_role
    raise "Unknown role #{current_role}. Must be one of #{VALID_ROLES.join(', ')}" unless VALID_ROLES.include?(current_role)
  rescue IndexError
    raise "Missing required role"
  end

  # validate that a host was specified
  def validate_ssh_hosts
    ssh_hosts = fetch_installer_config(current_role, :ssh_hosts) || fetch(:ssh_hosts)
    set(:ssh_hosts, ssh_hosts)
  rescue IndexError
    raise "Missing required ssh_hosts for role #{current_role}"
  end

  # validate that a ssh user was specified. Set the ssh_user to capistrano user field
  def validate_ssh_user
    ssh_user = fetch_installer_config(current_role, :ssh_user) || fetch(:ssh_user)
    set(:user, ssh_user)
  rescue IndexError  
    raise "Missing required ssh_user for role #{current_role}"
  end

  # Validate that credentials (password and/or key) was specified. Set the capistrano 'password' and 'ssh_options'
  # fields depending on the credentials specified.  Require at least one credential.
  #
  # TODO: Validators should not set things... O_O Fix this
  def validate_credentials
    ssh_password = fetch_installer_config(current_role, :ssh_password) || fetch(:ssh_password, nil)
    ssh_key = fetch_installer_config(current_role, :ssh_key) || fetch(:ssh_key, nil)

    # Set key and/or password values:
    set(:password, ssh_password) unless ssh_password.blank?
    set(:ssh_options, {keys: ssh_key}) unless ssh_key.blank?

    # Vars for helpful errors and warnings:
    ssh_hosts = fetch_installer_config(current_role, :ssh_hosts) || fetch(:ssh_hosts)
    ssh_user = fetch_installer_config(current_role, :ssh_user) || fetch(:ssh_user)

    # Complain loudly:
    if (ssh_password.blank? && ssh_key.blank? && !(ssh_user == 'root' && LOCALHOST.include?(ssh_hosts.downcase)))
      raise "Neither SSH password nor SSH key credentials were specified for #{ssh_user} on #{ssh_hosts}"
    end

    # Complain quietly:
    if (ssh_password.blank? && ssh_user != 'root')
      puts "WARNING: role #{current_role} does not have a password for #{ssh_user} on #{ssh_hosts}"
      puts "WARNING: this assumes that passwordless sudo is enabled for #{ssh_user} on #{ssh_hosts}"
    end
  end

  # validate that the host specified is not a value for localhost.
  def validate_host_not_localhost
    host = fetch_installer_config(current_role, :host) || fetch(:host)
    set(:host, host)
    raise "Value of host for role #{current_role} cannot be localhost" if LOCALHOST.include?(host.downcase)
  rescue IndexError
    raise "Missing required host for role #{current_role}"
  end

  # Upload and copy a initd script template to /etc/init.d
  #
  # @param [String] local_file the path to the local_file to upload
  # @param [String] remote_filename the name of the final file for the init.d
  def install_initd(local_file, remote_filename)
    upload_and_move(local_file, remote_filename, mode: '755', move_to_dir: SYSTEM_INITD_DIR)
    run "#{sudo} chkconfig --add #{remote_filename}"
  end

  # Upload and copy a logrotate script template to /etc/logrotate.d
  #
  # @param [String] local_file the path to the local_file to upload
  # @param [String] remote_filename the name of the final file for the logrotate
  def install_logrotate(local_file, remote_filename)
    upload_and_move(local_file, remote_filename, move_to_dir: SYSTEM_LOGROTATE_DIR)
  end

  # Validate the presence of field for a role in the yaml config file or fallback to the 
  # capistrano set field in the command line (using -s field=value options)
  #
  # @param [Symbol] role the current role
  # @param [field] field the name of the field in that role
  # @return [Any] value for that field. Raise a RuntimeError if field is not present
  def validate_field_presence(role, field)
    field_value = fetch_installer_config(role, field)
    # Only check nil, since false is a legal value:
    field_value = fetch(field) if field_value.nil?
    set(field, field_value)
  rescue IndexError
    raise "Missing #{role} #{field}"
  end

  # Validate that the selected port is available to run a server on.
  # NOTE: only works for localhost or an alias of same.
  #
  # @param [String|Integer] port Port number to check
  # @param [String] host Host name to bind to (or localhost)
  def validate_port_availability(port, host='localhost')
    TCPServer.new(host, port.to_i).close
  rescue
    raise "Port #{port} on #{host} not available"
  end

  # get the current_role 
  #
  # @return [Symbol] the current role 
  def current_role
    fetch(:role).to_sym
  end

  # backup a folder on the remote server by moving it with a timestamp
  # and recreate the original folder with appropriate user owner
  #
  # @param [String] folder the folder to backup
  # @param [String] user the owner of the folder
  def backup_folder_data(folder, user)
    if remote_file_exists?(folder)
      timestamp = Time.now.strftime BACKUP_TIMESTAMP_FORMAT
      run "#{sudo} mv #{folder} #{folder}_#{timestamp}"
      run "#{sudo} mkdir -p #{folder}"
      run "#{sudo} chown -R #{user}:#{user} #{folder}"
    end
  end

  # Perform an upload of a localfile to /tmp as remote_filename. Then move it to $HOME
  # or the move_to_dir options if specified. Also set the mode of the file [default 644].
  # Finally specify the role to use or use the current role
  #
  # @param [String] local_file path to the local file
  # @param [String] remote_filename name of the file remotely to use
  # @param [Hash] options  Extra options
  # @option mode: set the mode fo the file (chmod) [Default: 644]
  # @option role: the role to use to run the capistrano command
  # @options move_to_dir: Where to move the file once uploaded. [Default: $HOME] 
  def upload_and_move(local_file, remote_filename, options={})
    mode = options[:mode] || '644'
    roles = options[:roles] || current_role
    move_to_dir= options[:move_to_dir] || '$HOME'

    upload local_file, "/tmp/#{remote_filename}", mode: mode, roles: roles 
    run "#{sudo} mv /tmp/#{remote_filename} #{move_to_dir}", roles: roles
    run "#{sudo} chown root:root #{move_to_dir}/#{remote_filename}", roles: roles
  end

  # Generate based on a template and save it to a file. 
  # 
  # @param [String] template_filename path to the template file
  # @param [String] out_filename path to the generated file
  # @param [Hash] vars  variable that will be visible inside the template. Key become the variable to use
  def generate_from_template(template_filename, out_filename, vars={})
    raise "Could not find the template file #{template_filename}" unless File.exists?(template_filename)
    template = File.read(template_filename)

    begin
      current = binding
      vars.each_pair do |k,v| 
        if v.is_a?(String)
          current.eval("#{k.to_s} = '#{v}';")
        else
          current.eval("#{k.to_s} = #{v};")
        end
      end

      generated_content = ERB.new(template).result(current)
    rescue => e
      raise "Error while processing template: #{e.message}, #{e.backtrace[0]}"
    end

    File.open(out_filename, "w"){|f| f << generated_content}
  end
end

Capistrano.plugin :installer, Installer::Plugins::Base
