#
# Copyright 2016 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

module Installer::Plugins::Mongodb
  
  MONGODB_INIT_SCRIPT_NAME = 'collabnet-mongod'
  MONGODB_INIT_SCRIPT = File.join('/etc', 'init.d', MONGODB_INIT_SCRIPT_NAME)
  MONGODB_INIT_SCRIPT_TEMPLATE = File.join('config', 'templates', 'init.d', 'mongo', MONGODB_INIT_SCRIPT_NAME)
  
  MONGODB_LOGROTATE_NAME = 'collabnet-mongodb'
  MONGODB_LOGROTATE = File.join('config', 'templates', 'logrotate', 'mongodb')
  
  MONGODB_CONF = File.join('config', 'templates', 'mongodb', 'mongodb.conf')

  MONGODB_ACCESS_MAP = {
      :admin_user => {
          db: MONGODB_ADMIN_DB_NAME,
          roles: ['userAdminAnyDatabase', 'readWriteAnyDatabase', 'dbAdminAnyDatabase']
      },
      :backup_user => {
          db: MONGODB_APP_DB_NAME,
          roles: ['userAdmin', 'readWrite', 'dbAdmin']
      },
      :app_user => {
          db: MONGODB_APP_DB_NAME,
          roles: ['readWrite', 'dbAdmin']
      }
  }

  # Sets up mongodb.  Installs configuration files, creates databases, adds users, etc.
  def setup
    validate
    configure_initd
    configure_conf
    configure_logrotate

    start
    sleep 10

    add_user(:admin_user, fetch(:admin_password))
    add_user(:backup_user, fetch(:backup_password))
    add_user(:app_user, fetch(:app_password))

    stop
    enable_authentication

    start
    sleep 10
    check_status(fetch(:admin_user), fetch(:admin_password), fetch(:app_user), fetch(:app_password))
    verify_connection(fetch(:ssh_hosts),
                      fetch(:admin_user), fetch(:admin_password),
                      fetch(:app_user), fetch(:app_password))
  end

  # Gets the current role
  def current_role
    installer.current_role
  end

  # Validates the minimum needed values for this script to run
  def validate
    [:admin_user, :admin_password, :app_user, :app_password, :backup_user, :backup_password]
        .each { |field| installer.validate_field_presence(current_role, field) }
  end

  # Installs the init.d script file.
  def configure_initd
    installer.install_initd(MONGODB_INIT_SCRIPT_TEMPLATE, MONGODB_INIT_SCRIPT_NAME)
  end

  # Installs the logrotate configuration file.
  def configure_logrotate
    installer.install_logrotate(MONGODB_LOGROTATE, MONGODB_LOGROTATE_NAME)
  end

  # Installs the conf file for mongodb
  def configure_conf
    installer.upload_and_move(MONGODB_CONF, 'mongodb.conf', {move_to_dir: "#{MONGODB_HOME_DIR}/conf" })
  end

  # Starts mongodb process.
  def start
    run "#{sudo} #{MONGODB_INIT_SCRIPT} start"
  end

  # Stops mongodb process.
  def stop
    run "#{sudo} #{MONGODB_INIT_SCRIPT} stop"
  end

  # Adds a user, saves the password to a file.  Consults MONGODB_ACCESS_MAP for data.
  #
  # @param [Symbol] username_key Hash key for user to add
  # @param [String] pass Password for user
  def add_user(username_key, pass)
    user = fetch(username_key)
    db_name = MONGODB_ACCESS_MAP[username_key][:db]
    roles = MONGODB_ACCESS_MAP[username_key][:roles].to_s.tr('"', "'") # or "['#{x.join("','")}']"
    begin # Test for existing user
      login_and_get_db_name(user, pass, db_name)
    rescue
      run "#{MONGODB_HOME_DIR}/bin/mongo --eval \"db.getSiblingDB('#{db_name}').addUser({user: '#{user}', pwd: '#{pass}', roles: #{roles}})\""
      run "#{sudo} sh -c 'echo -e #{user}: #{pass} >> #{MONGODB_PASSWORDS_FILE}'"
      run "#{sudo} chmod 0600 #{MONGODB_PASSWORDS_FILE}"
    end
  end

  # Enables authentication
  def enable_authentication
    run "#{sudo} sed -i \"s/#auth = true/auth = true/g\" #{MONGODB_HOME_DIR}/conf/mongodb.conf"
  end

  # Checks the final database status by logging into each database using username and password.
  # @param [String] admin_user - the admin user to test
  # @param [String] admin_pass - the admin users' password
  # @param [String] app_user - the application user to test
  # @param [String] app_pass - the application users' password
  def check_status(admin_user, admin_pass, app_user, app_pass)
    login_and_get_db_name(admin_user, admin_pass, MONGODB_ADMIN_DB_NAME)
    login_and_get_db_name(app_user, app_pass, MONGODB_APP_DB_NAME)
  end

  # Logs into the given database using the given username and password.
  # Checks the output for the name of the connected database.
  # @param [String] user - the user to connect with
  # @param [String] password - the users' password
  # @param [String] db_name - the database name
  def login_and_get_db_name(user, pass, db_name)
    output = @config.capture "#{MONGODB_HOME_DIR}/bin/mongo #{db_name} -u '#{user}' -p '#{pass}' --eval \"db.getName()\""
    raise "Error checking admin account, #{output}" unless (output =~ /^(?=.*?(connecting to: #{db_name}))(?=.*?(#{db_name})).*$/)
  end

  # Backs up the mongodb data directory.
  def backup_data
    installer.backup_folder_data("#{MONGODB_HOME_DIR}/data", MONGODB_USER)
  end

  # Check the connection for MongoDB from localhost/application (where this 
  # script is running) to the remote server where Mongo was installed.
  # Will test each remote host for both application and 'admin' user account access
  # Will raise an error if authentication fails or if tcp connection is wrong/not available
  # FIXME: How exactly will it raise an exception?
  #
  # @param [String] host - the host to test for connection
  # @param [String] admin_user - the admin username of mongo
  # @param [String] admin_password - the admins password
  # @param [String] app_user - the application user of mongo
  # @param [String] app_password - the application users password
  def verify_connection(host, admin_user, admin_password, app_user, app_password)
    mongo_available?(host: host, port: '27017', database: MONGODB_ADMIN_DB_NAME, username: admin_user, password: admin_password)
    mongo_available?(host: host, port: '27017', database: MONGODB_APP_DB_NAME, username: app_user, password: app_password)
  end

  # Tests to see if Mongo is available by making a connection and authenticating.
  # @param [Hash] config - Configuration hash as used in producing the
  #               application's YAML file.
  # @return [Boolean] True if the connection worked, false otherwise
  def mongo_available?(config)
    Mongo::Connection.new(config[:host], config[:port]).db(config[:database]).authenticate(config[:username], config[:password])
    true

  rescue Exception => error
    puts 'Error encountered connecting to Mongo:'
    puts error.message
    return false
  end

end

Capistrano.plugin :mongo, Installer::Plugins::Mongodb
