#
# Copyright 2016 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

module Installer::Plugins::Nginx

  # entry point to setting up in correct order nginx
  def setup
    validate
    setup_logrotate
    setup_config_files
    if fetch(:use_ssl)
      copy_ssl_files
    end
  end

  # get the current cole
  def current_role
    installer.current_role
  end

  # global validation
  def validate
    validate_nginx_app_server
    validate_nginx_app_port
    validate_nginx_use_ssl

    if fetch(:use_ssl)
      validate_nginx_ssl_key_filename
      validate_nginx_ssl_crt_filename
      validate_nginx_ssl_ca_filename
    end
  end

  # validate the nginx app_server
  def validate_nginx_app_server
    installer.validate_field_presence(current_role, :app_server)
  end

  def validate_nginx_app_port
    port = installer.validate_field_presence(current_role, :app_port)
    installer.validate_port_availability(port)
  end

  # validate whether or not to use SSL
  def validate_nginx_use_ssl
    installer.validate_field_presence(current_role, :use_ssl)
  end

  # validate the nginx ssl_key_filename 
  def validate_nginx_ssl_key_filename
    installer.validate_field_presence(current_role, :ssl_key_filename)
  end

  # validate the nginx ssl_crt_filename 
  def validate_nginx_ssl_crt_filename
    installer.validate_field_presence(current_role, :ssl_crt_filename)
  end

  # validate the nginx ssl_ca_filename
  def validate_nginx_ssl_ca_filename
    # FIXME: not required?
    installer.validate_field_presence(current_role, :ssl_ca_filename)
  end

  # name of the initd script on the server
  def initd_remote_name
    'collabnet-nginx'
  end

  # install the logrotate config file to /etc/logrotate.d/rabbitmq-server
  def setup_logrotate
    installer.install_logrotate(logrotate_template_file, initd_remote_name)
  end

  # template to use for the logrotate file
  def logrotate_template_file
    File.join('config', 'templates', 'logrotate', 'nginx')
  end

  def setup_config_files
    template = File.join('config', 'templates', 'nginx', 'nginx.conf.erb')
    remote_filename = "collabnet-nginx.conf"
    template_vars = {app_server_name: fetch(:app_server),
                     app_server_port: fetch(:app_port),
                     use_ssl: fetch(:use_ssl)}

    local_file =  '/tmp/localproxy-nginx.conf'
    installer.generate_from_template(template, local_file,
                                      template_vars.merge(app_server: true))
    upload_config_file(local_file, remote_filename, current_role)
  end

  # copy the ssl files into the /opt/collabnet/nginx/conf/ssl
  # with correct name and rights
  def copy_ssl_files
    server_key = fetch(:ssl_key_filename)
    server_crt = fetch(:ssl_crt_filename)
    ca_crt = fetch(:ssl_ca_filename)

    run "#{sudo} cp #{server_key} #{NGINX_HOME_DIR}/conf/ssl/server.key"
    run "#{sudo} cp #{server_crt} #{NGINX_HOME_DIR}/conf/ssl/server.crt"
    run "#{sudo} sh -c 'cat #{ca_crt} >> #{NGINX_HOME_DIR}/conf/ssl/server.crt'" unless ca_crt.blank?
    run "#{sudo} chmod 600 #{NGINX_HOME_DIR}/conf/ssl/server.key"
    run "#{sudo} chmod 600 #{NGINX_HOME_DIR}/conf/ssl/server.crt"
  end

  # upload a local file to a remote server via a specific role. Wrapper around base
  #
  # @param [String] local_file the location of the local file to upload
  # @param [String] remote_filename the location of the file on the remote server
  # @param [Symbol] role the role to use to perform the operation (so list of server behind the scene)
  def upload_config_file(local_file, remote_filename, role)
    installer.upload_and_move(local_file, remote_filename, {roles: role, move_to_dir: "#{NGINX_HOME_DIR}/conf" })
  end

  # stop nginx
  def stop
    run "#{sudo} #{SYSTEM_INITD_DIR}/#{initd_remote_name} stop"
  end

  # start nginx
  def start
    run "#{sudo} #{SYSTEM_INITD_DIR}/#{initd_remote_name} start"
  end

  # restart nginx
  def restart
    stop
    start
  end

end

Capistrano.plugin :nginx_plugin, Installer::Plugins::Nginx
