#
# Copyright 2016 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

module Installer::Plugins::Yum

  # Assert that a repository is available for the packages provided, or that
  # those packages are already installed.  Raise an exception if neither is
  # the case.
  def assert_repo_available(packages)
    last_package = nil
    packages.each do |package|
      last_package = package
      begin
        # Is there something available for install or update?
        run "#{sudo} yum -d 0 list available #{package}"
      rescue
        # Maybe it's already installed?
        run "#{sudo} yum -d 0 list installed #{package}"
      end
    end
  rescue
    raise "No available yum repositories provide the package '#{last_package}'"
  end

  # Get the use home path
  #
  # @return [String] the home path for the current running user
  def user_home_path
    "/home/#{user}"
  end

  # Get the yum configuration directory path
  #
  # @return [String] defaault path to the yum repository config directory
  def default_yum_dir_config_path
    '/etc/yum.repos.d'
  end

  # Return a friendly Unix filename. 
  # 1) It trims extra white space, 
  # 2) remove non alpha numeric beside _ and -
  # 3) remove repeated \t,\n and white space between words
  # 4) replace \t,\n and white space to _
  #
  # @param [String] filename  a filename that might not have clean naming
  # @return [String] a clean filename for a Posix system
  def friendly_filename(filename)
    filename.strip
            .gsub(/[^\w\s_-]+/, '')
            .gsub(/(^|\b\s)\s+($|\s?\b)/, '\\1\\2')
            .gsub(/\s+/, '_')
  end
  
  # Update packages specified or all installed packages if none specified.
  #
  # @param [Array] packages  List of packages to update. Default: all
  def update(packages = [])
    packages = [ packages ] unless packages.is_a?(Array)
    run "#{sudo} yum -y update #{packages.join(" ")}"
  end
    
  # Remove packages specified 
  #
  # @param [Array] packages  List of packages to remove. 
  def remove(packages)
    packages = [ packages ] unless packages.is_a?(Array)
    run "#{sudo} yum -y remove #{packages.join(" ")}"  
  end
  
  # Install via yum.
  # 
  # @param [Array] packages Packages to install, either String (for single package) or Array
  # @param [Boolean] update_existing If package exists, where to yum update it, defaults to true  
  def install(packages, update_existing = true)    
    packages = [ packages ] unless packages.is_a?(Array) 
    
    assert_repo_available(packages)
    
    if update_existing
      installed_packages = packages_installed(packages)
      run "#{sudo} yum -y update #{installed_packages.join(" ")}" unless installed_packages.blank?
    end
    
    run "#{sudo} yum -y install #{packages.join(" ")}" unless packages.blank?
  end

  # Return the list of packages installed for the list specified
  def packages_installed(packages)
    installed_packages = []
      
    data = @config.capture "#{sudo} yum -d 0 list installed #{packages.join(" ")}"
    lines = data.split("\n")[1..-1]
    if lines.blank?
      logger.error data
    else
      installed_packages += lines.collect { |line| line.split(".").first }
    end
    packages & installed_packages 
  rescue
    []
  end
  
  # Clean yum
  def clean
    run "#{sudo} yum -y clean all"
  end
end

Capistrano.plugin :yum, Installer::Plugins::Yum
