#
# Copyright 2016 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

# questions.rb
# Defines the over-arching methods which can be called to ask the questions needed for the backend to configure things.

require File.expand_path('../../installer/config.rb', __FILE__)
require File.expand_path('../../asky.rb', __FILE__)
require 'etc'

module InstallerQuestions

  # This is a generic question set which can be applied to any role to acquire ssh information
  def ssh_select(type)
    child :ssh_config do
      if type.downcase == APP_NAME.downcase 
        # For consistency with other components; main application uses the variable with the singular, not the plural.
        @answers[:ssh_hosts] = @answers[:ssh_host] = 'localhost'
      else
        existing_system "Do you have an existing #{type} system you would like to use?", present, truthy, transform: boolean, default: 'No'
      end

      unless @answers[:existing_system]
        puts "\n  To set up the #{type} component for you, we will need to know some information about the target system:"
        unless @answers[:ssh_hosts]
          ssh_hosts "What is the SSH hostname of your #{type} host", present
        end

        auth = auth_info

        if auth
          @answers.merge! auth
        else
          ssh_user "What is the SSH username for your #{type} host (NOTE: This account should be root or have sudo access)", present, default: Etc.getlogin
          key_auth 'Does this account need a SSH key for authentication', present, truthy, transform: boolean, default: 'No'

          if @answers[:key_auth]
            ssh_key 'What is the full path to the SSH key you would like to use', present, exists, transform: truename, default: '~/.ssh/id_rsa'
          end
          # NOTE: root doesn't need a password for localhost, but *might* need a password for the remote host (unless
          # key auth is being used).
          unless (@answers[:ssh_user] == 'root' && @answers[:ssh_hosts] == 'localhost')
            ssh_password "What is the password for the `#{@answers[:ssh_user]}` user on #{@answers[:ssh_hosts]} (empty if passwordless sudo is enabled)", password: true
          else
            # Well, you must be root on localhost, but what the heck.  If you supplied it, we'll let it through.
            allow :ssh_password
          end

          if (@answers[:ssh_user] != 'root' && @answers[:ssh_password].blank?)
            puts "WARNING: No password was provided for `#{@answers[:ssh_user]}@#{@answers[:ssh_hosts]}`; expecting passwordless sudo enabled."
            # NOTE: still, we better have a key; this condition should be caught in the subsequent validations.
          end

          update_auth_info
        end
      end
    end
  end

  # Application questions
  def eventq 
    puts "\n#{APP_PRODUCT_NAME} Configuration:"

    child APP_NAME.to_sym do
      ssh_select(APP_PRODUCT_NAME)

      server "What virtual hostname should we use for the #{APP_PRODUCT_NAME} web service", present, default: `hostname`.strip
      port "What server port should #{APP_PRODUCT_NAME} listen on", present, numeric, default: APP_SERVICE_PORT

      use_ssl "Should #{APP_PRODUCT_NAME} use SSL?", present, truthy, transform: boolean, default: 'Yes'
      if @answers[:use_ssl]
        ssl_key_path "Where is the SSL key you want to use for #{APP_PRODUCT_NAME} located", present, exists, transform: truename
        ssl_cert_path "Where is the SSL certificate you want to use for #{APP_PRODUCT_NAME} located", present, exists, transform: truename
        ssl_ca_path "Where is the CA certificate chain you want to use for #{APP_PRODUCT_NAME} located (optional)", exists_if_specified, transform: truename
      end

      allow :operation
    end
  end

  # RabbitMQ questions
  def rabbitmq
    puts "\nRabbitMQ Configuration:"
    child :rabbitmq do
      ssh_answers = ssh_select('RabbitMQ')

      host 'What is FQDN or routable IP of your RabbitMQ host', present, not_localhost

      if ssh_answers[:existing_system]
        child :existing_system do
          port 'What port is RabbitMQ running on', present, numeric, default: '5672'
          vhost 'What is your RabbitMQ vhost', present, default: '/'
          username "Provide a RabbitMQ username #{APP_PRODUCT_NAME} should use to communicate with adapters", present, default: 'guest'
          password 'What is the password for this user', present, password: true, default: 'guest'
          admin_user "Provide a RabbitMQ username that has permission to create new RabbitMQ users with write access to #{APP_PRODUCT_NAME} queues", present
          admin_password 'What is the password for this user', present, password: true
        end
      end
    end
  end

  # MongoDB questions
  def mongodb
    puts "\nMongoDB Configuration:"
    child :mongodb do
      ssh_answers = ssh_select('MongoDB')

      if ssh_answers[:existing_system]
        child :existing_system do
          replset 'Is your Mongo clustered (replset)', truthy, transform: boolean, default: 'No'
          if @answers[:replset]
            hosts 'What are the hostnames and ports for your mongo cluster nodes (comma-separated) Ex: mongo1.example.com:27017, mongo2.example.com:27017', present
          else
            host 'What is the hostname for your mongo server', present
            port 'What port is MongoDB running on', present, numeric, default: '27017'
          end
          database 'What is the name of the database you wish to use', present, default: MONGODB_APP_DB_NAME
          username 'What username should we use to connect to this database', present, default: MONGODB_APP_DB_NAME
          password 'What is the password for this user', present, password: true
        end
      end
    end
  end

  # Teamforge questions
  def teamforge
    puts "\nCollabNet TeamForge Configuration:"

    child :teamforge do
      server 'What is the URL for your TeamForge server', uri_parseable, present
      allow :linked_app_id

      child :iaf do
        admin_user "What administrative user would you like to use to join #{APP_PRODUCT_NAME} to TeamForge", present
        admin_pass "What is the password for the TeamForge user `#{@answers[:admin_user]}`", present, password: true
        if DEBUG
          require 'securerandom'
          prefix "(DEBUG) What prefix should CTF use", present, default: "EVQ#{SecureRandom.hex(3)}"
          plug_name "(DEBUG) What plugin name should CTF use", present, default: 'RandomEventQPlugin'
        else
          allow :prefix
          allow :plug_name
        end
      end
    end
  end

  def nginx
    # No questions, but Asky needs this here or it gets unhappy
    child :nginx do
    end
  end
end
