#
# Copyright 2013 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

# Lexi provides a simple parser for content, allowing you to provide a set of
# regex/key pairs.  When #parse is called, the provided block will be called
# with the matched text, and the key for the matched regular
# expression.  The unmatched text will be ignored.
#
# If there is content left over after the last match is made, it will be passed
# to the block, along with a matched text of nil and a key of nil.  If no text
# is left over, it does not pass an empty string.
class Lexi
  MatchedRange = Struct.new(:from, :to, :key, :match_data)

  # Returns a new parser that will operate on [str], looking for patterns.
  # @param str The string to parse.
  # @param *patterns Pairs of [pattern, key], where pattern is a regex to match and key is a token identifier that
  # will be passed to the block along with any matched text.  Multiple patterns can have the same key.
  def initialize(str, patterns={})
    @whole = str.to_s
    @patterns = patterns
  end

  # Returns association remote id & corresponding sources or type symbols
  # Iterate over the content provided and return array of the hash of only matching value with regex & contents
  #
  # supports multiple cases for creating associations and extracting associations from the message
  def parse
    return [] if @patterns.empty?
    atoms = []

    # Extract the pattern values from the hash pair
    # to determine whether creating associations or extracting associations
    pattern_values = @patterns.values.map do | element |
      element.map(&:_type).uniq.first if element.is_a?(Array)
    end.compact.uniq

    # Checking whether this parsing is meant for creating association or linkifing association
    if (pattern_values & ActivitySource::SOURCE_TYPES).any?
      @whole.scan(/\[\s*([^\]]+)\s*\]/) do | m |
        # This is the case when we create the associations from message
        m.each do | n |
          atoms += get_association_list(n)
        end
      end
    else
      # This is the case when we extract the association from message
      atoms = get_linkifing_list unless @whole.blank?
    end
    atoms.uniq
  end

  # This is the case when we extract the association from message
  def get_linkifing_list
    cur_loc = 0
    atoms = []
    loop do
      element = next_range(@whole[cur_loc..-1], cur_loc)
      if element
        found_where = cur_loc + element.from
        found_until = cur_loc + element.to
        matched = @whole[found_where...found_until]
        cur_loc = found_until
        atoms << element
        break if matched.empty?
      else
        # Can't find match anymore then break the stop the loop
        break
      end
    end
    atoms
  end

  # This is the case when we create the associations from message
  def get_association_list(n)
    atoms = []
    n.scan(/#{@patterns.keys.join '|'}/i) do | atom |
      @patterns.keys.map do | regex |
        # Due to grouping by scan, there are some nil value return in array.
        # we need to purge the nil and make the array clean.
        # EX: message = "artf1234"
        #     atom will be ['artf1234', nil, nil]
        #     Those two nil values are from ReviewRequest regex group
        #
        #     message = "rv1:test-1"
        #     atom will be [nil, 'rv1:test-1', 'test-1']
        #     The one nil value is from WorkItem regex group
        atom = atom.uniq.try(:compact)
        next if atom.empty?
        if match = atom[0].match(regex)
          element = @patterns[regex].first.is_a?(ReviewServer) ? atom[1] : atom[0]
          atoms << { element => @patterns[regex] }
        end
      end
    end
    atoms
  end

  private
  def next_range(body, start_point=0)
    range = @patterns.map do |pattern|
      found = body.match(pattern.first)
      found ? found.offset(0) + [pattern.last, found] : nil
    end.compact.sort_by.with_index { |triple, idx| [triple.first, idx]}.first
    MatchedRange.new(*range) if range
  end
end
