#
# Copyright 2013 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#
module MigrationHelper

  def self.extended(base)
    base.set_write_concerns
  end

  # returns a MongoMapper database object
  def db
    @db ||= MongoMapper.database
  end

  # when dealing with replica sets we want to ensure that the writes propagated to the majority of nodes.
  def set_write_concerns
    if db.connection.is_a? Mongo::ReplSetConnection
      db.write_concern[:w] = 'majority'
    end
  end

  # allows to scope the operations by the collection
  # @param[Symbol] collection collection name in snake case, i.e :activities, :activity_summaries
  # you need to specify the block of code to run within the collection scope, ie
  # within :activities do |collection|
  #   collection.count
  # end
  def within(collection)
    if collection.is_a? Mongo::Collection
      yield collection
    elsif collection.is_a? Symbol
      yield db[collection]
    end
  end

  # returns the number of documents in a given collection that satisfy the search criteria
  # @param[Symbol] collection, collection name, e.g. :activities
  # @param[Hash] search_criteria, search  each parameters for the find operation, e.g.{'_type' => 'ReviewRequestSummary'}
  def document_count(collection, search_criteria = {})
    within(collection) do |collection|
      collection.find(search_criteria).count
    end
  end

  # allows to iterate over the cursor and yield each document to the supplied block of code
  # @param [Symbol] collection_name, collection name in snake case, i.e :activities, :activity_summaries
  # @param [Hash] search_criteria search parameters for the find operation, e.g.{'_type' => 'ReviewRequestSummary'}
  # @param [Hash] options options for the find operation, e.g.  {timeout: false, limit: 100, skip: 50, sort: [['_id', -1]]}
  # you need to specify the block of code to run for each document.
  def iterate_over(collection_name, search_criteria = {}, options ={})
    within collection_name do |collection|
      collection.find(search_criteria, options) do |cursor|
        cursor.each do |document|
          yield document
        end
      end
    end
  end

  # by default renames the fields according to the convention (e.g. foo becomes __foo__). That allows us isolate data
  # from the application logic and at the same time keep it if we need it later for some reason (data recovery for example).
  # if force flag is set to true, then the keys are removed permanently
  # @param [Symbol] collection_name, collection name
  # @param fields[Array<String>] fields
  def remove_fields(collection_name, fields, force = false)
    if force
      unset_query_hash = array_to_hash(fields) { '' }
      within collection_name do |collection|
        collection.update({}, {'$unset' => unset_query_hash}, multi: true)
      end
    else
      rename_hash = array_to_hash(fields) {|field| "__#{field}__"}
      rename_fields(collection_name, rename_hash)
    end
  end

  # renames the fields in the collection according to the provided old value - new value rename hash
  # @param [Symbol] collection_name
  # @param [Hash] rename_hash, a hash in the form old_value => new_value
  def rename_fields(collection_name, rename_hash)
    within collection_name do |collection|
      collection.update({}, {'$rename' => rename_hash}, multi: true)
    end
  end

  # maps array to hash where each key is a element of an array and the corresponding value is determined using a provided
  # block of code, e.g array_to_hash(['a','b','c']) {|elm| elm + 'b'} will return {'a' => 'ab', 'b' => 'bb', 'c' => 'cb'}
  # @param [Array]
  def array_to_hash(array)
    array.reduce({}) do |hash, element|
      hash.merge!({element => yield(element)})
    end
  end



  # Adds fields to all documents within the given collection
  # @param [Symbol] collection_name
  # @param fields[Array<String>] field names or [Hash] a hash with field_name => default_value pairs
  def add_fields(collection_name, fields)
    set_hash = fields.is_a?(Array) ? array_to_hash(fields) { '' } : fields
    within collection_name do |collection|
      collection.update({}, {'$set' => set_hash}, {upsert: false, multi: true})
    end
  end

end