#
# Copyright 2015 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#
# This file contains the work item source migration process specifically,
# the work item source has a connection to a tracker source that are outside of the project
#
require 'migrations/exceptions'
module OutsideProjectScenario

  extend MigrationHelper

  # Process all work item sources where work item's tracker source do not belong to the work item’s project
  def self.process_work_item_sources_where_tracker_source_outside_project
    new_sources = []; old_sources = []; new_trackers = []; update_trackers = []; old_new_sources_link = []

    wis_data = self.find_wis_on_different_proj_shared_tracker

    wis_data.each do | wis |
      new_source = self.create_new_wis(wis)
      new_sources << new_source
      old_new_sources_link << {"old_workitem_id" => wis["source_id"], "new_workitem_id" => new_source}

      unless wis['active']
        new_tracker = self.create_new_tracker(wis['tracker_id'], new_source)
        self.update_new_wis_with_tracker(new_source, new_tracker)
        new_trackers << {"tracker_id" => new_tracker, "new_source" => new_source}
      end

      # Reconnect activities & activity summaries
      self.reconnect_activity_summaries_to_new_source(wis['source_id'], new_source)
      self.reconnect_activities_to_new_source(wis['source_id'], new_source)


      old_sources << wis['source_id']
    end

    # Mark a source to delete
    self.mark_to_delete(old_sources)
  end

  private
  def self.find_wis_on_different_proj_shared_tracker
    new_project_source_linkages_hash = {}
    group_by_trackers = []; temp_different_project_ids = []; temp_source_ids = []
    collected_source_ids = []; reorganized_data = []

    # Get the list of Tracker_ID & source_id
    group_by_trackers = db[:activity_sources].aggregate( [ { '$group' => {
                                                               '_id' => "$tracker_id",
                                                               'wis_ids' => { '$push' => "$_id" }
                                                           }
                                                           } ] )

    # Get the list of source & project & pipeline connection
    iterate_over(:pipelines) do | pipeline |
      if pipe_steps = pipeline['pipeline_steps'].select { | ps | ps['type'] == 'work_item' }
        pipe_steps.each do | step |
          step['source_ids'].each do | source_id |
            next if ActivitySource.find(source_id).try(:to_delete)
            tracker_id = group_by_trackers.map { | tracker | tracker['_id'] if tracker['wis_ids'].include?(source_id) }.flatten.compact.first
            tracker_source = self.find_tracker_source(tracker_id)
            source_activation = tracker_source['project_id'] == pipeline['project_id']
            new_project_source_linkages_hash["#{source_id}"] = {
                "source_id" => source_id,
                "project_id" => pipeline['project_id'],
                "tracker_id" => tracker_id,
                "url" => tracker_source['url'] ,
                "active" => source_activation,
                "_type" => "WorkItemSource"
            }
          end
        end
      end
    end

    if group_by_trackers = group_by_trackers.reject { | gt | gt['_id'].nil? }
      group_by_trackers.each do | tracker |
        # Reset the temp array to hold ids
        temp_source_ids = []

        tracker['wis_ids'].each do | source_id |
          if source_linkage = new_project_source_linkages_hash["#{source_id}"]
            unless source_linkage["active"]
              temp_source_ids << source_id
            end
          end
        end

        collected_source_ids += temp_source_ids
      end
    end

    collected_source_ids.each do | source_id |
      old_source = ActivitySource.find(source_id)
      new_attributes = {"trigger_status" => old_source.trigger_status, \
                        "display_name" => old_source.display_name, \
                        "created_at" => old_source.created_at.try(:utc), \
                        "updated_at" => old_source.updated_at.try(:utc), \
                        "access_status" => old_source.access_status}.merge(new_project_source_linkages_hash[source_id.to_s])
      reorganized_data << {"source_id" => BSON::ObjectId("#{source_id}")}.merge(new_attributes)
    end
    reorganized_data
  end

  # Create a new work item source
  def self.create_new_wis(option={})
    option.merge!(ActivitySource.generated_columns)
    new_data = db[:activity_sources].insert(option.except('source_id').except('queue_server'))
    new_data
  end

  # Create a new tracker
  def self.create_new_tracker(old_tracker_id, source_id)
    old_tracker_hash = Tracker.find(old_tracker_id).serializable_hash.except('id')
    old_tracker_hash['work_item_source_id'] = source_id
    new_tracker = db[:trackers].insert(old_tracker_hash)
    new_tracker
  end

  # Update new work item source with tracker
  def self.update_new_wis_with_tracker(source_id, new_tracker_id)
    db[:activity_sources].update({'source_id' => source_id}, {'$set' => {'tracker_id' => new_tracker_id}})
  end

  # Need to pass BSON::ObjectId('8573653948') as tracker_id
  def self.find_tracker_source(tracker_id)
    tracker_source_id = nil
    db[:trackers].find({"_id" => tracker_id}) do |cursor|
      tracker_source_id = cursor.first["tracker_source_id"]
    end

    raise MongoMapper::DocumentNotFound.new ("Tracker not found") if tracker_source_id.nil?
    tracker_source = nil
    db[:activity_sources].find({"_id" => tracker_source_id}) do |cursor|
      raise MongoMapper::DocumentNotFound.new ("Multiple Tracker sources found") if cursor.count > 1
      if cursor.count == 1
        tracker_source = cursor.first
      end
    end
    tracker_source
  end

  # Reconnect activity summaries with new source
  def self.reconnect_activity_summaries_to_new_source(old_source_id, new_source_id)
    db[:activity_summaries].update({"source_id" => old_source_id}, {'$set' => {"source_id" => new_source_id}}, :multi => true)
  end

  # Reconnect activities with new source
  def self.reconnect_activities_to_new_source(old_source_id, new_source_id)
    db[:activities].update({"source_id" => old_source_id}, {'$set' => {"source_id" => new_source_id}}, :multi => true)
  end

  def self.mark_to_delete(source_ids=[])
    db[:activity_sources].update({'_id' => { '$in' => source_ids}}, {'$set' => {'to_delete' => true}}, :multi => true)
    source_ids.join(',')
  end
end