#
# Copyright 2015 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

module WithinProjectScenario

  extend MigrationHelper

  def self.process_sources_within_project
    # 1. Get all pipelines grouped by project
    pipelines_for_project_bson = db[:pipelines].aggregate( [ { '$group' => { '_id' => "$project_id", 'pipe_ids' => { '$push' => "$_id" } } } ] )
    pipelines_for_project = {}
    pipelines_for_project_bson.map{|hash| pipelines_for_project[hash["_id"]] = hash["pipe_ids"]}

    # 2. Get a map of sources to project
    source_to_project = {}
    pipelines_for_project.each do |project, pipelines|
      pipelines.each do |pipeline_id|
        steps = self.get_source_steps_for_pipeline(pipeline_id, 'work_item')
        steps.each do |each_step|
          source_ids = each_step['source_ids']
          source_ids.each {|source_id| source_to_project[source_id] = project}
        end
      end
    end

    # 3. Get all work item sources grouped by tracker id, to get
    #    source to tracker map
    source_to_tracker = {}
    group_by_trackers_bson = db[:activity_sources].aggregate( [ { '$group' => { '_id' => "$tracker_id", 'wis_ids' => { '$push' => "$_id" } } } ] )
    group_by_trackers_bson.map do |hash|
      hash["wis_ids"].map {|id| source_to_tracker[id] = hash["_id"]}
    end

    # 4. Create a temporary collection to use aggregation functionality from mongo
    #    to get all sources grouped by tracker and project
    db.create_collection("source_project_tracker")
    source_to_project.each do |source, project|
      if source.present? && project.present? && source_to_tracker[source].present?
        db[:source_project_tracker].insert({"source_id" => source, "tracker_id" => source_to_tracker[source], "project_id" => project })
      end
    end

    sources_grouped_by_proj_and_tracker_bson = db[:source_project_tracker].aggregate( [ { '$group' => { '_id' => {"proj" => "$project_id",
                                                                                                                  "tracker" => "$tracker_id"},
                                                                                                        'source_ids' => {'$push' => "$source_id"} }} ])

    # 5. Go through the list of sources (grouped by project and tracker)
    #    to a. Create a new Work Item Source (WIS)
    #       b. Create a new tracker
    #       c. Associate new WIS to tracker and project
    #       d. Merge activities and summaries if there are multiple WIS for a tracker, project combination
    #       e. As part of creating new WIS, new tracker, new Activities and Summaries, mark old sources,
    #          trackers, activities, summaries as 'to_delete'
    sources_grouped_by_proj_and_tracker_bson.map do |hash|
      hash['source_ids'].uniq!
      work_item_sources = []

      hash['source_ids'].map do |source_id|
        next unless old_work_item_source = find_source_by_id(source_id)
        work_item_sources << old_work_item_source
      end

      if sources_have_same_tracker_within_project?(hash['_id']['proj'], hash['_id']['tracker'])
        new_work_item_source_id = self.create_new_work_item_source_from_pipeline_sources(work_item_sources, hash['_id']['proj'], hash["_id"]["tracker"])
        self.merge_association_summaries_for_source(hash['source_ids'], new_work_item_source_id, 'work_item')
        mark_to_delete(hash['source_ids'])
      end
    end

    # 6. Remove the temporarily created collection
    db[:source_project_tracker].drop()
  end

  private

  def self.sources_have_same_tracker_within_project?(project_id, tracker_id)
    tracker_source = self.find_tracker_source(tracker_id)
    tracker_source['project_id'] == project_id
  end

  def self.create_new_work_item_source_from_pipeline_sources(work_item_sources, project_id, tracker_id)
    # Use first work item source hash
    new_work_item_source_hash = work_item_sources.first.dup
    project = find_project(project_id)
    new_work_item_source_hash['project_id'] = project['_id']
    new_work_item_source_hash.delete('_id')
    new_work_item_source_hash.merge!(ActivitySource.generated_columns)
    new_work_item_source_id = db[:activity_sources].insert(new_work_item_source_hash)
    new_tracker_id = self.create_tracker(new_work_item_source_id, tracker_id)
    tracker_source = self.find_tracker_source(tracker_id)
    mark_to_delete(tracker_source['_id'])
    db[:activity_sources].update({'_id' => new_work_item_source_id} ,
                                 {'$set' => {'url' => tracker_source['url'],
                                             'tracker_id' => new_tracker_id}})

    new_work_item_source_id
  end

  def self.find_tracker(tracker_id)
    tracker = nil
    db[:trackers].find({'_id' => tracker_id}) do |cursor|
      tracker = cursor.first
    end
    raise MongoMapper::DocumentNotFound.new ("Tracker not found") if tracker.nil?

    tracker
  end

  def self.find_project(project_id)
    project = nil
    db[:projects].find({'_id' => project_id}) do |cursor|
      project = cursor.first
    end
    raise MongoMapper::DocumentNotFound.new ("Project not found") if project.nil?
    project
  end

  def self.create_tracker(new_work_item_source_id, old_tracker_id)
    old_tracker_hash = find_tracker(old_tracker_id)
    old_tracker_hash['work_item_source_id'] = new_work_item_source_id
    old_tracker_hash.delete('tracker_source_id')
    old_tracker_hash.delete('_id')
    new_tracker_id = db[:trackers].insert(old_tracker_hash)

    new_tracker_id
  end

  def self.get_source_steps_for_pipeline(pipeline_id, type)
    pipeline = nil
    db[:pipelines].find({"_id" => pipeline_id}) do |cursor|
      pipeline = cursor.first
    end
    raise MongoMapper::DocumentNotFound.new ("Pipe not found") if pipeline.nil?
    pipeline["pipeline_steps"].select { | ps | ps['type'] == type }
  end

  def self.mark_to_delete(*ids_to_be_deleted)
    ids_to_be_deleted.each do | id |
      db[:activity_sources].update({'_id' => id}, {'$set' => {'to_delete' => true}})
    end
  end

  def self.find_tracker_source(tracker_id)
    tracker_source_id = nil
    db[:trackers].find({"_id" => tracker_id}) do |cursor|
      tracker_source_id = cursor.first["tracker_source_id"]
    end

    raise MongoMapper::DocumentNotFound.new ("Tracker not found") if tracker_source_id.nil?
    tracker_source = nil
    db[:activity_sources].find({"_id" => tracker_source_id}) do |cursor|
      raise MongoMapper::DocumentNotFound.new ("Multiple Tracker sources found") if cursor.count > 1
      if cursor.count == 1
        tracker_source = cursor.first
      end
    end

    tracker_source
  end

  def self.find_source_by_id(id)
    source = nil
    db[:activity_sources].find({"_id" => id}) do |cursor|
      source = cursor.first
    end

    source
  end

  def self.merge_association_summaries_for_source(dup_source_ids, new_source_id, type=nil)
    # Find activity summaries for each of the old work item sources
    # and group them by remote_ids. We will use the first summary as
    # the unique summary for the new work item source

    # Filter summaries that belong to sources (in dup sources)
    # Group them by remote_id and get a temp collection with
    # number of summaries in 'summary_ids'
    new_summaries_query = {}; new_summaries_query = {}
    if type == 'work_item'
      new_summaries_query = { '$group' => { '_id' => {"remote_id" => "$remote_id"},
                                            'summary_ids' => { '$push' => "$_id" },
                                            'group_count' => { '$sum' => 1 }
                                          }
                            }

      new_activities_query = { '$group' => { '_id' => {'create_year' => { '$year' => "$created_at" },
                                                       'create_month'=> { '$month' => "$created_at" },
                                                       'create_day'=> { '$dayOfMonth' => "$created_at" },
                                                       'create_hour'=> { '$hour' => "$created_at" },
                                                       'create_min'=> { '$minute' => "$created_at" },
                                                       'create_sec'=> { '$second' => "$created_at" },
                                                       "remote_id" => "$remote_id"},
                                             'activity_ids' => { '$push' => "$_id" }}
                             }
    elsif type == 'commit'
      new_summaries_query = { '$group' => { '_id' => {"revision_number" => "$revision_number"},
                                            'summary_ids' => { '$push' => "$_id" },
                                            'group_count' => { '$sum' => 1 }
                                          }
                            }

      new_activities_query = { '$group' => { '_id' => {"revision_number" => "$revision_number"},
                                             'activity_ids' => { '$push' => "$_id" }}
                             }

    else
      raise MongoMapper::DocumentNotFound.new ("Type : #{type} Not supported")
    end

    db[:activity_summaries].aggregate( [ { '$match' => { 'source_id' => { '$in' => dup_source_ids }}},
                                           new_summaries_query,
                                         {'$out' => "summaries_with_counts"} ])

    # Collect all summaries that need merging
    db[:summaries_with_counts].aggregate([
                                          { '$match' => { 'group_count' => {'$gt' => 1}} },
                                          { '$group' => { '_id' => {"group_count" => "$group_count"},
                                            'summary_ids_with_gt_1' => { '$push' => "$summary_ids" }}},
                                          {'$out' => "summaries_that_need_merging"}])

    old_summary_ids_to_new = {}
    to_be_merged_summaries = []
    db[:summaries_that_need_merging].find() do |cursor|
      cursor.each do |document|
        to_be_merged_summaries.push(*document['summary_ids_with_gt_1'])
      end
    end

    # In this loop, use the first summary as the summary to be retained
    # merge associated_ids and comments of the duplicates into the first summary
    # Merge Activities of the activity_summaries
    to_be_merged_summaries.each do |summary_ids|
      # Summaries need merging, since there are duplicate sources
      # Find all summaries where ids in summary_ids array, get the associated summary ids and comment ids
      associated_summaries_and_comments_of_dup_sources = db[:activity_summaries].find({"_id" => { '$in' => summary_ids}})

      merged_associated_ids = []
      merged_comments = []
      associated_summaries_and_comments_of_dup_sources.each do |summary|
        merged_associated_ids.push(*summary['associated_activity_summary_ids'])
        merged_comments.push(*summary['comments'])
      end
      merged_associated_ids.uniq!
      merged_comments.uniq!

      db[:activity_summaries].update({'_id' => summary_ids.first},
                                     {'$set' => {
                                                 'comments' => merged_comments,
                                                 'associated_activity_summary_ids' => merged_associated_ids
                                                 }})
      db[:activity_summaries].update({'_id' => summary_ids.first},
                                     {'$set' => {'source_id' => new_source_id}})

      summary_ids.from(1).each do |id|
        old_summary_ids_to_new[id] = summary_ids.first
        db[:activity_summaries].update({'_id' => id},
                                       {'$set' => {'to_delete' => true}})

      end
      # Update summaries activities to point to chosen summary id ie summary_ids.first
      # from the ones marked for delete ie summary_ids.from(1)
      db[:activities].update({'activity_summary_id' => { '$in' => summary_ids.from(1) }},
                             {'$set' => {'activity_summary_id' => summary_ids.first}},
                              :multi => true)
    end
    # This will update the remaining activity summaries (removed in bulk operation to get new source)
    dup_source_ids.each do |old_source_id|
       db[:activity_summaries].update({"source_id" => old_source_id}, {'$set' => {"source_id" => new_source_id}}, :multi => true)
    end

    # Fix associated summary ids for other summaries in associated
    new_source = ActivitySource.find(new_source_id)
    new_source.activity_summaries.each do |activity_summary|
      activity_summary.associated_activity_summary_ids.each do |other_summary_id|
        associated_ids = ActivitySummary.find(other_summary_id).associated_activity_summary_ids
        modified_associated_ids = []
        associated_ids.each do |id|
          if old_summary_ids_to_new[id].present?
            modified_associated_ids.push(old_summary_ids_to_new[id])
          else
            modified_associated_ids.push(id)
          end
        end
        db[:activity_summaries].update({'_id' => other_summary_id},
                                       {'$set' => {'associated_activity_summary_ids' => modified_associated_ids}})
      end
    end

    # Merge activities only if there are dup sources..
    # Otherwise update all the activity's source id to the new_source_id
    if dup_source_ids.count > 1
      # Merge activities with same remote and source id
      activities_of_source_grouped_by_remote_and_create_time_bson = \
          db[:activities].aggregate( [ { '$match' => { 'source_id' => { '$in' => dup_source_ids }}},
                                         new_activities_query ] )

      grouped_activities = []
      activities_of_source_grouped_by_remote_and_create_time_bson.each do |hash|
        grouped_activities << hash['activity_ids']
      end
      grouped_activities.each do |activity_ids|
        db[:activities].update({'_id' => activity_ids.first},
                               {'$set' => {'source_id' => new_source_id }})
        activity_ids.from(1).each do |id|
          db[:activities].update({'_id' => id},
                                 {'$set' => {'to_delete' => true}})

        end
      end
    else
      db[:activities].update({"source_id" => dup_source_ids.first}, {'$set' => {"source_id" => new_source_id}}, :multi => true)
    end

    # clear out temp collections
    db[:summaries_with_counts].drop()
    db[:summaries_that_need_merging].drop()

  end


end
