#
# Copyright 2013 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

# Reads configuration files based on a default file. If a user specific configuration exists, then that is merged with
# the default configuration.
class OverridableConfig
  class ConfigurationError < StandardError
  end

  # Loads the default configuration, then a user-specific override.
  #
  # @param [String] user The user whose configuration file we should merge into the default config, if any.
  def initialize(user = ENV['USER'])
    defaults_file = File.join(Rails.root, 'config', 'defaults.yml')
    unless File.exists?(defaults_file)
      raise ConfigurationError, "#{defaults_file} does not exist.  Please create this file and try again."
    end
    @loaded_configuration = parse_yaml(defaults_file)
    @loaded_configuration.merge!(parse_yaml(user_file(user))) if user
  end

  # Merge in another user's configuration.
  #
  # @param [String] new_user The user whose configuration we should merge into the currently constructed config
  def merge_user(new_user)
    @loaded_configuration.merge!(parse_yaml(user_file(new_user)))
  end

  # Extract the configuration for a subsection, scoped by the current Rails environment (staging, development, etc...)
  #
  # @param [String] base_name The configuration section to return.
  # @return [Hash] A hash of configuration entries from the passed in section, under the Rails environment subsection.
  def load(base_name)
    @loaded_configuration[base_name][Rails.env]
  end

  # Extract the entire configuration for a subsection
  #
  # @param [String] base_name The section to return.
  # @return [Hash] A hash of configuration entries from the passed in section.
  def load_raw(base_name)
    @loaded_configuration[base_name]
  end

  private
  # The location of the configuration file for the provided user.
  def user_file(user)
    File.join(Rails.root, 'config', 'user', "#{user}.yml")
  end

  # Parse a configuration file, first as an ERB template, then as a YAML file, returning a hash from the YAML.
  #
  # This passes a binding to the ERB parsing, even though nothing out of the current context is being used because this
  # can happen while another ERB process is running.  The outer ERB process uses the TOPLEVEL_BINDING.  However the
  # output appears to be tied to the binding, so if you don't pass a specific binding to #result it will attempt to ALSO
  # use the TOPLEVEL_BINDING which results in the parsed user configuration (this level's ERB) being prepended to the
  # outer level's ERB processing.  This is a long explanation for why you need to keep .result(binding) even though
  # nothing in the local binding is being used by the processed ERB.  --  Morgan
  def parse_yaml(file)
    if File.exist?(file)
      YAML.load(ERB.new(open(file).read).result(binding))
    else
      {}
    end
  end
end
