#
# Copyright 2016 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

require 'net/http'
require 'json'
require 'cgi'
require File.expand_path("../../application_constants", __FILE__)

# This Rabbit::Management module is independent from Rails due to our Upgrade script.
# So, this module should not contain rails-dependent code.
module Rabbit
  class Management
    GET = Net::HTTP::Get
    PUT = Net::HTTP::Put
    DELETE = Net::HTTP::Delete
    DEFAULT_PERMISSIONS = { configure: "^(#{OLD_APP_BASE_QUEUE_NAME}|#{APP_BASE_QUEUE_NAME}).*|amq.default|#{EVENT_HUB_NAME}", write: "^(#{OLD_APP_BASE_QUEUE_NAME}|#{APP_BASE_QUEUE_NAME}).*|amq.default|#{EVENT_HUB_NAME}", read: '' }

    attr_accessor :logger
    attr_writer :permissions

    # Construct a Management interface.
    #
    # @param [String] admin_user A user with the administration tag.
    # @param [String] admin_pass The password for the administration user.
    # @param [String] rabbit_uri RabbitMQ uri from the defaults.yml
    def initialize(admin_user, admin_pass, rabbit_uri)
      uri = URI.parse(rabbit_uri)
      uri.password = uri.user = nil
      @rabbit_host = uri.host
      @rabbit_port = 15672
      @vhost = (uri.path.empty? || uri.path.nil?) ? CGI.escape('/') : uri.path[1..-1]
      @admin_user = admin_user
      @admin_pass = admin_pass
      @http = Net::HTTP.new(@rabbit_host, @rabbit_port)
    end

    # Create a user on the RabbitMQ server
    #
    # @param [String] username The username to give to the RabbitMQ user
    # @param [String] password The password to assign to the RabbitMQ user
    # @return [Net::HTTPResponse] The HTTP response from the RabbitMQ API server.
    def create_user(username, password)
      req = api_request(PUT, "users/#{username}")
      req.body = {password: password, tags: APP_NAME}.to_json

      @http.request(req).is_a? Net::HTTPSuccess
    rescue => bad_request
      logger.warn "Could not create user #{username}: #{bad_request}" if logger
      false
    end

    # Limit the permissions of the given user; the default permissions allow
    # configure and write on APP_BASE_QUEUE_NAME.* for the adapters to optionally 
    # create the queue, and always write to it.  This should be made more general.
    #
    # @param [String] username The username to set permissions on.
    # @return [Net::HTTPResponse] The HTTP response from the RabbitMQ API server.
    def set_permissions(username, permissions = DEFAULT_PERMISSIONS)
      req = api_request(PUT, "permissions/#{@vhost}/#{username}")
      req.body = permissions.to_json

      @http.request(req).is_a? Net::HTTPSuccess
    rescue => bad_request
      logger.warn "Could not set permissions for #{username}: #{bad_request}" if logger
      false
    end

    # Delete a user from the RabbitMQ server
    #
    # @param [String] username The user to delete.
    def delete_user(username)
      req = api_request(DELETE, "users/#{username}")
      @http.request(req).is_a? Net::HTTPSuccess
    rescue => bad_request
      logger.warn "Could not delete #{username}: #{bad_request}" if logger
      false
    end

    # Retrieve a list of users from the server.
    # @return [Array<Hash[name, password_hash, tags]>] A list of hashes with user information.
    def users
      req = api_request(GET, 'users')
      @http.request(req) do |resp|
        return JSON.parse(resp.read_body)
      end
    rescue
      []
    end

    # Aliveness check; this will create a queue, post to it, and consume the post to verify that everything is working.
    #
    # @return [boolean] true if the server successfully performed its operations, false if any error occurred.
    def alive?
      req = api_request(GET, "aliveness-test/#{@vhost}")

      body = "{}"
      @http.request(req) do |resp|
        body = resp.read_body
      end
      JSON.parse(body)['status'] == 'ok'
    rescue => _
      false
    end

    private
    def api_request(kind, path)
      # Naively strip the leading slash if present
      path = path[1..-1] if path[0] == '/'
      kind.new("/api/#{path}").tap do |req|
        req.add_field('Content-Type', 'application/json') if kind == PUT
        req.basic_auth(@admin_user, @admin_pass)
      end
    end
  end
end

if __FILE__  == $0
  rmq = Rabbit::Management.new('guest', 'guest', 'localhost')
  puts "Alive: #{rmq.alive?}"
  puts "Users: #{rmq.users}"
  puts "Create a temporary user: #{rmq.create_user('mrs', 'password')}"
  puts "Users: #{rmq.users}"
  puts "Set its permissions: #{rmq.set_permissions('mrs')}"
  puts "Users: #{rmq.users}"
  puts "Delete the user: #{rmq.delete_user('mrs')}"
  puts "Users: #{rmq.users}"
  puts "Alive: #{rmq.alive?}"
end
