#
# Copyright 2016 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

# Application-side queue daemon.  Takes messages from queues and sends them to the proper parsers.

# Modify Dante to play nice with Foreman, and to not use 'puts'
module Dante
  class Runner
    def log(message)
      Rails.logger.info message
    end

    def redirect!
      redirect_output!
    end
  end
end

class QueueDaemon

  SHUTDOWN_SIGNALS = %w(INT TERM KILL)

  attr_accessor :daemon
  attr_accessor :out

  # Initialize the daemon defaults
  # The defaults here should change for production scenarios, but are fine for now
  # @param [String] runner_name The name of this runner
  def initialize(runner_name = 'queuerunner')
    self.out = $stdout

    @daemon = Dante::Runner.new('queuerunner', pid_path: "#{Rails.root}/tmp/pids/#{runner_name}.pid", log_path: "#{Rails.root}/log/#{runner_name}.log")
    @daemon.description = "#{I18n.t('app_name')}'s RabbitMQ Service"
    @daemon.with_options do |opts|
      options[:source_type] = runner_name.sub('_queue','')
      opts.on('-s', '--source-type SOURCE_TYPE', String, 'The source type to connect to') do |source_type|
        options[:source_type] = source_type
      end
    end
  end

  # Starts the main daemon loop
  def execute
    log_level = ActiveSupport::Logger::Severity::INFO
    Rails.logger = ActiveSupport::Logger.new(self.out, log_level)

    @daemon.execute do |opts|
      unless opts[:daemonize]
        @daemon.redirect!
      end

      trap_shutdown_signals

      source_type = lookup_source_type(opts)

      while true
        # EventMachine loop is in the daemon so we can handle signals properly.
        EventMachine.run do
          RabbitConsumer.new(source_type::QUEUE_NAME).start do |metadata, payload|
            consume(source_type, metadata, payload)
          end
        end
      end
    end
  end

  # Consumes the message and processes it based on the specifies source type
  # @param [Class] source_type the source parser for the messages
  # @param [JSON] metadata the message metadata
  # @param [JSON] payload the message data
  def consume(source_type, metadata, payload)
    MongoMapper::Plugins::IdentityMap.clear
    source_type.process_message(payload)
    metadata.ack
  rescue => e
    # Trap all exceptions so we don't die.
    error_msg = I18n.t('daemon.caught_exception', error: $!)
    Rails.logger.info(error_msg)
    Rails.logger.debug(e.backtrace.join("\n"))
    metadata.reject(requeue: false)
  end

  # Trap shutdown signals, so we can kill things nicely (and in the right order)
  def trap_shutdown_signals
    SHUTDOWN_SIGNALS.each { |sig| Signal.trap(sig) { shutdown } }
  end

  # Check to see if there is an item parser available for the source type
  # @param opts [Hash] the options passed to the daemon
  # @return [Class] the Class that matches the source types QueueItemParser
  def lookup_source_type(opts)
    if opts.key?(:source_type)
      begin
        source_type = "#{opts[:source_type].camelize}QueueItemParser".constantize
      rescue NameError
        shutdown(I18n.t('daemon.parser_not_found', source_type: opts[:source_type]), 2)
      end
      # The source type must be specified, or we can't run
    else
      shutdown(I18n.t('daemon.parser_not_specified'), 1)
    end
    source_type
  end

  # Shuts the process down gracefully.
  # @param [String] error_msg the message to log when shutting down
  # @param [Integer] exit_code the exit code to pass to the exit method
  def shutdown(error_msg = I18n.t('daemon.caught_shutdown'), exit_code = 0)
    EventMachine.stop if EventMachine.reactor_running?

    Rails.logger.info(error_msg)

    @daemon.stop if @daemon.daemon_running?
    exit(exit_code)
  end
end

