#
# Copyright 2013 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

require 'application_constants'

# Provides RabbitMQ consumer functionality
class RabbitConsumer
  # Constructor
  # @param [String] queue_name The name of the RabbitMQ queue we want to connect to
  def initialize(queue_name)
    @connected = false
    @connection_attempts = 0
    @queue_name = queue_name
    @config = Rails.configuration.local_settings.load('rabbitmq')
  end

  # Consumes messages and passes them to the provided block for processing
  def consume
    channel = AMQP::Channel.new(@rabbit_connection)

    # Create an exchange in RabbitMQ to support routing
    exchange = channel.topic(EVENT_HUB_NAME, durable: true)
    
    # Create a EventQ queue and bind since orchestrate queue are bind to event.hub
    queue = channel.queue(@queue_name, durable: true).bind(exchange, routing_key: @queue_name)

    AMQP::Consumer.new(channel, queue).consume.on_delivery(&@consume_block)
  end

  # Connects to Rabbit and sets up triggers for events
  # @param [Block] block Block to pass to the consumer method
  def start(&block)
    @consume_block = block
    @rabbit_connection = AMQP.connect(@config["uri"], on_tcp_connection_failure: method(:connection_failure))
    @rabbit_connection.on_recovery(&method(:set_connected))
    @rabbit_connection.on_open(&method(:set_connected))
    @rabbit_connection.on_tcp_connection_loss(&method(:connection_failure))
  end

  # Handles logging connections and resetting the sleep timer
  # @param [Varies] args Args which get dropped on the floor
  def set_connected(*args)
    if @connection_attempts == 0
      Rails.logger.info(I18n.t("api.initial_connection", uri: @config['uri']))
    else
      Rails.logger.info(I18n.t("api.recovered_connection", uri: @config['uri']))
    end

    @connection_attempts = 0
    @connected = true
    consume
  end

  # Handles logging and sleeping for connection issues (drops any arguments on the floor)
  # @param [Varies] args Args which get dropped on the floor
  def connection_failure(*args)
    timeout = compute_timeout

    if @connected
      Rails.logger.error(I18n.t("api.lost_connection", uri: @config['uri']))
      Rails.logger.error(I18n.t("api.retry_connection", timeout: timeout))

      @connected = false
      sleep(timeout)
      @rabbit_connection.reconnect(true)
    else
      Rails.logger.error(I18n.t("api.initial_connection_failure", uri: @config['uri']))
      Rails.logger.error(I18n.t("api.retry_connection", timeout: timeout))

      @connected = false
      sleep(timeout)
      start(&@consume_block)
    end
  end

  # Computes the proper time to sleep and then sleeps that amount
  def compute_timeout
    timeout_sleep = [2*@connection_attempts,30].min
    @connection_attempts += 1

    timeout_sleep
  end
end
