#
# Copyright 2014 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

require 'json'
require 'active_support/all'

# A few things this parser doesn't handle...
#   Optional Parameters
#   Parameters that can be any of a few different types
#   Looping :)
# Using JSON to pass hash and array objects makes semicolon-based line separation difficult, as it's tough to tokenize.
# Currently the language is purely newline separated.
# It would be possible to construct a RACC-style parser that would be more powerful to handle command separators, but
#   it would be potentially more difficult, and require some pre-processing before delivery.

class Language
  COMMAND_REGEX = /(?<command>[a-zA-z][a-zA-Z0-9_-]*)/
  STRING_REGEX = /(?<quote>["'])(?<parameter>((.(?!\k<quote>))*.?))(?:\k<quote>)/
  INTEGER_REGEX = /(?<parameter>\d+)/
  RFC3339_REGEX = /(\d{4})-(\d{2})-(\d{2})([Tt](\d{2}):(\d{2}):(\d{2})(\.\d+)?)?([Zz]|([+-])(\d{2}):(\d{2}))?/
  TIMERANGE_REGEX = /(\d+)\.(minutes?|hours?|days?|weeks?|months?|years?|seconds?)/
  FLEXTIME_REGEX = /((?i)now)|#{TIMERANGE_REGEX}\.(ago|from_now)(\.at_(midnight|noon|midday|((beginning|end)_of_(minute|hour|day|week|month|quarter|year))))?/
  TIME_REGEX = /(?<parameter>(?<rfc3339>(#{RFC3339_REGEX}))|(?<flextime>(#{FLEXTIME_REGEX})))/
  RANGE_REGEX = /(?<parameter>#{TIMERANGE_REGEX})/

  TYPE_MATCHERS = { string: STRING_REGEX, integer: INTEGER_REGEX, json: :everything, duration: RANGE_REGEX, time: TIME_REGEX }
  TYPE_CONVERTERS = { string: :to_s, integer: :to_i, json: :from_json, duration: :safe_eval, time: :from_time, choice: :to_s }

  attr_accessor :_query

  # Convert a parameter from json into a hash or array.  This is used to pass
  # hashes or arrays as parameters to commands.
  # @param [String] json The parameters passed into the command, must be in
  #                      JSON format or an exception gets thrown.
  # @return [Hash|Array] The Ruby-native equivalent of the JSON passed in.
  def from_json(json)
    JSON.parse(json)
  end

  # Executes a safely matched block of Ruby code in the current context.
  # @warning Only use this when the regular expression that matches it is VERY specific and
  # does not allow any excess code.  This is a useful escape hatch, but very dangerous if
  # not used carefully.
  # @param [String] code The Ruby expression to be evaluated.
  # @return [Object] Whatever the Ruby expression evaluates to.
  def safe_eval(code)
    #TODO: [cmiller] Is there anything that we can do to make this safer?
    eval(code)
  end

  # Converts a timelike string from either RFC3339 format or a mildly complex
  # flexible time format (like 3.weeks.ago.at_beginning_of_quarter or 1.week.ago.at_midnight)
  # into a native time value.
  # @param [String] timelike A time either in RFC3339 or Ruby flextime format.
  # @return [Time] The time represented by the timelike string.
  def from_time(timelike)
    match = timelike.match(TIME_REGEX)
    if match[:rfc3339]
      DateTime.rfc3339(match[:rfc3339]).to_time
    elsif match[:flextime]
      flextime = match[:flextime]
      flextime = "Time.now" if flextime.downcase == "now"
      eval(flextime).to_time
    end
  end

  def run(program)
    raise ArgumentError.new('Empty query') if program.size == 0
    tokens = tokenize(program)

    tokens.each do |command, raw_arguments, command_options|
      arguments = []
      arguments << convert_noun(self, raw_arguments, command_options[:type]) if command_options[:type]
      operation = command_options[:method] || command
      self._query = send(operation, *arguments)
    end
    _query
  end

  class << self
    # Add a new language command.
    # The command can have a :type (:string, :integer, :json, :time, :choice), a :method (defaults to the command name),
    # and (if type is choice) an array of acceptable :choices.
    # e.g.
    #   add_command :doit, type: :string, method: :do_something
    #   add_command :play, type: :choice, choices: %w(video music game)
    #
    # @param [Symbol] command The command itself, in symbol form, starting with an alphabetic character (case
    #                         insensitive), then any number of alphanumeric (or - or _) characters.
    # @param [Hash] options - A hash containing settings for this command, including :type [:string, :integer, ...?],
    #                         and :method which is the method on the child class to call to execute this command.
    def add_command(command, options = {})
      raise ArgumentError.new("Command must be a symbol") unless command.is_a?(Symbol)
      raise ArgumentError.new("Command must be alphanumeric with numbers, '-' and '_' only allowed after the first character") unless command.to_s.match(/\A#{COMMAND_REGEX}\z/)
      @commands ||= {}
      options[:choice_matcher] = create_choice_regex(options[:choices]) if options[:type] == :choice
      @commands[command] = options
    end

    def create_choice_regex(choices)
      Regexp.new("\\A\\s*(?<parameter>" + choices.map do |choice|
        if choice.is_a? Regexp
          choice.to_s
        else
          Regexp.quote(choice.to_s)
        end
      end.join("|") + ")\\s*\\z")
    end

    # Run a program in the language defined by add_command calls.
    # @param [String] program The newline-separated list of commands and parameters to those commands to execute.
    # @return [Object] The return value of the last command executed in the program.
    def run(program)
      new.run(program)
    end

    protected
    attr_accessor :commands
  end

  private
  def convert_noun(context, noun, param_type)
    converter = TYPE_CONVERTERS[param_type]
    if noun.respond_to? converter
      noun.send(converter)
    else
      context.send(converter, noun)
    end
  end

  # Recognize a line with just spaces, or with spaces then a '#' character as comments.
  def comment?(line)
    line.match(/\A\s*(#|\z)/) != nil
  end

  def tokenize(program)
    program.split("\n").map.with_index do |line, index|
      next if comment?(line)

      line_number = index + 1
      parsed = get_command(line_number, line)
      command, command_options = lookup_command(line_number, parsed)
      parameter = parse_parameter(line_number, command_options, parsed) if command_options[:type]

      [command, parameter, command_options]
    end.compact
  end

  def get_command(line_number, line)
    line.match(/\A\s*#{COMMAND_REGEX}(\s+(?<rest>.*))?\z/) ||
        raise(SyntaxError.new("Could not find a command in statement #{line_number}"))
  end

  def lookup_command(line_number, parsed)
    command = parsed[:command].to_sym
    command_options = self.class.instance_variable_get(:@commands)[command]
    raise SyntaxError.new("Unrecognized command '#{command}' in statement #{line_number}") unless command_options
    return command, command_options
  end

  def parse_parameter(line_number, command_options, parsed)
    param_type = command_options[:type]
    param_matcher = TYPE_MATCHERS[param_type] || command_options[:choice_matcher]
    return parsed[:rest] if param_matcher == :everything
    param = parsed[:rest].match(param_matcher) ||
        raise(SyntaxError.new("Bad parameter format in statement #{line_number}"))
    param[:parameter]
  end
end
