#
# Copyright 2014 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

# A very simple query language, based off of commands and parameters (sort of like verbs and nouns)
# to retrieve information from the database.
#
# A few notes on notation...
# @chains This method returns a proxy object, and you can add more conditionals on it.  It may resolve
#         the previous chain before doing so, but it still returns a chain-able proxy object.
# @resolves The return for this method will be something to pass as a return value.  It is unlikely to
#           be useful to put any commands after this.
# @requires If a command is going to query for a set of source_ids, be sure to send the source_id array
#           through the filter_by_source method, otherwise Mongo will create an OR condition and return all sources from
#           any previous commands (including the filter by project automatically used for User project authorization)
class QueryLanguage < Language
  SAMPLE_CODE = <<-EOCODE
    # Find all activities associated with this artifact
    associated_to ["artf192283"]
    # Find all activities associated to what we've found so far.
    associations
    # Trim that list back to only work items.
    restrict_to work_items

    before now
    after 2014-10-18T22:50:34+00:00
    fields ["remote_id", "title"]dd
  EOCODE

  add_command :restrict_to, type: :choice, choices: %w(builds commits reviews work_items custom)
  add_command :before, type: :time
  add_command :associated_to, type: :json
  add_command :in_repository, type: :string
  add_command :associations
  add_command :after, type: :time

  add_command :fields, type: :json
  add_command :count

  add_command :group_by_field, type: :string
  add_command :field_contains, type: :json
  add_command :sort_by, type: :json
  add_command :limit, type: :integer
  add_command :offset, type: :integer

  add_command :in_project, type: :json

  add_command :has_any, type: :string
  add_command :has_none, type: :string

  add_command :timeseries, type: :duration
  add_command :tagged_as, type: :json

  add_command :in_source, type: :json

  def initialize
    @_query = ActivitySummary.query
    super
  end

  # Return result set associated to given SAK(s)
  # @requires filter_by_source(source_id_array)
  def in_source(source_assn_keys)
    source_ids = ActivitySource.where(source_association_key: source_assn_keys).map(&:id).compact
    filter_by_source source_ids
  end


  # Restrict the result set to a particular activity type in the :choices list.
  # @chains
  def restrict_to(activity_type)
    _query.where(:_type => resolve_constant(activity_type).to_s)
  end

  # Restrict the result set to only those summaries whose source is tagged with at least one of the specified names
  # @param ["tag name"], eg: tagged_as ["Deploy", "Binary Repository"]
  # @requires filter_by_source(source_id_array)
  def tagged_as(tags_ary)
    tag_ids = ActivitySourceTag.where(:name => tags_ary).all.map(&:id)
    source_ids = CustomActivitySource.where(:activity_source_tag_ids => tag_ids).all.map(&:id)
    filter_by_source source_ids
  end

  # Restrict the result set to a particular repository.  Useful when querying the associations of a
  # particular commit, e.g.:
  #
  #    in_repository "reps12234"
  #    field_contains ["revision_number", ["GIT_SHA1"]]
  #    associations
  #
  # @param repo_id The repository ID to restrict to
  # @chains
  # @requires filter_by_source(source_id_array)
  def in_repository(repo_id)
    filter_by_source CtfRepository.where(ctf_id: repo_id).all.map(&:id)
  end

  # Find all ActivitySummary's which have any elements in the provided array in their association list, by
  # remote_id (revision_number for commits).
  # @chains
  def associated_to(ary)
    ids = ActivitySummary.where(:$or => [{ remote_id: ary }, { revision_number: ary }]).fields(:id).all
    if ids.empty?
      _query.where(:_id => []) # need a filter that will not match any elements, any better ideas for that?
    else
      _query.where(:associated_activity_summary_ids => ids.map(&:id))
    end
  end

  # Given the current list of activities, find all activities associated to them.
  # @chains
  def associations
    ids = _query.all.map(&:id)
    ActivitySummary.where(:associated_activity_summary_ids => ids)
  end

  # Anything created (event_time / time_created) before the given timestamp.
  # @chains
  def before(timestamp)
    _query.where(:time_created.lt => timestamp)
  end

  # Anything created (event_time / time_created) after the given timestamp.
  # @chains
  def after(timestamp)
    _query.where(:time_created.gt => timestamp)
  end

  # Sort the result set by the given field in the given order (default is ascending)
  # @chains
  def sort_by(ary)
    field = ary[0]
    sort_order = ary[1] || :asc
    _query.sort(field => sort_order)
  end

  # Limits the number of records returned in the given result to the integer parameter
  # using 0 as the limit_to nullifies the sort op
  # @chains
  def limit(int)
    _query.limit(int)
  end

  # Return a list of documents beyond the number that are requested to be skipped.
  # such that if your query returns 5 items, and skip(1) is used, it will then return 4, starting at the second
  # @chains
  def offset(int)
    _query.skip(int)
  end

  # Return the count of the current state of the query.
  # @resolves
  def count
    { 'count' => _query.count }
  end

  # Return a specific subset of the fields for each element in the current result set.
  # @resolves
  def fields(ary)
    _query.all.map { |row| ary.map { |field| row.fetch_attribute(field) } }
  end

  # Return a hash of activity arrays, keyed by a field value
  # @resolves
  def group_by_field(field)
    _query.all.group_by { |row| row.fetch_attribute field }
  end

  # Returns the subset of the current result set where the specified field contains the specified value
  # Params: ary[0] field, ary[1] array of values to filter on (e.g. field_contains [ "remote_id", ["ARTF1247", "ARTF1248"] ])
  # @chains
  # @requires filter_by_source(source_id_array) if field value eq :source_id
  def field_contains(ary)
    field = ary[0].to_sym
    values = ary[1]
    if field == :source_id
      source_ids = ActivitySource.where(:id => { :$in => values }).all.map(&:id)
      filter_by_source source_ids
    else
      _query.where(field => { :$in => values })
    end
  end

  # Scopes the summaries by the projects
  # @param [Array] projects, an array of ctf project ids
  # @requires filter_by_source(source_id_array)
  def in_project(projects)
    activity_sources = projects.reduce([]) do |sources, ctf_project_id|
      project = TeamForge::ProjectConfiguration.where(ctf_project_id: ctf_project_id).first.try(:project)
      if project
        sources << ActivitySource.in_project(project)
      else
        sources
      end
    end.flatten

    filter_by_source activity_sources
  end

  # Scopes to the summaries that have summaries of certain type in their first-order associations.
  # e.g. has_any('builds')
  def has_any(type)
    activity_type = resolve_constant(type)
    result_set = _query.all.select do |summary|
      summary.activity_summaries.any? { |x| x.is_a? activity_type }
    end
    ActivitySummary.where(:_id => result_set.map(&:id))
  end

  # Scopes to the summaries that do not have summaries of certain type in their first-order associations.
  # e.g. has_none('builds')
  def has_none(type)
    activity_type = resolve_constant(type)
    result_set = _query.all.reject do |summary|
      summary.activity_summaries.any? { |x| x.is_a? activity_type }
    end
    ActivitySummary.where(:_id => result_set.map(&:id))
  end

  # groups the result set according to the provided unit of time (1.day, 30.minutes etc.)
  # assumes that the query contains 'before' and 'after' command
  def timeseries(time_unit)
    start_time = _query[:time_created].try(:[], :$gt)
    end_time = _query[:time_created].try(:[], :$lt)
    unless (start_time && end_time)
      raise ArgumentError.new("'timeseries' command requires that both 'before' and 'after' commands were used")
    end
    summaries = _query.sort(time_created: :asc).all
    time_intervals = build_time_intervals(start_time, end_time, time_unit)
    {}.tap do |result|
      time_intervals.each { |key| result[key] = [] }
      summaries.each { |summary| key = range_for_activity(summary, time_intervals, time_unit); result[key] << summary }
    end
  end

  private
  def resolve_constant(name)
    activity_type = if name == 'reviews' then
                      'review_request'
                    else
                      name.to_s.singularize
                    end
    (activity_type + "_summary").camelize.constantize
  rescue NameError => badname
    nil
  end

  # figures out what time interval the activity summary belongs to and returns the start point for that interval
  def range_for_activity(activity_summary, time_intervals, time_unit)
    interval_index = ((activity_summary.time_created - time_intervals[0])/time_unit).floor
    time_intervals[interval_index]
  end

  # builds an array of endpoint.
  # @param [Time] start_point, start of the time interval.
  # @param [Time] end_point, end of the time interval.
  def build_time_intervals(start_time, end_time, time_unit)
    interval_endpoints = []
    while (start_time < end_time)
      interval_endpoints << start_time
      start_time = start_time + time_unit
    end
    interval_endpoints
  end


  # Considers existing query containing source_ids and creates AND condition to find the intersection
  # Otherwise, chained queries filtering on source_ids will return the superset
  def filter_by_source(additional_source_ids_ary)
    # we will filter the result set based on the source_ids returned from in_project
    pre_filtered_src_ids = _query[:source_id]
    if pre_filtered_src_ids.nil?
      _query.where(:source_id => additional_source_ids_ary)
    else
      _query.where(:$and => [{ :source_id => pre_filtered_src_ids }, { :source_id => additional_source_ids_ary }])
    end
  end


end
