#
# Copyright 2013 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

require 'application_constants'

# Parses review items and builds database items from them
class ReviewQueueItemParser
  
  QUEUE_NAME = REVIEW_QUEUE_NAME 
  DATA_SEGMENT = [:request_data]

  # Creates a new instance of the class and parses the review JSON
  # @param [String] message the JSON message from Rabbit
  def self.process_message(message)
    message_hash = JSON.parse(message).with_indifferent_access
    message_type = DATA_SEGMENT.find { |key| message_hash.include? key }

    unless message_type
      LogUtils.log_error_and_info(I18n.t('queue_parser.no_recognized_data'), message_hash)
      return
    end

    raise MissingFieldError.new(:api_version) if message_hash[:api_version].nil?
    raise MessageProcessingError.new(I18n.t('api.invalid.version')) if message_hash[:api_version] != '1'

    import_message(message_type, message_hash)

  rescue MissingFieldError => mfe
    LogUtils.log_error_and_info(I18n.t('api.required_field', field_name: mfe.message.to_s), message_hash)
  rescue NoSuchSourceAssociationKey => nssak
    LogUtils.log_error_and_info(I18n.t('api.invalid.source_association_key', source_association_key: nssak.message), message_hash)
  rescue JSON::ParserError
    LogUtils.log_error_and_info(I18n.t('api.json_required'), message)
  rescue InconsistentStateError => e
    LogUtils.log_error_and_info("#{I18n.t('api.errors_found')} #{e.message}", message_hash)
  rescue MessageProcessingError => e
    LogUtils.log_error_and_info("#{I18n.t('api.errors_found')} #{e.message}", message_hash)
  rescue SourceNotActive => e
    LogUtils.log_error_and_info(I18n.t('api.source_deactivated', attribute: e.message), message_hash)
  rescue ProjectNotActive => e
    LogUtils.log_error_and_info(e.message, message_hash)
  end

  def self.import_message(message_type, message_hash)
    sak = message_hash[:source_association_key]
    raise MissingFieldError.new(:source_association_key) if sak.nil?

    source = ActivitySource.find_by_source_association_key(sak)
    raise NoSuchSourceAssociationKey.new(sak) unless source

    # TODO: move this check into handle_message (refactor for all activity sources)
    raise SourceNotActive.new(sak) unless source.active
    source.handle_message(message_type, message_hash[message_type])
  end

  class << self
    include ::NewRelic::Agent::Instrumentation::ControllerInstrumentation
    add_transaction_tracer :process_message, category: :task, name: self.to_s
  end

end
