# -*- coding: utf-8 -*-
require 'yaml'

# Database-related rake tasks
namespace :db do
  desc "Set the database up for tests"
  task :migrate_and_prepare => [:migrate, 'db:index', 'db:test:prepare']
end

begin
  require 'rspec/core/rake_task'

  namespace :spec do
    desc "Runs normal CI tests"
    task :ci => [:prepare, :noint, :jasmine]

    desc "Runs all tests except PSR"
    task :all => [:prepare, :nopsr, :jasmine]

    desc "Setup"
    task :prepare => ['db:test:prepare', 'ci:setup:rspec']

    desc "Runs all RSpec tests except PSR or smoke"
    RSpec::Core::RakeTask.new(:nopsr) do |t|
      t.rspec_opts = ['--tag ~psr']
      t.rspec_opts << '--tag ~smoke_test'
    end

    namespace :jasmine do
      desc "Runs spec:jasmine task locally"
      task :local => [:run_local, :jasmine_fixtures, 'jasmine:ci']
      task :run_local do
        ENV['RUN_LOCAL'] = 'true'
      end
    end

    desc "Sets up the HTML fixtures for the JavaScript tests and runs the test in CI mode"
    task :jasmine => [:jasmine_fixtures, 'jasmine:ci']

    desc "Sets up the HTML fixtures for the JavaScript tests and starts the Jasmine server"
    task :jasmine_server do
      Rake::Task['spec:jasmine_fixtures'].execute
      Rake::Task['jasmine'].invoke
    end

    desc "Sets up the HTML fixtures"
    RSpec::Core::RakeTask.new(:jasmine_fixtures) do |t|
      FileUtils.rm_rf("spec/javascripts/fixtures/.", secure: true)
      t.rspec_opts = '--tag jasmine'
    end

    desc "Runs the specs that are not integration tests"
    RSpec::Core::RakeTask.new(:noint) do |t|
      t.rspec_opts = ['--tag ~integration']
      t.rspec_opts << '--tag ~psr'
    end

    desc "Runs the specs that are integration tests"
    RSpec::Core::RakeTask.new(:int) do |t|
      t.rspec_opts = ['--tag integration']
      t.rspec_opts << '--tag ~psr'
      t.rspec_opts << '--tag ~smoke_test'
    end

    desc "Runs PSR tests"
    RSpec::Core::RakeTask.new(:psr) do |t|
      t.rspec_opts = '--tag psr'
    end

    desc "Runs the smoke tests in the smoke_test environment"
    task :smoke_test do
      smoke_test_file = Rails.root.join('spec/integration/smoke_test_spec.rb')
      comment_in_test = "sed -i 's/pending/#pending/g' #{smoke_test_file}"
      comment_out_test = "sed -i 's/#pending/pending/g' #{smoke_test_file}"
      run_test = "rspec #{smoke_test_file}"
      puts run_test

      system comment_in_test
      system run_test
      system comment_out_test
    end

    desc "Runs all unit and integration tests in parallel and then the JavaScript tests"
    task :parallel_all => [:prepare, :parallel, 'jasmine:local']

    desc "Runs rspec in parallel"
    task :parallel do
      problem_tests = %w(build_api_spec review_api_spec smoke_test_spec)
      single_process = problem_tests.map {|test| "-s #{test}"}.join(" ")
      command = "parallel_test spec #{single_process} --type rspec"
      puts command
      system(command)
    end
  end
rescue LoadError
  # rspec rake tasks are only used in development mode
end

begin
  require 'gem-licenses'

  task :licenses do
    Gem.licenses.each do |license, gems|
      puts "#{license}"
      gems.sort_by {|gem| gem.name}.each do |gem|
        puts "* #{gem.name} #{gem.version} (#{gem.homepage}) - #{gem.summary}"
      end
    end
  end
rescue LoadError
  # gem-licenses is only used in development mode
end

# helper to locate user_auth_store.yml
def localauth_permissions_file
  Rails.root.join("gems", "omniauth-local", "lib", "omniauth", "strategies", "user_auth_store.yml")
end

namespace :db do

  desc "create demo projects and user permissions for demoing user authorization"
  task :permissions => :environment do
    Project.find_or_create_by_name(APP_PRODUCT_NAME)
    Project.find_or_create_by_name('Skunkworks')
    Project.find_or_create_by_name('Admin')

    User.apply_permissions(localauth_permissions_file)
  end

end
