#
# Copyright 2013 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#
# MongoDB database tasks
namespace :db do
  desc "Migrate the database through scripts in db/migrations. Target specific version with VERSION=x."
  task :migrate => :environment do
    migration_path = ENV["MIGRATION_PATH"]  || "db/migrations/"
    Migrator.migrate(migration_path, ENV["VERSION"] ? ENV["VERSION"].to_i : nil)
  end

  namespace :migrate do
    desc  'Rollbacks the database one migration and re migrate up. If you want to rollback more than one step, define STEP=x. Target specific version with VERSION=x.'
    task :redo => :environment do
      if ENV["VERSION"]
        Rake::Task["db:migrate:down"].invoke
        Rake::Task["db:migrate:up"].invoke
      else
        Rake::Task["db:rollback"].invoke
        Rake::Task["db:migrate"].invoke
      end
    end

    desc 'Runs the "up" for a given migration VERSION.'
    task :up => :environment do
      version = ENV["VERSION"] ? ENV["VERSION"].to_i : nil
      raise "VERSION is required" unless version
      migration_path = ENV["MIGRATION_PATH"]  || "db/migrations/"
      Migrator.run(:up, migration_path, version)
    end

    desc 'Runs the "down" for a given migration VERSION.'
    task :down => :environment do
      version = ENV["VERSION"] ? ENV["VERSION"].to_i : nil
      raise "VERSION is required" unless version
      migration_path = ENV["MIGRATION_PATH"]  || "db/migrations/"
      Migrator.run(:down, migration_path, version)
    end
  end

  desc 'Rolls the migrations back to the previous version. Specify the number of steps with STEP=n'
  task :rollback => :environment do
    step = ENV['STEP'] ? ENV['STEP'].to_i : 1
    migration_path = ENV["MIGRATION_PATH"]  || "db/migrations/"
    Migrator.rollback(migration_path, step)
  end

  desc "Load seed data provided by Chef"
  task :chef_seed  => :environment do
    env_seed_file = File.join(Rails.root, 'db', 'chef_seeds.rb')
    load(env_seed_file) if File.exist?(env_seed_file)
  end

  desc "Create database indexes"
  task :index => :environment do
    host, port = MongoMapper.connection.primary
    puts "Indexing database #{MongoMapper.database.name} on #{host}:#{port}"

    Activity.drop_indexes if Activity.exists?
    Activity.ensure_index [[:source_id, 1]]
    Activity.ensure_index [[:remote_id, 1]] #for work items
    # for pagination
    Activity.ensure_index [[:time_created, 1], [:source_id, 1]]
    Activity.ensure_index [[:time_created, 1], [:remote_id, 1]] #for work items
    Activity.ensure_index [[:source_id, 1], [:atomic, 1]]  # This might be backwards; it might be more efficient to reverse the combined index.
    Activity.ensure_index [[:atomic, -1]]
    # For item validators
    Activity.ensure_index [[:_type, 1], [:source_id, 1], [:revision_number, 1]] # SCM
    Activity.ensure_index [[:_type, 1], [:source_id, 1], [:remote_id, 1], [:time_created, 1]] # build and review

    ActivitySource.drop_indexes if ActivitySource.exists?
    ActivitySource.ensure_index :project_id
    ActivitySource.ensure_index :active

    ActivitySummary.drop_indexes if ActivitySummary.exists?
    ActivitySummary.ensure_index :opaque_id
    ActivitySummary.ensure_index :source_id
    ActivitySummary.ensure_index [[:_type, 1], [:source_id, 1], ['ctf_item.ctf_associations.item_id', 1], [:time_created, 1]] #ctf_associations
    ActivitySummary.ensure_index [[:_type, 1], [:source_id, 1], [:remote_id, 1]] #when matching RRSummary from the commit message
    ActivitySummary.ensure_index [[:_type, 1], [:source_id, 1], [:revision_number, 1]] # Matching commits by revision & source id.

    Project.drop_indexes if Project.exists?
    Project.ensure_index :active
    Project.ensure_index :name

    TeamForge::ProjectConfiguration.drop_indexes if TeamForge::ProjectConfiguration.exists?
    TeamForge::ProjectConfiguration.ensure_index :active
    TeamForge::ProjectConfiguration.ensure_index :ctf_project_id
    TeamForge::ProjectConfiguration.ensure_index :project_id
    TeamForge::ProjectConfiguration.ensure_index :project_name

    Tracker.drop_indexes if Tracker.exists?
    Tracker.ensure_index :work_item_source_id
    Tracker.ensure_index :remote_id

    User.drop_indexes if User.exists?
    User.ensure_index :email
    User.ensure_index :reset_password_token
    User.ensure_index :confirmation_token
    User.ensure_index :unlock_token
    User.ensure_index [[:provider, 1], [:uid, 1]]

    puts "Indexing done"
  end

  desc "Dumps the database"
  task :dump => :environment do
    MongoMapper.connect(Rails.env)
    host, port = MongoMapper.connection.primary
    puts "Dumping database #{MongoMapper.database.name} from #{host}:#{port} to tmp/dump"
    `rm -rf tmp/dump`
    Dir.mkdir 'tmp/dump'
    command_string = %{mongodump -h #{host} --port #{port} -o "tmp/dump" -d "#{MongoMapper.database.name}"}
    auth = MongoMapper.connection.auths[0].with_indifferent_access
    command_string += " -u #{auth["username"]} -p #{auth["password"]}" if auth
    `#{command_string}`
  end

  desc "Restores the database"
  task :restore => :environment do
    db_dir_name = Dir.entries('tmp/dump').find {|name| name !~ /\./}
    MongoMapper.connect(Rails.env)
    host, port = MongoMapper.connection.primary
    puts "Restoring database to #{MongoMapper.database.name} at #{host}:#{port} from tmp/dump/#{db_dir_name}"

    MongoMapper.database.collections.each do |collection|
      collection.drop unless collection.name.start_with? 'system'
    end

    command_string = %{mongorestore --drop --noIndexRestore -h #{host} --port #{port} -d "#{MongoMapper.database.name}"}
    auth = MongoMapper.connection.auths[0].with_indifferent_access if MongoMapper.connection.auths[0]
    command_string += " -u #{auth["username"]} -p #{auth["password"]}" if auth
    command_string += %{ "tmp/dump/#{db_dir_name}"}
    `#{command_string}`
  end

  desc "Copies the staging database to your development database"
  task :copy_staging do
    Rails.env = 'staging'
    Rake::Task['db:dump'].invoke
    Rails.env = 'development'
    Rake::Task['db:restore'].invoke
    Rake::Task['db:index'].invoke
  end
end
