#!/opt/collabnet/ruby/bin/ruby

ENV['RAILS_ENV'] = "production"

require 'getoptlong'
require 'readline'
require 'etc'

RAILS_ENVIRONMENT = '/opt/collabnet/orchestrate/config/environment.rb'

class CommitAssociationSurgeon

  # Instantiate and grab all command line arguments
  def initialize
    @debug = @list = false
    @opaque_id = @message = nil
    @add = []
    @remove = []
    file = nil

    begin
      getopts = GetoptLong.new(
          ['-d', '--debug', GetoptLong::NO_ARGUMENT],
          ['-h', '--help', GetoptLong::NO_ARGUMENT],
          ['-l', '--list', GetoptLong::NO_ARGUMENT],
          ['-i', '--opaqueid', GetoptLong::REQUIRED_ARGUMENT],
          ['-f', '--file', GetoptLong::REQUIRED_ARGUMENT],
          ['-m', '--message', GetoptLong::REQUIRED_ARGUMENT],
          ['-a', '--add', GetoptLong::REQUIRED_ARGUMENT],
          ['-r', '--remove', GetoptLong::REQUIRED_ARGUMENT],
      )

      getopts.each do |opt, arg|
        case opt
          when '-d'
            @debug = true
          when '-h'
            usage
          when '-l'
            @list = true
          when '-i'
            @opaque_id = arg
          when '-f'
            file = arg
          when '-m'
            @message = arg
          when '-a'
            @add << arg
          when '-r'
            @remove << arg
        end
      end
    rescue GetoptLong::MissingArgument => e
      usage e.message
    rescue GetoptLong::InvalidOption => e
      usage e.message
    end

    # Eh, try to take the message from the remaining args if missing
    @message ||= ARGV[0]

    usage "No --opaqueid specified" unless @opaque_id
    #usage "Either --file or --message must be specified" unless file or @message

    warn "both --file and --message specified; ignoring --file" if file and @message

    if !@message and file
      usage "No such file #{file}" unless File.file? file
      begin
        @message = File.read(file)
      rescue Exception => e
        usage "Unable to read from file: #{e.message}"
      end
    end

    load_rails
  end

  # Load up rails environment, handle user or permissions exceptions
  def load_rails
    begin
      require RAILS_ENVIRONMENT
    rescue Exception => e
      debug e.inspect if @debug
      unless ['root', 'orchestrate'].include? Etc.getlogin
        error "must be root or orchestrate to run this script"
      else
        error "#{e.message}"
      end
    end
  end

  # Find the best description for a given summary
  def best_description(summary, dump=true)
    desc = summary.title
    desc ||= summary.description if summary.respond_to?(:description)
    desc ||= summary.status_name if summary.respond_to?(:status_name)
    # If dumped, strip the leading and trailing quotation marks
    desc = desc.dump.chomp('"')[1..-1] if dump
    desc
  end

  # List associations of selected activity summary.  Optionally number selections for choosing.
  def list_associations(summary, number_selections=false)
    summary.reload
    choices = {}
    if summary.activity_summaries.empty?
       puts "\n<No associations for #{summary._type} '#{summary.opaque_id}'>\n\n"
    else
      format = "%-4.4s %-4.8s | %-16.16s | %-30.30s | %-40.40s\n"
      puts "\nCurrent associations for #{summary._type} '#{summary.opaque_id}':"
      puts "(\"#{ellipsize(best_description(summary), 101)}\")\n\n"
      printf(format, '', 'id:', 'type:', 'remote_id:', 'description:')
      puts '=' * 105
      n = 0
      summary.activity_summaries.each do |fs|
        n += 1
        sel = number_selections ? "[#{n}]" : ''
        # Find a textual description:
        desc = best_description(fs)
        remote_id = fs.respond_to?(:remote_id) ? fs.remote_id : fs.revision_number
        printf(format, sel, fs.opaque_id, fs._type, ellipsize(remote_id, 30), ellipsize(desc, 40))
        choices[n.to_s] = fs
      end
      puts
    end
    choices
  end

  # Select an item (association) from a hash of selector: item
  def select_associations(choices)
    selections = []
    return selections if choices.empty?
    puts "\nSelect association to remove (<return> to finish selection, q to quit):"
    while line = Readline.readline('> ').strip
      break if line =~ /^\s*$/
      exit if line =~ /^\s*q\s*$/i
      if choices.keys.include?(line)
        selections << choices[line] unless selections.include? choices[line]
      else
        puts 'Invalid selection'
      end
    end
    selections
  end

  # Remove the association between two activity summaries
  def remove_association(s1, s2)
    info "Removing association between '#{s1.opaque_id}' and '#{s2.opaque_id}'"
    s1.associated_activity_summary_ids.delete?(s2.id) and update "'#{s2.opaque_id}' removed from '#{s1.opaque_id}'"
    s2.associated_activity_summary_ids.delete?(s1.id) and update "'#{s1.opaque_id}' removed from '#{s2.opaque_id}'"
    s1.save! and s2.save! and update "Removals saved"
  end

  def add_association(s1, s2)
    info "Adding association between '#{s1.opaque_id}' and '#{s2.opaque_id}'"
    s1.associated_activity_summary_ids.add?(s2.id) and update "'#{s2.opaque_id}' added to '#{s1.opaque_id}'"
    s2.associated_activity_summary_ids.add?(s1.id) and update "'#{s1.opaque_id}' added to '#{s2.opaque_id}'"
    s1.save! and s2.save! and update "Additions saved"
  end

  # Update the commit message for a CommitSummary
  def update_message(summary, message)
    error "CommitSummary '#{id}' does not contain one and only one activity" if summary.activities.count != 1
    commit = summary.activities.last
    error "no Commit found for '#{id}'." unless commit

    original_message = commit.message
    if original_message != message
      commit.message = message
      debug "replacing original message:\n#{original_message}"
      debug "new message:\n#{message}"
      commit.save! and CommitSummary.create_or_update(commit) and update "Updated commit message for '#{summary.opaque_id}'"
    else
      info "commit message unchanged"
    end
  end

  # Truncate and ellipsize a string to a given maximum length
  def ellipsize(string, max, ellipsis='...')
    string.size > max ? string[0..(max-(ellipsis.size+1))] + ellipsis : string
  end

  # Output method:
  def debug(message)
    puts "DEBUG: #{message}" if @debug
  end

  # Output method:
  def update(message)
    puts "UPDATE: #{message}"
  end

  # Output method:
  def info(message)
    puts "INFO: #{message}"
  end

  # Output method:
  def error(message)
    puts "ERROR: #{message}"
    abort
  end

  # Output method:
  def warn(message)
    puts "WARNING: #{message}"
  end

  def main
    # Fetch the summary and its commit, make sure they're legal:
    summary = CommitSummary.where(opaque_id: @opaque_id).last
    error "no CommitSummary '#{@opaque_id}' found." unless summary

    list_associations(summary) and exit if @list

    additions = ActivitySummary.where(opaque_id: @add).all
    removals = ActivitySummary.where(opaque_id: @remove).all + select_associations(list_associations(summary, true))

    removals.each { |selection| remove_association(summary, selection) }
    additions.each { |selection| add_association(summary, selection) }

    update_message(summary, @message) if @message

    # Really, only removals matter for this:
    (additions + removals).each { |selection|
      selection.reload
      warn "#{selection._type} '#{selection.opaque_id}' has no more associations" if selection.associated_activity_summary_ids.empty?
    }

    list_associations(summary)
  end

  def usage(message=nil)
    scriptname = File.basename(__FILE__)
    puts "\nERROR: #{message}\n\n" if message
    puts "Overview:"
    puts "  List and interactively remove associations from a given CommitSummary,"
    puts "  and re-form a commit's associations based on a new commit message or"
    puts "  other supplied arguments.  This script directly modifies data in the"
    puts "  orchestrate database, but does NOT modify records in the sources or in"
    puts "  other data stores."
    puts
    puts "  *** USE THIS TOOL WITH EXTREME CAUTION ***"
    puts
    puts "Usage:"
    puts "  #{scriptname} -i <opaque_id> [-f <file> | -m <message...>]"
    puts "  #{scriptname} -i <opaque_id> [-a <opaque_id>] [-r <opaque_id>]"
    puts "  #{scriptname} -i <opaque_id> -l"
    puts
    puts "Options:"
    puts " -i, --opaqueid    The <opaque_id> of the commit to act upon"
    puts "                   (NOTE: opaque_ids are case-sensitive)"
    puts " -a, --add         Additional <opaque_id> to create association with"
    puts " -r, --remove      Additional <opaque_id> to remove association with"
    puts " -f, --file        File containing new commit message"
    puts " -m, --message     Commit message to use (precludes -f)"
    puts " -l, --list        List current associations of <opaque_id> and exit"
    puts " -d, --debug       Display debugging info"
    puts " -h, --help        Show this message and exit"
    puts
    exit
  end
end

CommitAssociationSurgeon.new.main()
