#!/opt/collabnet/ruby/bin/ruby

ENV['RAILS_ENV'] = "production"

require 'getoptlong'
require 'readline'
require 'etc'

RAILS_ENVIRONMENT = '/opt/collabnet/eventq/config/environment.rb'

class FixProjectIaf

  # for a given IAF (active)
  # Find all projects with names like the IAF project name
  # and choose by earliest/latest, and change the association
  # when prompted for choice of projects
  # Order them chronologically
  # Show how many activities are in each project


  # Instantiate and grab all command line arguments
  def initialize
    @debug = @list = @summary = false
    @activate = @apply = false

    begin
      getopts = GetoptLong.new(
          ['-d', '--debug', GetoptLong::NO_ARGUMENT],
          ['-h', '--help', GetoptLong::NO_ARGUMENT],
          ['-s', '--summary', GetoptLong::NO_ARGUMENT],
          ['-l', '--list', GetoptLong::REQUIRED_ARGUMENT],
          ['-a', '--activate', GetoptLong::REQUIRED_ARGUMENT],
          ['-A', '--apply', GetoptLong::REQUIRED_ARGUMENT],
      )

      getopts.each do |opt, arg|
        case opt
          when '-d'
            @debug = true
          when '-h'
            usage
          when '-l'
            @list = arg
          when '-s'
            @summary = true
          when '-a'
            @activate = arg
          when '-A'
            case arg
              when 'oldest_creation'
                @apply = :oldest_creation
              when 'most_sources'
                @apply = :most_sources
              else
                raise GetoptLong::InvalidOption.new("invalid -A parameter: #{arg}")
            end
        end
      end
    rescue GetoptLong::MissingArgument => e
      usage e.message
    rescue GetoptLong::InvalidOption => e
      usage e.message
    end

    load_rails
  end

  # Load up rails environment, handle user or permissions exceptions
  def load_rails
    begin
      require RAILS_ENVIRONMENT
    rescue Exception => e
      debug e.inspect if @debug
      unless ['root', 'eventq'].include? Etc.getlogin
        error "must be root or eventq to run this script"
      else
        error "#{e.message}"
      end
    end
  end

  # Truncate and ellipsize a string to a given maximum length
  def ellipsize(string, max, ellipsis='...')
    string.size > max ? string[0..(max-(ellipsis.size+1))] + ellipsis : string
  end

  # Output method:
  def debug(message)
    puts "DEBUG: #{message}" if @debug
  end

  # Output method:
  def update(message)
    puts "UPDATE: #{message}"
  end

  # Output method:
  def info(message)
    puts "INFO: #{message}"
  end

  # Output method:
  def error(message)
    puts "ERROR: #{message}"
    abort
  end

  # Output method:
  def warn(message)
    puts "WARNING: #{message}"
  end

  def main
    @configs_and_projects = TeamForge::ProjectConfiguration.all.inject({}) { |memo, pc| memo.merge({ pc => Project.where(active: false, name: pc.project_name).all}) }

    child_projects = @configs_and_projects.values
    orphan_projects = nil # Project.where(active:false, id: {"$nin" => child_projects}).all

    if @list
      config = find_config_by_project_name(@list)
      display_config_summary(config)
      warn "No project found" if config.empty?
    end
    if @summary
      @configs_and_projects.each do |config_hsh|
        display_config_summary(config_hsh)
      end
      if orphan_projects.present?
        info "Orphan projects:"
        display_projects_summary(orphan_projects)
      end
    end
    if @activate.present?
      config_name, project_index = @activate.to_s.split(':')
      error("Incorrect value given") unless config_name && project_index
      config = find_config_by_project_name(config_name)
      error("Project not found: #{config_name}") unless config
      cfg = config[0]

      if project_index.to_i
        prjs = config[1]
        expected_project = prjs[project_index.to_i] #to select
      else # we assume it is an orphaned project to look up by name (in the future)
        error "No index provided"
      end

      activate_project_for_config(cfg, expected_project) if expected_project

      info "Activation complete - Please verify."
      display_config_summary(config)
    end
    if @apply
      info "Applying #{@apply} to all projects"
      @configs_and_projects.each do |config, projects|
        previous_id = config.project_id
        project = projects.reduce(config.project) do |memo, project|
          if @apply == :most_sources && project.sources.count > memo.sources.count
            project
          elsif @apply == :oldest_creation && project.created_at < memo.created_at
            project
          else
            memo
          end
        end
        if project
          status = previous_id == project.id ? "Unchanged" : "Changed"
          info "#{config.project_name}: #{status}"
          activate_project_for_config(config, project)
        else
          warn "Did not find any suitable project to activate for #{config.project_name}"
        end
      end
      info "Complete! Please verify"
    end
  end

  def activate_project_for_config(config, project)
    config.project_id = project.id
    config.save
    project.active = true
    project.save
  end

  def find_config_by_project_name(query_name)
    @configs_and_projects.find { |k,v| query_name == k.project_name }
  end

  def display_config_summary(config_hsh)
    config_hsh.tap do |key, value|
      info "TeamForge Project: #{key.project_name}"
      display_projects_summary(value)
    end
  end

  def display_projects_summary(projects)
    projects.each_with_index do |project, index|
      info "#{index}: #{project.name} #{project.active? ? "Active" : "      "} Sources:#{project.sources.count} Activities:#{ActivitySummary.where("source_id" => project.sources.map(&:id)).count} Created:#{project.created_at}"
    end
  end

  def usage(message=nil)
    scriptname = File.basename(__FILE__)
    puts "\nERROR: #{message}\n\n" if message
    puts <<-eom
Overview:
  This script allows you to change the mapping of an EventQ project
  to a TeamForge project.

  *** BACK UP DATA BEFORE USING THIS TOOL ***
  *** USE THIS TOOL WITH EXTREME CAUTION ***

Usage:
  #{scriptname} -c

Options:
 -l, --list <project_name> Display the current status of eq project mappings for a project
 -s, --summary     Display the status of all projects the database
 -a, --activate <project_name:index_number> Activate a different eq project in the list for a given project
 -A, --apply <decision> Apply the decision to all projects. oldest_creation, most_sources
 -d, --debug       Display debugging info
 -h, --help        Show this message and exit

    eom
    exit
  end
end

FixProjectIaf.new.main()
