#
# Copyright 2013 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#


require 'omniauth'
require 'savon'

module OmniAuth
  module Strategies
    class Teamforge
      include OmniAuth::Strategy

      # Configured when this omniauth provider is included.
      # :server_root is the root of the CTF installation
      # :linked_app_id is the ID of the site-wide linked application on the CTF installation
      args [:server_root, :linked_app_id]

      option :server_root, nil
      option :linked_app_id, nil

      # Called by OmniAuth when a request to authenticate with this provider is made.
      def request_phase
        url = URI(options.server_root)
        url.path = '/sf/linkedapplication/do/viewSiteLinkedApplication'
        url.query = "id=#{options.linked_app_id}"
        redirect url.to_s
      end

      # Called by OmniAuth with the credentials to validate. Fails if the user is not valid.
      def callback_phase
        return fail!(:invalid_credentials) unless valid_user
        super
      end

      # Validates the user with TeamForge. Handles getting a request when requested via a site-wide linked app
      # and using SSO. Also handles username/password authentication if calling this provider via a login form.
      # If the authentication is successful, loads the user's data from TeamForge.
      def valid_user
        if request.params['sfUsername'] && request.params['sfLoginToken']
          web_and_session_token = authenticate_with_token
          @username = request.params['sfUsername']
          @token = web_and_session_token.split("&")[1] if web_and_session_token
        else
          @username = request.params['username']
          @token = authenticate_with_credentials
        end

        @user = get_user_data(@token, @username) if @token
      end

      # Called by OmniAuth to set the 'uid' in the authentication hash.
      uid do
        @user[:username]
      end

      # Called by OmniAuth to set the user's 'info' in the authentication hash.
      info do
        { name: @user[:full_name], email: @user[:email] }
      end

      # Called by OmniAuth to set the credentials we use to access TeamForge going forward.
      credentials do
        { token: @token }
      end

      # Authenticates to TeamForge with the given one-time token.
      # @return [String] a web session token and a SOAP session token (one string separated by an &) if the login
      #   token is valid. Returns nil if authentication is not valid.
      def authenticate_with_token
        one_time_token = request.params['sfLoginToken']
        return nil unless one_time_token

        ctf.request(:get_web_and_soap_session_id) do
          soap.body = { "oneTimeToken" => one_time_token }
        end.body[:get_web_and_soap_session_id_response][:get_web_and_soap_session_id_return]
      rescue => e
        log(:warn, "Failed login with token: #{e}")
        nil
      end

      # Authenticates the user with TeamForge using username and password credentials.
      #
      # @return [String] the session token or nil if the user did not successfully log in.
      def authenticate_with_credentials
        username = request.params['username']
        password = request.params['password']
        ctf.request(:login) do
          soap.body = { "userName" => username, "password" => password}
        end.body[:login_response][:login_return]
      rescue => e
        log(:warn, "Failed login: #{e}")
        nil
      end

      # Loads the user's information from TeamForge.
      #
      # @return [Hash] the user data from TeamForge
      def get_user_data(token, username)
        ctf.request(:get_user_data) do
          soap.body = { "sessionId" => token, "username" => username }
        end.body[:multi_ref]
      end

      # Creates a connection to the CTF server.
      #
      # @return [Savon::Client] a new soap client
      def ctf
        endpoint_url = "#{options.server_root}/ce-soap60/services/CollabNet"
        @ctf ||= Savon::Client.new("#{endpoint_url}?wsdl").tap do |ep|
          ep.wsdl.endpoint = endpoint_url
          # TODO: we will need to handle SSL certs properly
          ep.http.auth.ssl.verify_mode = :none
        end
      end
    end
  end
end
