require 'omniauth/strategies/teamforge'
require 'rack/test'
require File.dirname(__FILE__) + '/fixtures'

describe OmniAuth::Strategies::Teamforge do

  include OmniAuth::Test::StrategyTestCase
  include Rack::Test::Methods
  include OmniAuth::Strategies::Teamforge::Fixtures

  let(:server_root) { 'https://orc-test-ctf7.sp.corp.collab.net' }
  let(:linked_app_id) { 'lapp1001' }

  def strategy
    [OmniAuth::Strategies::Teamforge, server_root, linked_app_id]
  end

  describe '#request_phase' do
    it 'redirects to teamforge for authentication' do
      get '/auth/teamforge'
      last_response.header['Location'].should == "#{server_root}/sf/linkedapplication/do/viewSiteLinkedApplication?id=#{linked_app_id}"
    end
  end

  describe '#callback_phase' do
    let(:auth_hash) { last_request.env['omniauth.auth' ]}
    let(:username) { 'admin' }
    let(:email) { 'root@example.com' }
    let(:full_name) { 'TeamForge Administrator' }
    let(:session_key) { 'some_session_id' }

    before(:each) { intercept_savon }
    after(:each) { reset_savon }

    describe 'with sfUsername and sfLoginToken authentication' do
      let(:token) { "web_token&#{session_key}" }
      let(:login_token) { 'some_one_time_token' }

      before(:each) do
        post '/auth/teamforge/callback', sfUsername: username, sfLoginToken: login_token
      end

      describe 'with valid credentials' do
        let(:request_params) do
          {
            getWebAndSoapSessionId: { "oneTimeToken" => login_token },
            getUserData: { 'sessionId' => session_key, 'username' => username }
          }
        end
        let(:response_bodies) do
          {
            getWebAndSoapSessionId: fixture(:success_web_and_soap_token, token: token),
            getUserData: fixture(:user_data, username: username, email: email, full_name: full_name)
          }
        end

        it { last_response.status.should == 404 }
        it { auth_hash['uid'].should == username }
        it { auth_hash['info']['email'].should == email }
        it { auth_hash['info']['name'].should == full_name }
      end

      describe 'with invalid credentials' do
        let(:request_params) { { getWebAndSoapSessionId: {"oneTimeToken" => login_token} } }
        let(:response_bodies) { { getWebAndSoapSessionId: fixture(:failed_auth) }}

        it 'redirects to a failure error page' do
          last_response.should be_redirect
          last_response.location.should == '/auth/failure?message=invalid_credentials&strategy=teamforge'
        end
      end
    end

    describe 'with username/password authentication' do
      let(:request_params) do
        {
          login: { 'userName' => username, 'password' => password },
          getUserData: { 'sessionId' => session_key, 'username' => username }
        }
      end

      before(:each) do
        post '/auth/teamforge/callback', username: username, password: password
      end

      describe 'with valid credentials' do
        let(:password) { 'admin' }
        let(:response_bodies) do
          {
              login: fixture(:success_auth, session_id: session_key),
              getUserData: fixture(:user_data, username: username, email: email, full_name: full_name)
          }
        end

        it { last_response.status.should == 404 }
        it { auth_hash['uid'].should == username }
        it { auth_hash['info']['email'].should == email }
        it { auth_hash['info']['name'].should == full_name }
      end

      describe 'with invalid credentials' do
        let(:password) { 'bogus' }
        let(:session_key) { nil }
        let(:response_bodies) { { login: fixture(:failed_auth) } }

        it 'redirects to a failure error page' do
          last_response.should be_redirect
          last_response.location.should == '/auth/failure?message=invalid_credentials&strategy=teamforge'
        end
      end

    end
  end

  # Configures savon hooks so we don't make actual calls to the server.
  def intercept_savon
    Savon.config.hooks.define(:spec_action, :soap_request) do |_, request|
      method_name = request.soap.input[1]
      HTTPI::Response.new(200, {}, response_bodies[method_name])
    end

    Savon.config.hooks.define(:spec_body, :soap_request) do |_, request|
      method_name = request.soap.input[1]
      actual = request.soap.body
      expected = request_params[method_name]
      raise Exception, "expected #{expected.inspect} to be sent, got: #{actual.inspect}" unless actual == expected
      HTTPI::Response.new(200, {}, response_bodies[method_name])
    end
  end

  # Resets savon between tests.
  def reset_savon
    Savon.config.hooks.reject(:spec_body)
    Savon.config.hooks.reject(:spec_action)
  end

end