# Copyright 2013 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.

require 'savon'

# Client for a CTF service
class TeamForge::Service

  attr_reader :session_id

  # Creates an client for a CTF service
  # @param [String] root_url the URL of the CTF server
  # @param [String] service_name the name of the service
  # @param [String] session_id the session ID of the service user
  # @param [String] directory where the WSDL is being stored
  def initialize(root_url, service_name, session_id, wsdl_directory = nil)
    @root_url = root_url
    @service_name = service_name
    @session_id = session_id
    @wsdl_directory = wsdl_directory
  end

  # Creates an endpoint, and caches it, for this Service.
  # @return [Savon::Client] the API endpoint
  def endpoint
    return @endpoint if @endpoint
    endpoint_url = if @service_name == 'OrchestrateSyncListener'
                     URI.join(@root_url, '/ce-soap/services/OrchestrateSyncListener').to_s
                   else
                     URI.join(@root_url, "/ce-soap60/services/#{@service_name}").to_s
                   end
    wsdl_location = !@wsdl_directory.nil? ? File.join(@wsdl_directory, "#{@service_name}.xml") : "#{endpoint_url}?wsdl"
    @endpoint = Savon::Client.new(wsdl_location).tap do |ep|
      ep.wsdl.endpoint = endpoint_url
      # TODO: we will need to handle SSL certs properly
      ep.http.auth.ssl.verify_mode = :none
    end
  end

  # Sends a request to the CTF service.
  # Use this method if you don't want the session ID automatically passed to the service method.
  #
  # @param [String, Symbol] method the method name
  # @param [Hash] args the named arguments
  # @return [Hash] the response body
  # @raise [TeamForge::ServiceError] if there's an error sending the request
  def request(method, args = {})
    body = endpoint.request(method) do
      soap.body = args
    end.body

    response = TeamForge::Service.strip_empty_nodes!(body)
    extract_simple_response(method, response)

  rescue Savon::SOAP::Fault, SocketError => e
    converted = TeamForge::ServiceError.new e.message
    converted.set_backtrace e.backtrace
    raise converted
  end

  # Extracts simple responses if one is available.
  # @param [Symbol] method the method being called
  # @param [Hash] full_response the full response returned by the server
  # @return the simple response if there is one or the full response if not
  def extract_simple_response(method, full_response)
    response_sym = "#{method}_response".to_sym
    return_sym = "#{method}_return".to_sym
    response = full_response
    # the typical response from CTF (after being converted to hash by Savon) has the following form. Here we called #login on a ctf session object:
    # {:login_response=>{:login_return=>"soapFD4EEB11A842A9A481329E1C2963F7CA"}}
    if full_response.is_a?(Hash) && full_response[response_sym] && full_response[response_sym].has_key?(return_sym)
      response = full_response[response_sym][return_sym]
      response.strip! if response.respond_to? :strip! #strip whitespace if response is a string (not always the case, can also be an Array for example)
      response ||= full_response
    end
    response
  end

  # Calls the SOAP service with the current session ID if the method is a SOAP action
  def method_missing(method, *args, &block)
    actions = endpoint.wsdl.soap_actions
    return super unless actions.include?(method)

    request_args = {}
    request_args.merge!(session_id: session_id) if session_id
    request_args.merge!(args[0]) if args[0].is_a? Hash

    request(method, request_args)
  end

  # Fixes the empty nodes from a SOAP response.
  # Empty nodes are represented as a hash of the attributes of the empty XML node
  # due to a bug in nori: https://github.com/rubiii/nori/issues/12
  # The empty node values are replaced with nil.
  #
  # @param [Hash] response the soap response
  # @return [Hash] the stripped soap response
  def self.strip_empty_nodes!(response)
    case response
      when Hash
        response.reject! { |k, v| k.to_s[0] == '@' || k == :data_rows }
        response.each do |key, value|
          stripped_value = strip_empty_nodes!(value)
          response[key] = (stripped_value unless stripped_value.blank?)
        end

      when Array
        response.each { |node| strip_empty_nodes! node }
        response.reject! &:blank?
    end

    response
  end

end
