# Copyright 2013 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.

require 'savon'

# The top level TeamForge SOAP session. This is where you log in, get a session ID, and create other
# services.
class TeamForge::Session

  # The TeamForge SOAP session ID.
  attr_reader :session_id

  # Creates a new Session.
  # @param [Hash] opts  hash with the following keys: :ctf_root_url -required( e.g https://forge.collab.net.), :session_id optional, :wsdl_directory optional
  def initialize(opts)
    @ctf_root_url = opts[:ctf_root_url]
    @session_id = opts[:session_id]
    @wsdl_directory = opts[:wsdl_directory]
  end

  # Log in to the CTF server with the given credentials.
  # @param [String] username the username
  # @param [String] password the password
  # @return [String] the session ID
  def login(username, password)
    @session_id = teamforge.request(:login, user_name: username, password: password)
    invalidate_sessions
    @session_id
  end

  # Logs into a CTF server with the given token.
  # @param [String] token the token with which to log in
  # @return [String] the session ID
  def token_login(token)
    @session_id = teamforge.request(:get_session_id, token: token)
    invalidate_sessions
    @session_id
  end

  # Clears the current sessions.
  # @return [TeamForge::Session] self
  def invalidate_sessions
    @teamforge, @scm, @pluggable, @sync = nil, nil, nil, nil
    self
  end

  # Creates a connection to the CollabNet SOAP service.
  # @see http://help.collab.net/index.jsp?topic=/teamforge.api.610/javadoc/com/collabnet/ce/soap60/webservices/cemain/package-summary.html
  # @return [TeamForge::Service] a service that can be called with methods from the CollabNet service
  def teamforge
    @teamforge ||= TeamForge::Service.new(@ctf_root_url, "CollabNet", session_id, @wsdl_directory)
  end

  # Creates a connection to the PluggableApp SOAPservice.
  # @see http://help.collab.net/index.jsp?topic=/iaf-611/javadoc/com/collabnet/ce/soap60/webservices/pluggable/package-summary.html
  # @return [TeamForge::Service] a service that can be called with methods from the PluggableApp service
  def pluggable
    @pluggable ||= TeamForge::Service.new(@ctf_root_url, "PluggableApp", session_id, @wsdl_directory)
  end

  # Creates a connection to the RbacApp SOAPservice. (Rbac = "Role based access control")
  # @return [TeamForge::Service] a service that can be called with methods from the PluggableApp service
  def rbac
    @rbac ||= TeamForge::Service.new(@ctf_root_url, "RbacApp", session_id, @wsdl_directory)
  end

  # Creates a connection to the ScmApp SOAP service.
  # @see http://help.collab.net/index.jsp?topic=/teamforge.api.610/javadoc/com/collabnet/ce/soap60/webservices/scm/package-summary.html
  # @return [TeamForge::Service] a service that can be called with methods from the ScmApp service
  def scm
    @scm ||= TeamForge::Service.new(@ctf_root_url, "ScmApp", session_id, @wsdl_directory)
  end

  # Creates a connection to the SimpleFileStorageApp SOAP service.
  # @see http://help.collab.net/index.jsp?topic=/teamforge.api.610/javadoc/com/collabnet/ce/soap60/webservices/filestorage/package-summary.html
  # @return [TeamForge::Service] a service that can be called with methods from the SimpleFileStorageApp service
  def simple_file_storage
    @simple_file_storage ||= TeamForge::Service.new(@ctf_root_url, "SimpleFileStorageApp", session_id, @wsdl_directory)
  end

  # Creates a connection to the OrchestrateSyncListener SOAP service.
  # @return [TeamForge::Service] a service that can be called with methods from the OrchestrateSyncListener  service
  def sync
    @sync ||= TeamForge::Service.new(@ctf_root_url, "OrchestrateSyncListener", session_id, @wsdl_directory)
  end

end
