# Copyright 2013 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.

require 'spec_helper'
require 'teamforge'
require 'ostruct'

describe TeamForge::Service do
  before(:each) do
    @root_url = 'https://forge.collab.net'
    @service_name = 'CollabNet'
    @session_id = 'sessionID'
    @endpoint_url = "#{@root_url}/ce-soap60/services/#{@service_name}"
    @service = TeamForge::Service.new(@root_url, @service_name, @session_id)
  end

  it 'creates a connection to the correct API endpoint' do
    expected = expect_create_endpoint
    actual = @service.endpoint
    actual.should be_equal(expected)
  end

  it 'creates a connected to the endpoint created for internal use if the service name is OrchestrateSyncListener' do
    sync_service = 'OrchestrateSyncListener'
    # CTF folks exposed an OrchestrateSyncListener service on a different endpoint
    @end_point_url= "#{@root_url}/ce-soap/services/#{@service_name}"
    @service = TeamForge::Service.new(@root_url, sync_service, @session_id)
  end

  it 'only creates the endpoint once' do
    expect_create_endpoint

    endpoint1 = @service.endpoint
    endpoint2 = @service.endpoint
    endpoint1.should be_equal(endpoint2)
  end

  it 'uses the local file to read wsdl from if passed the wsdl directory' do
    service = TeamForge::Service.new(@root_url, @service_name, @session_id, "/foo/wsdl")
    service.endpoint.wsdl.document.should == "/foo/wsdl/CollabNet.xml"
  end

  it 'infers external location for the wsdl from root_url and service_name if the local wsdl directory is not passed' do
    service = TeamForge::Service.new(@root_url, @service_name, @session_id)
    service.endpoint.wsdl.document.should == "https://forge.collab.net/ce-soap60/services/CollabNet?wsdl"
  end

  describe 'request' do

    before(:each) do
      some_url = 'http://orc-test-ctf/ce-soap60/services/CollabNet?wsdl'
      stub_request(:get, some_url).
          with(:headers => {'Accept'=>'*/*', 'User-Agent'=>'Ruby'}).
          to_return(:status => 200, :body => '', :headers => {})

      @client = Savon::Client.new(some_url) { wsdl.endpoint = some_url }
      @service.should_receive(:endpoint).and_return(@client)
    end

    it 'calls the endpoint with the given method' do
      savon.expects(:the_method).returns(code: 200, headers: {}, body: empty_response)

      @service.request(:the_method)
    end

    it 'adds the given arguments' do
      args = { arg1: 'value1', arg2: 'value2'}

      savon.expects(:the_method).with(args).returns(code: 200, headers: {}, body: empty_response)

      @service.request(:the_method, args)
    end

    it 'raises a TeamForge::ServiceError if some SOAP fault happens' do
      http = OpenStruct.new(code: 200, headers:{}, body: fault_response)
      @client.should_receive(:request).with(:the_method).and_raise(Savon::SOAP::Fault.new(http))

      lambda do
        @service.request(:the_method)
      end.should raise_error(TeamForge::ServiceError)
    end

    it 'strips simple results to their return values' do
      method = :the_method
      result = 'some response'
      savon.expects(method).returns(code: 200, headers: {}, body: simple_response(method, result))
      actual = @service.request(method)
      actual.should == result
    end
  end

  describe 'method_missing' do
    it 'does not try to call methods not defined in the wsdl' do
      endpoint = expect_create_endpoint
      expect_soap_actions(endpoint)

      lambda do
        @service.some_method
      end.should raise_error(NoMethodError)
    end

    it 'calls methods on the API that are in the wsdl' do
      endpoint = expect_create_endpoint
      expect_soap_actions(endpoint, :some_method)

      @service.should_receive(:request).with(:some_method, session_id: @session_id)

      @service.some_method
    end

    it "doesn't pass the session id if there isn't one on the service" do
      service = TeamForge::Service.new(@root_url, @service_name, nil)
      endpoint = expect_create_endpoint
      expect_soap_actions(endpoint, :some_method)

      service.should_receive(:request).with(:some_method, {})

      service.some_method
    end

    # Configures the mocks to expect to find the available SOAP actions in the WSDL.
    # @param [Mock] endpoint the endpoint that should get the wsdl call
    # @param [Array] actions the actions that will be returned
    def expect_soap_actions(endpoint, *actions)
      wsdl = mock()
      endpoint.should_receive(:wsdl).and_return(wsdl)
      wsdl.should_receive(:soap_actions).and_return(actions)
    end
  end

  describe 'strip_empty_nodes!' do
    it 'fixes the representation of empty nodes' do
      response = {:some_node => {:'@attribute' => nil}, :non_empty_scalar => 'asdf', :another_node => {:'@xsi:type' => 'xsd:string'} }

      TeamForge::Service.strip_empty_nodes!(response).should == { :some_node => nil, :non_empty_scalar => 'asdf', :another_node => nil }
    end

    it "doesn't change non-empty array nodes" do
      response = {:multi_ref => [1, 2, 3]}

      TeamForge::Service.strip_empty_nodes!(response).should == { :multi_ref => [1, 2, 3] }
    end

    it 'does fix nested objects' do
      response = {
          :object1 => { :empty => {:'@xsi:type' => 'string' } },
          :object2 => { :multi_ref => {:'@xsi:size' => nil} }
      }

      TeamForge::Service.strip_empty_nodes!(response).should == { :object1 => { :empty => nil }, :object2 => { :multi_ref => nil } }
    end
  end

  # Configures the mocks to expect to create a SOAP endpoint.
  # @return the mock endpoint
  def expect_create_endpoint
    endpoint = mock(Savon::SOAP::Request)
    Savon::Client.should_receive(:new).and_return(endpoint)
    wsdl = mock()
    endpoint.should_receive(:wsdl).and_return(wsdl)
    wsdl.should_receive(:endpoint=).with(@endpoint_url)

    ssl = mock()
    endpoint.stub_chain(:http, :auth, :ssl).and_return(ssl)
    ssl.should_receive(:verify_mode=).with(:none)

    endpoint
  end

  # Returns an empty SOAP Fault response.
  # @return [String] an empty SOAP Fault document
  def empty_response
    <<-HERE
<?xml version="1.0" encoding="utf-8"?>
<soap:Envelope xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
xmlns:xsd="http://www.w3.org/2001/XMLSchema" xmlns:soap="http://
schemas.xmlsoap.org/soap/envelope/">
  <soap:Body>
  </soap:Body>
</soap:Envelope>
    HERE
  end

  def simple_response(method, a_string)
    soap_method_name = method.to_s
    <<-HERE
<?xml version="1.0" encoding="utf-8"?>
<soap:Envelope xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
xmlns:xsd="http://www.w3.org/2001/XMLSchema" xmlns:soap="http://
schemas.xmlsoap.org/soap/envelope/">
  <soap:Body>
    <#{soap_method_name}Response>
      <#{soap_method_name}Return>
        #{a_string}
      </#{soap_method_name}Return>
    </#{soap_method_name}Response>
  </soap:Body>
</soap:Envelope>
    HERE
  end

  # Returns an empty SOAP response.
  # @return [String] an empty SOAP document
  def fault_response
    <<-HERE
<?xml version="1.0" encoding="utf-8"?>
<soap:Envelope xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
xmlns:xsd="http://www.w3.org/2001/XMLSchema" xmlns:soap="http://
schemas.xmlsoap.org/soap/envelope/">
  <soap:Body>
    <soap:Fault>
      <faultcode>Server</faultcode>
      <faultstring>Some error happened</faultstring>
      <faultactor>The user</faultactor>
      <detail>Some app specific detail</detail>
    </soap:Fault>
  </soap:Body>
</soap:Envelope>
    HERE
  end

end
