#
# Copyright 2013 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

# Represents a Team Forge item association.
class CtfAssociation
  # Child of [CtfItem].
  include MongoMapper::EmbeddedDocument
  include Comparable

  # Matches CTF item IDs (core and IAF).
  ITEM_REGEX = /((?<type>[a-z]+)(?<id>\d+)|(?<type>\w+)_(?<id>\w+))/

  # The item's key in the format of artf1234, cmmt789, BUILD_kkDD, etc.
  key :item_id, String

  # Give the ID key an alias that matches the parent CtfItem.
  alias :ctf_id :item_id

  # Returns the associated commits for this CTF association
  # @param [project] project to look for commits
  # @return [Array<Commit>] associated commits
   def associated_commit_summaries(project)
     CommitSummary.find_commit_summaries_by_ctf_association(self, project)
   end


  # Defines the natural sort order of [CtfAssociation]s. First sorted alphabetically by type, then numerically
  # by item ID.
  # @param [CtfAssociation] other the association to compare to this one
  # @return -1 if this is less than the other, 0 if they're equal, +1 if this is greater than the other
  def <=>(other)
    parts = ITEM_REGEX.match(item_id)
    other_parts = ITEM_REGEX.match(other.item_id)
    if numeric?(parts[:id]) && numeric?(other_parts[:id])
      # compare IDs numerically if both are numbers
      [parts[:type].downcase, parts[:id].to_i] <=> [other_parts[:type].downcase, other_parts[:id].to_i]
    else
      # else use string compare for IDs
      [parts[:type].downcase, parts[:id]] <=> [other_parts[:type].downcase, other_parts[:id]]
    end
  end

  private
  # determine whether the string is numeric or not
  # @return [Boolean]
  def numeric?(str)
    str =~ /\A\d+\Z/
  end

end