require 'json/jwt'
require 'net/http'

module CtfApiAuthentication

  JWT_ISSUER = 'urn:ctf:iss:tf-idp'
  JWT_AUDIENCE = 'soap60'

  # Simulate the OmniAuth Hash structure
  UserObject = Struct.new(:uid, :provider, :info)
  UserInfo = Struct.new(:name, :email)

  # Make sure that the user is authenticated and has the necessary permissions
  def authenticate_with_teamforge_and_get_permissions
    authenticate_with_teamforge
    get_user_permissions unless @ctf_user_data.nil?
  rescue TeamForge::ServiceError => ex
    logger.warn(ex.message)
    head status: 404
  end

  # Parse the token from request and authenticate user
  def authenticate_with_teamforge
    encoded_token = extract_session_from_request(request)
    token = JSON::JWT.decode encoded_token, keyset

    if is_token_valid?(token)
      validate_cached_user_data(token[:sub], encoded_token)
    else
      raise AuthenticationError.new "token invalid #{token.as_json}"
    end
  rescue JSON::JWT::InvalidFormat, AuthenticationError => invalid
    logger.warn invalid.message
    head status: 401
  end

  # Check if the token in received request is valid
  #
  # @param [Hash] token Decoded keys and value
  # @param [Integer] leeway seconds of leeway to apply
  # @param [Time] compare_time time to compare with
  # @return [Boolean] whether the token is currently valid or not
  def is_token_valid?(token, leeway = 30.seconds, compare_time = Time.now)
    token[:iss] == JWT_ISSUER &&
    token[:aud].include?(JWT_AUDIENCE) &&
    token[:sub].present? &&
    Time.at(token[:exp]) > compare_time - leeway
  end

  # Retrieve the keyset to decode the encoded token from redis or get the keyset from
  # Teamforge if it is not already cached.
  def keyset
    keyset = $redis_jwt_keyset.get('keyset')

    if keyset.nil?
      keyset = get_keyset_from_teamforge
      raise AuthenticationError.new "Unable to fetch keyset from teamforge" unless keyset
      $redis_jwt_keyset.set('keyset', keyset)
    end

    JSON::JWK::Set.new(JSON.parse(keyset))
  end

  # Get the keyset from Teamforge
  def get_keyset_from_teamforge
    base_url = Rails.configuration.teamforge[:server]
    url = URI.parse(base_url + '/sf/auth/token/keyset')
    http = HttpConfiguration.create_connection url
    response = http.start do |http|
      get_keyset = Net::HTTP::Get.new url
      http.request get_keyset
    end
    return response.body
  end


  # User's data from CTF can be saved on the cache. If it is cached, use it
  # else refresh the user data from Teamforge and cache it.
  #
  # @param [String] username Key of the record cached on redis
  # @param [String] encoded_token Encoded token that can be used to get TF session
  def validate_cached_user_data(username, encoded_token)

    cache_ctf_user_data = $redis_user_data.get(username)

    if cache_ctf_user_data.nil?
      @ctf_session = CtfService.create_teamforge_session(encoded_token)
      teamforge_service = @ctf_session.teamforge
      @ctf_user_data = teamforge_service.get_current_user_data
      $redis_user_data.set(username, @ctf_user_data.to_json)
      $redis_user_data.expire(username, 86400) # Set expiry period to 1 day
    else
      @ctf_user_data = JSON.parse(cache_ctf_user_data).deep_symbolize_keys
    end

  end


  # Tease the session from one of the possible request headers
  #
  # @param [ActionDispatch::Request] request Request object
  # @return [String] Session key
  def extract_session_from_request(request)
    if request.headers['Authorization']
      request.headers['Authorization'].sub(/^Bearer\s+/i, "")
    else
      request.headers['X-EventQ-Session'] || request.headers["X-Orchestrate-Session"]
    end
  end

  # Projects that the user has access to when data is requested via the Reporting API
  def set_reporting_access
    @privileged_projects = []
    @privileged_projects = get_a_permission_for_user(current_user, 'reporting_api_projects') unless current_user.super_user
  end

  # Initiates the permissions flow for current user. New users are created if not already found in EQ database
  def get_user_permissions
    @ctf_user = @ctf_user_data[:multi_ref] || @ctf_user_data
    auth = UserObject.new.tap do |u|
      u.uid = @ctf_user[:username]
      u.provider = 'teamforge'
      u.info = UserInfo.new(@ctf_user[:full_name], @ctf_user[:email])
    end
    @current_user = User.find_or_create_user(auth)
    ctf_permissions_check
  end

  def check_post_schema_permissions_for_user
    if !(@current_user.super_user || project_ids_with_permission(:pipeline_create, @current_user).count > 0)
      raise Exception.new("POST_SCHEMA_PERMISSIONS_EXCEPTION. #{@current_user.inspect}")
    end
  rescue Exception => ex
    logger.error(ex.message)
    head status: 404
  end

  # Source API differs in its required permissions, all actions require:
  # :site_admin || :pipeline_edit (for current proj) || :pipeline_create (for current_proj) || :project_admin (for current_proj)
  def require_crud_permissions
    require_api_permission('SOURCE')
  end

  def privileged_projects
    @privileged_projects
  end

  # common permission permutation
  # @return [Boolean] true or false
  # :site_admin || :pipeline_edit (for current proj) || :pipeline_create (for current_proj)
  def can_access_resources?
    (current_user.super_user || has_write_permission_for_project_id?(current_project.id, current_user))
  end

  # Tool API differs in its required permissions, all actions require:
  # :site_admin || :pipeline_edit (for current proj) || :pipeline_create (for current_proj)
  def can_crud_tools?
    require_api_permission('TOOL')
  end

  # Tracker API differs in its required permissions, all actions require:
  # :site_admin || :pipeline_edit (for current proj) || :pipeline_create (for current_proj)
  def require_crud_tracker
    require_api_permission('TRACKER')
  end

  # common api user permission
  def require_api_permission(resource='API')
    unless can_access_resources?
      raise Exception.new("#{resource}_PERMISSIONS_EXCEPTION. #{current_user.inspect}")
    end
  rescue Exception => ex
    logger.error(ex.message)
    head status: 404
  end

  # super_user permission check
  def require_super_user
    unless current_user.super_user?
      raise Exception.new("PERMISSIONS_EXCEPTION. #{current_user.inspect}")
    end
  rescue Exception => ex
    logger.error(ex.message)
    head status: 403
  end

end
