#
# Copyright 2016 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

require 'application_constants'

# Configures TeamForge installations with an IAF connection to this application instance.
class CtfConfigurer
  extend CtfAssociationPrefixes

  APP_NAME_IN_CTF = "#{APP_PRODUCT_NAME} App"

  # The default options for configuring TeamForge for this application.
  # @return [Hash] the default options
  DEFAULT_OPTS =
    {
      ctf_url: Rails.configuration.teamforge[:server],
      orc_url: "#{Rails.configuration.teamforge[:server]}/orc",
      plug_name: APP_CTF_PLUG_NAME,
      prefix: ctf_association_prefixes.join(',')
    }.merge(Rails.configuration.teamforge[:iaf]).with_indifferent_access

  TRANSLATION_KEYS = {iaf_description: 'l10n.application.description',
                      source_association_key: {
                          title: 'l10n.association_key.title',
                          description: 'l10n.association_key.description'
                      },
                      queue_server: {
                          title: 'l10n.queue_server.title',
                          description: 'l10n.queue_server.description'
                      },
                      queue_username: {
                          title: 'l10n.queue_username.title',
                          description: 'l10n.queue_username.description'
                      },
                      queue_password: {
                          title: 'l10n.queue_password.title',
                          description: 'l10n.queue_password.description'
                      },
                      queue_name: {
                          title: 'l10n.queue_name.title',
                          description: 'l10n.queue_name.description'
                      }
  }

  RESOURCE_BUNDLES = [{locale: 'en',
                       resources: [
                           {names: {
                               item: %w(l10n.association_key.description l10n.association_key.title l10n.queue_server.title l10n.queue_server.description l10n.queue_username.title l10n.queue_username.description l10n.queue_password.title l10n.queue_password.description l10n.queue_name.title l10n.queue_name.description l10n.application.description)
                           },
                            values: {
                                item: [
                                    'Values populated automatically',
                                    'Source Association Key',
                                    'Queue Server',
                                    'Values populated automatically',
                                    'Queue Username',
                                    'Values populated automatically',
                                    'Queue Password',
                                    'Values populated automatically',
                                    'Queue Name',
                                    'Values populated automatically',
                                    APP_NAME_IN_CTF 
                                ]
                            },
                            :attributes! => {
                                'xmlns:impl' => "http://schema.open.collab.net/sfee50/soap60/type"
                            }
                           }
                       ],
                       :attributes! => {
                           resources: {
                               'xmlns' => "http://schema.open.collab.net/sfee50/soap60/type"
                           }
                       }
                      }]

  # these are displayed on the 'iaf edit' page inside CTF in the project context
  IAF_CONFIG_PARAMS = {param: [
      {title: 'l10n.association_key.title',
       name: 'tracker_source_association_key',
       description: 'l10n.association_key.description',
       defaultValue: 'DO NOT EDIT',
       displayType:'TEXT',
       editable: 'false',
       valueType: 'String',
       maxLength: '100'
      },
      {title: 'l10n.queue_server.title',
       name: 'tracker_queue_uri',
       description: 'l10n.queue_server.description',
       defaultValue: 'DO NOT EDIT',
       displayType:'TEXT',
       editable: 'false',
       valueType: 'String',
       maxLength: '300'
      },
      {title: 'l10n.queue_username.title',
       name: 'tracker_queue_username',
       description: 'l10n.queue_username.description',
       defaultValue: 'DO NOT EDIT',
       displayType:'TEXT',
       editable: 'false',
       valueType: 'String',
       maxLength: '300'
      },
      {title: 'l10n.queue_password.title',
       name: 'tracker_queue_password',
       description: 'l10n.queue_password.description',
       defaultValue: 'DO NOT EDIT',
       displayType:'TEXT',
       editable: 'false',
       valueType: 'String',
       maxLength: '100'
      },
      {title: 'l10n.queue_name.title',
       name: 'tracker_queue_name',
       description: 'l10n.queue_name.description',
       defaultValue: 'DO NOT EDIT',
       displayType:'TEXT',
       editable: 'false',
       valueType: 'String',
       maxLength: '100'
      }
  ]}


  def initialize(opts = {})
    @options = DEFAULT_OPTS.merge(opts)
    session = CtfService.create_teamforge_session
    session.login(@options[:admin_user], @options[:admin_pass])
    @pluggable = session.pluggable
    @teamforge = session.teamforge
    @simple_file_storage = session.simple_file_storage
  end


  # Configures a TeamForge instance with this application.
  def configure
    create_site_wide_linked_application
    create_iaf_application
  end

  # Uploads the application icon into TeamForge for use in the project toolbar
  def upload_icon
    @simple_file_storage.start_file_upload.tap do |file_storage_id|
      icon_path = Rails.root + "app/assets/images/icons/orchestrate-icon.png"
      icon_encoded = Base64.encode64(File.binread(icon_path))
      @simple_file_storage.write(file_storage_id: file_storage_id, data: icon_encoded)
      @simple_file_storage.end_file_upload(file_storage_id: file_storage_id)
    end
  end

  # application permissions in the form expected by teamforge
  def app_permissions
    UserPermission::PERMISSIONS.map do |perm|
      {permission: perm.humanize, is_admin_tool: 'false'}.tap do |result|
        if perm == 'pipeline_read'
          result.merge!(dap_mapped_to: 'View')
        end
      end
    end
  end

  # iaf configuration object as seen by CTF
  def plugguble_config
    {
      baseurl:  @options[:orc_url],
      resource_bundles: { 'IntegratedAppResourceBundleSoapDO' => RESOURCE_BUNDLES },
      config_parameters: IAF_CONFIG_PARAMS,
      description: TRANSLATION_KEYS[:iaf_description],
      enabled: 'true',
      endpoint: "#{@options[:orc_url]}/services/team_forge",
      gourl: "#{@options[:orc_url]}/go/%o",
      id_pattern: '\\w+',
      name: @options[:plug_name],
      permissions: { 'PluggablePermissionSoapDO' => app_permissions },
      prefix: @options[:prefix],
      require_page_component: 'false',
      require_prefix_project: 'false',
      'requireSCMIntegration' => 'false',
      service_type: 'REST',
      version: '0'
    }
  end

  # Creates the application inside TeamForge. This is the first step in adding
  # this application to a project.
  # @return [String] the application's TeamForge ID
  def create_iaf_application
    icon_file_key = upload_icon

    @pluggable.create_integrated_application4(
        'soapDO' => plugguble_config,
        icon_file_key: icon_file_key,
        icon_file_name: 'orchestrate-icon.png',
        icon_mime_type: 'image/png'
    )

    application_id = @pluggable.get_plug_id_by_base_url(base_url: @options[:orc_url])

    @pluggable.set_pluggable_app_message_resource(
        pluggable_id: application_id,
        locale: 'en',
        key: TRANSLATION_KEYS[:iaf_description],
        value: APP_NAME_IN_CTF)

    application_id
  end

  # Adds the application to the project with the given ctf_project_id.
  # @param [String] application_id the application's TeamForge ID
  # @param [String] project_path the TeamForge Project path to add this
  #                 application to
  def add_to_project(application_id, project_path)
    project_id = @teamforge.get_project_by_path('projectPath' => "projects.#{project_path}")[:multi_ref][:id]

    @pluggable.enable_pluggable_component(
        project_id: project_id,
        pluggable_id: application_id,
        params: [],
        prefix: @options[:prefix])

    TeamForge::ProjectConfiguration.find_by_ctf_project_id(project_id).tap do |project|
      project.link_plug_id = @pluggable.get_link_plug_id_by_plug_id(
          project_id: project_id,
          pluggable_id: application_id)
      project.server_url = @options[:ctf_url]
      project.application_id = application_id
      project.app_prefix = @options[:prefix]
      project.save!
    end
  end

  # Adds this appliication to CTF as a site-wide linked application
  def create_site_wide_linked_application
    @teamforge.create_site_wide_linked_app(
        title: @options[:plug_name],
        application_url: "#{@options[:orc_url]}/users/auth/teamforge/callback",
        application_type: 'same',
        single_signon: true
    )
  end

  # Used for upgrading instances of this application in CTF
  def edit_iaf_application
    current_iaf_configuration = @pluggable.get_integrated_application_by_name3(plug_name: @options[:plug_name])[:multi_ref].first
    application_id = current_iaf_configuration[:id]
    icon_key = current_iaf_configuration[:icon_key]
    @pluggable.edit_integrated_application3(
        'soapDO' => plugguble_config.merge({id: application_id, icon_key: icon_key})
    )
    update_iaf_icon application_id

  end

  # New CTF Soap Method added to upload a new icon for a given integrated application id
  # @param [String] application_id the application's TeamForge ID, defaults to edit_iaf_application.appId
  # @param present so that a bulk replace can be made of appliication icons by appId
  def update_iaf_icon(ctf_application_id)
    icon_file_key = upload_icon
    @pluggable.set_integrated_application_icon(
        integrated_app_id: ctf_application_id,
        icon_file_key: icon_file_key,
        icon_file_name: 'orchestrate-icon.png',
        icon_mime_type: 'image/png')
  end

end
