# This module separates out the ctf communication concerns from the user and
module CtfUserAuthorization
  include CtfApiAuthentication

  VALID_PERMISSIONS = %w(pipeline_create pipeline_edit pipeline_read reporting_api pipeline_api pipeline_trusted_association)

  def self.included(base)
    base.extend CtfUserAuthorization
  end

  # Called by #cache_permissions_for_user, this populates the user's permissions
  # cache with permissions retrieved from CTF.
  # @param [TeamForge::Session] ctf_session A CTF SOAP session for current user
  # with teamforge and rbac SOAP endpoints.
  def teamforge_refresh_permissions(ctf_session, user)

    ctf_permissions_map = {}
    VALID_PERMISSIONS.each do |permission|
      ctf_permissions_map[permission.to_s] = []
    end

    # Construct from first principles the user's list of permissions to CTF projects.
    # Gets the list of project ids for this user from CTF, and then for each of them
    # asks for the user's permissions to that project.  The resultant permissions set
    # is converted to a hash with permission names as keys. The hash is saved to Redis.
    #
    # It is implemented as a simple map/reduce operation, mapping from
    # projects:Array[permissions]:Array[UserPermissions] and reducing that to a single array
    # for each permission
    iaf_projects_ids(ctf_session.teamforge, user).reduce([]) do |accum, ctf_project_id|
      project_config = TeamForge::ProjectConfiguration.find_by_ctf_project_id(ctf_project_id)
      accum + iaf_permissions_for_project(ctf_session.rbac, ctf_project_id).map do |perm|
        if !ctf_permissions_map[perm.to_s]
          ctf_permissions_map[perm.to_s] = []
        end
        ctf_permissions_map[perm.to_s] << project_config.project_id
      end
    end

    ctf_permissions_map
  end

  # Tests if the user has any provided permission on a given object (or nil if a project level permission)
  def ctf_general_permission_for_project?(ctf_session, project_id, operation, object_id = nil)
    ctf_session.teamforge.has_general_permission(username: uid,
                                                 project_id: project_id,
                                                 operation: operation,
                                                 object_id: object_id) == true
  end

  # Test whether this user has a given permission on a project
  # @param [String] permission to test
  # @param [ProjectConfiguration] project_config to test
  # @param [TeamForge::Session] ctf_session to use
  # @return [Boolean] true or false, with errors logged
  def ctf_permission_for_project?(permission, plug_id, project_config, ctf_session)
    logger.debug "Testing permission '#{permission}' on #{project_config.try(:link_plug_id)}"
    operation = ["PluggableLink", plug_id, permission.humanize, permission.humanize].join('.')

    #noinspection RubySimplifyBooleanInspection
    # this client method returns boolean true or a soap response hash, which we want to interpret as false.
    project_id = project_config.ctf_project_id
    object_id = project_config.link_plug_id

    ctf_general_permission_for_project?(ctf_session, project_id, operation, object_id)
  rescue => e
    logger.warn "Unable to test permission '#{permission}' on (#{project_config.try(:ctf_project_id)} / #{project_config.try(:link_plug_id)})"
    logger.warn e.message
    false
  end

  # Pings CTF for IAF permissions and returns an array of permissions
  # @param [String] ctf_rbac, CTF session id
  # @param [String] ctf_project_id, CTF project id
  # @return [Array] array of permission names if the user has any permissions in this project
  # @return [nil] if the  user does not have any permissions in this project
  def iaf_permissions_for_project(ctf_rbac, ctf_project_id)
    link_id = TeamForge::ProjectConfiguration.find_by_ctf_project_id(ctf_project_id).get_link_plug_id(ctf_rbac.session_id)
    permissions = ctf_rbac.get_integrated_app_permissions_for_project(project_id: ctf_project_id, link_id: link_id)[:get_integrated_app_permissions_for_project_return]
    if permissions
      permissions_array = ([] << permissions).flatten #need this because CTF returns a string for a single permission or array for multiple
      permissions_array.map { |perm| perm.downcase.gsub(/ +/, "_") }
    else
      []
    end
  end

  # Returns the ctf project_ids for the projects that user is a member of and which also have ORC as an IAF app
  # @param [TeamForge::Service] ctf_service a teamforge CTF service
  # @return [Array]
  def iaf_projects_ids(ctf_service, user=nil)
    ctf_projects_ids(ctf_service, user) & Project.active_projects_ctf_ids
  end

  # Queries the CTF for the projects that the user is a member of and returns an array of their ids
  # @param [TeamForge::Service] ctf_service a teamforge CTF service
  # @return [Array] the ctf projects ids
  def ctf_projects_ids(ctf_service, user=nil) #TODO extract
    user ||= current_user
    ctf_projects = ctf_service.get_project_list_for_user(user_id: user[:uid], fetch_hierarchy_path: false, include_group_membership: false)[:multi_ref]
    ctf_projects.to_a.map { |project| project[:id] }
  end


  # Reads the cache and returns the cached permissions
  # @param [String] permission - the name to read from cache and return the value for
  # @param [String] user - the name to read from cache and return the value for
  # @return [Array] - The projects that have a permission for the user if permission name is passed
  def get_a_permission_for_user(user, permission)

    cache_permissions_for_user(user) if permissions_not_cached?(user)
    current_user_permissions = get_all_permissions_for_user(user)
    current_user_permissions.keys.include?(permission.to_s) ? current_user_permissions[permission.to_s] : update_cache_and_return_value(permission.to_s, user)

  end

  # Reads the cache and returns the cached permissions
  # @param [String] user - the name to read from cache and return the value for
  # @return [Hash] - The hash with all permission names as keys and array with project IDs as values
  def get_all_permissions_for_user(user)

    cache_permissions_for_user(user) if permissions_not_cached?(user)
    JSON.parse($redis_user_permissions.get(user.uid))

  end

  # Checks if an entry already exists for the user in the Redis cache.
  # @return [Boolean] true if user entry is not present in the cache.
  # @return [Boolean] false if user entry is present in the cache.
  def permissions_not_cached?(user)
    $redis_user_permissions.get(user.uid).nil?
  end

  # Refresh permissions for CTF and populate cache. Triggers method to compute
  # and cache project access information for Reporting API. Cache structure is as below.
  # cached_perms = { 'userName': {
  #                   pipeline_create: [prj01, prj03],
  #                   pipeline_edit: [prj02, prj03],
  #                   pipeline_read: [prj03, prj05],
  #                   reporting_api: [prj02, prj05],
  #                   pipeline_api: [prj04, prj03],
  #                   pipeline_trusted_association: [prj04, prj01],
  #                   reporting_api_projects: [prj03, prj05, prj02]} }
  def cache_permissions_for_user(user)

    return if user.super_user

    username = user.uid
    user_permissions = teamforge_refresh_permissions(@ctf_session, user)
    $redis_user_permissions.set(username, user_permissions.to_json)
    cache_projects_for_reporting(user)
  end

  # Updates the cache with list of projects that the user has Reporting API access to
  def cache_projects_for_reporting(user)
    username = user.uid
    user_permissions = get_all_permissions_for_user(user)
    user_permissions['reporting_api_projects'] = get_reporting_projects(user)
    $redis_user_permissions.set(username, user_permissions.to_json)
  end

  # Returns a list of projects with their CTF Project IDs - for which the user has either
  # Pipeline Read or Reporting API permissions
  # @return [Array] CTF Project IDs
  def get_reporting_projects(user=nil)
    user_privileged_projects = []
    ctf_projects = TeamForge::ProjectConfiguration
    # Associations viewer in CTF needs to be functioning on either Pipeline Read permission or Reporting API permission,
    # and not just the latter one.
    read_access_projects = project_ids_with_permission(:pipeline_read, user).uniq
    read_access_projects.each { |orc_proj_id| user_privileged_projects << ctf_projects.where(:project_id => orc_proj_id).first.ctf_project_id } unless read_access_projects.empty?
    reporting_access_projects = project_ids_with_permission(:reporting_api, user).uniq
    reporting_access_projects.each { |orc_proj_id| user_privileged_projects << ctf_projects.where(:project_id => orc_proj_id).first.ctf_project_id } unless reporting_access_projects.empty?

    user_privileged_projects.flatten.uniq
  end

  # Retrieve a list of project_ids for projects the user has the provided permissions on.
  #
  # @params [Symbol] permission The permission to check against.
  # @params [user] object
  # @return [Array[id]] The list of project ids for projects the user has the provided permission in.
  def project_ids_with_permission(permission, user=nil)

    user ||= current_user
    if user.super_user
      Project.active_projects.map{ | proj | proj.id.to_s}
    else
      imply_permissions_and_get_project_list(get_all_permissions_for_user(user), permission, user)
    end

  end

  # Read permission is implied if the user has Edit and Create permissions
  # Edit permission is implied for Create permission
  # Create permission is specific and does not imply any additional permission
  # @param [Hash] all_permissions User's cached permission hash
  # @return [Array] permission The permission to be check and implied if required
  def imply_permissions_and_get_project_list(all_permissions, permission, user=nil)

    user ||= current_user
    case permission.to_s
      when 'pipeline_read'
        all_permissions['pipeline_read'] + all_permissions['pipeline_edit'] + all_permissions['pipeline_create']
      when 'pipeline_edit'
        all_permissions['pipeline_edit'] + all_permissions['pipeline_create']
      when 'pipeline_create'
        all_permissions['pipeline_create']
      else
        (all_permissions.keys.include?(permission.to_s)) ? all_permissions[permission.to_s] : update_cache_and_return_value(permission.to_s, user)
    end

  end

  def update_cache_and_return_value(new_permission, user=nil, project_id=nil)

    if VALID_PERMISSIONS.include?(new_permission)
      username = user.uid
      current_user_permissions = get_all_permissions_for_user(user)
      project_id.nil? ? current_user_permissions[new_permission] = [] : current_user_permissions[new_permission] = [] << project_id
      $redis_user_permissions.set(username, current_user_permissions.to_json)

      get_a_permission_for_user(user, new_permission)
    else
      []
    end
  end

  # Process the CTF user permission notification message
  # @param [Hash] message contains new user permission set per project
  #
  # {
  #   "event_time":"2016-07-27T08:53:10-04:00",
  #   "project_id":"proj1234",
  #   "username": "username1",
  #   "user_permission":["pipeline_read", "pipeline_write"]
  # }
  def handle_message(message)

    if user = User.find_by_uid(message['username'])
      if updated_user_permissions = calculate_user_permissions(message)

        # set the redis cache
        $redis_user_permissions.set(user.uid, updated_user_permissions.to_json)

        # update cache project_ids for reporting api
        cache_projects_for_reporting(user)
      else
        Rails.logger.info("Nothing to set for user permissions")
      end
    else
      Rails.logger.info("No user found in EventQ")
    end
  end


  def calculate_user_permissions(message)
    username = message['username']
    current_user_permissions =  if redis_user_data = $redis_user_permissions.get(username)
                                  JSON.parse(redis_user_data)
                                else
                                  Hash[VALID_PERMISSIONS.map {|x| [x, []]}]
                                end
    if project_config = TeamForge::ProjectConfiguration.find_by_ctf_project_id(message['project_id'])
      if project_config.project
        project_id = project_config.project_id.to_s
        new_user_permissions = message['user_permission'].map { |up| up.parameterize('_') }

        # updating permissions
        VALID_PERMISSIONS.each { | perm |
          if new_user_permissions.include?(perm)
            if current_user_permissions.key?(perm)
              current_user_permissions[perm] << project_id unless current_user_permissions[perm].include?(project_id)
            else
              current_user_permissions[perm] = [project_id]
            end
          else
            current_user_permissions[perm] -= [project_id] if current_user_permissions.key?(perm)
          end
        }
      else
        # removing permissions
        VALID_PERMISSIONS.each { | perm |
          current_user_permissions[perm] -= [project_config.project_id.to_s] if current_user_permissions.key?(perm)
        }
      end
    else
      logger.info "Project - #{message['project_id']} Not Found"
      current_user_permissions = nil
    end
    current_user_permissions
  end

  # Find out if a permission exists for the user on a project.
  #
  # @param [Symbol] permission The permission to check.
  # @params [BSON id] project_id The project to check.
  # @return [Boolean] Asserts if the user have the permission on the specified project.
  def has_permission_for_project_id?(permission, project_id, user=nil)
    project_ids_with_permission(permission, user).uniq.include? project_id.to_s
  end

  def has_write_permission_for_project_id?(project_id, user=nil)
    has_permission_for_project_id?(:pipeline_create, project_id, user) || has_permission_for_project_id?(:pipeline_edit, project_id, user)
  end

  # Test whether the user has the permission to create trusted associations on
  # a given project.
  #
  # @params [BSON id] project_id The project to check.
  # @return [Boolean] Asserts if the user have the permission on the specified project.
  def create_trusted_associations?(project_id, user=nil)
    has_permission_for_project_id?(:pipeline_trusted_association, project_id, user)
  end
end
