#
# Copyright 2016 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

class PostCommitMessageHandler
  include CtfAssociationPrefixes

  def initialize(ctf_message)
    @ctf_message = ctf_message
  end


  # extracts data from the CTF post-commit, transforms into the format 
  # this application can understand and passes it to the server to create
  def process_message
    commit_data = translate_data
    commit_servers = repositories_for_commit(@ctf_message["repositoryId"])
    unless commit_servers.present?
      LogUtils.log_error_and_info(I18n.t('commit_server.errors.no_repository_for_ctf_id', ctf_id: @ctf_message["repositoryId"]), @ctf_message)
    end
    commit_servers.each { |server| server.handle_message(:commit_data, commit_data) }
  end


  # translates data into our format
  # @return [Hash]
  def translate_data
    associations = extract_associations(@ctf_message['associationsCreated'])

    {
      revision_id: extract_revision_number(@ctf_message["filesChanged"]).to_s,
      message: @ctf_message["commitMessage"],
      changes: extract_path_changes(@ctf_message["filesChanged"]),
      event_time: @ctf_message["commitDate"],
      ctf_item: {ctf_id: @ctf_message["commitId"], ctf_associations: associations},
      created_by: @ctf_message["userName"],
      type: "post-commit"
    }.with_indifferent_access
  end

  # extracts commit revision number and path changes from the reported "filesChanged"
  # @param [Hash] or [Array<Hash>] reported_changes
  def extract_path_changes(reported_changes)
    return [] unless reported_changes
    changed_files = reported_changes.is_a?(Hash) ? [reported_changes] : reported_changes
    # remove the meta file entry reported for Git commits
    # {"fileName"=>"GIT Commit", "version"=>"d266473d9d96809fff1d299073cc2fc11af37788", "status"=>"GIT PUSH"}
    # and the files with no fileName (happens for SVN commits for property changes(e.g. a file is added to the ignored list))
    changed_files.delete_if { |file| file["status"] =~ /Git Push/i || file["fileName"].blank? }
    changed_files.map do |file|
      {path: file["fileName"], action: convert_status(file["status"])}
    end
  end

  # extracts commit revision number from the reported "filesChanged"
  # @param [Hash] or [Array<Hash>] reported_changes
  def extract_revision_number(reported_changes)
    return nil unless reported_changes.present?
    changed_files = reported_changes.is_a?(Hash) ? [reported_changes] : reported_changes
    changed_files.last["version"]
  end


  # CTF reports deleted files as 'Deleted from ', so here we convert them to our accepted values.
  # @param [String] status the CTF-reported status
  # @return [String] the API-accepted status value
  def convert_status(status)
    case status.downcase
      when /^deleted/ then 'deleted'
      when /^added/ then 'added'
      when /^modified/ then 'modified'
      when /^copied/ then 'copied'
      else
        status.downcase
    end
  end


  # fetches all active ctf repositories matching the repository id for this commit from the database
  # @param [String] repo_id the repositoryId of the message hash
  def repositories_for_commit(repo_id)
    # Because of data migration with 8.1, there are old source data with to_delete state.
    # We need to exclude those data.
    # We will remove those data with next release.
    CtfRepository.where(ctf_id: repo_id, active: true, to_delete: nil).all
  end

  private

  # Convert the data passed from CTF into a list of CtfAssociation objects.  CTF passes a single string (not an array
  # with one element) in the case of a single association, but an array of strings if more than one association has been
  # drawn.  We normalize for that, and convert to CtfAssociation objects. Associations using our
  # IAF prefixes are ignored.
  #
  # @param [String|Array[String]] associations Either a single CTF object id as a string, or an array of strings with object ids.
  # @return [Array[CtfAssociation]] A list of associations, one for each association passed in.
  def extract_associations(associations)
    if associations.present?
      associations_list = associations.is_a?(Array) ? associations : [associations]
      associations_list.reject! { |association| ctf_id_has_orc_iaf_prefix?(association) }
      associations_list.map { |association| CtfAssociation.new(item_id: association) }
    else
      []
    end
  end

  # Evaluates a ctf item id, returning true if the item
  # uses an IAF prefix reserved by this app instance
  #
  # @return [Boolean] is this an iaf id for this application instance?
  def ctf_id_has_orc_iaf_prefix?(ctf_id)
    id_parts = CtfAssociation::ITEM_REGEX.match(ctf_id)
    ctf_association_prefixes.include? id_parts[:type]
  end
end

