#
# Copyright 2013 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

# this class handles soap calls to the CTF's ScmApp endpoint

class ScmService
  include UriHelper

  # session is a TeamForge::Session object, which should be available in the controllers as an instance variable @ctf_session
  # (it is created as a part of the authentication process, see TeamforgeApplication#ping_teamforge for details).
  def initialize(session)
    @service = session.scm
  end

  # returns repository information for a ctf project
  #
  # @param [String] ctf_project_id, currently lives in the Project_configuration model (ctf_project_id key)
  # @return [Array<Hash>] an array of hashes {repository_name: "foo", ctf_id: "reps101"}
  # @return [Array] empty array if either there are no repos in the project or the user has no read permissions for the
  # repos (in which case TeamForge will throw a ServiceError)
  def repositories_for_ctf_project(ctf_project_id)
    repositories = get_code_repositories(@service.get_repository_list(project_id: ctf_project_id))
    repository_data(repositories)
    rescue TeamForge::ServiceError
      []
  end

  # retrieves and parses the checkout command to get a repository url
  #
  # @param [String] repository_id ctf repository id
  # @@return [String] formatted url (with the username and password password removed)
  def repository_url(repository_id)
    user_name = Rails.configuration.teamforge[:iaf][:admin_user]
    repository_from_checkout_command(@service.get_scm_checkout_command(repository_id: repository_id, user_name: user_name))
  end

  protected

  # Extract the code repository data from a CTF repository list
  # CTF always reports the default repo for publishing HTML content, need to exclude it
  # @param [Hash] hash of SOAP data from get_repository_list
  # @return [Array<Hash>] SOAP data for repositories
  def get_code_repositories(ctf_answer)
    ctf_answer[:multi_ref].reject { |repo_hash| repo_hash[:path] =~ /scm\.Publishing/ }
  end

  # Map repository information from the CTF repository list
  # @param [Array<Hash>] CTF repository entries
  # @return [Array<Hash>] the relevant attributes for CommitSource
  def repository_data(repo_hashes)
    id_map = repo_hashes.inject({}) {|accum, repo_hash| accum.merge(repo_hash[:id] => (repository_url(repo_hash[:id]) rescue nil))}
    repo_hashes.select! {|repo_hash| id_map[repo_hash[:id]]}

    repo_hashes.map do |repo_hash|
      CtfRepositoryData.new(repo_hash[:title], repo_hash[:id], id_map[repo_hash[:id]])
    end.sort_by{ |entry| entry.name.downcase }
  end

  # Tease the VCS URI out of the full checkout command
  # Currently there is no way to just get this from CTF, so we extract it where we can
  # @param [String] command string. Example: 'svn checkout --username admin https://example.com/svn/repository-internal/application-publishing'
  # @return [String] the repository URI from the command string
  def repository_from_checkout_command(command)
    uri = URI.parse(command.split()[-1]) # URL is the last parameter of the command
    normalize_repository_uri(uri)
  end

  # convenience class, objects represent ctf_repos as they reported by CTF
  class CtfRepositoryData
    attr_accessor :name, :ctf_id, :url

    def initialize(name, ctf_id, url)
      self.name = name
      self.ctf_id = ctf_id
      self.url = url
    end
  end
end
